<?php
/* ********************************************************************
 * Interleave
 * Copyright (c) 2001-2016 Atomos Applications The Netherlands - www.atomos.nl - info@atomos.nl - www.interleave.nl - info@interleave.nl
 * Licensed under the GNU GPL. For full terms see the file COPYING.
 *
 * This file contains all database-type-specific functions
 *
 * Check http://www.interleave.nl/ for more information
 *
 * BUILD / RELEASE :: 7.2.0.20180308
 *
 **********************************************************************
 */


/* dit is tijdelijk om makkelijk te kunnen wisselen tussen utf8 en utf8mb4 */
$GLOBALS['charset'] = "utf8mb4";

function db_only_connect($user, $pass, $host) {
	$t = explode(":", $host);
	$dbhost = $t[0];
	$port = 3306;
	if (is_positive_integer($t[1])) {
		$port = $t[1];
	}
	if ($db = mysqli_connect($dbhost, $user, $pass, false, $port)) {

        mcq("SET NAMES " . $GLOBALS['charset'], $db);

        /* HF20160425 added mysql mode to avoid issues with STRICT NO_ZERO_IN_DATE,NO_ZERO_DATE etc */

        mcq("SET SESSION sql_mode = 'NO_ENGINE_SUBSTITUTION'", $db);

		return $db;
	}
	return false;
}
function db_only_select($database, $db_ident=false) {
	global $db;
	if (!$db_ident) {
		$db_ident = $db;
	}
	if (mysqli_select_db($db_ident, $database)) {
//JH 2017-08-29 already done by db_only_connect()
//		mcq_simple("SET NAMES " . $GLOBALS['charset'], $db_ident);
//		mcq_simple("SET SESSION sql_mode = 'NO_ENGINE_SUBSTITUTION'", $db_ident);
		return true;
	} else {
		return false;
	}
}
function db_base_connect($user, $pass, $host, $database) {
	if ($db = db_only_connect($user, $pass, $host)) {
		if (db_only_select($database, $db)) {
			return $db;
		} else {
			db_CloseConnection($db);
			return false;
		}
	} else {
		return false;
	}
}
function mcq_simple($query, $db_ident=false) {
	// deze functie wordt soms gebruikt om queries uit te voeren die niet door mcq mogen, bv. het loggen van foutmeldingen
	global $db;
	if (!$db_ident) {
		$db_ident = $db;
	}
//print '<pre>' . htme($query) . '</pre>';
	$a = mysqli_query($db_ident, $query) or handle_error(mysqli_error($db_ident) . " (mcq_simple)", mysqli_errno($db_ident), $query);
	return $a;
}
//vanwege memory gebruik (downloaden van grote tabel via xml) zouden we MYSQLI_USE_RESULT moeten gebruiken maar er worden nog queries uitgevoerd tijdens het afbeelden van de lijst en dan kan dat niet; MySQL Error: Commands out of sync; you can't run this command now
function mcq($query, $db_ident=false, $MYSQLI_USE_RESULT=false) {
//function mcq($query, $db_ident=false) {
	global $db;

	if (!$db_ident) {
		$db_ident = $db;
	}
//print '<pre>' . htme($query) . '</pre>';

	$tr = 0;
	if ($GLOBALS['logqueries']) {
		$tr = microtime_float();
	}


	$a = NULL;
	$try = 0;


	$resultmode = MYSQLI_STORE_RESULT;
	if ($MYSQLI_USE_RESULT)
	{
		$resultmode = MYSQLI_USE_RESULT;
	}


	while (true)
	{
		$try++;
		$success = true;

		$a = mysqli_query($db_ident, $query, $resultmode) or $success = false;
		//$a = mysqli_query($db_ident, $query) or $success = false;

		if ($success)
		{
			if ($try > 3) /* HF20170603 set to >3 instead of >1 */
			{
				//cannot use log_msg() beceause the query could be an INSERT and we might need the db_GetInsertId(), which would be overwritten by log_msg()
				//log_msg("ERROR: query successful after " . $try . " tries: " . $query);
				mail("root@localhost", "Query successful after " . $try . " tries " . $_SERVER['HTTP_HOST'], $query);
			}

			//query executed successfully; break from while loop
			break;
		}

		//1213 Deadlock found when trying to get lock
		//1062 Duplicate entry for key 'PRIMARY' error => percona cluster can cause these errors. Interleave hardly ever inserts primary key values.
		//1205 Lock wait timeout exceeded; try restarting transaction
		$errno = mysqli_errno($db_ident);
		if (($errno == 1213) || ($errno == 1062) || ($errno == 1205))
		{
			if ($try >= 100)
			{
				//we allow for a maximum of 100 consecutive deadlocks, before a fatal crash
				handle_error(mysqli_error($db_ident) . " (AFTER " . $try . " TRIES)", $errno, $query);
				//handle_error already exits, so this break is not needed
				break;
			}
			//sleeps 5 to 25ms and try again
			usleep(mt_rand(5000, 25000));
		}
		else
		{
			//a different kind of database error. this is fatal.
			handle_error(mysqli_error($db_ident), $errno, $query);
			//handle_error already exits, so this break is not needed
			break;
		}
	}

	$GLOBALS['mysql_query_counter']++;

	/* only when enabled, log this query */
	if ($GLOBALS['logqueries']) {
		$tm = microtime_float() - $tr;

		/* we're logging to a file called querylog.txt */
		$fp = fopen("querylog.txt","a");
		fputs($fp, $GLOBALS['mysql_query_counter'] . "\t" . $tm . "\t" . $query . " == " . $_SERVER['REQUEST_URI'] . "\n");
		fclose($fp);
		if ($GLOBALS['qlog_onscreen']) {
			$GLOBALS['tracelog'] .=  "SQL: " . $query . "\n";
		}
	}

	return $a;
}
function mres($str, $db_ident=false) {
	global $db;
	if (!$db_ident) {
		$db_ident = $db;
	}

	return mysqli_real_escape_string($db_ident, $str);
}
function db_GetAssocArray($query, $db_ident=false) {
	$ret = array();
	$res = mcq($query, $db_ident);
	while ($row = mysqli_fetch_array($res)) {
		$ret[$row[0]] = $row[1];
	}
	mysqli_free_result($res);
	return $ret;
}
function db_GetValue($query, $db_ident=false) {
	if (!preg_match("/ LIMIT \d+$/i", $query)) {
		$query .= " LIMIT 1";
	}
	$res = mcq($query, $db_ident);
	$ret = mysqli_fetch_array($res);
	mysqli_free_result($res);
	return $ret[0];
}
function db_GetRow($query, $db_ident=false) {
	if (!preg_match("/ LIMIT \d+$/i", $query)) {
		$query .= " LIMIT 1";
	}
	$res = mcq($query, $db_ident);
	$ret = mysqli_fetch_array($res);
	mysqli_free_result($res);
	return $ret;
}
function db_GetArray($query, $db_ident=false) {
	$ret = array();
	$res = mcq($query, $db_ident);
	while ($row = mysqli_fetch_array($res)) {
		$ret[] = $row;
	}
	mysqli_free_result($res);
	return $ret;
}
function db_GetInsertId($db_ident=false) {
	global $db;
	if (!$db_ident) {
		$db_ident = $db;
	}
	return mysqli_insert_id($db_ident);
}
function db_CloseConnection($db_ident=false) {
	global $db;
	if (!$db_ident) {
		$db_ident = $db;
	}
	mysqli_close($db_ident);
}
function db_GetFlatArray($query, $db_ident=false) {
	$ret = array();
	$res = mcq($query, $db_ident);
	while ($row = mysqli_fetch_array($res)) {
		$ret[] = $row[0];
	}
	mysqli_free_result($res);
	return $ret;
}
function DB_Connect($repository) {
	global $db;
	include $GLOBALS['CONFIGFILE'];
	ClearAllRunningCache();
	if ($repository == "")
	{
		$repository = 0;
	}
	if ($host[$repository] != "") {

		qlog(INFO, "Connecting to repository " . $repository . "");

		$chosen_host = $host[$repository];

		$db_conn_str = $chosen_host . $repository;
		if ($GLOBALS['LAST_DB_CONN'] == $db_conn_str && is_resource($db)) {
			//qlog(INFO, "Skipping DB connection setup. Connection should be there already. (" . $t_user . "@" . $t_host . ")");
			$GLOBALS['TIMES_DB_CONNECTS_SKIPPED']++;

//JH 2017-08-29 already done by db_only_connect()
//			mcq("SET NAMES " . $GLOBALS['charset'], $db);
            /* HF20160425 added mysql mode to avoid issues with STRICT NO_ZERO_IN_DATE,NO_ZERO_DATE etc */
//            mcq("SET SESSION sql_mode = 'NO_ENGINE_SUBSTITUTION'", $db);

			return $db;
		} else {
			if ($db = db_only_connect($user[$repository], $pass[$repository], $chosen_host)) {

				$GLOBALS['TIMES_DB_CONNECTS']++;
				$GLOBALS['LAST_DB_CONN'] = $db_conn_str;

				if (db_only_select($database[$repository])) {

//JH 2017-08-29 already done by db_only_connect()
//					mcq("SET NAMES " . $GLOBALS['charset'], $db);
//                			mcq("SET SESSION sql_mode = 'NO_ENGINE_SUBSTITUTION'", $db);

					return $db;

				} else {
					//print "SELECT FAIL " . $database[$repository] . "";
					return false;
				}
			} else {
				//print "CONNECT FAIL " . $database[$repository] . "";
				return false;
			}

		}
	}
}

?>
