/* interleave functions */
function __log(e, data) {
  log.innerHTML += "<br>" + e + " " + (data || '');
}

var audio_context;
var recorder;
var startTime;
var analyserContext = null;
var rafID = null;
var canvasWidth, canvasHeight;
var analyserNode;

function startUserMedia(stream) {

  inputPoint = audio_context.createGain();
  var input = audio_context.createMediaStreamSource(stream);

  audioInput = input;
  audioInput.connect(inputPoint);

  analyserNode = audio_context.createAnalyser();
  analyserNode.fftSize = 2048;
  inputPoint.connect( analyserNode );


  recorder = new Recorder(input, {
                numChannels: 1
              });

    __log('Ready. Press \'record\' to start recording.');

    zeroGain = audio_context.createGain();
    zeroGain.gain.value = 0.0;
    inputPoint.connect( zeroGain );
    zeroGain.connect( audio_context.destination );

    updateAnalysers();

}

function startRecording(button) {
  recorder && recorder.record();
  button.disabled = true;
  button.nextElementSibling.disabled = false;
  __log('Recording...');
  startTime = new Date();

}

function stopRecording(button) {
  recorder && recorder.stop();
  button.disabled = true;
  button.previousElementSibling.disabled = false;
  __log('Stopped recording.');

  // create WAV download link using audio data blob
   createDownloadLink();
  recorder.clear();
}

function createDownloadLink() {
  recorder && recorder.exportWAV(function(blob) {

  });
}

window.onload = function init() {
  try {
    // webkit shim
    window.AudioContext = window.AudioContext || window.webkitAudioContext;
    navigator.getUserMedia = ( navigator.getUserMedia ||
                     navigator.webkitGetUserMedia ||
                     navigator.mozGetUserMedia ||
                     navigator.msGetUserMedia);
    window.URL = window.URL || window.webkitURL;

    audio_context = new AudioContext;
    //__log('Audio context set up.');
    //__log('navigator.getUserMedia ' + (navigator.getUserMedia ? 'available.' : 'not present!'));
  } catch (e) {
    alert('No web audio support in this browser!');
  }

  navigator.getUserMedia({audio: true}, startUserMedia, function(e) {
    //__log('No live audio input: ' + e);
    __log('Could not access microphone or other error. Check your local settings.');
  });
};

function gettimediff(startTime, visual) {
    // later record end time
    var endTime = new Date();

    // time difference in ms
    var timeDiff = endTime - startTime;

    // strip the miliseconds
    timeDiff /= 1000;

    // get seconds
    var seconds = Math.round(timeDiff % 60);

    // remove seconds from the date
    timeDiff = Math.floor(timeDiff / 60);

    // get minutes
    var minutes = Math.round(timeDiff % 60);

    // remove minutes from the date
    timeDiff = Math.floor(timeDiff / 60);

    // get hours
    var hours = Math.round(timeDiff % 24);

    // remove hours from the date
    timeDiff = Math.floor(timeDiff / 24);

    // the rest of timeDiff is number of days
    var days = timeDiff;

    //$(".time").text(days + " days, " + hours + ":" + minutes + ":" + seconds);
    if (visual) {
        return hours + 'h ' + minutes + 'm ' + seconds + 's';
    } else {
        return endTime - startTime;
    }
}


function cancelAnalyserUpdates() {
    window.cancelAnimationFrame( rafID );
    rafID = null;
}

function updateAnalysers(time) {

    if (!analyserContext) {
        var canvas = document.getElementById("analyser");
        canvasWidth = canvas.width;
        canvasHeight = canvas.height;
        analyserContext = canvas.getContext('2d');
    }

    // analyzer draw code here
    {
//        __log('Drawing...' + canvasWidth);
        var SPACING = 3;
        var BAR_WIDTH = 1;
        var numBars = Math.round(canvasWidth / SPACING);
        var freqByteData = new Uint8Array(analyserNode.frequencyBinCount);

        analyserNode.getByteFrequencyData(freqByteData);

        analyserContext.clearRect(0, 0, canvasWidth, canvasHeight);
        analyserContext.fillStyle = '#F6D565';
        analyserContext.lineCap = 'round';
        var multiplier = analyserNode.frequencyBinCount / numBars;

        // Draw rectangle for each frequency bin.
        for (var i = 0; i < numBars; ++i) {
            var magnitude = 0;
            var offset = Math.floor( i * multiplier );
            // gotta sum/average the block, or we miss narrow-bandwidth spikes
            for (var j = 0; j< multiplier; j++)
                magnitude += freqByteData[offset + j];
            magnitude = magnitude / multiplier;
            var magnitude2 = freqByteData[i * multiplier];
            analyserContext.fillStyle = "hsl( " + Math.round((i*360)/numBars) + ", 100%, 50%)";
            analyserContext.fillRect(i * SPACING, canvasHeight, BAR_WIDTH, -magnitude);
        }
    }

    rafID = window.requestAnimationFrame( updateAnalysers );
}
/* recordmp3js */



(function(window){

  var WORKER_PATH = 'lib/recordmp3js/js/recorderWorker.js';
  var encoderWorker = new Worker('lib/recordmp3js/js/mp3Worker.js');

  var Recorder = function(source, cfg){
    var config = cfg || {};
    var bufferLen = config.bufferLen || 4096;
    var numChannels = config.numChannels || 2;
    this.context = source.context;
    this.node = (this.context.createScriptProcessor ||
                 this.context.createJavaScriptNode).call(this.context,
                 bufferLen, numChannels, numChannels);
    var worker = new Worker(config.workerPath || WORKER_PATH);
    worker.postMessage({
      command: 'init',
      config: {
        sampleRate: this.context.sampleRate,
        numChannels: numChannels
      }
    });
    var recording = false,
      currCallback;

    this.node.onaudioprocess = function(e){
      if (!recording) return;

      $('#recording_time').html(gettimediff(startTime, true));
      updateAnalysers();

      if (gettimediff(startTime, false) > 120000) {
          /* we automatically restart after two minutes (120000 ms)*/
          log.innerHTML += "<br>" + "Uploading two minutes of audio...";
          stopRecording(document.getElementById('JS_stoprecording'));
          startRecording(document.getElementById('JS_startrecording'));

      }

      var buffer = [];
      //for (var channel = 0; channel < numChannels; channel++){
      /* HF20161121 mono, single channel */
      channel = 0;
          buffer.push(e.inputBuffer.getChannelData(channel));
      //}

      worker.postMessage({
        command: 'record',
        buffer: buffer
      });
    }

    this.configure = function(cfg){
      for (var prop in cfg){
        if (cfg.hasOwnProperty(prop)){
          config[prop] = cfg[prop];
        }
      }
    }

    this.record = function(){
      recording = true;
    }

    this.stop = function(){
      recording = false;
    }

    this.clear = function(){
      worker.postMessage({ command: 'clear' });
    }

    this.getBuffer = function(cb) {

      currCallback = cb || config.callback;
      worker.postMessage({ command: 'getBuffer' })
    }

    this.exportWAV = function(cb, type){

      currCallback = cb || config.callback;
      type = type || config.type || 'audio/wav';
      if (!currCallback) throw new Error('Callback not set');
      worker.postMessage({
        command: 'exportWAV',
        type: type
      });
    }

	//Mp3 conversion
    worker.onmessage = function(e){
      var blob = e.data;
	  //console.log("the blob " +  blob + " " + blob.size + " " + blob.type);

	  var arrayBuffer;
	  var fileReader = new FileReader();

	  fileReader.onload = function(){
		arrayBuffer = this.result;
		var buffer = new Uint8Array(arrayBuffer),
        data = parseWav(buffer);

        /* console.log(data); */
		console.log("Converting to Mp3, do not close this window.");
		log.innerHTML += "<br>" + "Converting to Mp3 (this may take a while, do not close this popup)";
        SWID('MainContent');

        encoderWorker.postMessage({ cmd: 'init', config:{
            mode : 3,
			channels:1,
			samplerate: data.sampleRate,
			bitrate: data.bitsPerSample
        }});

        encoderWorker.postMessage({ cmd: 'encode', buf: Uint8ArrayToFloat32Array(data.samples) });
        encoderWorker.postMessage({ cmd: 'finish'});
        encoderWorker.onmessage = function(e) {
            if (e.data.cmd == 'data') {

				console.log("Done converting to Mp3");
				log.innerHTML += "<br>" + "Done converting to Mp3";
                HWID('MainContent');

				/*var audio = new Audio();
				audio.src = 'data:audio/mp3;base64,'+encode64(e.data.buf);
				audio.play();*/

				//console.log ("The Mp3 data " + e.data.buf);

				var mp3Blob = new Blob([new Uint8Array(e.data.buf)], {type: 'audio/mp3'});
				uploadAudio(mp3Blob);

				/*
                var url = 'data:audio/mp3;base64,'+encode64(e.data.buf);
				var li = document.createElement('li');
				var au = document.createElement('audio');
				var hf = document.createElement('a');

				au.controls = true;
				au.src = url;
				hf.href = url;
				hf.download = 'audio_recording_' + new Date().getTime() + '.mp3';
				hf.innerHTML = hf.download;
				li.appendChild(au);
				li.appendChild(hf);
                */
			 /*	recordingslist.appendChild(li);*/

            }
        };
	  };

	  fileReader.readAsArrayBuffer(blob);

      currCallback(blob);
    }


	function encode64(buffer) {
		var binary = '',
			bytes = new Uint8Array( buffer ),
			len = bytes.byteLength;

		for (var i = 0; i < len; i++) {
			binary += String.fromCharCode( bytes[ i ] );
		}
		return window.btoa( binary );
	}

	function parseWav(wav) {
		function readInt(i, bytes) {
			var ret = 0,
				shft = 0;

			while (bytes) {
				ret += wav[i] << shft;
				shft += 8;
				i++;
				bytes--;
			}
			return ret;
		}
		if (readInt(20, 2) != 1) throw 'Invalid compression code, not PCM';
		if (readInt(22, 2) != 1) throw 'Invalid number of channels, not 1';
		return {
			sampleRate: readInt(24, 4),
			bitsPerSample: readInt(34, 2),
			samples: wav.subarray(44)
		};
	}

	function Uint8ArrayToFloat32Array(u8a){
		var f32Buffer = new Float32Array(u8a.length);
		for (var i = 0; i < u8a.length; i++) {
			var value = u8a[i<<1] + (u8a[(i<<1)+1]<<8);
			if (value >= 0x8000) value |= ~0x7FFF;
			f32Buffer[i] = value / 0x8000;
		}
		return f32Buffer;
	}

	function uploadAudio(mp3Data){
		var reader = new FileReader();

        /* alert('uploadadio loopt ' + $('#JS_table').val() + ' record ' + $('#JS_record').val()); */
		reader.onload = function(event){
            log.innerHTML += "<br>Uploading recording";
			var fd = new FormData();
			var mp3Name = encodeURIComponent('audio_recording_' + new Date().getTime() + '.mp3');
			console.log("mp3name = " + mp3Name);
			fd.append('audio_recording_name', mp3Name);
			fd.append('audio_recording_mp3', event.target.result);
            fd.append('table_to_upload_to', $('#JS_table').val());
            fd.append('record_to_upload_to', $('#JS_record').val());
			$.ajax({
				type: 'POST',
				url: 'upload.php',
				data: fd,
				processData: false,
				contentType: false
			}).done(function(data) {
				//console.log(data);
				// log.innerHTML += "<br>" + data;
                result =  $.parseJSON(data);
                WhosMyDaddy().ListSetFileFolder(result["table"], result["folder"]);

                if (result["message"] != '')
                {
                    alert(result["message"]);
                }
                log.innerHTML += "<br>Done uploading. You can now close this window or start a new recording.";

			});
		};
		reader.readAsDataURL(mp3Data);
	}

    source.connect(this.node);
    this.node.connect(this.context.destination);    //this should not be necessary
  };

  /*Recorder.forceDownload = function(blob, filename){
	console.log("Force download");
    var url = (window.URL || window.webkitURL).createObjectURL(blob);
    var link = window.document.createElement('a');
    link.href = url;
    link.download = filename || 'output.wav';
    var click = document.createEvent("Event");
    click.initEvent("click", true, true);
    link.dispatchEvent(click);
  }*/

  window.Recorder = Recorder;

})(window);
