<?php
/* ********************************************************************
* Interleave
* Copyright (c) 2001-2016 Atomos Applications The Netherlands - www.atomos.nl - info@atomos.nl - www.interleave.nl - info@interleave.nl
* Licensed under the GNU GPL. For full terms see http://www.gnu.org/
*
* This script contains record functions
*
* Check http://www.interleave.nl/ for more information
*
* BUILD / RELEASE :: 7.2.0.20180308
*
**********************************************************************
*/
function process_record_save($table, $record, $data, $current_record = array(), $access = '', $skip_log_todo = false) // returns list($success, $record, $messages);
{

    /*
        process_record_save: the standard record add/update handler

        $table				DS tablename
        $record				Record id or empty for add
        $data				Associative array of data to change (could be a copy of $_POST)
        $current_record		The current record. If given empty, it will be fetched
        $access				"ok" only. If not set, it will be figured out.

        returns:

        array(
            [0]	success bool
            [1] recordid
            [2] array with messages
        )

    */

    /* array to hold return messages */
    $messages = array();

    $fields_to_update = array();
    $sql_fieldpart = array();
    $posted_fields = array();
    $sql_datapart = array();
    $field_todos = array();

    $to_journal = '';
    $to_datajournal = array();

    $update = false;
    $add = false;

    if (ipi($record)) {
        $update = true;
    } else {
        $add = true;
    }

    if ($access != 'ok' || (count($current_record) == 0 && ipi($record))) {

        /* $access should have been given to this function and it should be "ok" but if it's not given we have to check ourselves */

        $access = 'nok';

        if (!ipi($record) && $table != '' && $GLOBALS['UC']['PersonalDataStructure'][$table]['allowed_to_add']) {

            /* a plain add, with access */
            $access = 'ok';

        } elseif (ipi($record) && is_array($GLOBALS['UC']['PersonalDataStructure'][$table])) {

            $fields_to_fetch = array();
            $selectallfields = false;
            $bypass_record_access_by_sql = false;
            foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]['fields']) as $field) {
                $fields_to_fetch[] = $table.'-'.$field;
            }

            // load the current_record variable
            $current_record = getrecord($table, $record, $selectallfields, $bypass_record_access_by_sql, $fields_to_fetch);

            if ($current_record['type_of_access'] == 'ok') {

        		if ($current_record['record_locked_by'] == 0)
        		{
        	                /* nothing special, access granted, all fine */
        	                $access = 'ok';

				if (GetSetting('EnableEntityLocking') == 'Yes')
				{
	        	                /* insert a lock in the lock table */
    			                //$sql = 'INSERT INTO CRMentitylocks(lockby, lockon, locktable, timestamp_last_change) VALUES ('.$GLOBALS['USERID'].",'".mres($record)."','".mres($table)."', NOW()) ON DUPLICATE KEY UPDATE timestamp_last_change=NOW()";
	        	                //mcq($sql);
					/* strikt genomen moeten we nu een lock aanleggen en deze daarna weer weggooien maar nu blijft-ie achter en dat is niet goed HF20170922 */
				}
        		}
        		else
        		{
        			$access = 'readonly';
        		}

            } elseif ($current_record['type_of_access'] == 'readonly') {

                /* read-only, various reasons */
                $access = 'readonly';
            }
        } else {
            $messages[] = 'Unclear what to do. (' . $GLOBALS['UC']['PersonalDataStructure'][$table]['allowed_to_add'] . ')';

            return array(false, $record, $messages);
        }
    }
    $saveonadd = false;
    if ($add) {
        if (GetAttribute('flextable', 'CreateEmptyRecordOnAdd', str_replace('t', '', $table)) == 'Yes') {
            $saveonadd = true;
        }
    }
    foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]['fields']) as $field) {
        $dsname = $table.'-'.$field;
        $dsname_in_post = '';
        $message = '';
        $value = '';

        if (!ipi($record)) {
            $record = '_new_';
        }
        $dsname_in_post = $table.'-'.$field.'-'.$record;
        /* print 'Check for ' . $dsname_in_post . '<br>'; */

        if (isset($data[$dsname_in_post]) && $field != 'recordid') {

            /* we only allow setting timestamp_last_change when adding a new records on opening the form */
            if ($field != 'timestamp_last_change' || $saveonadd) {

                /* check if this field is writeable by this user. log error if it isn't */
                if (($current_record['type_of_access'] == 'ok' && $current_record['record_locked_by'] == 0 && $current_record['*'.$dsname] == 1) || $add) {

                    /* some corrections */
                    if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['fieldtype'] == 'date') {
                        if ($current_record[$dsname] == '0000-00-00 00:00:00') {
                            $current_record[$dsname] = '';
                        }
                        if ($current_record[$dsname] == '0000-00-00') {
                            $current_record[$dsname] = '';
                        }
                    } elseif ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'] != '' && $current_record[$dsname] == 0) {
                        $current_record[$dsname] = '';
                    }

                    if (array_key_exists($dsname_in_post, $data) && (trim($data[$dsname_in_post]) !== $current_record[$dsname] || $add || $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['multiselect'])) {

                        // Validate field input here

                        $result = 'nok';

                        /* check for multiselect and diary fields (strange ducks in the byte) */

                        if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['fieldtype_in_extrafields_table'] == 'diary') {

                            /* diary additions can be submitted also (next to ajax) */

                            /* get current diary content. it's in the database as a serialized array (ouch) */
                            $current_diary = array();

                            if ($current_record[$dsname] != '') {
                                $current_diary = unserialize($current_record[$dsname]);
                            }

                            if ($data[$dsname_in_post] != '') {
                                /* add array with new value */
                                    $current_diary[] = array(date('U'), $GLOBALS['USERID'], $data[$dsname_in_post]);
                            }

                            /* compose stuff to put into the database */
                            $value = serialize($current_diary);

                            if (is_array(unserialize($value))) {
                                $result = 'ok';
                            }
                        } else {

                            $value = $data[$dsname_in_post];

                            if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['fieldtype_in_extrafields_table'] == 'date (iso)' || $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['fieldtype_in_extrafields_table'] == 'date') {

                                /* reformat date fields */

                                $hv = false;
                        		if ($value != '') $hv = true;
                        		$value = INTLDate2NLDate(StrDate2IsoDate($value));
                        		if ($value == '' && $hv) {
                        			/* it HAD a value, but now it's gone, must have been wrong. trick old validator with bogus value to trigger error */
                        			$value = 'dl3b';
                        		}

                            } elseif ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['fieldtype_in_extrafields_table'] == 'date/time') {
                                /*
                                    do nothing. inconsistent, but ValidateFieldInputDataStructureOnly expects date/time fields to be user-formatted
                                */
                            }
                            /* validate the field value */
                            list($result, $message, $value) = ValidateFieldInputDataStructureOnly($dsname, $value, true, $record);



                            if ($result == 'ok') {

                                /* we don't support corrections on multiselect fields, so we use the original array, since it validated */
                                if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['multiselect']) {

                                        /*
                                            serialize it, this is the way it has to be put into the database (legacy)
                                            multiselect fields will always have 1 empty element. we need to remove those using FlattenArray
                                        */
                                        if (count(FlattenArray($data[$dsname_in_post])) > 0) {
                                            $value = serialize(FlattenArray($data[$dsname_in_post]));
                                        } else {
                                            $value = '';
                                        }

                                } elseif ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['fieldtype_in_extrafields_table'] == 'date (iso)') {

                                    /* the validator returns dutch formatted dates. for iso, we need to convert it to yyyy-mm-dd */
                                    $value = NLDate2INTLDate($value);
                                }
                            } else {
                                //$messages[] =  'ValidateFieldInputDataStructureOnly returned ' . $result . '';
                            }
                        }

                        if ($result == 'ok') {
                            if ($add) {
                                // Add to-do for triggers
                                /*
                                if (ipi(str_replace("f", "", $field))) {
                                    $field_todos[] = array("EFID" . str_replace("f", "", $field), $value);
                                }
                                */

                                if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'] != '' && $value == '0') {
                                    /* when adding a new record, a reference field containing "0" means not set at all. as this is database default, we can safely ignore it here to make sure it isn't logged and no rules are kicked  */
                                        $value = '';
                                }

                                /*
                                    only set something if it's given. as we're ADDING a record, the current value is always empty
                                    also process fields which have a non-empty default value, but were posted empty
                                */
                                if ($value != '' || ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['defaultvalue'] != '' && $value == '')) {

                                    /* add to to-do query parts */
                                    $sql_fieldpart[] = '`'.$GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['name_in_database'].'`';
                                    $sql_datapart[] = mres($value);

                                    /* add to-do for rules. store in array, the record number is unknown */
                                    /* DIT STOND EERST NET BUITEN DEZE IF LUS. GEVAARLIJKE CHANGE */
                                    $field_todos[] = array($table.'-'.$field, $value);

                                    $to_datajournal[] = array($value, $field);
                                    $to_journal .= 'Field '.$field.' ['.$GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['name'].'] updated from [';

                                    if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'] == 'users') {
                                        $to_journal .= $current_record[$dsname].':'.GetUserName($current_record[$dsname]);
                                    } else {
                                        $to_journal .= $current_record[$dsname];
                                    }

                                    $to_journal .= '] to [';

                                    if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'] == 'users') {
                                        $to_journal .= GetUserName($value);
                                    } else {
                                        $to_journal .= $value;
                                    }

                                    $to_journal .= "]\n";

                                    /* we need to pass this array to add_default_values_for_fields to make sure they're skipped when adding default values */
                                    $posted_fields[] = $field;
                                    //print '<h2>Add ' . $field . ' to posted_fields because the value is "' . htme($value) . '" and the defaultvalue is "' . $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]['defaultvalue'] . '</h2>';
                                } else {
                                    //print '<h2>Do NOT add ' . $table . '-' . $field . ' to posted_fields because the value is "' . htme($value) . '" and the defaultvalue is "' . $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]['defaultvalue'] . '</h2>';
                                }
                            } elseif ($update) {

                                /* TODO: DEZE FUNCTIE LOGGED EEN ONTERECHTE CHANGE OP EEN DIARY VELD IN ALLE GEVALLEN OOK ALS DEZE NIET GEUPDATE IS! */

                                /* correct for proper comparison with getrecord output */
                                $cmpval = $value;
                                if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['fieldtype_in_extrafields_table'] == 'date (iso)' || $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['fieldtype_in_extrafields_table'] == 'date') {

                                        /* reformat date fields */
                                        $cmpval = NLDate2INTLDate($value);
                                }

                                /* check if it was changed */
                                if ($cmpval !== $current_record[$dsname]) {

                                    /* add to list of fields to update */
                                    $fields_to_update[$GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['name_in_database']] = $value;

                                    /* log a todo */
                                    if (!$skip_log_todo) {
                                        AddToDo($table.'-'.$field, $value, $record, $current_record[$dsname]);
                                    }

                                    /* log a message in the datajournal */
                                    DataJournal($record, $current_record[$dsname], $value, $field, ds_tablename_to_datajournal_tabletype($table));

                                    /* add text for journal */
                                    $to_journal .= 'Field '.$field.' ['.$GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['name'].'] updated from [';

                                    /* some formatting for the journal. this is quite expensive, but much more friendly for the user */
                                    if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'] != '') {
                                        $to_journal .= $current_record[$dsname].':'.get_refer_layout_for_record($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'], $current_record[$dsname]);
                                    } else {
                                        $to_journal .= $current_record[$dsname];
                                    }

                                    $to_journal .= '] to [';

                                    if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'] != '') {
                                        $to_journal .= $value.':'.get_refer_layout_for_record($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'], $value);
                                    } else {
                                        $to_journal .= $value;
                                    }

                                    $to_journal .= "]\n";
                                }
                            }
                        } else {
                            $messages[] = 'Error in '.$field.'. Record not added or updated: '.$message.'.';

                            return array(false, $record, $messages);
                        }
                    }
                } else {
                    /* this is not fatal; the record was saved, but this field was excluded */
                    log_msg('ERROR: Field ['.$field."] was posted, but the user doesn't have access to write to this field.");

                    if (GetSetting('DevelopmentOrProduction') == 'Development') {
                        $messages[] = 'ERROR: Field ['.$field."] was posted, but the user doesn't have access to write to this field.";
                    }
                }
            }
        }
    }

    //qlog(INFO, string_r($field_todos));

    if ($update && $access == 'ok') {
        $sql = 'UPDATE `CRM'.$GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'].'` SET ';
        $sqlparts = array();

        foreach ($fields_to_update as $field => $value) {
            $sqlparts[] = '`CRM'.$GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'].'`.`'.$field."` = '".mres($value)."' ";
        }

        $sql .= implode(',', $sqlparts);

        $sql .= ' WHERE '.$GLOBALS['UC']['PersonalDataStructure'][$table]['primary'].' = '.$record;

        if (count($sqlparts) > 0) {

            /* execute the query */
            mcq($sql);

            /* check if there are any admin actions to perform like DS drop or mysql trigger re-create */
            check_admin_changes_after_change($table);

            if ($to_journal != '') {
                journal($record, $to_journal, ds_tablename_to_journal_tabletype($table));
            }

            /* create to-do's for rules */
            if (!$skip_log_todo) {
                AddToDo($table.'-changed', '', $record, '');
            }

            $fields_to_fetch = array();
            $selectallfields = false;
            $bypass_record_access_by_sql = false;
            foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]['fields']) as $field) {
                $fields_to_fetch[] = $table.'-'.$field;
            }

            /* reload the current_record variable */
            $current_record = getrecord($table, $record, $selectallfields, $bypass_record_access_by_sql, $fields_to_fetch);

        } elseif ($update) {

            $update = false;
            /* in 99,9% of the cases this will be a form submit with not a single value *change*. not a real error */

        }
    } elseif ($add && $access == 'ok') {

        /* add timestamp_last_change field in case *nothing* was given to save (to prevent a crashing query) */
        if (!in_array('`timestamp_last_change`', $sql_fieldpart)) {
            $sql_fieldpart[] = '`timestamp_last_change`';
            $sql_datapart[] = date('Y-m-d H:i:s');
        }

        $sql = 'INSERT INTO `CRM'.$GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'].'` ('.implode(',', $sql_fieldpart).") VALUES('".implode("','", $sql_datapart)."')";

        mcq($sql);

//		print '<h1>' . $sql . '</h1>';

        $record = db_GetInsertID();

        /* journal this event */
        journal($record, 'Record created', ds_tablename_to_journal_tabletype($table));

        DataJournal($record, '-', '-', 'record created', ds_tablename_to_datajournal_tabletype($table));

        if ($to_journal != '') {
            journal($record, $to_journal, ds_tablename_to_journal_tabletype($table));
        }
        foreach ($to_datajournal as $dj) {
            DataJournal($record, '{{add}}',  $dj[0], $dj[1], ds_tablename_to_datajournal_tabletype($table));
        }

        /* add all default values */
        add_default_values_for_fields($table, $record, $posted_fields);

        /* we now know the record number, add todo's SHOULD NOT USE TODO'S - PROCESSRULES SHOULD BE CALLED DIRECTLY */
        if (!$skip_log_todo) {

            foreach ($field_todos as $ftd) {
                AddToDo($ftd[0], $ftd[1], $record, '');
            }

            /* log todo for later processing of rules */
            AddToDo($table.'-added', '', $record, '');

        }

        /* check if there are any admin actions to perform like DS drop or mysql trigger re-create */
        check_admin_changes_after_change($table);

    } else {

        /* no access; return false */
        return array(false, $record, $messages);
    }

    /* we changed a record; we have to unset the materialized record access subquery (cache) */
    unset($GLOBALS['PageCache']['php_subquery_materialization']);

    return array(true, $record, $messages);
}

function get_parsed_template($table, $record, $passthru, $rec, $force_autoform, $access)
{
    /*
     *   This function should be able to parse any template type. It fetches the template content
     *   and determines the type. The rest is done by get_parsed_template_with_given_template.
     */
    $template_contents = '';
    $template_type = '';

    /* let's see if this table has a template set */
    $form = $GLOBALS['UC']['PersonalDataStructure'][$table]['form-template-id'];

    /* this overrules the above */
    if (ipi($passthru['template_to_use'])) {
        /* using the ftu var, a form to use can be given which overrules */
        $form = $passthru['template_to_use'];
    }

    if ($force_autoform) {
        $template_contents .= ParseAutoForms('@AUTOHEADER@ @AUTOCHILDTABLESINNAV@', $table);
        $template_type = 'TEMPLATE_HTML_FORM';
    } else {
        $template_type = ParseTemplateInclusions(GetTemplateType($form));

        if ($template_type != 'TEMPLATE_HTML_FORM' && $template_type != 'TEMPLATE_HTML' && $template_type != 'TEMPLATE_XML' && $template_type != 'TEMPLATE_REPORT_PDF' && $template_type != 'TEMPLATE_PLAIN') {
            return 'Unsupported template type';
        }
        if ($template_type == 'TEMPLATE_HTML_FORM') {
            $template_contents .= ParseAutoForms(ParseTemplateInclusions(GetTemplate($form)), $table);
        } else {
            $template_contents .= ParseTemplateInclusions(GetTemplate($form));
        }
    }

    return get_parsed_template_with_given_template($table, $record, $passthru, $rec, $force_autoform, $access, $template_contents, $template_type);
}

function get_parsed_template_with_given_template($table, $record, $passthru, $rec, $force_autoform, $access, $template_contents, $template_type)
{
    global $lang;

    $locked = false;
    $lockedby = '';
    $html = '';

    if (!ipi($record) && $table != '') {
        $record = '_new_';
    }

    /* define what to do per type. this DOES NOT INFLUENCE THE ORDER IN WHICH THEY HAPPEN */

    $todo = array();
    switch ($template_type) {

        case 'TEMPLATE_HTML':
        case 'TEMPLATE_PLAIN':
        case 'TEMPLATE_XML':
            $todo = array('generic', 'php', 'modules', 'listtags', 'language', 'usergroup', 'languageif');
            if ($table != '' && $record != '') {
                $todo[] = 'data';
            }
        break;
        case 'TEMPLATE_REPORT_PDF':
            $todo = array('autopdf', 'generic', 'php', 'modules', 'listtags', 'language', 'usergroup', 'languageif');
            if ($table != '' && $record != '') {
                $todo[] = 'data';
            }
        break;
        case 'DEFAULTVALUE':
            $todo = array('generic', 'php', 'modules', 'language', 'usergroup', 'languageif');
            if ($table != '') {
                $todo[] = 'data';
            }
        break;

        case 'TEMPLATE_HTML_FORM':
            $todo = array('generic', 'php', 'modules', 'listtags', 'language', 'usergroup', 'languageif');
            if ($table != '') {
                $todo[] = 'form';
                $todo[] = 'data';
            }
            if (ipi($record)) {
                $todo[] = 'listtags';
            }
        break;

        case 'TEMPLATE_DASHBOARD':
            $todo = array('generic', 'php', 'modules', 'listtags', 'usergroup', 'language', 'dashboard', 'languageif');
        break;

        case 'TEMPLATE_MENU':
            $todo = array('generic', 'php', 'modules', 'listtags', 'usergroup', 'language', 'languageif');
        break;

        default:
            log_msg('ERROR: get_parsed_template_with_given_template called with no or unkown tempate type.');
        break;

    }
    //DA($todo, true);

    /* we also support parsing templates not bound to a table */
    if (in_array('form', $todo)) {
	    //2017-04-26 only display the locked message if this record would be writable if it wasn't locked
        if (($rec['type_of_access'] == 'ok') && ($rec['record_locked_by'] > 0)) {
            $locked = true;
            $lockedby = $rec['record_locked_by'];
        }

        $save_after_validate = 'false';
        if ($passthru['BYAJAX']) {
            $save_after_validate = 'true';
        }
        //print 'debug rf.php:502<br>';
        //$save_after_validate = "true";
        $html = '';

        $html .= '<div id="lostlockmsg" class="warnmsg hide_content">'.htme($lang['lostlock']).'</div><div id="form-'.$table.'-'.$record.'"><form class="recordform" novalidate autocomplete="off" enctype="multipart/form-data" id="'.$table;

	//JH: the save button directly after the <form> is opened forces the browser to submit this button when enter is pressed in a text box

        //* this is where we submit the form. disable all elements in a list */

        $html .= '-'.$record.'" method="post" action="" onsubmit="

        if (general_form_submit(this.id, \''.$record.'\', '.$save_after_validate.')) {
            $(\'.listcontainer :input\').prop(\'disabled\', true);
            return true;
        } else {
            return false;
        }
        " onkeyup="LockCounter=1;"><input type="submit" value="' . htme($lang['save']) . '" style="position: absolute; left: -5000px;">'.$template_contents;

    } else {

        $html = $template_contents;

    }
    if (in_array('autopdf', $todo)) {
        $html = str_replace('@AUTO@', ReturnDefaultPDFTemplateForTable($table, $record), $html);
        $html = str_replace('@AUTO_WITH_CHILD_TABLES@', ReturnDefaultPDFTemplateForTable($table, $record, true), $html);
    }


    if (in_array('generic', $todo)) {

        /*
        *
        * Handles:
        *
        * Time, user, userid, group, groupid tags
        * Calendar tags
        * Info tags
        * Inline add form tags
        * Default value tags
        * Template inclusions
        * Default Report Tags
        * User and group fields
        * Absolutely addressedd ifields
        *
        */

        $html = ParseTemplateGeneric($html);
    }
    if (in_array('usergroup', $todo)) {
        $html = StripGroupOnlyTags($html);
    }
    if (in_array('languageif', $todo)) {
        $html = StripLanguageOnlyTags($html);
    }
    /* replace language tags */
    if (in_array('language', $todo)) {
        // Replace language tags like [[save]] and [[efid12]]
        $html = ParseTemplateLanguageTags($html);
    }

    if (in_array('form', $todo)) {
        if ($access != 'readonly') {
            $html = StripROOnlyTags($html);
        } elseif ($access == 'readonly') {
            $html = StripRWOnlyTags($html);
        }
        if (ipi($record)) {
            $html = StripNewOnlyTags($html);
        } else {
            $html = StripExistingOnlyTags($html);
        }
    }

    /* we also support parsing templates not bound to a table */
    if (in_array('data', $todo)) {

        /* now parse the record */

        $html = ParseTemplateRecordData($html, $table, $record, $passthru['duplicate']);
        if (in_array('form', $todo)) {

            /* replace hidden fields. syntax: #HEFID500[1]# */
            /* HF20170424 we stopped supporting hidden fields */

            $savebtn = '';
            if ($access == 'ok') {
                $savebtn = '<input type="submit" id="JS_submitbutton" name="btnsubmit" value="'.htme($lang['save']).'">';
            }

            $html = str_replace('#SAVEBUTTON#', escape_interleave_characters($savebtn), $html);

            $delbtn = '';

            /* passthru['no_deletes'] is important when opening a record using the link in a ref-to-flex field. if a user would have a delete button (if the parent is a new record, no relation exists) evrything goes haywire */

            if ($record != '_new_' && $access == 'ok' && $passthru['no_deletes'] == '' && $GLOBALS['UC']['PersonalDataStructure'][$table]['allowed_to_delete']) {
                if (CheckFunctionAccess('AllowedToUseDescendingDeletes') != 'ok' && CheckIfRecordHasChilds($table, $record) && !is_administrator()) {
                    $delbtn = "<input type='button' disabled value='".htme($lang['delete'])."' name='delete_record_including_childs' class='deletebutton'>";
                } else {
                    if ($GLOBALS['webapp']) {
                        $delbtn = "<input type='button' value='".htme($lang['delete'])."' name='delete_record_including_childs' onclick=\"delete_by_ajax(" . $record.",'".$table."-recordid');\" class='deletebutton'>";
                    } else {
                        $delbtn = "<input type='button' value='".htme($lang['delete'])."' name='delete_record_including_childs' onclick=\"PopFancyBoxDimensioned('','record.php?record=".$record.'&amp;table='.$table.'&amp;delete=1&amp;nonavbar=1&amp;ParentAjaxHandler='.htme($passthru['ParentAjaxHandler']).'&amp;fromlist='.htme($passthru['fromlist'])."&amp;deleting_myself=1&amp;BYAJAX=" . htme($passthru['byajax']) . "', 300, 500);\" class='deletebutton'>";
                    }
                }
            }

            $html = str_replace('#DELETEBUTTON#', escape_interleave_characters($delbtn).' ', $html);
            $html = str_replace('#DELETEINCLUDINGCHILDSBUTTON#', escape_interleave_characters($delbtn).' ', $html);

            /* some variables we need to pass along */

            $pass_along = array();
            $pass_along['form_submitted'] = '0';
            $pass_along['e_button'] = '';
            $pass_along['concerningrecord'] = $record;
            $pass_along['concerningtable'] = $table;
            $pass_along['hash'] = randomstring(32);

            if (ipi($passthru['duplicate'])) {
                $pass_along['duplicate_of'] = $passthru['duplicate'];
            }
            if (ipi($passthru['template_to_use'])) {
                $pass_along['template_to_use'] = $passthru['template_to_use'];
            }

            foreach ($pass_along as $name => $value) {
                $html .= '<input type="hidden" id="JS_'.$name.'" name="'.$name.'" value="'.htme($value).'">';
            }

            $html .= '</form></div>';
        }

        if (stristr($html, '@BACKTOLIST@')) {
            $html = str_replace('@BACKTOLIST@', escape_interleave_characters(backtolist($passthru['fromlist'])), $html);
        }

        if (stristr($html, '@FAMILYTREE@')) {
            $html = str_replace('@FAMILYTREE@', escape_interleave_characters(familytree($table, $record, $passthru['starttable'], $passthru['startrecordid'], $passthru['fromlist'])), $html);
        }
    }

    /* evaluate modules when they are part of this template */
    if (in_array('modules', $todo)) {
        if (stristr($html, '@MODULE')) { /* some reference to a module exists */
            $x = GetModules();
            foreach ($x as $module) {
                if (stristr($html, '@MODULE'.$module['mid'].'@')) {
                    $ajax = GetAttribute("module", "RunInAjax", $module['mid']);
                    if ($ajax == 'Yes') {
                        $ContentsOfDivWhileLoadingViaAjax = GetAttribute('module', 'ContentsOfDivWhileLoadingViaAjax', $module['mid']);
                        $x_html = '<div id="JS_RunModule' . $module['mid'] . '">' . htme(ParseTemplateGeneric($ContentsOfDivWhileLoadingViaAjax)) . '</div>';
                        $x_html .= '<script type="text/javascript">$(document).ready(function() {';
                        $x_html .=  "refresh_contents_of_div('JS_RunModule" . $module['mid'] . "', 'assist.php', 'Function=RunModule&ExecInsModule=" . $module['mid'] . "&recordid=" . $record . "&table=" . $table . "', true, 'POST', '', false);";
					    $x_html .= '});</script>';
                        $html = str_replace('@MODULE'.$module['mid'].'@', $x_html, $html);
                    } else {
                        if ($table != '') {
                            $html = str_replace('@MODULE'.$module['mid'].'@', escape_interleave_characters(RunModule($module['mid'], false, true, $record, str_replace('t', '', $table))), $html);
                        } else {
                            $html = str_replace('@MODULE'.$module['mid'].'@', escape_interleave_characters(RunModule($module['mid'], false, true, '', '')), $html);
                        }
                    }
                    qlog(INFO, 'Module '.$module['mid'].' was ran and included in this template');
                }
            }
        }
    }

    // Replace list tags e.g. #LIST|table=t2#
    if (in_array('listtags', $todo)) {

        /* TODO: we moeten hier nog iets verzinnen waardoor waardeloze list tags (sublist maar nog geen record) niet getoond worden */

        /* replace the tags */
        $html = ParseTemplateListTags($html);

        /* parse inline-add tags */
        $html = ParseInlineAddFormTags($html);
    }

    if (in_array('generic', $todo)) {
        // Replace report tags
        $html = ParseTemplateCreateDefaultReportTags($html);
    }

    /* we also support parsing templates not bound to a table */
    if (in_array('form', $todo)) {

        // Inline add forms
        $template = ParseInlineAddFormTags($template);

        // Arrow tag
	$listid = $_REQUEST["listid"];
        $html = parse_arrow_tag($html, $table, $record, $listid);

        // The lock icon
        if (!$locked) {
            $html = StripLockedOnlyTags($html);
            $html = str_replace('#LOCKICON#', '', $html);
        } else {
            $html = str_replace('#LOCKICON#', '<span class="fa fa-lock"></span>', $html);
            $html = str_replace('@LOCKBY@', escape_interleave_characters(GetUserName($lockedby)), $html);
            $html = str_replace('@LOCKEDONLY@', '', $html);
            $html = str_replace('@ENDLOCKEDONLY@', '', $html);
        }

        // Report icon (only for MINV) -- LEGACY
        if (strpos($html, '#REPORTICON#') !== false) {
            $flextableid = '';
            $flextablerecord = '';
            if (ipi(str_replace('t', '', $table))) {
                $flextableid = str_replace('t', '', $table);
                $flextablerecord = $record;
            }

            $reporticon = "<a href=\"javascript:poplittlewindow('entityreport.php?FlexTableRecord=".$flextablerecord.'&amp;FlexTable='.$flextableid."&amp;nonavbar=1')\">";
            $reporticon .= '<span class="fa fa-file-word-o"></span>';
            $reporticon .= '</a>';

            $html = str_replace('#REPORTICON#', escape_interleave_characters($reporticon), $html);
        }

        // The file box and list
        if (strpos($html, '#FILEBOX#') !== false) {
            $new = FileUploadBox($table, $record);
            if (ipi($record)) {
                $new .= DisplayFileList($record, $table);
            }
            $html = str_replace('#FILEBOX#', escape_interleave_characters($new), $html);
        }
        if (strpos($html, '#FILEBOXNOVERSIONING#') !== false) {
            $new = FileUploadBox($table, $record, 0, true);
            if (ipi($record)) {
                $new .= DisplayFileList($record, $table);
            }
            $html = str_replace('#FILEBOXNOVERSIONING#', escape_interleave_characters($new), $html);
        }
        if (strpos($html, '#FILELIST#') !== false) {
            $replace = '';
            if (ipi($record)) {
                $replace = escape_interleave_characters(DisplayFileList($record, $table));
            }
            $html = str_replace('#FILELIST#', $replace, $html);
        }

        // The journal icon + popuplink

        // The journal icon. Don't show if not allowed.
        $journalicon = '';
        if ((CheckFunctionAccess('DenyJournalAccess') != 'ok' || is_administrator()) && $record != '_new_') {
            $journalicon = '<div id="journal-icon"><a title="Journal" href="journal.php?table='.$table.'&amp;recordid='.htme($record).'&amp;nonavbar=1" onclick="PopFancyBoxLarge(\'\', this.href); return false;">';
            $journalicon .= '<span class="fa fa-book"></span>';
            $journalicon .= '</a></div>&nbsp;';
        }

        $html = str_replace('#JOURNALICON#', escape_interleave_characters($journalicon), $html);

        /* the lock message */

        if ($template_type == 'TEMPLATE_HTML_FORM') {
            if (!$locked) {

                /* remove text only for when locked */
                $html = StripLockedOnlyTags($html);

                if ($access == 'ok') {
                    /* insert javascript to update the lock on this record  */
                    $html .= ReturnLockJavascript($record, $table);
                }

            } else {
                /* leave lock message in place */
                $html = str_replace('@LOCKBY@', GetUserName($lockedby), $html);
            }

        }
    } // end if todo form

    if (in_array('dashboard', $todo)) {
        if (strstr($html, '#CLOCK#')) {
            $html = str_replace('#CLOCK#', ShowClock(), $html);
        }
    }

    if (in_array('php', $todo)) {
        if (stristr($html, '<?')) { // PHP exists in template
            $html = EvaluateTemplatePHP($html, false, $table, $record);
        }
    }

    /* we also support parsing templates not bound to a table */
    if (in_array('form', $todo)) {

        /* show inline and raw edit icons */
        if (($GLOBALS['SesMem']['InlineEditMode_InlineLinks'] == 'On' && is_administrator()) && ($template_type == 'TEMPLATE_HTML_FORM')) {

            // Show raw-edit icon
            if (ipi($record)) {
			    $html = '<div class="InlineEditModeDivContainer"><div class="InlineEditModeDiv"><a title="Edit this record raw" href="record.php?raw=1&amp;record='.htme($record).'&amp;table='.htme($table).'"><span class="fa fa-wrench"></span></a></div></div>'.$html;
            }
        }

        /* now add form-specific javascript. we do this last, so we don't bother other functions with a bunch of javascript in the template */
        if (!$locked) {
            $html .= '<script type="text/javascript">';
            $html .= PrintExtraFieldForceJavascript($table);
            $html .= '</script>';
        }
    }

    return unescape_interleave_characters($html);
}

function process_open_todos($table, $record)
{
    /*
        This function gets all todo's for rules (using fully qualified onchange field tags) and processes them. It ignores todo's made for triggers.
    */
    //trace("process_open_todos($table, $record)");
    $sql = "SELECT * FROM `CRMtodo` WHERE `CRMtodo`.`user`='".mres($GLOBALS['USERID'])."' AND LEFT(`CRMtodo`.`onchange`, ".strlen($table).") = '".mres($table)."' AND `CRMtodo`.`worker_hash` = '' AND `CRMtodo`.`eid`='".$record."'";

    $Todo = db_GetArray($sql);

    $sql = "SELECT COUNT(*) FROM `CRMrules` WHERE `CRMrules`.`table`='".mres($table)."'";
    $rules = mcq($sql);

    if (count($Todo) > 0 && $rules > 0) {
        if (!is_array($GLOBALS['SystemDataStructure'])) {
            $GLOBALS['SystemDataStructure'] = GetSystemDataStructure();
        }
        // SLORDIG en memory intensief; Aan Jeroen vragen of dit netter kan.

        // Bypass PDS security by overwriting it to SDS
        $PDSPARK = $GLOBALS['UC']['PersonalDataStructure'];
        $GLOBALS['UC']['PersonalDataStructure'] = $GLOBALS['SystemDataStructure'];

        /*
            To prevent ProcessRules from having to get_records over and over we get it here.
        */

        /* determine the fields to select by checking ruleactions, ruleconditions and ruleactionconditions */
        $sql = "SELECT `CRMruleconditions`.`field` FROM `CRMruleconditions` WHERE `CRMruleconditions`.`field` LIKE '".$table."-%' UNION SELECT `CRMruleactionconditions`.`field` FROM `CRMruleactionconditions` WHERE `CRMruleactionconditions`.`field` LIKE '".$table."-%' UNION SELECT `CRMruleactions`.`field` FROM `CRMruleactions` WHERE `CRMruleactions`.`field` LIKE '".$table."-%'";

        /* now add all primaries to the fields to fetch */
        $fields_to_fetch_tmp = db_GetFlatArray($sql);
        $fields_to_fetch = array();

        foreach ($fields_to_fetch_tmp as $ftf) {

            /* add current always */
            $fields_to_fetch[] = $ftf;

            $elements = explode('-', $ftf);
            $element_table = $elements[count($elements) - 2];

            /* last table is not our own table, add primary */
            if ($element_table != $table) {
                unset($elements[count($elements) - 1]);
                $path = implode('-', $elements).'-'.$GLOBALS['UC']['PersonalDataStructure'][$element_table]['primary'];
                if (!in_array($path, $fields_to_fetch)) {
                    $fields_to_fetch[] = $path;
                }
            }
        }

        $selectallfields = false;
        $bypass_record_access_by_sql = false;

        $current_record = getrecord($table, $record, $selectallfields, $bypass_record_access_by_sql, $fields_to_fetch);

        // Put PDS back
        $GLOBALS['UC']['PersonalDataStructure'] = $PDSPARK;
        unset($PDSPARK);

	//we use strtolower because this database column in not binary. to prevent cases where worker_hash "abcde" would also match "ABCDE" or "ABcdE", etc
        $worker_hash = strtolower(randomstring(32));

        /* update the table to mark (lock) records for processing */
        $sql = "UPDATE `CRMtodo` SET `CRMtodo`.`worker_hash` = '".$worker_hash."' WHERE `CRMtodo`.`user`='".mres($GLOBALS['USERID'])."' AND LEFT(`CRMtodo`.`onchange`, ".strlen($table).") = '".mres($table)."' AND `CRMtodo`.`eid`='".$record."' AND `CRMtodo`.`worker_hash` = ''";
        mcq($sql);

        /* next, select the records we just marked. the order by clause is important; further down we fetch the records one by one */
        $sql = "SELECT * FROM `CRMtodo` WHERE `CRMtodo`.`worker_hash`='".$worker_hash."' ORDER BY `CRMtodo`.`eid`";
        $Todo = db_GetArray($sql);

        /* array to hold todo-id's which were processed and can be deleted */
        $Todelete = array();

        /* loop through all todo's */
        foreach ($Todo as $do) {
            if ($do['onchange'] == $table.'-added') {
                /* an added record */
                //trace("Processrules($table, $record, '', \"added\", " . $do['to_value'] . "");
                Processrules($table, $record, '', 'added', $do['to_value'], $current_record);
                $Todelete[] = $do['todoid'];
            } elseif ($do['onchange'] == $table.'-changed') {
                /* a changed record */
                Processrules($table, $record, '', 'changed', $do['to_value'], $current_record);
                $Todelete[] = $do['todoid'];
            } elseif ($do['onchange'] == $table.'-deleted') {
                /* a deleted record */
                //trace("Processrules($table, $record, '', \"deleted", " . $do['to_value'] . "");
                Processrules($table, $record, '', 'deleted', $do['to_value'], $current_record);
                $Todelete[] = $do['todoid'];
            } else {

                /* a generic change-of-a-field */
                /* get the field */
                list($ignore, $todofield) = explode('-', $do['onchange']);
                unset($ignore);

                /* check if the field is in the data structure */
                if (is_array($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$todofield])) {

                    //trace("Processrules($table, $record, \"$todofield\", \"changed\", " . $do['to_value'] . ");");
                    Processrules($table, $record, $todofield, 'changed', $do['to_value'], $current_record);
                    $Todelete[] = $do['todoid'];
                }
            }
        }

        /* delete the processed todo's */
        if (count($Todelete) > 0) {
            $sql = 'DELETE FROM `CRMtodo` WHERE `CRMtodo`.`todoid` IN('.implode(',', $Todelete).')';
            mcq($sql);
        }
    } else {
        /* no work for me */
    }

}

function update_field_value($fqn, $recordid, $value, $bypass_security = false, $dont_kick_rules = false, $journal_addition = '(autosave)')
{
    /*
        This function updates a single field to a new value. The value must be given in RAW form; exactly in the way it is supposed to be put into the database!
    */
    $messages = '';
//	print "update_field_value($fqn, $recordid, " . string_r($value) . ", $bypass_security=false, $dont_kick_rules";

    if (!ipi($recordid)) {
        $messages .= "F: Empty record id\n";
        log_msg('ERROR: update_field_value called with empty record id');

        return $messages;
    } else {
        list($table, $field) = explode('-', $fqn);

        /* get the record, but only for the field we're looking for ($fqn) */
        $current_record = getrecord($table, $recordid, false, $bypass_security, array($fqn));

        if (($current_record['type_of_access'] == 'ok' && $current_record['record_locked_by'] == 0 && $current_record['*'.$fqn] == 1) || $bypass_security) {

            if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['multiselect']) {
                /*
                    serialize it, this is the way it has to be put into the database (legacy)
                    multiselect fields will always have 1 empty element. we need to remove those using FlattenArray
                */
                if (count(FlattenArray($value)) > 0) {
                    $value = serialize(FlattenArray($value));
                } else {
                    $value = '';
                }


            }

            /* if the database value of this field is not equal to the submitted value ... */

            if ($current_record[$fqn] !== $value) {

                /* ... update it */
                $sql = 'UPDATE `CRM'.$GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'].'` SET `CRM'.$GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'].'`.`'.$GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['name_in_database']."` = '".mres($value)."' WHERE `CRM".$GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'].'`.`'.$GLOBALS['UC']['PersonalDataStructure'][$table]['primary'].'` = '.$recordid;

                mcq($sql);

                /* journal this event in both journals */
                DataJournal($recordid, $current_record[$fqn], $value, $field, ds_tablename_to_datajournal_tabletype($table));

                $journalvalue_to = $value;
                if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'] == 'users') {
                    $journalvalue_to = $journalvalue_to.':'.GetUserName($journalvalue_to);
                }
                $journalvalue_from = $current_record[$fqn];
                if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'] == 'users') {
                    $journalvalue_from = $journalvalue_from.':'.GetUserName($journalvalue_from);
                }
                journal($recordid, 'Field '.$field.' ['.$GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['name'].'] updated from ['.$journalvalue_from.'] to ['.$journalvalue_to.'] '.$journal_addition, ds_tablename_to_journal_tabletype($table));

                if (!$dont_kick_rules) {

                    /* add to-do's for rules to maybe process rules later */
                    AddToDo($fqn, $value, $recordid, $current_record[$fqn]);
                    AddToDo($table.'-changed', '', $recordid, '');

                    if ($table == 'users') {
                        AddToDo('user_change', '', $recordid); // Add entity change to todo list if it wasn't there
                    }

                    /* add todo's for triggers on fields */
                    if (ipi(str_replace('f', '', $field))) {
                        AddToDo($fqn, $value, $recordid);
                    }
                }

                /* check if there are any admin actions to perform like DS drop or mysql trigger re-create */
                check_admin_changes_after_change($table);
            //	$messages .= "D: " . string_r(db_GetArray("SELECT * FROM CRMtodo"));
            } else {
                // $messages .= "Field value not changed, not saving anything";
            }
        }

        return $messages;
    }
}
function get_selectfield_options($fqn, $recordid)
{
    global $user;
    global $pass;
    global $host;
    global $database;
    /* ignore undeclared-and-unused-var-checker.php about this $db global */
    global $db;

    /* array to hold additional classes to return */
    $classes = array();

    $is_assoc = false;
    /*
        This function returns an associative array with options for a select field. Source is datastructure, PHP code, or SQL query
    */
    list($table, $field) = explode('-', $fqn);

    /* the id of the field */
    $field_id = str_replace('f', '', $field);

    if (GetAttribute('extrafield', 'RepopulateByAjax', $field_id) == 'Yes') {
        $classes[] = 'calculation';
    }

    /* array to hold the eventual options */
    $options_list = array();

    /* fetch the complete field row - we need it because not everything is in the datastructure */
    $extrafield = db_GetRow("SELECT * FROM `CRMextrafields` WHERE `CRMextrafields`.`id`='".$field_id."'");

    /* start checking field variants. start with a field populated by PHP code */
    if (substr($extrafield['options'], 0, 18) == '%POPULATE_BY_CODE%') {

        /* an PHP code evaluation should return an array called $result being an array, could be assiociative and plain */
        $result = array();

        /* get the code */
        $code = str_replace('%POPULATE_BY_CODE%', '', $extrafield['options']);

        /* code almost always contains tags, parse it */
        $code = unescape_interleave_characters(ParseTemplateRecordData($code, $table, $recordid));

        /* messages from the undeclared-and-unused-var-checker SHOULD BE IGNORED. variables can be used inside eval() */
        $flextableid = '';
        $flextablerecord = '';
        $eid = '';

        /* set some vars for backwards compatibility */
        if (ipi(str_replace('t', '', $table))) {

            /* messages from the undeclared-and-unused-var-checker SHOULD BE IGNORED. variables can be used inside eval() */
            $flextableid = str_replace('t', '', $table);
            $flextablerecord = $recordid;
        }
        /* evaluate the code */
        @eval($code);

        /* loop through the results and check if it's 1 or 2 columns and put it into options_list */
        foreach ($result as $row) {
            //if ($row[1] != "") {
            if (array_values($row)[1] != '') {
                $is_assoc = true;
                $options_list[$row[0]] = $row[1];
            } else {
                $options_list[] = $row;
            }
        }
    } elseif ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['fieldtype_in_extrafields_table'] == 'SQL Query') {

        /* a field which options are populated by the output of an SQL query. If it returns 1 column than value=text; two columns means [0]=value [1]=text */

        /* the database name for the query is stored in a weird way: a serialized array */
        $o = @unserialize($extrafield['options']);
        $dbase = $o[0];
        $query = $o[1];

        /* replace tags in the query */
        $query = ParseTemplateGeneric($query);
        $query = ParseTemplateRecordData($query, $table, $recordid);
        $query = unescape_interleave_characters($query);
        $query = str_replace("\n", ' ', $query);

        if ($dbase != '') {
            /* check to see if the database name also has a host (syntax: database_name@hostname) */
            if (strpos($dbase, '@') !== false) {

                /* it does, connect to it */
                $tmp = explode('@', $dbase);
                $dbase = $tmp[0];
                $tmphost = $tmp[1];
                $db = db_only_connect($user[$GLOBALS['repository']], $pass[$GLOBALS['repository']], $tmphost);
            }

            /* select the given database */
            if (!db_only_select($dbase)) {
                /* couldn't connect; put error INSIDE field options. not a nice way to do this but it's clear to the user */
                $options_list[0] = 'Could not select database ['.$dbase.']';
            }
        }

        $res = array();
        /* run the query */
        if (trim($query) != '') {
            $res = db_GetArray($query);
        }
        /* loop through the results and check if it's 1 or 2 columns and put it into options_list */
        foreach ($res as $row) {
            if ($row[1] != '') {
                $is_assoc = true;
                $options_list[$row[0]] = $row[0].': '.$row[1];
                //print 'add key ' . $row[0] . ' value ' . $row[1] . '<br>';
            } else {
                $options_list[] = $row[0];
            }
        }

        if ($dbase != '') {
            /* connect back to our interleave database */
            $db = db_base_connect($user[$GLOBALS['repository']], $pass[$GLOBALS['repository']], $host[$GLOBALS['repository']], $database[$GLOBALS['repository']]);

            /* load environment. is this really nescessary?? */
            SwitchToRepos($GLOBALS['repository']);
        }
    } else {

        /* it's a normal options field - options are in the data structure, always non-associative */
        $options_list = $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['options'];
    }

    /* if the field properties say so, sort the array */
    if ($extrafield['sort'] == 'y') {
        natcasesort($options_list);
    }

    $return = array();

    /* clean up the options_list */
    if (!$is_assoc) {
        foreach ($options_list as $key => $value) {
            unset($key);
            $return[$value] = $value;
        }
    } else {
        $return = $options_list;
    }
    //DA($return, false);
    /* return the result */
    return array($return, $classes);
}

function get_elements_to_update_on_field_change($fqn)
{
    /*
        This function checks which other fields on a form need to be refreshed when a field is changed. This could
        be due to the attributes RefreshByAjaxOnChangeOfField and RepopulateByAjax, but also field requirement
        conditions could have a different result because of a change.
    */

    list($table, $field) = explode('-', $fqn);

    /* the array to hold the result */
    $result = array();
    //$result[] = "CALLED WITH $fqn";

    /* the field id of the field in the fields table */
    $field_id = str_replace('f', '', $field);

    /* get all field-ids (.entity) and CSV-lists of field (.value) of all fields. this query could be more specific */
    $sql = "SELECT `CRMattributes`.`entity`, `CRMattributes`.`value` FROM `CRMattributes` WHERE `CRMattributes`.`attribute` ='RefreshByAjaxOnChangeOfField' AND `CRMattributes`.`identifier` ='extrafield' AND `CRMattributes`.`value` != ''";
    $lor = db_GetArray($sql);

    /* loop through all fields having something set */
    foreach ($lor as $row) {

        /* explode the value (it's stored like 1,2,4) */
        $lof = explode(',', $row['value']);

        /* if the field which was updated is in this list, this field needs to be refreshed */
        if (in_array($field_id, $lof) || $lof[0] == 'All') {

            /* add this field to the result set */
            $result[] = get_ds_tablename_for_field($row['entity']).'-f'.$row['entity'];
        }
    }

    /* get all field-ids (.entity) and CSV-lists of field (.value) of all fields. this query could be more specific */
    $sql = "SELECT `CRMattributes`.`entity` FROM `CRMattributes` WHERE `CRMattributes`.`attribute` ='RepopulateByAjax' AND `CRMattributes`.`identifier` ='extrafield' AND `CRMattributes`.`value`='Yes'";
    $lor = db_GetArray($sql);

    /* loop through the matched fields - they need to be refreshed always regardless of what happened */
    foreach ($lor as $row) {
        $dstablename = get_ds_tablename_for_field($row['entity']);

        /* check if the field is in "my" table. if so, add to result set */
        if ($dstablename == $table) {
            $result[] = $table.'-f'.$row['entity'];
        }
    }

    /* slordig! soms komt de "eigen" fqn mee, dat gaat fout! */
    $endresult = array();
    foreach ($result as $field) {
        if ($field != $fqn) {
            $endresult[] = $field;
        }
    }

    return $result;
}
function ef_tabletype_to_ds_tablename($ef_tabletype)
{
    /*
        This functions converts a  table type used in the fields table to a datastructure tablename
    */
    if (ipi($ef_tabletype)) {
        return 't'.$ef_tabletype;
    } elseif ($ef_tabletype == 'user') {
        return 'users';
    } elseif ($ef_tabletype == 'group') {
        return 'groups';
    } elseif ($ef_tabletype == 'files') {
        return 'files';
    }
}
function ds_tablename_to_journal_tabletype($ds_tablename)
{
    /*
        This functions converts a datastructure tablename to the table type used in the journal table
    */
    if (ipi(str_replace('t', '', $ds_tablename))) {
        return str_replace('t', 'flextable', $ds_tablename);
    } elseif ($ds_tablename == 'users') {
        return 'user';
    } elseif ($ds_tablename == 'groups') {
        return 'group';
    } else {
        return $ds_tablename;
    }
}
function ds_tablename_to_datajournal_tabletype($ds_tablename)
{
    /*
        This functions converts a datastructure tablename to the table type used in the datajournal table
    */
    if (ipi(str_replace('t', '', $ds_tablename))) {
        return str_replace('t', '', $ds_tablename);
    } elseif ($ds_tablename == 'users') {
        return 'user';
    } elseif ($ds_tablename == 'groups') {
        return 'group';
    } else {
        return $ds_tablename;
    }
}
function ds_tablename_to_binfiles_table($ds_tablename)
{
    /*
        This functions converts a datastructure tablename to the table used in the binfiles table
    */
    if (ipi(str_replace('t', '', $ds_tablename))) {
        return str_replace('t', 'flextable', $ds_tablename);
    } else {
        return $ds_tablename;
    }
}
function binfiles_table_to_ds_tablename($binfiles_tabletype)
{
    /*
        This functions converts a binfiles tabletype to the tablename used in the datasctructure
    */
    if (ipi(str_replace('flextable', '', $binfiles_tabletype))) {
        return str_replace('flextable', 't', $binfiles_tabletype);
    } else {
        return $binfiles_tabletype;
    }
}
function delete_record($table, $recordid, $skip_rules = false)
{
    /* No authentication. If not sure, don't use this function. */
    $sql = '';
    if (!is_array($GLOBALS['SystemDataStructure'])) {
        $GLOBALS['SystemDataStructure'] = GetSystemDataStructure();
    }

    if ($GLOBALS['SystemDataStructure'][$table]['logical_delete']) {

        /* this means the table uses the field "deleted" y|n to mark a record as deleted. */

        $sql = 'UPDATE `CRM'.$GLOBALS['SystemDataStructure'][$table]['name_in_database'].'` SET `CRM'.$GLOBALS['SystemDataStructure'][$table]['name_in_database']."`.`deleted` = 'y' ";

        $sql .= ' WHERE `CRM'.$GLOBALS['SystemDataStructure'][$table]['name_in_database'].'`.`'.$GLOBALS['SystemDataStructure'][$table]['primary']."`='".$recordid."'";

        /* we don't delete journal information so we even journal this event */
        DataJournal($recordid, '?', 'y', 'deleted', ds_tablename_to_datajournal_tabletype($table));

        log_msg('DELETE UNDO ' . $table . ':' . $record . ': UPDATE `CRM'.$GLOBALS['SystemDataStructure'][$table]['name_in_database'].'` SET `CRM'.$GLOBALS['SystemDataStructure'][$table]['name_in_database']."`.`deleted` = 'n' WHERE `CRM".$GLOBALS['SystemDataStructure'][$table]['name_in_database'].'`.`'.$GLOBALS['SystemDataStructure'][$table]['primary']."`='".$recordid."'");

    } else {

	    /* in this case, records are physically deleted from a table */
        $sql = 'DELETE FROM `CRM'.$GLOBALS['SystemDataStructure'][$table]['name_in_database'].'` WHERE `CRM'.$GLOBALS['SystemDataStructure'][$table]['name_in_database'].'`.`'.$GLOBALS['SystemDataStructure'][$table]['primary']."`='".$recordid."'";

        /* we don't delete journal information so we even journal this event */
        DataJournal($recordid, '?', 'y', 'physically-deleted', ds_tablename_to_datajournal_tabletype($table));
    }

    mcq($sql);

    /* delete files along with this record. files are always deleted physically */
    $sql = "DELETE `CRMbinfiles`, `CRMblobs` FROM `CRMbinfiles`, `CRMblobs` WHERE `CRMbinfiles`.`fileid` = `CRMblobs`.`fileid` AND `CRMbinfiles`.`koppelid` = '".$recordid."' AND `CRMbinfiles`.`type` = '".ds_tablename_to_binfiles_table($table)."'";

    /* put a message in the journal */
    journal($recordid, 'This record was deleted', ds_tablename_to_journal_tabletype($table));

    if (!$skip_rules) {
        /* add to-do's for rules */
        AddToDo($table.'-changed', '', $recordid, '');
        AddToDo($table.'-deleted', '', $recordid, '');
    }

    /* check if there are any admin actions to perform like DS drop or mysql trigger re-create */
    check_admin_changes_after_change($table);
}
function descendingdelete($table, $ids, $return_list_only = false, $skip_rules = false)
{
    global $lang;

    //ids is array because this function is also used for massdelete
    $filter = array();
        $filter[$table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"]] = $ids;
    list($descendingdelete_allowed, $found_ids) = descendingdelete_recursive($table, $filter);
    //the recursive function may return duplicate id's, so we run array_unique. array_merge_recursive() does not remove duplicates
    foreach (array_keys($found_ids) as $table_from_which_to_delete)
    {
    	$found_ids[$table_from_which_to_delete]['all'] = array_unique($found_ids[$table_from_which_to_delete]['all']);
    	$found_ids[$table_from_which_to_delete]['ok'] = array_unique($found_ids[$table_from_which_to_delete]['ok']);
    	$found_ids[$table_from_which_to_delete]['nok'] = array_unique($found_ids[$table_from_which_to_delete]['nok']);
    }


    $return_array = array();

    if (!$return_list_only)
    {
    	if ($descendingdelete_allowed)
    	{
    	    foreach (array_keys($found_ids) as $table_from_which_to_delete)
    	    {
        		foreach ($found_ids[$table_from_which_to_delete]['ok'] as $record_to_delete)
        		{
        //			print "delete " . $table_from_which_to_delete . " " . $record_to_delete . "<br>";
        		    delete_record($table_from_which_to_delete, $record_to_delete, $skip_rules);
        		}
        	}

    	}
    }
    else
    {
    	//these could be records to which this user has no access, or records in tables this user has no access to
    	$invisible_records_prevent_delete = false;
    	foreach (array_keys($found_ids) as $table_from_which_to_delete)
    	{
    	    if ($found_ids[$table_from_which_to_delete]['number_invisible'] > 0)
    	    {
    		$invisible_records_prevent_delete = true;
    		break;
    	    }
    	}
    	if ($invisible_records_prevent_delete)
    	{
            /* re-set the array. we don't want to report anything which could be secret for the user */
            $return_array = array();
    	}
    	else
    	{

    	    foreach (array_keys($found_ids) as $table_from_which_to_delete)
    	    {
        		//we KNOW by now this user may see this table; otherwise number_invisible would be greater than 0
                $return_array[$GLOBALS['UC']['PersonalDataStructure'][$table_from_which_to_delete]['plural']] = array();

                $return_array[$GLOBALS['UC']['PersonalDataStructure'][$table_from_which_to_delete]['plural']]["ok"] = count($found_ids[$table_from_which_to_delete]['ok']);

                $return_array[$GLOBALS['UC']['PersonalDataStructure'][$table_from_which_to_delete]['plural']]["nok"] = count($found_ids[$table_from_which_to_delete]['nok']);

        	}
    	}
    }
    return array($descendingdelete_allowed, $return_array);
}

function descendingdelete_recursive($table, $filter)
{
    //this function is somewhat similar to CheckIfRecordHasChilds()

    //this function could exit immediately if it turns out a user doesn't have descending delete rights. At the moment the function continues to run to provide more feedback.

    $descendingdelete_allowed = true;

    if (!is_array($GLOBALS['SystemDataStructure']))
    {
	$GLOBALS['SystemDataStructure'] = GetSystemDataStructure();
    }

//	print '<h3>' . $table . '</h3>';
//	print_r($filter);

    //we need to use the SystemDataStructure because this user might not have access to this table
    $primary = $table . "-" . $GLOBALS['SystemDataStructure'][$table]["primary"];

    $search = "";
    $searchperfield = array();
    $start = 0;
    $perpage = 0;
    $sortdirection = "asc";
    $filtercolumn = "";
    $filtertype = array(); //defaults to an IS filter. IMPORTANT!
    $listlayout = array();
    $connecting_field = "";
    $use_pre_select_records = false;
    $pre_select_records = array();
    $order_by_pre_select = false;
    $search_parents = true;
    $search_childs = true;
    $search_files = true;
    $type_of_list = "mainlist";
    $bypass_record_access_by_sql = false;
    $tabletype = "normal";
    $childtable = "";
    $childfield = "";
    $childrecordid = 0;
    $already_joined = array();
    $hiddenfilter = array();
    $hiddenfiltertype = array();
    $listid = "";

    //this is a count query; we're not interested in write access
    $do_not_check_write_access = true;

    $sortby = $primary;
    $listlayout[] = $primary;
    //needed for recursive delete of folders; only do this if we're deleting a specific file (which could be a folder) OR if we're requesting a list of files in a folder. In all other cases we request ALL the files connected to a specific record, regardless in which folder they are saved
    if (($table == "files") && ((is_array($filter["files-fileid"])) || (is_array($filter["files-folder"]))))
    {
	    $listlayout[] = "files-type";
	    $listlayout[] = "files-koppelid";
	    $listlayout[] = "files-filetype";
    }
    $RemoveExtraSelectCondition = true;

    $found_ids = array();

    $PDSPARK = $GLOBALS['UC']['PersonalDataStructure'];
    $GLOBALS['UC']['PersonalDataStructure'] = $GLOBALS['SystemDataStructure'];
    //we begin to check if there are any records that are matching based on the system datastructure
    $return_countquery = true;
    $query = make_query($table, $search, $searchperfield, $start, $perpage, $sortby, $sortdirection, $filtercolumn, $filter, $filtertype, $listlayout, $RemoveExtraSelectCondition, $connecting_field, $use_pre_select_records, $pre_select_records, $order_by_pre_select, $search_parents, $search_childs, $search_files, $return_countquery, $type_of_list, $bypass_record_access_by_sql, $tabletype, $childtable, $childfield, $childrecordid, $already_joined, $hiddenfilter, $hiddenfiltertype, $listid, $do_not_check_write_access);
//	pre($query);
    $count_all_records = db_GetValue($query);
    $GLOBALS['UC']['PersonalDataStructure'] = $PDSPARK;


    //if there are no records we do not need to check anything else. It is important to use this order, because a filter on a field that this user may not see is not applied which means this user sees all the records in that table. because of that $found_ids[$table]['all'] could be higher than $count_all_records. TODO; WE NEED TO CHANGE THIS SO A FILER LIKE THIS filter[t2-bla][]=boe produces 0 results instead of ALL results
    if ($count_all_records > 0)
    {
	//check if the user has access to this table; we will check allowed_to_delete further down because it is not necessary to have this privilege if there are no records in this table connected to the parent record
	if (is_array($GLOBALS['UC']['PersonalDataStructure'][$table]))
	{
	    $return_countquery = false;
	    $do_not_check_write_access = false;

	    $query = make_query($table, $search, $searchperfield, $start, $perpage, $sortby, $sortdirection, $filtercolumn, $filter, $filtertype, $listlayout, $RemoveExtraSelectCondition, $connecting_field, $use_pre_select_records, $pre_select_records, $order_by_pre_select, $search_parents, $search_childs, $search_files, $return_countquery, $type_of_list, $bypass_record_access_by_sql, $tabletype, $childtable, $childfield, $childrecordid, $already_joined, $hiddenfilter, $hiddenfiltertype, $listid, $do_not_check_write_access);
//			pre($query);

	    //check if deleting the current record(s) is allowed
	    foreach (db_GetArray($query) as $record)
	    {
		$found_ids[$table]['all'][] = $record[$primary];

		//if you are not allowed to delete records from THIS table, but there aren't any records connected to the parent record, you are still alowed to delete the parent record
		if (($GLOBALS['UC']['PersonalDataStructure'][$table]["allowed_to_delete"]) && ($record["type_of_access"] == "ok") && ($record['record_locked_by'] == 0))
		{
		    $found_ids[$table]['ok'][] = $record[$primary];
		}
		else
		{
		    $found_ids[$table]['nok'][] = $record[$primary];
		    $descendingdelete_allowed = false;
    //				print "DESCENDING DELETE NOT ALLOWED: not allowed to delete record " . $record[$primary] . " from table " . $table . "<br>";
		}

		//we would only need to descend if access == ok, but for display purposes we descend always
		if (($table == "files") && ($record["files-filetype"] == "{{{folder}}}"))
		{
			$child_table = "files";

			$child_filter = array();
			$child_filter["files-koppelid"][] = $record["files-koppelid"];
			$child_filter["files-type"][] = $record["files-type"];
			$child_filter["files-folder"][] = $record[$primary];

			list($child_descendingdelete_allowed, $child_found_ids) = descendingdelete_recursive($child_table, $child_filter);
			if (($descendingdelete_allowed) && (!$child_descendingdelete_allowed))
			{
				$descendingdelete_allowed = false;
			}

			$found_ids = array_merge_recursive($found_ids, $child_found_ids);
		}
	    }
	}


	//check if there are invisible records
	$number_invisible = $count_all_records - count($found_ids[$table]['all']);
	if ($number_invisible > 0)
	{
	    $found_ids[$table]['number_invisible'] = $number_invisible;
	    $descendingdelete_allowed = false;
//		print "DESCENDING DELETE NOT ALLOWED: there are " . $number_invisible . " invisible records in table " . $table . "<br>";
	}


	//deleted users remain visible in Interleave, they are logically-logically deleted. For this reason we do not descend from the users table. This also means we don't have to exclude uselog, journal, datajournal and mailqueue from the child tables
	if ($table != "users")
	{
	    //check if deleting files is allowed
	    if ($GLOBALS['UC']['PersonalDataStructure'][$table]["HasFileAttachments"])
	    {
		$child_table = "files";

		$child_filter = array();
		$child_filter["files-koppelid"] = $found_ids[$table]['all'];
		$child_filter["files-type"][] = $GLOBALS['UC']['PersonalDataStructure'][$table]["name_in_database"];

		list($child_descendingdelete_allowed, $child_found_ids) = descendingdelete_recursive($child_table, $child_filter);
		if (($descendingdelete_allowed) && (!$child_descendingdelete_allowed))
		{
		    $descendingdelete_allowed = false;
		}
		$found_ids = array_merge_recursive($found_ids, $child_found_ids);
	    }

	    //check if deleting child records is allowed; we use the SystemDataStructure to also check tables/fields this user has no access to
	    foreach (array_keys($GLOBALS['SystemDataStructure'][$table]["childs"]) as $child_table)
	    {
		foreach ($GLOBALS['SystemDataStructure'][$table]["childs"][$child_table] as $child_field)
		{
		    if (!$GLOBALS['SystemDataStructure'][$child_table]["fields"][$child_field]["multiselect"])
		    {
			$child_filter = array();
			$child_filter[$child_table . "-" . $child_field] = $found_ids[$table]['all'];

			list($child_descendingdelete_allowed, $child_found_ids) = descendingdelete_recursive($child_table, $child_filter);
			if (($descendingdelete_allowed) && (!$child_descendingdelete_allowed))
			{
				$descendingdelete_allowed = false;
			}
			$found_ids = array_merge_recursive($found_ids, $child_found_ids);
		    }
		}
	    }
	}
    }

    return array($descendingdelete_allowed, $found_ids);
}



function check_admin_changes_after_change($table, $identifier = '')
{
    if (is_administrator()) {
        /* check mysql triggers */

        if (!function_exists("return_mysql_trigger_queries")) {
            /* WATCH OUT: a few line below we also need this include! */
            require 'adminfunctions.php';
        }

        $tables = array('fields', 'tables');
        if (in_array($table, $tables)) {
            if (GetSetting('DevelopmentOrProduction') == 'Development') {
                SetAttribute('system', 'TriggerQueriesForDeploy', return_mysql_trigger_queries(true, true, true), 1, false);
            }
            require_once 'adminfunctions.php';
            $triggers = return_mysql_trigger_queries();
            foreach ($triggers as $sql) {
                mcq($sql);
            }
        }

        /* check datastructure drop */

        $tables = array('fields', 'tables', 'fieldconditions', 'languages', 'settings', 'users', 'profiles', 'extendedprivsgroup', 'extendedprivsuser', 'attributes', 'tableaccessrules', 'templateaccessrules', 'moduleaccessrules', 'calendaraccessrules', 'fieldaccessrules', 'calendaraccessrules');

        if (in_array($table, $tables)) {

            /* drop all DS. we should do this smarter. we KNOW adminfunctions.php is already included (see above) */
            AdminChangesDropCache();
        }
    }
}
function add_default_values_for_fields($table, $recordid, $skip_fields)
{

    /*
        adds default values for fields just after creating a new record
    */

    /* use system data structure, we need ALL fields, not only visiable */
    CheckAndCreateSystemDataStructure();

    /* it's now in the databases */
    $sds = unserialize(db_GetValue("SELECT datastructure FROM CRMdatastructures WHERE userid=0 AND datastructure != ''"));

    $fields_to_update = array();
    $to_datajournal = array();
    $to_journal = '';

    /* loop through all fields in this table */
    foreach ($sds[$table]['fields'] as $dsname => $field) {
        if (!in_array($dsname, $skip_fields)) {
            if ($field['defaultvalue'] != '') {
                $val = get_parsed_template_with_given_template($table, $recordid, array(), '', false, '', $field['defaultvalue'], 'DEFAULTVALUE');

                if ($val != '') {
		    $result = ''; //JH 2017-06-24 result is ALWAYS ok. remove var?
		    $value = '';

                    if ($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$dsname]['fieldtype_in_extrafields_table'] != 'date/time' && $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$dsname]['fieldtype_in_extrafields_table'] != 'date (iso)') {

                        /* validate the field value */
                        //list($result, $message, $value) = ValidateFieldInputDataStructureOnly($table.'-'.$dsname, $val, true, $recordid);

		    /* HF20170301 uitgezet omdat velden die niet in de DS van de gebruiker zitten fout gaan. TODO we moeten hier WEL valideren! */
		    $result = "ok";
		    $value = $val;

                    } else {

                        $result = 'ok';
                        $value = $val;
                    }

                    /* everything is ok */
                    if ($result == 'ok') {
                        $fields_to_update[$dsname] = $value;

                        $to_datajournal[] = array($value, $dsname);
                        $to_journal .= 'Default value of '.$dsname.' ['.$sds[$table]['fields'][$dsname]['name'].'] set to [';

                        if ($sds[$table]['fields'][$dsname]['relation_to'] == 'users') {
                            $to_journal .= GetUserName($value);
                        } else {
                            $to_journal .= $value;
                        }

                        $to_journal .= "]\n";
                    } else {
                        $to_journal .= "Default value didn't validate: [".$value . "]\n";
                    }
                }
            }
        } else {
            //		print "SKIP FIELD " . $dsname . " WAS ALREADY POSTED";
        }
    }

    if (count($fields_to_update) > 0) {

        $sql = 'UPDATE `CRM'.$sds[$table]['name_in_database'].'` SET ';
        $sqlparts = array();

        foreach ($fields_to_update as $field => $value) {
            $sqlparts[] = '`CRM'.$sds[$table]['name_in_database'].'`.`'.$sds[$table]['fields'][$field]['name_in_database']."` = '".mres($value)."' ";
        }

        $sql .= implode(',', $sqlparts);

        $sql .= ' WHERE '.$sds[$table]['primary'].' = '.$recordid;

        mcq($sql);

        if ($to_journal != '') {
            journal($recordid, $to_journal, ds_tablename_to_journal_tabletype($table));
        }
        foreach ($to_datajournal as $dj) {
            DataJournal($recordid, '{{def}}',  $dj[0], $dj[1], ds_tablename_to_datajournal_tabletype($table));
        }

    } else {
	/* journal nevertheless */
        if ($to_journal != '') {
            journal($recordid, $to_journal, ds_tablename_to_journal_tabletype($table));
        }
    }
}
function run_isolated_php_code($code, $table = '', $record = '', $pass = array())
{
    global $lang;

    foreach ($pass as $varname => $varvalue) {
        $$varname = $varvalue;
    }

    if (ipi($record) && is_array($GLOBALS['UC']['PersonalDataStructure'][$table])) {

        /* parse the code */
        $code = unescape_interleave_characters(ParseTemplateRecordData($code, $table, $record));

        /* compatibility */
        /* messages from the undeclared-and-unused-var-checker SHOULD BE IGNORED. variables can be used inside eval() */

        $recordid = $record;
        $eid = $record;
        $flextablerecord = $record;

        if (ipi(str_replace('f', '', $table))) {
            $flextable = str_replace('f', '', $table);
            $flextableid = str_replace('f', '', $table);
        }
    }

    /* execute the code */
    /* messages from the undeclared-and-unused-var-checker SHOULD BE IGNORED. variables can be used inside eval() */
    $result = false;

    @eval($code);

    return $result;
}

function validate_field_input($fqn, $value, $record, $check_for_required_fields = false)
{
    global $lang;

    if ($GLOBALS['DEVELOPMENT']) {

        /* to test only server-side validation */
        if ($_POST['form_submitted'] != 1) {
            $return = array();
            $return['result'] = 'ok';
            $return['message'] = ' niet gechecked dude ';
            $return['corrected_value'] = $value;

            return array($return['result'], $return['message'], $return['corrected_value']);
        }
    }
    /*
        The general field value validator

    */

    /* explode fqn */
    list($table, $field) = explode('-', $fqn);

    /* we ALWAYS trim leading and trailing spaces from values */
    if (!is_array($value)) {
        $value = trim($value);
    }

    /* copy (by ref) the interesting part of the data structure */
    $prop = &$GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field];

    /***************************************************************************************/
    /*
     * 					LEGACY, TODO/FIX/REMOVE
     */

    /* convert dates and date/time to ISO, so we can change the callers later on as soon as ValidateFieldInput is eliminated */
    if ($prop['fieldtype'] == 'date') {

        /* dates are given as dd-mm-yyyy */
        if ($prop['dateformat_in_database'] == 'yyyy-mm-dd' || $prop['dateformat_in_database'] == 'dd-mm-yyyy') {
            $value = NLDate2INTLDate($value);
        }
    }
    /* BELOW THIS LINE ALL DATES ARE TREATED AS ISO */
    /***************************************************************************************/

    /* set initial response */
    $return = array('result' => 'undetermined', 'message' => $fqn.' failed validation type:'.$prop['fieldtype'].' name: '.$prop['name'], 'corrected_value' => $value);

    /* we don't validate buttons in this function */
    /* why not? TODO */
    if ($prop['formelementtype'] == 'button') {
        $return = array('result' => 'ok', 'message' => '', 'corrected_value' => '');
    } else {

        /* check to see if there is a custom validation php set */
        if ($prop['CustomValidationFunctionPHP'] != '') {

            /* run code isolated / in own scope to prevent other vars being changed */
            $result = run_isolated_php_code($prop['CustomValidationFunctionPHP'], $table, $record, array('value' => $value));

            /* this is nasty. the code can result in two types;

                $result as array having the same structure as in this function
                $result as string simply  containing "ok" or "nok
            */

            if (is_array($result)) {
                if ($result[0] == 'ok' || $result[0] == 'nok') {
                    $return['result'] = $result[0];
                    $return['message'] = $result[1];
                    if (isset($result[2])) {
                        $return['corrected_value'] = $result[2];
                    }
                } else {
                    $return['result'] = 'nok';
                    $return['message'] = 'Custom validation function returned invalid response: '.string_r($result);
                }
            } else {
                if ($result == 'ok' || $result == 'nok') {
                    $return['result'] = $result[0];
                } else {
                    $return['result'] = 'nok';
                    $return['message'] = 'Custom validation function returned invalid response: '.$result;
                }
            }
        }

        /* determine if the submitted value is empty */
        $is_empty = true;

        /* $value is already trimmed */
        if (!is_array($value) && $value !== '') {

            /* plain values */
            $is_empty = false;
        } elseif (is_array($value)) {

            /* multiselect values */
            foreach ($value as $val) {

                /* we need to trim $val here */
                if (trim($val) !== '') {
                    $is_empty = false;
                }
            }
        }

        /* now, check requirement. if a field is required, but not entered, nothing else matters */
        if ($return['result'] == 'undetermined' && $check_for_required_fields && $is_empty && $prop['required'] && $prop['formelementtype'] != 'checkbox') {
            $return['result'] = 'nok';
        }

        if ($return['result'] == 'undetermined' && $is_empty && !$prop['required']) {

            /* the value is empty, and it's not required in any way, and CustomValidationFunctionPHP didn't NOK it. we're fine */
            $return['result'] = 'ok';
            $return['message'] = '';

        /* at this point we know it's a normally entered field. earlier checks on CustomValidationFunctionPHP and
           requirement didn't result in a NOK, and it's not empty */
        } elseif ($return['result'] == 'undetermined') {

            /* first, check if the field has options set in the datastructure. if so, use them as final check and ignore anything else */
            if (is_array($prop['options']) && count($prop['options']) > 0) {
                $some_failed = false;
                $failed = array();

                /* this catches all fields who have options set */

                if ($prop['multiselect']) {
                    foreach ($value as $val) {
                        if (!in_array($val, $prop['options']) && $val != '') {
                            $some_failed = true;
                            $failed[] = $val;
                        }
                    }
                } else {
                    if (!in_array($value, $prop['options'])) {
                        $some_failed = true;
                        $failed[] = $value;
                    }
                }
                if ($some_failed) {
                    $return['result'] = 'nok';
                    $return['message'] = 'Given value '.implode(',', $failed).' not found in options';
                } else {
                    $return['result'] = 'ok';
                    $return['message'] = '';
                    //$return["message"] = $field . ": " . $prop['fieldtype_in_extrafields_table'] . ": approved";
                }
            } elseif ($prop['multiselect'] && !$prop['relation_to']) {

                /*
                  this MUST be either "Drop-down (populate by code multiselect)" or "SQL Query (multiselect)".
                  we do not validate these types

                */
                $return['result'] = 'ok';
                $return['message'] = '';
                //$return["message"] = $field . ": " . $prop['fieldtype_in_extrafields_table'] . ": ignored";
            } elseif ($prop['relation_to'] != '') {

                /* this catches all fields referring to another table: ref-to-flex, user, group */
                $check_if_exists_recordids = false;

                if (!$prop['multiselect']) {

                    /* select_refer_field_options always needs an array with selected values */
                    $check_if_exists_recordids = array($value);
                } else {

                    /* multiselect values are given as an array */
                    $check_if_exists_recordids = $value;
                }

                /* check if it's an array in case of multiselect. single select will always be an array, see above */
                if (!is_array($check_if_exists_recordids)) {

                    /* not an array, but it should be. will have something to do with a multiselect not being passed as an array */
                    log_msg('ERROR: Field '.$field.' is a multiselect but the value given to validate_field_input is not an array');
                    $return['result'] = 'nok';
                    $return['message'] = "Given value should be an array, but it isn't";
                } else {

                    /* get possible values. select_refer_field_options will add the currently selected value(s) if they are not in the result set */
                    $possible_values = select_refer_field_options($table, $field, $record, false, $check_if_exists_recordids);

                    /* cycle trough options and check if all selected values are in the result set */
                    $found = true;
                    $failed = array();
                    foreach ($check_if_exists_recordids as $check_if_exists_recordid) {
                        if (ipi($check_if_exists_recordid) && !array_key_exists($check_if_exists_recordid, $possible_values)) {

                            /* given value not found; the user is not allowed to select this value; deny */
                            $found = false;
                            $failed[] = $check_if_exists_recordid;
                        }
                    }

                    /* if all went well... */
                    if ($found) {
                        $return['result'] = 'ok';
                        $return['message'] = '';
                    } else {

                        /* no message, defaults to default message */
                        //$return["message"] = "One of the given values not found in result of select_refer_field_options.";

                        $return['result'] = 'nok';
                        $return['message'] = '';

                        /* log this, it shoudn't happen */
                        log_msg('ERROR: A user selected a value in a reference field, the value is not in the list of options.');
                    }
                }
            } elseif ($prop['fieldtype'] == 'text') {

                /*

                these DS text-types are not validated but treated as text:

                fieldtype_in_extrafields_table == 'Drop-down (populate by code)'
                fieldtype_in_extrafields_table == 'SQL Query'

                 */

                if (ipi($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['MaximumLength']) && strlen($value) > $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['MaximumLength']) {

                    /* the length of the given value exceedes the maximum length of the field */
                    $return['result'] = 'nok';

                    /* no message, defaults to default message */
                    $return['message'] = '';
                } else {

                    /* hyperliks */
                    if ($prop['fieldsubtype'] == 'link') {
                        if (substr($value, 0, 7) != 'http://' && substr($value, 0, 8) != 'https://') {
                            $value = 'http://'.$value;
                        }

                        if (IsValidURL($value)) {
                            $return['result'] = 'ok';
                            $return['message'] = '';
                            $return['corrected_value'] = $value;
                        } else {
                            $return['result'] = 'nok';
                            $return['message'] = '';
                        }

                    /* e-mail addresses */
                    } elseif ($prop['fieldsubtype'] == 'mail') {
                        if (ValidateEmail($value)) {
                            $return['result'] = 'ok';
                            $return['message'] = '';
                            $return['corrected_value'] = $value;
                        } else {
                            $return['result'] = 'nok';
                            $return['message'] = '';
                        }

                    /* telephone numbers */
                    } elseif ($prop['fieldsubtype'] == 'tel') {
                        list($ok, $ret) = IsValidPhonenumber($value);

                        if ($ok) {
                            $return['result'] = 'ok';
                            $return['message'] = '';
                            $return['corrected_value'] = $ret;
                        } else {
                            $return['result'] = 'nok';
                            $return['message'] = '';
                        }

                    /* nothing more to check */
                    } else {
                        $return['result'] = 'ok';
                        $return['message'] = '';
                    }
                }
            } elseif ($prop['fieldtype'] == 'number') {
                if (is_numeric($value)) {

                    /* check for minimum and maximum value */
                    if (
                        /* maximum */
                        (is_numeric($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['MaximumValue']) && $value > $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['MaximumValue'])
                        ||
                        /* minimum */
                        (is_numeric($GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['MinimumValue']) && $value < $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['MinimumValue'])
                       ) {
                        $return['result'] = 'nok';
                        /* no message, defaults to default message */
                        $return['message'] = '';

                        /* we use HARD CODED decimal char "." here. IS THIS CORRECT? */
                        $return['corrected_value'] = number_format($value, $prop['decimals'], '.', '');
                    } else {
                        $return['result'] = 'ok';
                        $return['message'] = '';

                        /* we use HARD CODED decimal char "." here. IS THIS CORRECT? */
                        $return['corrected_value'] = number_format($value, $prop['decimals'], '.', '');
                    }
                } else {

                    /* not even numeric */
                    $return['result'] = 'nok';

                    /* no message, defaults to default message */
                    $return['message'] = '';
                }
            } elseif ($prop['fieldtype'] == 'date') {

                /* check if it's date or date-time. yyyy-mm-dd and dd-mm-yyyy mean date */
                if ($prop['dateformat_in_database'] == 'yyyy-mm-dd' || $prop['dateformat_in_database'] == 'dd-mm-yyyy') {
                    if (is_valid_iso_date($value)) {
                        $return['result'] = 'ok';
                        $return['message'] = '';
                        $return['corrected_value'] = from_iso_datetime($value);
                    } else {
                        $return['result'] = 'nok';
                        $return['message'] = '';
                    }

                /* the only option which can refer to date/time */
                } elseif ($prop['dateformat_in_database'] == 'yyyy-mm-dd hh:mm:ss') {

                    /* a date-and-time field */
                    $test = StrDateTime2IsoDateTime($value);

                    if ($test != '') {
                        $return['result'] = 'ok';
                        $return['message'] = '';
                        //$return["message"] = 'OK waarde: ' . $value . ' result: ' . $test;
                        $return['corrected_value'] = trim($test);
                    } else {
                        $return['result'] = 'nok';
                        /* no message, defaults to default message */
                        $return['message'] = '';
                    }
                } else {
                    $return['result'] = 'nok';
                    $return['message'] = 'Unknown format ('.$prop['dateformat_in_database'].')';
                }
            }
        }
    }

    //$return["message"] .= "run end req:" . $check_for_required_fields;
    if ($return['result'] == 'undetermined') {
        $return['result'] = 'nok';
        $return['message'] .= ' (UNDETERMINED)';
    }
    if ($return['result'] == 'nok') { // && $GLOBALS['SesMem']['debug']) {
        $return['message'] .= '	FIELD '.$field.' TABLE '.$table.' VALUE '.$value.' TYPE '.$prop['fieldtype'];
    }

    return array($return['result'], $return['message'], $return['corrected_value']);
}

function generate_hideandshow_javascript($sourcedstable = '')
{
    /* TODO deze functie hobbelt nu 3x door de lijst met rules heen. dat kan makkelijk naar 1x */

    /* now we will loop through all visual_only extended privileges */
    $sqladd = '';
    $functions_created = array();

    /* add group number if the group id > 0 */
    if (ipi($GLOBALS['GROUPID'])) {
        $sqladd .= ','.$GLOBALS['GROUPID'];
    }

    /* admin rules (groupid=0) are appliccable to everybody */
    $sql = "SELECT * FROM `CRMextendedprivsgroup` WHERE `CRMextendedprivsgroup`.`visual_only` = 'y' AND `CRMextendedprivsgroup`.`only_apply_for_field` > 0 AND `CRMextendedprivsgroup`.`groupid` IN (0".$sqladd.')';
    $list = db_GetArray($sql);
    $rules = array();
    $todo = array();

    /* flatten this array */
    foreach ($list as $row) {
        $fqn = get_ds_tablename_for_field($row['only_apply_for_field']).'-f'.$row['only_apply_for_field'];

        /* declare array if not already so */
        if (!is_array($rules[$fqn])) {
            $rules[$fqn] = array();
        }
        if (!is_array($rules[$fqn][$row['rule']])) {
            $rules[$fqn][$row['rule']] = array();
        }

        $forfqn = get_ds_tablename_for_field($row['field']).'-f'.$row['field'];
        list($d1, $d2) = explode('-', $forfqn);
        if ($sourcedstable == '' || $sourcedstable == $d1) {
            /* check security : we don't want rules of unaccessible fields to be visible */
            if (is_array($GLOBALS['UC']['PersonalDataStructure'][$d1]['fields'][$d2])) {

                if ($GLOBALS['UC']['PersonalDataStructure'][$d1]['fields'][$d2]['relation_to'] != '' && trim($row['value']) == '0') {
                    /* here we correct a match to 0 for a reference field. 0 is correct, but on client side we can only match against '' because the form element won't contain a 0 at any time */
                    $row['value'] = '';
                }

                /* group by field first, type of rule (invisible or readonly) next */

                $rules[$fqn][$row['rule']][] = array('field' => $forfqn, 'operand' => $row['operand'], 'value' => $row['value'], 'rule' => $row['rule']);
                if (!in_array($forfqn, $todo)) {

                        $todo[] = $forfqn;
                }
            }
        }
    }

    /* compose javascript */
    $js = '';
    $output = '';
    $extrajs = '';
    $js .= "

function hideshowallelements() {
	/* rules for invisibility - THIS IS NOT MEANT TO BE SECURITY - these are visual-only rules intended to be used for improving useability */

	var currenttable = $('#JS_concerningtable').val();
	var currentrecord = $('#JS_concerningrecord').val();

	var made_readonly = [];


	if (currenttable != '' && currentrecord != '') {

";

    /* in this loop we create some js which will fetch the value of the fields, but only if we're looking
       at fields in the same table as the form is for. we only know this client-side. */
    $js .= "		/* get the field values, as long as the field is in the same table as the form */\n";
    $done = array();
    foreach ($todo as $field) {
        $done[] = $field;

        list($dstable, $dsfield) = explode('-', $field);

        $js_element_to_check = str_replace('-', '_', $field);

        $js .= "		if (currenttable == '".$dstable."' && typeof ".$js_element_to_check." == 'undefined') {
             var ".$js_element_to_check." = find_field_value('".$field."', currentrecord);
             ";
             if ($GLOBALS['UC']['PersonalDataStructure'][$dstable]['fields'][$dsfield]['relation_to'] != '') {
                 $js .= "
                    if (".$js_element_to_check." == '0') {
                        /* ref to flex field (serverside condition) */
                        ".$js_element_to_check." = '';
                    }

                 ";
             }
             $js .= "
             /*
             alert ('" . $field . " is ' + " . $js_element_to_check . ");
             */
         }\n";

    }
    //$js .= string_r($rules);
    /* rule example

        [t4-f182] => Array
            (
                [invisible] => Array
                    (
                        [0] => Array
                            (
                                [field]  =>  t4-f52
                                [operand]  =>  IS NOT
                                [value]  =>  Wacht op dagboekbericht
                                [rule]  =>  invisible
                            )

                        [1] => Array
                            (
                                [field]  =>  t4-f52
                                [operand]  =>  IS NOT
                                [value]  =>  Dagboekbericht geplaatst
                                [rule]  =>  invisible
                            )

                        [2] => Array
                            (
                                [field]  =>  t4-f52
                                [operand]  =>  IS NOT
                                [value]  =>  Afgerond
                                [rule]  =>  invisible
                            )

                        [3] => Array
                            (
                                [field]  =>  t4-f76
                                [operand]  =>  IS
                                [value]  =>  geen
                                [rule]  =>  invisible
                            )

                    )

            )
    */
    $megaif = array();
    foreach ($rules as $field => $rule) {
        $megaif[$field]['invisible'] = array('is' => array(), 'is not' => array());
        $megaif[$field]['readonly'] = array('is' => array(), 'is not' => array());

        foreach ($rule['invisible'] as $eval) {
            if ($eval['operand'] == 'IS NOT') {
                $megaif[$field]['invisible']['is not'][$eval['field']][] = inline_jsencode($eval['value']);
            } elseif ($eval['operand'] == 'IS') {
                $megaif[$field]['invisible']['is'][$eval['field']][] = inline_jsencode($eval['value']);
            }
        }
        foreach ($rule['readonly'] as $eval) {
            if ($eval['operand'] == 'IS NOT') {
                $megaif[$field]['readonly']['is not'][$eval['field']][] = inline_jsencode($eval['value']);
            } elseif ($eval['operand'] == 'IS') {
                $megaif[$field]['readonly']['is'][$eval['field']][] = inline_jsencode($eval['value']);
            }
        }
    }
//	$js .= string_r($megaif);
    /* megaif sample

    [t21-f305] => Array
           (
               [invisible] => Array
                   (
                       [is]  =>
                       [is not]  =>
                   )

               [readonly] => Array
                   (
                       [is]  =>
                       [is not] => Array
                           (
                               [t21-f395] => Array
                                   (
                                       [0]  =>  Yes
                                   )

                               [t21-f318] => Array
                                   (
                                       [0]  =>  1
                                   )

                               [t21-f302] => Array
                                   (
                                       [0]  =>  RORO
                                   )

                           )

                   )

           )

      */

    foreach ($megaif as $fqn => $data) {
        $k = false;
        $f = false;
        if (count($data['readonly']['is not']) > 0 || count($data['readonly']['is']) > 0) {
            $js .= '
				if (';
            if (count($data['readonly']['is not']) > 0) {
                $js .= '(';
                $k = false;
                foreach ($data['readonly']['is not'] as $me => $values) {
                    if ($k) {
                        $js .= ' || ';
                    }
                    $k = true;
                    $js .= ' (';
                    $f = false;
                    foreach ($values as $value) {
                        if ($f) {
                            $js .= ' && ';
                        }
                        $f = true;
                        $js .= ' !match('.str_replace('-', '_', $me).", '".inline_jsencode($value)."') /* hier1 */";
                    }
                    $js .= ')';
                }
                $js .= ')';
            }

            if (count($data['readonly']['is']) > 0) {
                if ($k) {
                    $js .= '|| ';
                }

                $k = false;
                foreach ($data['readonly']['is'] as $me => $values) {
                    if ($k) {
                        $js .= ' || ';
                    }
                    $k = true;

                    $f = false;
                    foreach ($values as $value) {
                        if ($f) {
                            $js .= ' || ';
                        }
                        $f = true;
                        $js .= ' match('.str_replace('-', '_', $me).",'".inline_jsencode($value)."')  /* hier2 */";
                    }
                }
            }

            $js .= ')';
            /*
                we hide/show the *id* but also the *class*. the latter can be used
                for anything thats needs to be hidden also
            */
            $js .= " { /* hier1 */
                    if ($('#ffc-" .$fqn."').length) {
                        $('#ffc-" .$fqn." :input').prop(\"disabled\", true);
                        $('#ffc-" .$fqn."').hide(250);
                    }
                    if ($('.ffc-" .$fqn."')[0]){
                        $('.ffc-" .$fqn." :input').prop(\"disabled\", true);
	  	       	        $('.ffc-" .$fqn."').hide(250);
                    }
                    made_readonly.push('" .$fqn."');

	  			} else {
                    if ($('#ffc-" .$fqn."').length) {
					    $('#ffc-" .$fqn." :input').prop(\"disabled\", false);
                        $('#ffc-" .$fqn."').show(250);
                    }
	  	        	if ($('.ffc-" .$fqn."')[0]){
                        $('.ffc-" .$fqn."').show(250);
                    }

				}";
        }
    }
    foreach ($megaif as $fqn => $data) {
        $k = false;
        $f = false;
        if (count($data['invisible']['is not']) > 0 || count($data['invisible']['is']) > 0) {
            $js .= ' if (';

            if (count($data['invisible']['is not']) > 0) {
                $js .= '(';
                $k = false;
                foreach ($data['invisible']['is not'] as $me => $values) {
                    if ($k) {
                        $js .= ' || ';
                    }
                    $k = true;
                    $js .= ' (';
                    $f = false;
                    foreach ($values as $value) {
                        if ($f) {
                            $js .= ' && ';
                        }
                        $f = true;
                        $js .= ' !match('.str_replace('-', '_', $me).", '".inline_jsencode($value)."')  /* hier3 */";
                    }
                    $js .= ')';
                }
                $js .= ')';
            }

            if (count($data['invisible']['is']) > 0) {
                if ($k) {
                    $js .= ' || ';
                }
                $k = false;

                foreach ($data['invisible']['is'] as $me => $values) {
                    if ($k) {
                        $js .= ' || ';
                    }
                    $k = true;

                    $f = false;
                    foreach ($values as $value) {
                        if ($f) {
                            $js .= ' || ';
                        }
                        $f = true;
                        $js .= ' match('.str_replace('-', '_', $me).",'".inline_jsencode($value)."')  /* hier4 */";
                    }
                }
            }
            $js .= ')';

            /*
                we hide/show the *id* but also the *class*. the latter can be used
                for anything thats needs to be hidden also
            */

            $js .= " {
                if ($('#ffc-" .$fqn."').length) {
				    $('#ffc-" .$fqn." :input').prop(\"disabled\", true);
	        	    $('#ffc-" .$fqn."').hide();
                }
                if ($('.ffc-" .$fqn."')[0]){
                    $('.ffc-" .$fqn." :input').prop(\"disabled\", true);
	        	    $('.ffc-" .$fqn."').hide();
                }
			} else {
				if ($.inArray('" .$fqn."', made_readonly) == -1) {
					$('#ffc-" .$fqn." :input').prop(\"disabled\", false);
				}
                if ($('#ffc-" .$fqn."').length) {
				    $('#ffc-" .$fqn."').show(250);
                }
                if ($('.ffc-" .$fqn."')[0]){
                    $('.ffc-" .$fqn."').show(250);
                }
			}";
        }
    }

    $js .= '


    hideshowautonavtabs();
    activate_iltabs();


	}
}
$(document).ready(function() { hideshowallelements(); });
';
    //$js = preg_replace("/ {2,}/", " ", $js);
    return $js;
}
?>
