<?php
/* ********************************************************************
 * Interleave
 * Copyright (c) 2001-2016 Atomos Applications The Netherlands - www.atomos.nl - info@atomos.nl - www.interleave.nl - info@interleave.nl
 * Licensed under the GNU GPL. For full terms see the file COPYING.
 *
 * listfunctions.php
 * contains functions for:
 * - creating data structures
 * - sql security
 * - making list queries
 * - creating a list in html, xml, tsv or Excel format
 *
 * Check http://www.interleave.nl/ for more information
 *
 * BUILD / RELEASE :: 7.2.0.20180308
 *
 **********************************************************************
 */

function get_data_structure()
{
	//to prevent a fixed set of records to be used inside the DS queries; it would be cleaner to use an additional function argument; but we resort to a GLOBAL for now
	$GLOBALS["DO_NOT_USE_php_subquery_materialization"] = true;

	$t1 = microtime(true);

	list($GLOBALS['UC']['PersonalDataStructure'], $reasons) = get_data_structure_worker();

	//we halen alle beschikbare velden per table op; dus ook alle parent velden; dit is vrij duur
	foreach (array_keys($GLOBALS['UC']['PersonalDataStructure']) as $table)
	{
		$query_parts = make_query_parts($table);
		$GLOBALS['UC']['PersonalDataStructure'][$table]["allfields"] = $query_parts["select_alias"];
	}

	//pas HIER kan parse_old_template_to_new_template() gedaan worden want die heeft de volledige DS nodig om zijn werk te doen
	foreach (array_keys($GLOBALS['UC']['PersonalDataStructure']) as $table)
	{
		//dit is zoals de tabel zichzelf presenteert
		$GLOBALS['UC']['PersonalDataStructure'][$table]["layout"] = parse_old_template_to_new_template($GLOBALS['UC']['PersonalDataStructure'][$table]["layout"], $table);

		foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]) as $field)
		{
			if ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["layout"] != "")
			{
				if ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["multiselect"])
				{
					//we nemen field niet mee; we doen alsof we de reference niet kennen; dit is slordig maar een workaround
//					$GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["layout"] = parse_old_template_to_new_template($GLOBALS['UC']['PersonalDataStructure'][$GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["relation_to"]]["layout"], $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["relation_to"]);

					$GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["layout"] = parse_old_template_to_new_template($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["layout"], $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["relation_to"]);
				}
				else
				{
					$GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["layout"] = parse_old_template_to_new_template($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["layout"], $table . "-" . $field);
				}
			}
			if ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["ListOptionsForParentTable"] != "")
			{
				$GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["ListOptionsForParentTable"] = parse_old_template_to_new_template($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["ListOptionsForParentTable"], $table);
			}
		}
	}

	//hier bouwen we de tags array op; dat is sneller om 1x te doen dan op vele plekken in de code
	foreach (array_keys($GLOBALS['UC']['PersonalDataStructure']) as $table)
	{
		$GLOBALS['UC']['PersonalDataStructure'][$table]["layout_tags"] = get_tags($GLOBALS['UC']['PersonalDataStructure'][$table]["layout"]);
		foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]) as $field)
		{
			$GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["layout_tags"] = get_tags($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["layout"]);
		}
	}

	//write access moet ook bepaald worden, want je mag alleen bestanden wissen als je write access op het bovenliggen record hebt
	//wordt bij het toevoegen/deleten van een bestand goed gecheckt of het bovenliggende record beschrijfbaar is?
	//security toevoegen voor openen/verwijderen van files; MUST BE OWNER of file?
/*
* write access bepalen om te weten of iemand een file mag wissen
* highlighten van searches vanaf de parent (highlight-in-form)
* oude code deleten uit assist vanwege security issue
*/
	//we kunnen HIER pas de security query opbouwen voor de toegang tot bestanden, omdat we daarvoor een valide datastructure nodig hebben
	if (!is_administrator())
	{
		$reasons[] = "not admin, so now putting security on files, journal and datajournal table";

		$securityparts_files = array();
		$securityparts_journal = array();
		$securityparts_datajournal = array();

		//standaard nergens toegang toe; dit moet omdat onderstaande lus alleen tabellen doorloopt waartoe iemand toegang heeft. Stel dat iemand tot geen enkele tabel toegang zou hebben, dan zou hij hierdoor wel ALLE bestanden en journal records zien
//		$securityparts_files[] = "(FALSE)";
		//iedereen heeft toegang tot general purpose files en public files (koppelid = 0), alleen admins hebben toegang tot "system" files, maar admins mogen sowieso alle bestanden zien
		$securityparts_files[] = "(`files`.`type` IN ('generalpurpose', 'public'))";
		$securityparts_journal[] = "(FALSE)";
		$securityparts_datajournal[] = "(FALSE)";

		foreach (array_keys($GLOBALS['UC']['PersonalDataStructure']) as $table)
		{
			//bestanden kunnen alleen aan deze tabellen hangen. Als er bestanden aan andere tabellen hangen dan zijn die niet zichtbaar voor normale users. voor admin wel
			if ($GLOBALS['UC']['PersonalDataStructure'][$table]["datatable"])
			{
				$search = "";
				$searchperfield = array();
				$start = 0;
				$perpage = 0; //belangrijk
				$sortby = "";
				$sortdirection = "asc";
				$filtercolumn = "";
				$filter = array();
				$filtertype = array();
				$listlayout = array();
				$listlayout[] = $table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"];

				$RemoveExtraSelectCondition = true; //belangrijk
				$connecting_field = "";
				$use_pre_select_records = false;
				$pre_select_records = array();
				$order_by_pre_select = false;
				$search_parents = true;
				$search_childs = true;
				$search_files = true;
				$return_countquery = false;
				$type_of_list = "mainlist";
				$bypass_record_access_by_sql = false;
				$tabletype = "normal";
				$childtable = "";
				$childfield = "";
				$childrecordid = 0;
				$already_joined = array();
				$hiddenfilter = array();
				$hiddenfiltertype = array();
				$listid = "";
				$do_not_check_write_access = true;

				$sql_files = "(`files`.`type`='" . $GLOBALS['UC']['PersonalDataStructure'][$table]["name_in_database"] . "'";
				$sql_journal = "(`journal`.`type`='" . $GLOBALS['UC']['PersonalDataStructure'][$table]["name_in_database"] . "'";

				$table_for_datajournal = $table;
				if (preg_match("/^t(\d+)$/", $table, $match))
				{
					$table_for_datajournal = $match[1];
				}
				$sql_datajournal = "(`datajournal`.`table`='" . $table_for_datajournal . "'";
				//dit is een beetje lomp; als de output van get_recursive_access_querypart() leeg is dan ligt er geen security op deze tabel en hoeven we dus niet te limiteren op fileid
				//is die NIET leeg dan moeten we nog steeds door make_query() heen voor de volledige query

				if (get_recursive_access_querypart($table, "", "read") != "")
				{
					$subquery = make_query($table, $search, $searchperfield, $start, $perpage, $sortby, $sortdirection, $filtercolumn, $filter, $filtertype, $listlayout, $RemoveExtraSelectCondition, $connecting_field, $use_pre_select_records, $pre_select_records, $order_by_pre_select, $search_parents, $search_childs, $search_files, $return_countquery, $type_of_list, $bypass_record_access_by_sql, $tabletype, $childtable, $childfield, $childrecordid, $already_joined, $hiddenfilter, $hiddenfiltertype, $listid, $do_not_check_write_access);

					$sql_files .= " AND `files`.`koppelid` IN (" . $subquery . ")";
					$sql_journal .= " AND `journal`.`eid` IN (" . $subquery . ")";
					$sql_datajournal .= " AND `datajournal`.`record` IN (" . $subquery . ")";
				}
				$sql_files .= ")";
				$sql_journal .= ")";
				$sql_datajournal .= ")";

				//als ie geen bestanden "mag" hebben dan querien we er ook niet op
				if ($GLOBALS['UC']['PersonalDataStructure'][$table]["HasFileAttachments"])
				{
					$securityparts_files[] = $sql_files;
				}
				$securityparts_journal[] = $sql_journal;
				$securityparts_datajournal[] = $sql_datajournal;
			}
		}

		$table = "files";
		//een user kan ook geen toegang tot de files tabel hebben wanneer de files tabel een parent heeft die die user niet mag zien
		if (is_array($GLOBALS['UC']['PersonalDataStructure'][$table]))
		{
			$type_of_access = "read";
			//er is altijd minimaal 1 securityparts_files vanwege de generalpurpose en public files
			$GLOBALS['UC']['PersonalDataStructure'][$table]["record_access_restrictions"][$type_of_access]["fileid"]["query"] = "SELECT `files`.`fileid` FROM `CRMbinfiles` `files` WHERE " . implode(" OR ", $securityparts_files);
		}

		$table = "journal";
		if (is_array($GLOBALS['UC']['PersonalDataStructure'][$table]))
		{
			$type_of_access = "read";
			//er is altijd minimaal 1 securityparts_files vanwege de (FALSE)
			$GLOBALS['UC']['PersonalDataStructure'][$table]["record_access_restrictions"][$type_of_access]["id"]["query"] = "SELECT `journal`.`id` FROM `CRMjournal` `journal` WHERE " . implode(" OR ", $securityparts_journal);
			//in get_data_structure_worker() is write access al ontnomen voor IEDEREEN (inc admin), niemand kan journal records wissen of aanpassen
		}

		$table = "datajournal";
		if (is_array($GLOBALS['UC']['PersonalDataStructure'][$table]))
		{
			$type_of_access = "read";
			//er is altijd minimaal 1 securityparts_files vanwege de (FALSE)
			$GLOBALS['UC']['PersonalDataStructure'][$table]["record_access_restrictions"][$type_of_access]["id"]["query"] = "SELECT `datajournal`.`id` FROM `CRMdatajournal` `datajournal` WHERE " . implode(" OR ", $securityparts_datajournal);
			//in get_data_structure_worker() is write access al ontnomen voor IEDEREEN (inc admin), niemand kan journal records wissen of aanpassen
		}
	}

	$t2 = microtime(true);
	$t = $t2 - $t1;

	$reasons[] = "runtime get_data_structure(): " . $t . " seconds";

	SetAttribute("user", "DataStructureAccessReasons", $reasons, $GLOBALS['USERID']);

	$GLOBALS["DO_NOT_USE_php_subquery_materialization"] = false;

	return $GLOBALS['UC']['PersonalDataStructure'];
}
function get_data_structure_worker()
{
	global $lang;

	$reasons = array();

	$INTERACTIVEFIELDS = @unserialize(db_GetValue("SELECT `INTERACTIVEFIELDS` FROM `CRMuserprofiles` WHERE `id`='" . $GLOBALS['GROUPID'] . "'"));
	if (!is_array($INTERACTIVEFIELDS))
	{
		$INTERACTIVEFIELDS = array();
	}

	$tables_allowed_to_read = array();
	$tables_allowed_to_write = array();
	$tables_allowed_to_add = array();
	$tables_allowed_to_delete = array();

	//deze array defineert de records die iemand mag zien of editten.
	//als de array leeg is dan mag iemand alle records zien/editten
	//als de array gevuld is dan mag die persoon alleen die records zien/editten die voldoen aan de "filter" in deze tabel
	$record_access_restrictions = array();

	//vaak nodig dus lokaal opslaan helpt
	$is_admin = is_administrator();

	$MaySelectColumns = false;
	//CheckFunctionAccess doet geen admin check
	if (($is_admin) || ((CheckFunctionAccess("MaySelectColumns") == "ok")))
	{
		//op flextable niveau KAN dit nog overrulet, maar je moet in ieder geval MaySelectColumns aan hebben staan
		$MaySelectColumns = true;
	}

	$download_allowed = false;
	//CheckFunctionAccess doet geen admin check
	if (($is_admin) || (CheckFunctionAccess("HideListExportIcons") == "nok"))
	{
		//dit kan op ft niveau nog overrulet worden
		$download_allowed = true;
	}

	foreach (db_GetArray("SELECT `recordid` FROM `CRMflextabledefs` ORDER BY `ordering`, `recordid`") as $flextable)
	{
		$table = "t" . $flextable['recordid'];

		$allowed_to_read = true;
		$allowed_to_write = true;
		$allowed_to_add = true;
		$allowed_to_delete = true;

		if (!$is_admin)
		{
			if (db_GetValue("SELECT COUNT(*) FROM `CRMtableaccessrules` WHERE `table`=" . $flextable['recordid']) > 0)
			{
				$allowed_to_read = false;
				$allowed_to_write = false;
				$allowed_to_add = false;
				$allowed_to_delete = false;
				//warning triggeren op meer dan 1 record
				$row = db_GetRow("SELECT * FROM `CRMtableaccessrules` WHERE `table`=" . $flextable['recordid'] . " AND `group`='" . $GLOBALS['GROUPID'] . "' ORDER BY `id`");
				if ($row['read'] == 1)
				{
					$allowed_to_read = true;
				}
				if ($row['write'] == 1)
				{
					$allowed_to_write = true;
				}
				if ($row['add'] == 1)
				{
					$allowed_to_add = true;
				}
				if ($row['delete'] == 1)
				{
					$allowed_to_delete = true;
				}
			}
		}

		$reasons[] = "table access " . $table . " read/write/add/delete is " . ($allowed_to_read ? "1" : "0") . "/" . ($allowed_to_write ? "1" : "0") . "/" . ($allowed_to_add ? "1" : "0") . "/" . ($allowed_to_delete ? "1" : "0");

		//deze hebben we ook voor admin nodig; vandaar dat we deze lus doorlopen
		$tables_allowed_to_read[$table] = $allowed_to_read;
		$tables_allowed_to_write[$table] = $allowed_to_write;
		$tables_allowed_to_add[$table] = $allowed_to_add;
		$tables_allowed_to_delete[$table] = $allowed_to_delete;

		if (!$is_admin)
		{
			if ($allowed_to_read)
			{
				$tmp = GetAttribute("group", "SecuritySubselectForRecordAccessReadForFlextable" . $flextable["recordid"], $GLOBALS['GROUPID']);
				if ($tmp != "")
				{
					$reasons[] = "SecuritySubselectForRecordAccessReadForFlextable" . $flextable["recordid"] . ": " . $tmp;
					$record_access_restrictions["read"][$table]["recordid"]["query"] = $tmp;
				}

				$tmp = GetAttribute("group", "SecuritySubselectForRecordAccessWriteForFlextable" . $flextable["recordid"], $GLOBALS['GROUPID']);
				if ($tmp != "")
				{
					$reasons[] = "SecuritySubselectForRecordAccessWriteForFlextable" . $flextable["recordid"] . ": " . $tmp;
					$record_access_restrictions["write"][$table]["recordid"]["query"] = $tmp;
				}

				if (!$allowed_to_write)
				{
					$record_access_restrictions["write"][$table]["recordid"]["IS"][] = 0;
				}
			}
		}
	}

	if (!$is_admin)
	{
		if (is_positive_integer($GLOBALS['GROUPID']))
		{
			//TODO 2015-08-31; deze hele lus waarbij er data in de $record_access_restrictions array wordt toegevoegd, moet naar onder, wanneer de hele DS al is opgebouwd, dan kan die code bovendien worden samengevoegd met de plek waar `only_apply_for_field`>0
			//zelfde code staat 2x in lf, moet nog samenvoegen
			foreach (db_GetArray("SELECT `field`, `operand`, `value`, `rule` FROM `CRMextendedprivsgroup` WHERE `groupid`=" . $GLOBALS['GROUPID'] . " AND `only_apply_for_field`=0 AND `visual_only`='n'") as $row)
			{
				$table = get_ds_tablename_for_field($row['field']);
				$field = "f" . $row['field'];
				$value = $row['value'];
				$operand = $row['operand'];

				$type_of_access = "read"; //invisible
				if ($row['rule'] == "readonly")
				{
					$type_of_access = "write";
				}

				//LET OP: precies tegenovergesteld opgeslagen in record_access_restrictions want die houdt bij wat je WEL mag zien
				if ($operand == "IS NOT")
				{
					$operand = "IS";
				}
				else
				{
					$operand = "NOT";
				}

				$value = unescape_interleave_characters(ParseTemplateUserAndGroupFields($value));

				//2015-08-31
				if ($value == "")
				{
					//dit is erg vies; we hebben nog geen DS dus we moeten hier zelf de fieldtype beredeneren; zie opmerking hierboven; deze hele code moet hier weg en NAAR ONDER!
					$fieldtype = db_GetValue("SELECT `fieldtype` FROM `CRMextrafields` WHERE `deleted`='n' AND `id`=" . $row['field']);
					$number_format = db_GetValue("SELECT `number_format` FROM `CRMextrafields` WHERE `deleted`='n' AND `id`=" . $row['field']);

					if (($fieldtype == "date") || ($fieldtype == "date (iso)") || ($fieldtype == "date/time"))
					{
						//dit matcht ook op een date/time veld in mysql waar eigenlijk 0000-00-00 00:00:00 in staat
						$value = "0000-00-00";
					}
					else if ($fieldtype == "numeric")
					{
						if ($number_format != "don't format")
						{
							$value = 0;
						}
					}
				}

				$logline = "Record access rule " . $table . "-" . $field . " ";
				if ($operand == "IS")
				{
					$logline .= "must be";
				}
				else
				{
					$logline .= "must NOT be";
				}
				$logline .= " \"" . $value . "\" otherwise NO " . $type_of_access . " access";
				$reasons[] = $logline;
				$record_access_restrictions[$type_of_access][$table][$field][$operand][] = $value;
			}
		}

		//UseThisFieldForAccessRestriction START
		//access to this field is NOT checked; so this user does not need to be able to view this field in the users table
		foreach (db_GetArray("SELECT `id`, `fieldtype`, `options` FROM `CRMextrafields` WHERE `deleted`='n' AND `tabletype`='user' AND `fieldtype` IN ('Reference to FlexTable', 'Reference to FlexTable (multiselect)') ORDER BY `ordering`, `id`") as $extrafield)
		{
			$relation_to = "";
			if (is_positive_integer($extrafield["options"]))
			{
				$relation_to = "t" . $extrafield["options"];
			}

			if ($relation_to != "")
			{
				$UseThisFieldForAccessRestriction = GetAttribute("extrafield", "UseThisFieldForAccessRestriction", $extrafield["id"]);
				$type_of_access = "";
				if ($UseThisFieldForAccessRestriction == "Read access")
				{
					$type_of_access = "read";
				}
				else if ($UseThisFieldForAccessRestriction == "Write access")
				{
					$type_of_access = "write";
					//er is geen enkele write toegang tot deze tabel; daarom willen we deze security niet toevoegen anders krijg je WHERE CRMflextable2 IN (0,15,29) -> en dan heb je dus WEL write toegang
					if ((is_array($record_access_restrictions[$type_of_access][$relation_to]["recordid"]["IS"])) && ($record_access_restrictions[$type_of_access][$relation_to]["recordid"]["IS"][0] === 0))
					{
						$type_of_access = "";
					}
				}

				if ($type_of_access != "")
				{
					$value = db_GetValue("SELECT `EFID" . $extrafield["id"] . "` FROM `CRMloginusers` WHERE `id`=" . $GLOBALS['USERID']);
					$allowed_values = array();
					if ($extrafield["fieldtype"] == "Reference to FlexTable")
					{
						$allowed_values[] = $value;
					}
					else //Reference to FlexTable (multiselect)
					{
						$allowed_values = unserialize($value);
					}

					foreach ($allowed_values as $recordid)
					{
						if (is_positive_integer($recordid))
						{
							$reasons[] = $type_of_access . " limit for " . $relation_to . ": " . $recordid . " (UseThisFieldForAccessRestriction)";
							$record_access_restrictions[$type_of_access][$relation_to]["recordid"]["IS"][] = $recordid;
						}
					}
				}
			}
		}
		//UseThisFieldForAccessRestriction END
	}

	$DS = array();

	//because templates refer to users, users to groups and groups to templates we needed to isolate the template layout. We need to use it before the template table is declared in $DS
	//2017-08-21 templates does not refer to users any longer; can we remove this now?
	$layout_of_template_table = "@templates-templateid@: @templates-templatename@";
	//de "t" is hier niet altijd goed want kan ook aan user, group en files refereren; maar dat is dan maar zo
	$layout_of_field_table = "t@fields-tabletype@-f@fields-id@: @fields-name@";

	$allow_useradmin = false;
	if (($is_admin) || (CheckFunctionAccess("UserAdmin") == "ok"))
	{
		$allow_useradmin = true;
	}

	$table = "groups";
	$DS[$table]["primary"] = "id";
	$DS[$table]["name_in_database"] = "userprofiles";
	$DS[$table]["openlink"] = "useradmin.php?EditGroup=";
	$DS[$table]["MainTable"] = true; //voor admin interface
	$DS[$table]["addlink"] = "";
	$DS[$table]["addlinktext"] = "Add";
	$DS[$table]["allowed_to_write"] = $is_admin; //alleen admins mogen groups aanpassen; als je UserAdmin hebt mag je ALLEEN users aanpassen
	$DS[$table]["allowed_to_add"] = false; //want dit gaat via een custom methode
	$DS[$table]["allowed_to_delete"] = $is_admin;
	$DS[$table]["allow_direct_view"] = $is_admin;
	$DS[$table]["record_access_restrictions"]["read"] = $record_access_restrictions["read"][$table];
	$DS[$table]["record_access_restrictions"]["write"] = $record_access_restrictions["write"][$table];
	$DS[$table]["default_sortby"] = $table . "-name";
	$DS[$table]["default_sortdirection"] = "asc";
	$DS[$table]["IncludeInSystemWideSearches"] = false;
	$DS[$table]["singular"] = $lang["group"];
	$DS[$table]["plural"] = $lang["groups"];
	//op tabel niveau is het de layout zoals de tabel zichzelf wil presenteren
	$DS[$table]["layout"] = "@" . $table . "-name@";
	$DS[$table]["perpage"] = 50;
	$DS[$table]["download_allowed"] = $download_allowed;
	$DS[$table]["download_limit"] = 0;
	$DS[$table]["search_parents"] = false; //hij heeft ook geen parents
	$DS[$table]["search_childs"] = false;
	$DS[$table]["HasFileAttachments"] = false;
	$DS[$table]["search_files"] = false;
	$DS[$table]["ShowInlineDeleteLink"] = true;
	$DS[$table]["open_records_in_popup_mainlist"] = 0;
	$DS[$table]["open_records_in_popup_sublist"] = 0;
	$DS[$table]["MaySelectColumns"] = $MaySelectColumns;

	$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");
	$DS[$table]["fields"]["name"] = array("name" => $DS[$table]["singular"], "fieldtype" => "text");

	if ($is_admin)
	{
		$DS[$table]["fields"]["HIDEFROMASSIGNEEANDOWNERLISTS"] = array("name" => "Hide from lists", "fieldtype" => "text", "filter" => true, "options" => array("n", "y"), "hide_field_when_in_child_table" => true);

//		$DS[$table]["fields"]["active"] = array("name" => $lang['active'], "fieldtype" => "text", "filter" => true, "options" => array("no", "yes"), "hide_field_when_in_child_table" => true);

		$DS[$table]["fields"]["CLLEVEL"] = array("name" => "Clearance level", "fieldtype" => "text", "serialized" => true, "hide_field_when_in_child_table" => true);

		$relation_to = "templates";
		$DS[$table]["fields"]["MENUTOUSE"] = array("name" => "Menu", "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $layout_of_template_table, "filter" => true, "hide_field_when_in_child_table" => true);

		if (GetSetting('WebAppEnabled') == 'Yes')
		{
			$relation_to = "templates";
			$DS[$table]["fields"]["webappmenu"] = array("name" => "Webapp menu", "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $layout_of_template_table, "filter" => true, "ExtraSelectConditionForSelectingNewRecords" => "`" . $relation_to . "`.`version_belonging_to`=0", "hide_field_when_in_child_table" => true);
		}

		$relation_to = "templates";
		$DS[$table]["fields"]["DASHBOARDFILEID"] = array("name" => "Dashboard", "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $layout_of_template_table, "filter" => true, "hide_field_when_in_child_table" => true);

		$DS[$table]["fields"]["INTERACTIVEFIELDS"] = array("name" => "Interactive fields", "fieldtype" => "text", "serialized" => true, "hide_field_when_in_child_table" => true);

		$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true, "hide_field_when_in_child_table" => true);
	}

	$table = "users";
	$DS[$table]["primary"] = "id";
	$DS[$table]["name_in_database"] = "loginusers";
	$DS[$table]["openlink"] = "useradmin.php?EditUser=";
	$DS[$table]["addlink"] = "";
	$DS[$table]["addlinktext"] = "Add";
	$DS[$table]["allowed_to_write"] = $allow_useradmin;
	$DS[$table]["allowed_to_add"] = false; //want dit gaat via een custom methode
	$DS[$table]["allowed_to_delete"] = $allow_useradmin;
	$DS[$table]["allow_direct_view"] = $allow_useradmin; //BELANGRIJK! als je geen useradmin hebt zal je deze tabel nooit integraal mogen zien
	$DS[$table]["record_access_restrictions"]["read"] = $record_access_restrictions["read"][$table];
	$DS[$table]["record_access_restrictions"]["write"] = $record_access_restrictions["write"][$table];
	$DS[$table]["default_sortby"] = $table . "-FULLNAME";
	$DS[$table]["default_sortdirection"] = "asc";
	$DS[$table]["IncludeInSystemWideSearches"] = false;
	$DS[$table]["singular"] = $lang["user"];
	$DS[$table]["plural"] = $lang["users"];
	$DS[$table]["layout"] = "@" . $table . "-FULLNAME@";
	$DS[$table]["perpage"] = 50;
	$DS[$table]["download_allowed"] = $download_allowed;
	$DS[$table]["download_limit"] = 0;

	//we gaan niet joinen naar userprofiles; dat geeft een hele hoop joins
	//userprofiles is de enige tabel waarvoor deze var uit staat; dit voorkomt een hoop joins naar deze tabel in een search
	$DS[$table]["search_parents"] = false;
	$DS[$table]["search_childs"] = false;
	$DS[$table]["HasFileAttachments"] = false;
	$DS[$table]["search_files"] = false;

	$DS[$table]["open_records_in_popup_mainlist"] = 0;
	$DS[$table]["open_records_in_popup_sublist"] = 0;
	$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
	$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");
	$DS[$table]["fields"]["FULLNAME"] = array("name" => $lang["name"], "fieldtype" => "text");

	//als HideUserNames aan staat mag je alleen je eigen username zien; deze logica zit zowel in de het opbouwen van de DS als ook in GetUserName()
	if (CheckFunctionAccess("HideUserNames") == "ok")
	{
		//alle usernames zijn leeg behalve de eigen; dit is anders dan hoe GetUserName() werkte want die retourneerde "[n/a]"
		$DS[$table]["fields"]["FULLNAME"]["field_access_restrictions"]["read"]["id"]["IS"][] = $GLOBALS['USERID'];
	}

	$relation_to = "groups";
	$DS[$table]["fields"]["PROFILE"] = array("name" => $DS[$relation_to]["singular"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "filter" => true, "ExtraSelectConditionForSelectingNewRecords" => "`" . $relation_to . "`.`active`='yes'", "InheritWriteAccessFromParent" => false);

	if ($allow_useradmin)
	{
		$DS[$table]["fields"]["name"] = array("name" => $lang["username"], "fieldtype" => "text", "hide_field_when_in_child_table" => true);
		$DS[$table]["fields"]["password"] = array("name" => $lang["password"], "fieldtype" => "text", "hide_from_list" => true, "hide_from_form" => true, "hide_field_when_in_child_table" => true);
		$DS[$table]["fields"]["EMAIL"] = array("name" => $lang["emailaddress"], "fieldtype" => "text", "fieldsubtype" => "mail", "hide_field_when_in_child_table" => true);
		$DS[$table]["fields"]["HIDEFROMASSIGNEEANDOWNERLISTS"] = array("name" => "Hide from lists", "fieldtype" => "text", "filter" => true, "options" => array("n", "y"), "hide_field_when_in_child_table" => true);
		$DS[$table]["fields"]["active"] = array("name" => $lang['active'], "fieldtype" => "text", "filter" => true, "options" => array("no", "yes"), "hide_field_when_in_child_table" => true);
		$DS[$table]["fields"]["LASTPASSCHANGE"] = array("name" => "Last password change", "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true, "hide_field_when_in_child_table" => true);
		$DS[$table]["fields"]["FORCEPASSCHANGE"] = array("name" => "Force password change", "fieldtype" => "text", "filter" => true, "options" => array("n", "y"), "hide_field_when_in_child_table" => true);
		$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true, "hide_field_when_in_child_table" => true);
		$DS[$table]["fields"]["administrator"] = array("name" => "Administrator", "fieldtype" => "text", "filter" => true, "options" => array("no", "yes"), "hide_field_when_in_child_table" => true);
	}

	//even though this user has user admin rights; these fields are invisible
	if ($is_admin)
	{
		$DS[$table]["fields"]["CLLEVEL"] = array("name" => "Clearance level", "fieldtype" => "text", "serialized" => true, "hide_field_when_in_child_table" => true);
		$DS[$table]["fields"]["TRACE"] = array("name" => "Trace", "fieldtype" => "text", "filter" => true, "options" => array("Off", "On"), "hide_field_when_in_child_table" => true);
	}

	if (($allow_useradmin) && (!$is_admin))
	{
		//not allowed to edit administrators
		$DS[$table]["record_access_restrictions"]["write"]["administrator"]["IS"][] = "no";
		//only allowed to edit users with a profile
		$DS[$table]["record_access_restrictions"]["write"]["PROFILE"]["NOT"][] = 0;
		//only allowed to edit active users
		$DS[$table]["record_access_restrictions"]["write"]["active"]["IS"][] = "yes";
		//not allowed to edit own user account
		$DS[$table]["record_access_restrictions"]["write"]["id"]["NOT"][] = $GLOBALS['USERID'];

		//not allowed to write administrator field; altough this field must be visible because of the "record_access_restrictions" on this field
		$DS[$table]["fields"]["administrator"]["allowed_to_write"] = false;
	}

	//2016-06-09 deze wordt veel gebruikt dus in een var gezet
	$ExtraSelectConditionForSelectingNewRecords_for_user_refs = "`users`.`active`='yes' AND `users`.`HIDEFROMASSIGNEEANDOWNERLISTS`='n'";
	$hidegroups = db_GetFlatArray("SELECT `id` FROM `CRMuserprofiles` WHERE `HIDEFROMASSIGNEEANDOWNERLISTS`='y'");
	if (count($hidegroups) > 0)
	{
		$ExtraSelectConditionForSelectingNewRecords_for_user_refs .= " AND `users`.`PROFILE` NOT IN (" . implode(", ", $hidegroups) . ")";
	}

	$table = "files";
	$DS[$table]["primary"] = "fileid";
	$DS[$table]["name_in_database"] = "binfiles";
	$DS[$table]["openlink"] = "csv.php?fileid=";
	$DS[$table]["addlink"] = "";
	$DS[$table]["addlinktext"] = "";
	$DS[$table]["allowed_to_write"] = true;
	$DS[$table]["allowed_to_add"] = false;
	$DS[$table]["allowed_to_delete"] = true;
	$DS[$table]["allow_direct_view"] = true;
	$DS[$table]["record_access_restrictions"]["read"] = $record_access_restrictions["read"][$table]; //deze wordt later gezet in get_data_structure()
	$DS[$table]["record_access_restrictions"]["write"] = $record_access_restrictions["write"][$table]; //deze wordt later gezet in get_data_structure()
	$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
	$DS[$table]["default_sortdirection"] = "asc";
	$DS[$table]["IncludeInSystemWideSearches"] = false;
	$DS[$table]["singular"] = $lang["file"];
	$DS[$table]["plural"] = $lang["files"];
	$DS[$table]["layout"] = "@" . $table . "-filename@";
	$DS[$table]["perpage"] = 50;
	$DS[$table]["download_allowed"] = $download_allowed;
	$DS[$table]["download_limit"] = 0;
	$DS[$table]["search_parents"] = true;
	$DS[$table]["search_childs"] = false;
	$DS[$table]["HasFileAttachments"] = false;
	$DS[$table]["search_files"] = false;
	$DS[$table]["open_records_in_popup_mainlist"] = 0;
	$DS[$table]["open_records_in_popup_sublist"] = 0;
	$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
	$DS[$table]["ShowInlineDeleteLink"] = true;

		$DS[$table]["fields"]["fileid"] = array("name" => $lang["id"], "fieldtype" => "id");

		//hierin staat flextable12, generalpurpose of system
		$DS[$table]["fields"]["type"] = array("name" => "Table", "fieldtype" => "text", "filter" => true, "textrelation" => true);

		$DS[$table]["fields"]["koppelid"] = array("name" => "Record " . $lang["id"], "fieldtype" => "id");
		$DS[$table]["fields"]["filename"] = array("name" => $lang["filename"], "fieldtype" => "text");
		//, "interactive" => true, "interactive-alwaysinline" => false);

		$DS[$table]["fields"]["filetype"] = array("name" => "Filetype", "fieldtype" => "text", "filter" => true);
		$DS[$table]["fields"]["filesize"] = array("name" => $lang["filesize"], "fieldtype" => "number", "add_up_total" => true, "filter" => true); //, "postfix" => "bytes"

		$relation_to = "users";
		$DS[$table]["fields"]["username"] = array("name" => $lang["createdby"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "filter" => true, "ExtraSelectConditionForSelectingNewRecords" => $ExtraSelectConditionForSelectingNewRecords_for_user_refs, "filter" => true, "InheritWriteAccessFromParent" => false);

		if (GetSetting("ENABLEFILEVERSIONING") == "Yes")
		{
			//dit is een referentie naar zichzelf
			$DS[$table]["fields"]["version_belonging_to"] = array("name" => "Version belonging to", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["version_no"] = array("name" => $lang["version"], "fieldtype" => "number");
		}

		if (GetSetting("AllowFoldersInFilelists") == "Yes")
		{
			//dit is een referentie naar zichzelf
			$DS[$table]["fields"]["folder"] = array("name" => "Folder", "fieldtype" => "id");
		}

		if ($is_admin)
		{
			$DS[$table]["fields"]["extractedascii"] = array("name" => "Extracted ascii", "fieldtype" => "text");
			$DS[$table]["fields"]["offloaded"] = array("name" => "Offloaded", "fieldtype" => "text", "filter" => true, "options" => array("n", "y"));
			$DS[$table]["fields"]["resized"] = array("name" => "Resized", "fieldtype" => "text", "filter" => true, "options" => array("n", "y"));
			$DS[$table]["fields"]["coordinates"] = array("name" => "GPS Coordinates", "fieldtype" => "text");
		}
		$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["date"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

//op dit moment niet in gebruik
//		$DS[$table]["fields"]["checked"] = array("name" => "Checked", "filter" => true, "options" => array("in", "out"));
//		$relation_to = "users";
//		$DS[$table]["fields"]["checked_out_by"] = array("name" => "Checked out by", "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "ExtraSelectConditionForSelectingNewRecords" => $ExtraSelectConditionForSelectingNewRecords_for_user_refs, "filter" => true);


	if ($is_admin)
	{
		//niemand heeft toegang tot deze tabel nodig; omdat binaire data toch rechtstreeks gequeried wordt
		$table = "blobs";
		$DS[$table]["primary"] = "fileid";
		$DS[$table]["openlink"] = "csv.php?fileid=";
		$DS[$table]["addlink"] = "";
		$DS[$table]["addlinktext"] = "";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = false;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "File content";
		$DS[$table]["plural"] = "File contents";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
			//dit is eigenlijk een 1-op-1 koppeling naar files. maar dat ondersteund make_list niet goed omdat dit tevens de primary id is. hierdoor is make_list niet op de hoogte van de relatie tussen blobs en files maar dat maakt niet uit want er gebeurt toch niks met blobs
			$DS[$table]["fields"]["fileid"] = array("name" => $lang["id"], "fieldtype" => "id");

			$DS[$table]["fields"]["gz"] = array("name" => "Compressed", "fieldtype" => "text", "filter" => true, "options" => array("n", "y"));
			//dit moet "binary" zijn; maar daar heeft de DS geen support voor
			$DS[$table]["fields"]["content"] = array("name" => "Data", "fieldtype" => "text", "hide_from_list" => true, "hide_from_form" => true);
			//dit moet "binary" zijn; maar daar heeft de DS geen support voor
			$DS[$table]["fields"]["thumbnail"] = array("name" => "Thumbnail", "fieldtype" => "text", "hide_from_list" => true, "hide_from_form" => true);
			//dit moet "binary" zijn; maar daar heeft de DS geen support voor
			$DS[$table]["fields"]["minithumbnail"] = array("name" => "Mini thumbnail", "fieldtype" => "text", "hide_from_list" => true, "hide_from_form" => true);
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["date"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "templates";
		$DS[$table]["primary"] = "templateid";
		$DS[$table]["openlink"] = "admin.php?templates=1&editHTMLtemplate=";
		$DS[$table]["addlink"] = "admin.php?templates=1&editHTMLtemplate=duplicate&add=1";
		$DS[$table]["addlinktext"] = "Create a new template";
		$DS[$table]["MainTable"] = true; //voor admin interface
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = true;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Template";
		$DS[$table]["plural"] = "Templates";
		$DS[$table]["layout"] = $layout_of_template_table;
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = false; // was true aangepast door HF 20131223
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
		$DS[$table]["ShowInlineDuplicateLink"] = true;
		$DS[$table]["ShowInlineDeleteLink"] = true;
			$DS[$table]["fields"]["templateid"] = array("name" => $lang["id"], "fieldtype" => "id");
			$DS[$table]["fields"]["templatename"] = array("name" => "Name", "fieldtype" => "text");
			$DS[$table]["fields"]["templatetype"] = array("name" => "Type", "fieldtype" => "text", "filter" => true, "interactive" => true, "interactive-alwaysinline" => true, "options" => array('TEMPLATE_HTML','TEMPLATE_HTML_FORM','TEMPLATE_DASHBOARD','TEMPLATE_MENU','TEMPLATE_REPORT_PDF','TEMPLATE_CSS','TEMPLATE_XML','TEMPLATE_PLAIN','TEMPLATE_JAVASCRIPT'));
			$DS[$table]["fields"]["page_id"] = array("name" => "Page id", "fieldtype" => "text"); // toegevoegd door HF 20131223

			$DS[$table]["fields"]["template_subject"] = array("name" => "Subject", "fieldtype" => "text");

//JH 2017-10-10 switched of to prevent circular join
			//dit is een circulaire referentie; "ExtraSelectConditionForSelectingNewRecords" zou type == "css" moeten zijn
//			$relation_to = "templates";
//			$DS[$table]["fields"]["stylesheet"] = array("name" => "Stylesheet", "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "filter" => true);
			$DS[$table]["fields"]["stylesheet"] = array("name" => "Stylesheet", "fieldtype" => "id", "filter" => true);

			$DS[$table]["fields"]["orientation"] = array("name" => "Orientation", "fieldtype" => "text", "filter" => true, "options" => array("P", "L"));
			$DS[$table]["fields"]["content"] = array("name" => "Content", "fieldtype" => "text", "multiline" => true);
			//moet binary zijn; maar daar heeft de DS geen support voor
			$DS[$table]["fields"]["binary_content"] = array("name" => "Binary content", "fieldtype" => "text");
			$DS[$table]["fields"]["version_belonging_to"] = array("name" => "Version belonging to", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "templateaccessrules";
		$DS[$table]["primary"] = "id";
		$DS[$table]["name_in_database"] = "templateaccessrules";
		$DS[$table]["openlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = true;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Template access rule";
		$DS[$table]["plural"] = "Template access rules";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
		$DS[$table]["ShowInlineDuplicateLink"] = true;
		$DS[$table]["ShowInlineDeleteLink"] = true;
			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");
			$relation_to = "groups";
			$DS[$table]["fields"]["group"] = array("name" => $DS[$relation_to]["singular"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "filter" => true, "ExtraSelectConditionForSelectingNewRecords" => "`" . $relation_to . "`.`active`='yes'");
			$relation_to = "templates";
			$DS[$table]["fields"]["template"] = array("name" => $DS[$relation_to]["singular"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "filter" => true, "ExtraSelectConditionForSelectingNewRecords" => "`" . $relation_to . "`.`version_belonging_to`=0");
			$DS[$table]["fields"]["read"] = array("name" => "Read", "fieldtype" => "text", "options" => array("0", "1"), "filter" => true, "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes", "defaultvalue" => "0", "fieldtype_in_extrafields_table" => "checkbox", "formelementtype" => "checkbox");
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		//2016-10-12 nu ook alleen nog voor admin
		$table = "messages";
		$DS[$table]["primary"] = "id";
		$DS[$table]["name_in_database"] = "internalmessages";
		$DS[$table]["openlink"] = "index.php?UserMessage&ReadInternalMessage=";
		$DS[$table]["addlink"] = "index.php?UserMessage&ComposeNewMessage=true";
		$DS[$table]["addlinktext"] = "Compose new message";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = true;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
/*		if (!$is_admin)
		{
			$DS[$table]["record_access_restrictions"]["read"]["to"]["IS"][] = $GLOBALS['USERID'];
		}*/
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "desc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Message";
		$DS[$table]["plural"] = "Messages";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$popupsize = getpopupsize($table);
		$DS[$table]["open_records_in_popup_mainlist"] = $popupsize;
		$DS[$table]["open_records_in_popup_sublist"] = $popupsize;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
		$DS[$table]["ShowInlineDuplicateLink"] = false;
		$DS[$table]["ShowInlineDeleteLink"] = true;
			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");

			$relation_to = "users";
			$DS[$table]["fields"]["to"] = array("name" => "To", "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "ExtraSelectConditionForSelectingNewRecords" => $ExtraSelectConditionForSelectingNewRecords_for_user_refs, "filter" => true);

			$relation_to = "users";
			$DS[$table]["fields"]["from"] = array("name" => "From", "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "ExtraSelectConditionForSelectingNewRecords" => $ExtraSelectConditionForSelectingNewRecords_for_user_refs, "filter" => true);

			$DS[$table]["fields"]["subject"] = array("name" => "Subject", "fieldtype" => "text");
			$DS[$table]["fields"]["read"] = array("name" => "Read", "fieldtype" => "text", "filter" => true, "options" => array("n", "y"));
			$DS[$table]["fields"]["shown_in_popup"] = array("name" => "Shown", "fieldtype" => "text", "filter" => true, "options" => array("n", "y"));
			$DS[$table]["fields"]["dub_count"] = array("name" => "Times received", "fieldtype" => "number");
			$DS[$table]["fields"]["body"] = array("name" => "Body", "fieldtype" => "text", "multiline" => true);
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "modules";
		$DS[$table]["primary"] = "mid";
		$DS[$table]["openlink"] = "modules.php?action=edit&mid=";
		$DS[$table]["addlink"] = "modules.php?action=addonline";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["MainTable"] = true; //voor admin interface
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = true;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Module";
		$DS[$table]["plural"] = "Modules";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@: @" . $table . "-module_name@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
		$DS[$table]["ShowInlineDuplicateLink"] = false;
		$DS[$table]["ShowInlineDeleteLink"] = true;
			$DS[$table]["fields"]["mid"] = array("name" => $lang["id"], "fieldtype" => "id");
			$DS[$table]["fields"]["module_name"] = array("name" => "Name", "fieldtype" => "text");
			$DS[$table]["fields"]["module_description"] = array("name" => "Description", "fieldtype" => "text");
			$DS[$table]["fields"]["module_add_date"] = array("name" => $lang["date"], "fieldtype" => "date", "dateformat_in_database" => "epoch", "filter" => true);
			$DS[$table]["fields"]["module_last_run_result"] = array("name" => "Last run result", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["module_last_run_date"] = array("name" => "Last run date", "fieldtype" => "date", "dateformat_in_database" => "epoch", "filter" => true);
			$DS[$table]["fields"]["module_code"] = array("name" => "Code", "fieldtype" => "text", "multiline" => true);
			$DS[$table]["fields"]["module_list_html"] = array("name" => "List html", "fieldtype" => "text");
			$DS[$table]["fields"]["version_belonging_to"] = array("name" => "Version belonging to", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["for_table"] = array("name" => "For table", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "moduleaccessrules";
		$DS[$table]["primary"] = "id";
		$DS[$table]["name_in_database"] = "moduleaccessrules";
		$DS[$table]["openlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = true;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Module access rule";
		$DS[$table]["plural"] = "Module access rules";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 3;
		$DS[$table]["open_records_in_popup_sublist"] = 3;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
		$DS[$table]["ShowInlineDuplicateLink"] = true;
		$DS[$table]["ShowInlineDeleteLink"] = true;
		$DS[$table]["dqs_enabled"] = true;
			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");
			$relation_to = "groups";
			$DS[$table]["fields"]["group"] = array("name" => $DS[$relation_to]["singular"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "filter" => true, "ExtraSelectConditionForSelectingNewRecords" => "`" . $relation_to . "`.`active`='yes'");
			$relation_to = "modules";
			$DS[$table]["fields"]["module"] = array("name" => $DS[$relation_to]["singular"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "filter" => true, "ExtraSelectConditionForSelectingNewRecords" => "`" . $relation_to . "`.`version_belonging_to`=0");
			$DS[$table]["fields"]["execute"] = array("name" => "Execute", "fieldtype" => "text", "options" => array("0", "1"), "filter" => true, "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes", "defaultvalue" => "0", "fieldtype_in_extrafields_table" => "checkbox", "formelementtype" => "checkbox");
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "settings";
		$DS[$table]["primary"] = "settingid";
		$DS[$table]["openlink"] = "admin.php?editsetting=";
		$DS[$table]["addlink"] = "";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = false;
		$DS[$table]["allowed_to_delete"] = false;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-setting";
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Setting";
		$DS[$table]["plural"] = "Settings";
		$DS[$table]["layout"] = "@" . $table . "-setting@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
			$DS[$table]["fields"]["settingid"] = array("name" => $lang["id"], "fieldtype" => "id");
			$DS[$table]["fields"]["setting"] = array("name" => "Setting", "fieldtype" => "text");
			$DS[$table]["fields"]["value"] = array("name" => "Value", "fieldtype" => "text", "multiline" => true); //some values are multiline

			$DS[$table]["fields"]["value"]["optionclass"]["No"] = "errmsg";
			$DS[$table]["fields"]["value"]["optionclass"]["Yes"] = "okmsg";

			//spelfout in database
			$DS[$table]["fields"]["discription"] = array("name" => "Description", "fieldtype" => "text");
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);
	}

	if (($is_admin) || (CheckFunctionAccess("DenyJournalAccess") != "ok")) //Deny access to all journals
	{
		$table = "journal";
		$DS[$table]["primary"] = "id";
		$DS[$table]["openlink"] = "";
		$DS[$table]["addlink"] = "";
		$DS[$table]["addlinktext"] = "Add";
		//journal records zijn niet aanpasbaar voor niemand niet. ook niet voor admin
		$DS[$table]["allowed_to_write"] = false;
		$DS[$table]["allowed_to_add"] = false;
		$DS[$table]["allowed_to_delete"] = false;
		$DS[$table]["allow_direct_view"] = true;

		$DS[$table]["record_access_restrictions"]["read"] = array();
		if ((!$is_admin) && (CheckFunctionAccess("ShowOnlyOwnJournalRecords") == "ok")) //When showing journals, show only own updates
		{
			$DS[$table]["record_access_restrictions"]["read"]["user"]["IS"][] = $GLOBALS['USERID'];
		}
		//journal records zijn niet aanpasbaar voor niemand niet. ook niet voor admin
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["record_access_restrictions"]["write"]["id"][] = 0;

		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "desc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Journal record";
		$DS[$table]["plural"] = "Journal records";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 10;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");
			$DS[$table]["fields"]["type"] = array("name" => "Table", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["eid"] = array("name" => "Record " . $lang["id"], "fieldtype" => "id");

			$relation_to = "users";
			$DS[$table]["fields"]["user"] = array("name" => $lang["createdby"], "Type", "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "ExtraSelectConditionForSelectingNewRecords" => $ExtraSelectConditionForSelectingNewRecords_for_user_refs, "filter" => true);

			$DS[$table]["fields"]["message"] = array("name" => "Message", "fieldtype" => "text", "multiline" => true);
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "datajournal";
		$DS[$table]["primary"] = "id";
		$DS[$table]["openlink"] = "";
		$DS[$table]["addlink"] = "";
		$DS[$table]["addlinktext"] = "Add";
		//journal records zijn niet aanpasbaar voor niemand niet. ook niet voor admin
		$DS[$table]["allowed_to_write"] = false;
		$DS[$table]["allowed_to_add"] = false;
		$DS[$table]["allowed_to_delete"] = false;
		$DS[$table]["allow_direct_view"] = true;

		$DS[$table]["record_access_restrictions"]["read"] = array();
		if ((!$is_admin) && (CheckFunctionAccess("ShowOnlyOwnJournalRecords") == "ok")) //When showing journals, show only own updates
		{
			$DS[$table]["record_access_restrictions"]["read"]["by_user"]["IS"][] = $GLOBALS['USERID'];
		}
		$DS[$table]["record_access_restrictions"]["write"] = array();
		//journal records zijn niet aanpasbaar voor niemand niet. ook niet voor admin
		$DS[$table]["record_access_restrictions"]["write"]["id"][] = 0;

		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "desc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Data journal record";
		$DS[$table]["plural"] = "Data journal records";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 10;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");
			//textrelation
			$DS[$table]["fields"]["table"] = array("name" => "Table", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["record"] = array("name" => "Record " . $lang["id"], "fieldtype" => "id");
			//textrelation
			$DS[$table]["fields"]["field"] = array("name" => "Field", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["from_val"] = array("name" => "From value", "fieldtype" => "text");
			$DS[$table]["fields"]["to_val"] = array("name" => "To value", "fieldtype" => "text", "sometimes_serialized" => true);

			$relation_to = "users";
			$DS[$table]["fields"]["by_user"] = array("name" => "User", "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "ExtraSelectConditionForSelectingNewRecords" => $ExtraSelectConditionForSelectingNewRecords_for_user_refs, "filter" => true);

			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);
	}

	if ($is_admin)
	{
		$table = "locks";
		$DS[$table]["primary"] = "lockon";
		$DS[$table]["compoundprimarykey"] = true; //PRIMARY KEY (`lockon`,`lockby`,`locktable`),
		$DS[$table]["name_in_database"] = "entitylocks";
		$DS[$table]["openlink"] = "";
		$DS[$table]["addlink"] = "";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = false;
		$DS[$table]["allowed_to_delete"] = false; //false because of compound key
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Lock";
		$DS[$table]["plural"] = "Locks";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
			$DS[$table]["fields"]["lockon"] = array("name" => "On", "fieldtype" => "id");

			$relation_to = "users";
			$DS[$table]["fields"]["lockby"] = array("name" => "By", "Type", "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "ExtraSelectConditionForSelectingNewRecords" => $ExtraSelectConditionForSelectingNewRecords_for_user_refs, "filter" => true);

			//textrelation
			$DS[$table]["fields"]["locktable"] = array("name" => "Table", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "tables";
		$DS[$table]["primary"] = "recordid";
		$DS[$table]["name_in_database"] = "flextabledefs";
		$DS[$table]["openlink"] = "flextable.php?EditFlexTable=";
		$DS[$table]["addlink"] = "flextable.php?EditFlexTable=new";
		$DS[$table]["MainTable"] = true; //voor admin interface
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = true;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Table";
		$DS[$table]["plural"] = "Tables";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@: @" . $table . "-tablename@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
			$DS[$table]["fields"]["recordid"] = array("name" => $lang["id"], "fieldtype" => "id");
			$DS[$table]["fields"]["ordering"] = array("name" => "Order", "fieldtype" => "number", "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes");
			$DS[$table]["fields"]["tablename"] = array("name" => "Name", "fieldtype" => "text");

			$relation_to = "templates";
			$DS[$table]["fields"]["formid"] = array("name" => "Template", "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $layout_of_template_table, "filter" => true);

			$DS[$table]["fields"]["refer_field_layout"] = array("name" => "Refer layout", "fieldtype" => "text");
			$DS[$table]["fields"]["comment"] = array("name" => "Remarks", "fieldtype" => "text", "multiline" => true, "PerFieldSearch" => true, "interactive" => true);
			$DS[$table]["fields"]["maxrowsperpage"] = array("name" => "Per page", "fieldtype" => "number");
			$DS[$table]["fields"]["headerhtml"] = array("name" => "Header", "fieldtype" => "text");
			$DS[$table]["fields"]["addlinktext"] = array("name" => "Add link", "fieldtype" => "text", "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes");
			$DS[$table]["fields"]["add_in_popup"] = array("name" => "Open records in popup for sublists", "fieldtype" => "text", "options" => array("n", "y"), "filter" => true);

			$relation_to = "fields";
			$DS[$table]["fields"]["sort_on"] = array("name" => "Sort", "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $layout_of_field_table, "filter" => true);

			$DS[$table]["fields"]["sort_direction"] = array("name" => "Sort direction", "fieldtype" => "text", "options" => array("Ascending", "Descending"), "filter" => true);
			$DS[$table]["fields"]["tableheaderrepeat"] = array("name" => "Header repeat", "fieldtype" => "number");
			$DS[$table]["fields"]["stayinformaftersave"] = array("name" => "Stay in form after save", "fieldtype" => "text", "options" => array("n", "y"), "filter" => true);
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "downloadrules";
		$DS[$table]["primary"] = "id";
		$DS[$table]["name_in_database"] = "downloadrules";
		$DS[$table]["openlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = true;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Download rule";
		$DS[$table]["plural"] = "Download rules";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 3;
		$DS[$table]["open_records_in_popup_sublist"] = 3;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
		$DS[$table]["ShowInlineDeleteLink"] = true;
			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");

			$relation_to = "tables";
			$DS[$table]["fields"]["table"] = array("name" => $DS[$relation_to]["singular"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "filter" => true, "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes");
			$relation_to = "groups";
			$DS[$table]["fields"]["group"] = array("name" => $DS[$relation_to]["singular"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "filter" => true, "ExtraSelectConditionForSelectingNewRecords" => "`" . $relation_to . "`.`active`='yes'");

			$DS[$table]["fields"]["max_rows"] = array("name" => "Maximum rows", "fieldtype" => "number", "required" => true, "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes");
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "tableaccessrules";
		$DS[$table]["primary"] = "id";
		$DS[$table]["name_in_database"] = "tableaccessrules";
		$DS[$table]["openlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = true;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Table access rule";
		$DS[$table]["plural"] = "Table access rules";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 3;
		$DS[$table]["open_records_in_popup_sublist"] = 3;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
		$DS[$table]["ShowInlineDuplicateLink"] = true;
		$DS[$table]["ShowInlineDeleteLink"] = true;
		$DS[$table]["dqs_enabled"] = true;
			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");
			$relation_to = "groups";
			$DS[$table]["fields"]["group"] = array("name" => $DS[$relation_to]["singular"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "filter" => true, "ExtraSelectConditionForSelectingNewRecords" => "`" . $relation_to . "`.`active`='yes'");
			$relation_to = "tables";
			$DS[$table]["fields"]["table"] = array("name" => $DS[$relation_to]["singular"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "filter" => true);
			$DS[$table]["fields"]["read"] = array("name" => "Read", "fieldtype" => "text", "options" => array("0", "1"), "filter" => true, "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes", "defaultvalue" => "0", "fieldtype_in_extrafields_table" => "checkbox", "formelementtype" => "checkbox");
			$DS[$table]["fields"]["write"] = array("name" => "Write", "fieldtype" => "text", "options" => array("0", "1"), "filter" => true, "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes", "defaultvalue" => "0", "fieldtype_in_extrafields_table" => "checkbox", "formelementtype" => "checkbox");
			$DS[$table]["fields"]["add"] = array("name" => "Add", "fieldtype" => "text", "options" => array("0", "1"), "filter" => true, "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes", "defaultvalue" => "0", "fieldtype_in_extrafields_table" => "checkbox", "formelementtype" => "checkbox");
			$DS[$table]["fields"]["delete"] = array("name" => "Delete", "fieldtype" => "text", "options" => array("0", "1"), "filter" => true, "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes", "defaultvalue" => "0", "fieldtype_in_extrafields_table" => "checkbox", "formelementtype" => "checkbox");
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);


		$table = "fields";
		$DS[$table]["primary"] = "id";
		$DS[$table]["name_in_database"] = "extrafields";
		$DS[$table]["logical_delete"] = true;
		$DS[$table]["openlink"] = "extrafields.php?editextrafield=";
		$DS[$table]["addlink"] = "";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["MainTable"] = true; //voor admin interface
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = false;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-ordering";
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Field";
		$DS[$table]["plural"] = "Fields";
		$DS[$table]["layout"] = $layout_of_field_table;
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
		$DS[$table]["ShowInlineDuplicateLink"] = false;
		$DS[$table]["ShowInlineDeleteLink"] = true;
			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");
			$DS[$table]["fields"]["ordering"] = array("name" => "Order", "fieldtype" => "number", "interactive" => true, "interactive-alwaysinline" => true);
			$DS[$table]["fields"]["tabletype"] = array("name" => "Table", "fieldtype" => "text", "filter" => true, "size_width" => 4);
			$DS[$table]["fields"]["deleted"] = array("name" => "Deleted", "fieldtype" => "text", "options" => array("n", "y"), "filter" => true);
			$DS[$table]["fields"]["fieldtype"] = array("name" => "Field type", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["name"] = array("name" => "Name", "fieldtype" => "text", "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes");
			$DS[$table]["fields"]["displaylistname"] = array("name" => "List name", "fieldtype" => "text", "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes");
			$DS[$table]["fields"]["options"] = array("name" => "Options", "fieldtype" => "text", "sometimes_serialized" => true, "filter" => true);
			$DS[$table]["fields"]["optioncolors"] = array("name" => "Option colors", "fieldtype" => "text", "serialized" => true);
			$DS[$table]["fields"]["forcing"] = array("name" => "Required", "fieldtype" => "text", "options" => array("n", "y"), "filter" => true);
			$DS[$table]["fields"]["defaultval"] = array("name" => "Default value", "fieldtype" => "text");
			$DS[$table]["fields"]["sort"] = array("name" => "Sort OR Select from table popup instead of drop-down", "fieldtype" => "text", "options" => array("n", "y"), "filter" => true);
			$DS[$table]["fields"]["size"] = array("name" => "Size", "fieldtype" => "number");
			$DS[$table]["fields"]["showsearchbox"] = array("name" => "Show searchbox", "fieldtype" => "text", "options" => array("n", "y"), "filter" => true);
			$DS[$table]["fields"]["israwhtml"] = array("name" => "HTML", "fieldtype" => "text", "options" => array("n", "y"), "filter" => true);
			$DS[$table]["fields"]["remarks"] = array("name" => "Remarks", "fieldtype" => "text", "multiline" => true, "PerFieldSearch" => true);
			$DS[$table]["fields"]["calculation"] = array("name" => "Calculation", "fieldtype" => "text", "multiline" => true, "PerFieldSearch" => true);
			$DS[$table]["fields"]["calculation_type"] = array("name" => "Calculation type", "fieldtype" => "text", "options" => array("database trigger", "calculated on display"), "filter" => true);
			$DS[$table]["fields"]["showasradio"] = array("name" => "Show as radio", "fieldtype" => "text", "options" => array("n", "y"), "filter" => true);
			$DS[$table]["fields"]["excludefromfilters"] = array("name" => "No filter", "fieldtype" => "text", "options" => array("n", "y"), "filter" => true);
			$DS[$table]["fields"]["underwaterfield"] = array("name" => "Hide from lists and forms", "fieldtype" => "text", "options" => array("n", "y"), "filter" => true);
			$DS[$table]["fields"]["sum_column"] = array("name" => "Add up total", "fieldtype" => "text", "options" => array("n", "y"), "filter" => true);
			$DS[$table]["fields"]["number_format"] = array("name" => "Number format", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["log_stats"] = array("name" => "Log statistics", "fieldtype" => "text", "options" => array("n", "y"), "filter" => true);
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "fieldaccessrules";
		$DS[$table]["primary"] = "id";
		$DS[$table]["name_in_database"] = "fieldaccessrules";
		$DS[$table]["openlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = true;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Field access rule";
		$DS[$table]["plural"] = "Field access rules";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 3;
		$DS[$table]["open_records_in_popup_sublist"] = 3;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
		$DS[$table]["ShowInlineDuplicateLink"] = true;
		$DS[$table]["ShowInlineDeleteLink"] = true;
		$DS[$table]["dqs_enabled"] = true;
			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");
			$relation_to = "groups";
			$DS[$table]["fields"]["group"] = array("name" => $DS[$relation_to]["singular"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "filter" => true, "ExtraSelectConditionForSelectingNewRecords" => "`" . $relation_to . "`.`active`='yes'");
			$relation_to = "fields";
			$DS[$table]["fields"]["field"] = array("name" => $DS[$relation_to]["singular"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $layout_of_field_table, "filter" => true);
			$DS[$table]["fields"]["read"] = array("name" => "Read", "fieldtype" => "text", "options" => array("0", "1"), "filter" => true, "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes", "defaultvalue" => "0", "fieldtype_in_extrafields_table" => "checkbox", "formelementtype" => "checkbox");
			$DS[$table]["fields"]["write"] = array("name" => "Write", "fieldtype" => "text", "options" => array("0", "1"), "filter" => true, "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes", "defaultvalue" => "0", "fieldtype_in_extrafields_table" => "checkbox", "formelementtype" => "checkbox");
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);
	}

	if ($is_admin)
	{
		$table = "languages";
		$DS[$table]["primary"] = "id";
		$DS[$table]["openlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlinktext"] = "Add language tag";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = true;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Language tag";
		$DS[$table]["plural"] = "Language tags";
		$DS[$table]["layout"] = "@" . $table . "-LANGID@ @" . $table . "-TEXTID@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 2;
		$DS[$table]["open_records_in_popup_sublist"] = 2;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
		$DS[$table]["ShowInlineDuplicateLink"] = true;
		$DS[$table]["ShowInlineDeleteLink"] = false;
			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");
			$DS[$table]["fields"]["LANGID"] = array("name" => "Language", "fieldtype" => "text", "filter" => true, "options" => db_GetFlatArray("SELECT DISTINCT(LANGID) FROM CRMlanguages ORDER BY LANGID"), "required" => true, "fieldtype_in_extrafields_table" => "drop-down", "formelementtype" => "select");
			$DS[$table]["fields"]["TEXTID"] = array("name" => "Tag", "fieldtype" => "text", "filter" => true, "required" => true);
			$DS[$table]["fields"]["TEXT"] = array("name" => "Translation", "fieldtype" => "text", "required" => true);
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "sessions";
		$DS[$table]["primary"] = "id";
		$DS[$table]["openlink"] = "";
		$DS[$table]["addlink"] = "";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = false;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Session";
		$DS[$table]["plural"] = "Sessions";
		$DS[$table]["layout"] = "@" . $table . "-temp@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");
			$relation_to = 'users';
			$DS[$table]["fields"]["userid"] = array("name" => $DS[$relation_to]["singular"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "required" => true, "filter" => true);

			$DS[$table]["fields"]["temp"] = array("name" => "Session id", "fieldtype" => "text");
			//hierin staat niet de expiry time maar de last activity; hiervoor zou ook timestamp_last_change gebruikt kunnen worden
			$DS[$table]["fields"]["exptime"] = array("name" => "Last activity", "fieldtype" => "date", "dateformat_in_database" => "epoch", "filter" => true);
			$DS[$table]["fields"]["noexp"] = array("name" => "Webapp session", "fieldtype" => "text", "options" => array("n", "y"), "filter" => true);
			$DS[$table]["fields"]["sessioncache"] = array("name" => "Cache", "fieldtype" => "text", "serialized" => true);
			$DS[$table]["fields"]["ip"] = array("name" => "IP-address", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "rules";
		$DS[$table]["primary"] = "id";
		$DS[$table]["openlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlinktext"] = "Add rule";
		$DS[$table]["MainTable"] = true; //voor admin interface
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = true;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Rule";
		$DS[$table]["plural"] = "Rules";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@: @" . $table . "-name@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = true;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
		$DS[$table]["ShowInlineDuplicateLink"] = true;
		$DS[$table]["ShowInlineDeleteLink"] = true;
		$DS[$table]["dqs_enabled"] = true;
			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");
			$DS[$table]["fields"]["table"] = array("name" => "Table", "fieldtype" => "text", "filter" => true, "use_populator_function" => true, "showsearchbox" => true, "required" => true);

			$DS[$table]["fields"]["enabled"] = array("name" => "Enabled", "fieldtype" => "text", "options" => array("No", "Yes"), "filter" => true, "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes", "defaultvalue" => "Yes", "fieldtype_in_extrafields_table" => "checkbox", "formelementtype" => "checkbox");

			$DS[$table]["fields"]["log_passed"] = array("name" => "Log messages of passed conditions", "fieldtype" => "text", "options" => array("No", "Yes"), "filter" => true, "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes", "defaultvalue" => "Yes", "fieldtype_in_extrafields_table" => "checkbox", "formelementtype" => "checkbox", "FormComment" => "Log messages about passed conditions. Even when enabled, these journal messages will be discarded after two months.");

			$DS[$table]["fields"]["log_failed"] = array("name" => "Log messages of failed conditions", "fieldtype" => "text", "options" => array("No", "Yes"), "filter" => true, "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes", "defaultvalue" => "Yes", "fieldtype_in_extrafields_table" => "checkbox", "formelementtype" => "checkbox", "FormComment" => "Log messages about failed conditions. Even when enabled, these journal messages will be discarded after one week.");

			$DS[$table]["fields"]["enabled"]["optionclass"]["No"] = "errmsg";
			$DS[$table]["fields"]["enabled"]["optionclass"]["Yes"] = "okmsg";

			$DS[$table]["fields"]["log_passed"]["optionclass"]["Yes"] = "errmsg";
			$DS[$table]["fields"]["log_passed"]["optionclass"]["No"] = "okmsg";
			$DS[$table]["fields"]["log_failed"]["optionclass"]["Yes"] = "errmsg";
			$DS[$table]["fields"]["log_failed"]["optionclass"]["No"] = "okmsg";

			$DS[$table]["fields"]["operator"] = array("name" => "Conditions operator", "fieldtype" => "text", "options" => array("AND", "OR"), "required" => true, "defaultvalue" => "AND", "filter" => true);

			$DS[$table]["fields"]["name"] = array("name" => "Name", "fieldtype" => "text", "required" => true, "filter" => true, "size_width" => 50);
			$DS[$table]["fields"]["remarks"] = array("name" => "Remarks", "fieldtype" => "text", "multiline" => true, "interactive" => true);
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "ruleconditions";
		$DS[$table]["primary"] = "id";
		$DS[$table]["openlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["MainTable"] = true; //voor admin interface
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = true;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Rule condition";
		$DS[$table]["plural"] = "Rule conditions";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
		$DS[$table]["ShowInlineDuplicateLink"] = true;
		$DS[$table]["ShowInlineDeleteLink"] = true;
		$DS[$table]["dqs_enabled"] = true;
			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");
			$relation_to = "rules";
			$DS[$table]["fields"]["rule"] = array("name" => "Rule", "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "required" => true, "filter" => true);
			$DS[$table]["fields"]["enabled"] = array("name" => "Enabled", "fieldtype" => "text", "options" => array("No", "Yes"), "filter" => true, "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes", "defaultvalue" => "Yes", "fieldtype_in_extrafields_table" => "checkbox", "formelementtype" => "checkbox");
			$DS[$table]["fields"]["enabled"]["optionclass"]["No"] = "errmsg";
			$DS[$table]["fields"]["enabled"]["optionclass"]["Yes"] = "okmsg";
			$DS[$table]["fields"]["field"] = array("name" => "Field", "fieldtype" => "text", "filter" => true, "use_populator_function" => true, "showsearchbox" => true);

			$DS[$table]["fields"]["condition"] = array("name" => "Condition", "fieldtype" => "text", "use_populator_function" => true, "required" => true, "FormComment" => "Combine a '[field] must be [matchvalue]' condition with a '[field] must be changed' condition if you need an evaluation on a field getting a certain value.");

			$DS[$table]["fields"]["matchvalue"] = array("name" => "Match value", "fieldtype" => "text", "multiline" => true, "FormComment" => "You can use generic tags like @ISODATE@, %-tags and PHP code");

			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "ruleactions";
		$DS[$table]["primary"] = "id";
		$DS[$table]["openlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["MainTable"] = true; //voor admin interface
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = true;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Rule action";
		$DS[$table]["plural"] = "Rule actions";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = true;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
		$DS[$table]["ShowInlineDuplicateLink"] = true;
		$DS[$table]["ShowInlineDeleteLink"] = true;
		$DS[$table]["dqs_enabled"] = true;

			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");
			$relation_to = "rules";
			$DS[$table]["fields"]["rule"] = array("name" => "Rule", "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "filter" => true);
			$DS[$table]["fields"]["enabled"] = array("name" => "Enabled", "options" => array("No", "Yes"), "filter" => true, "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes", "defaultvalue" => "Yes", "fieldtype_in_extrafields_table" => "checkbox", "formelementtype" => "checkbox");
			$DS[$table]["fields"]["enabled"]["optionclass"]["No"] = "errmsg";
			$DS[$table]["fields"]["enabled"]["optionclass"]["Yes"] = "okmsg";
			$DS[$table]["fields"]["operator"] = array("name" => "Action conditions operator", "fieldtype" => "text", "options" => array("AND", "OR"), "required" => true, "defaultvalue" => "AND", "filter" => true);
			$DS[$table]["fields"]["action"] = array("name" => "Action", "fieldtype" => "text", "use_populator_function" => true);

			$relation_to = "templates";
			$DS[$table]["fields"]["template"] = array("name" => $DS[$relation_to]["singular"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $layout_of_template_table, "filter" => true, "ExtraSelectConditionForSelectingNewRecords" => "`" . $relation_to . "`.`version_belonging_to`=0 AND `" . $relation_to . "`.`templatetype` IN ('TEMPLATE_HTML', 'TEMPLATE_REPORT_PDF')", "showsearchbox" => true, "FormComment" => "The template used for sending an e-mail or SMS or generating or attaching a report. In case of SMS, only the template subject will be used.");

			$relation_to = "modules";
			$DS[$table]["fields"]["module_to_run"] = array("name" => "Module", "fieldtype" => "text", "use_populator_function" => true, "filter" => true, "FormComment" => "The module to be ran (if you selected [run module] as action). Your module needs to have a RunAsUser setting.", "relation_to" => $relation_to);

			$DS[$table]["fields"]["field"] = array("name" => "Field", "fieldtype" => "text", "filter" => true, "showsearchbox" => true, "use_populator_function" => true);

			$DS[$table]["fields"]["attach"] = array("name" => "File(s) to attach to mail", "fieldtype" => "text", "use_populator_function" => true);
			$DS[$table]["fields"]["attach_matchvalue"] = array("name" => "Attach-matchvalue", "fieldtype" => "text", "FormComment" => "You can use tags, but only generic tags like @USER@ and @TODAY@.");
			$DS[$table]["fields"]["value"] = array("name" => "Value", "fieldtype" => "text", "multiline" => true, "FormComment" => "You can use tags, but only raw-style: %....% for matching and also generic tags for e-mail addresses (like @ADMEMAIL@). E-mail addresses and SMS numbers need to be entered comma-seperated. PHP is supported.");
			$DS[$table]["fields"]["message_to_display"] = array("name" => "Message to display", "fieldtype" => "text", "multiline" => true, "FormComment" => "You can use tags.");
			$DS[$table]["fields"]["alternate_from_address"] = array("name" => "Alternate from-address", "fieldtype" => "text", "FormComment" => "When emailing, use this address instead of any other address (this setting overrules everything, but NOT the global UNIFIED_FROM setting). Form: 'Firstname Lastname <firstname@lastname.com>' or just firstname@lastname.com. You can use tags.", "interactive" => true);

			$DS[$table]["fields"]["remarks"] = array("name" => "Remarks", "fieldtype" => "text", "multiline" => true, "interactive" => true);

			$DS[$table]["fields"]["order"] = array("name" => "Order", "fieldtype" => "number", "filter" => false, "FormComment" => "The order in which rule actions will be executed", "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes", "size_width" => 4);

			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "ruleactionconditions";
		$DS[$table]["primary"] = "id";
		$DS[$table]["openlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["MainTable"] = true; //voor admin interface
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = true;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Rule action condition";
		$DS[$table]["plural"] = "Rule action conditions";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
		$DS[$table]["ShowInlineDuplicateLink"] = true;
		$DS[$table]["ShowInlineDeleteLink"] = true;
		$DS[$table]["dqs_enabled"] = true;

			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");

			$relation_to = "ruleactions";
			$DS[$table]["fields"]["ruleaction"] = array("name" => $DS[$relation_to]["singular"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "required" => true, "filter" => true);
			$DS[$table]["fields"]["enabled"] = array("name" => "Enabled", "fieldtype" => "text", "options" => array("No", "Yes"), "filter" => true, "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes", "defaultvalue" => "Yes", "fieldtype_in_extrafields_table" => "checkbox", "formelementtype" => "checkbox");
			$DS[$table]["fields"]["enabled"]["optionclass"]["No"] = "errmsg";
			$DS[$table]["fields"]["enabled"]["optionclass"]["Yes"] = "okmsg";
			$DS[$table]["fields"]["field"] = array("name" => "Field", "fieldtype" => "text", "filter" => true, "use_populator_function" => true, "showsearchbox" => true, "required" => true);

			$DS[$table]["fields"]["condition"] = array("name" => "Condition", "fieldtype" => "text", "use_populator_function" => true, "required" => true);

			$DS[$table]["fields"]["matchvalue"] = array("name" => "Match value", "fieldtype" => "text", "multiline" => true, "FormComment" => "You can use generic tags like @ISODATE@, %-tags and PHP code");
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);


		$table = "calendars";
		$DS[$table]["primary"] = "id";
		$DS[$table]["openlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlinktext"] = "Create new calendar";
		$DS[$table]["MainTable"] = true; //voor admin interface
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = true;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-sortorder";
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Calendar";
		$DS[$table]["plural"] = "Calendars";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@: @" . $table . "-description@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
		$DS[$table]["ShowInlineDuplicateLink"] = true;
		$DS[$table]["ShowInlineDeleteLink"] = true;

			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");
			$DS[$table]["fields"]["description"] = array("name" => "Description", "fieldtype" => "text", "required" => true);
			$DS[$table]["fields"]["sortorder"] = array("name" => "Sort order", "fieldtype" => "number");

			//isrelationtodo to fields
			$DS[$table]["fields"]["ref_start_date_time"] = array("name" => "Start date/time", "fieldtype" => "id", "filter" => true, "showsearchbox" => true, "use_populator_function" => true);

			//isrelationtodo to fields
			$DS[$table]["fields"]["ref_end_date_time"] = array("name" => "End date/time", "fieldtype" => "id", "required" => true, "filter" => true, "showsearchbox" => true, "use_populator_function" => true);

			$DS[$table]["fields"]["appointment_click_action"] = array("name" => "Action on click", "fieldtype" => "text", "options" => array('None','Open record in popup','Open record in current window'), "required" => true, "filter" => true);
			$DS[$table]["fields"]["show_calendar_in_legend"] = array("name" => "Show in legend", "fieldtype" => "text", "options" => array('Yes','No'), "required" => true, "filter" => true);
			$DS[$table]["fields"]["summary"] = array("name" => "Summary", "fieldtype" => "text", "multiline" => true, "required" => true, "FormComment" => "The main appointment description. Use tags, html is allowed.");

			//isrelationtodo to fields
			$DS[$table]["fields"]["ref_organiser"] = array("name" => "Organiser", "fieldtype" => "id", "filter" => true, "showsearchbox" => true, "use_populator_function" => true);

			$DS[$table]["fields"]["resource"] = array("name" => "Resource", "fieldtype" => "text");
			$DS[$table]["fields"]["comment"] = array("name" => "Comment", "fieldtype" => "text");
			$DS[$table]["fields"]["location"] = array("name" => "Location", "fieldtype" => "text");
			$DS[$table]["fields"]["display_popup"] = array("name" => "Appointment popups", "fieldtype" => "text", "options" => array('Yes','No'), "required" => true, "filter" => true);
			$DS[$table]["fields"]["minutes_before_csv"] = array("name" => "Popups minutes before", "fieldtype" => "text", "FormComment" => "A comma-separated list: '60,30,10'");
			$DS[$table]["fields"]["color_in_legend"] = array("name" => "Color", "fieldtype" => "text", "filter" => true, "fieldtype_in_extrafields_table" => "color");
			$DS[$table]["fields"]["listoptions"] = array("name" => "List options", "fieldtype" => "text", "multiline" => true, "FormComment" => "List-tag-style, e.g. hiddenfilter_t12-f18=Done");
			$DS[$table]["fields"]["calendar_object"] = array("name" => "Object", "fieldtype" => "text", "FormComment" => "Can be used as iCal identifier. If empty id is used.");
			$DS[$table]["fields"]["show_add_option"] = array("name" => "Show add-option in calendar", "fieldtype" => "text", "options" => array('Yes','No'), "required" => true, "filter" => true, "FormComment" => "Enable to be able to add records to this table directly from the calendar.");
			$DS[$table]["fields"]["add_option_url_extras"] = array("name" => "Extras for add-url", "fieldtype" => "text", "required" => false, "filter" => false, "FormComment" => "Add extra values to add url in this field, e.g. '&t11-f12-_new_=addedfromcalendar'. Use @CHOSENDATE@ to refer to the date chosen.");
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "calendaraccessrules";
		$DS[$table]["primary"] = "id";
		$DS[$table]["name_in_database"] = "calendaraccessrules";
		$DS[$table]["openlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = true;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Calendar access rule";
		$DS[$table]["plural"] = "Calendar access rules";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
		$DS[$table]["ShowInlineDeleteLink"] = true;
		$DS[$table]["ShowInlineDuplicateLink"] = true;
		$DS[$table]["dqs_enabled"] = true;
			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");
			$relation_to = "groups";
			$DS[$table]["fields"]["group"] = array("name" => $DS[$relation_to]["singular"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "filter" => true, "ExtraSelectConditionForSelectingNewRecords" => "`" . $relation_to . "`.`active`='yes'");
			$relation_to = "calendars";
			$DS[$table]["fields"]["calendar"] = array("name" => $DS[$relation_to]["singular"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "filter" => true);
			$DS[$table]["fields"]["read"] = array("name" => "Read", "fieldtype" => "text", "options" => array("0", "1"), "filter" => true, "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes", "fieldtype_in_extrafields_table" => "checkbox", "formelementtype" => "checkbox", "defaultvalue" => "0");

			$DS[$table]["fields"]["read"]["optionclass"]["1"] = "okmsg";
			$DS[$table]["fields"]["read"]["optionclass"]["0"] = "errmsg";

			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);


		$table = "attributes";
		$DS[$table]["primary"] = "entity";
		$DS[$table]["compoundprimarykey"] = true; //PRIMARY KEY (`entity`,`identifier`,`attribute`)
		$DS[$table]["openlink"] = "";
		$DS[$table]["addlink"] = "";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = false;
		$DS[$table]["allowed_to_delete"] = false; //because of compound key
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Attribute";
		$DS[$table]["plural"] = "Attributes";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
			$DS[$table]["fields"]["entity"] = array("name" => "Entity", "fieldtype" => "id");
			$DS[$table]["fields"]["identifier"] = array("name" => "Identifier", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["attribute"] = array("name" => "Attribute", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["value"] = array("name" => "Value", "fieldtype" => "text", "sometimes_serialized" => true);
			//deze is ALTIJD geserialized, toch?
			$DS[$table]["fields"]["allowed_values"] = array("name" => "Allowed values", "fieldtype" => "text", "serialized" => true);
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "uselog";
		$DS[$table]["primary"] = "id";
		$DS[$table]["openlink"] = "";
		$DS[$table]["addlink"] = "";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = false;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Uselog record";
		$DS[$table]["plural"] = "Uselog records";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");
			$DS[$table]["fields"]["ip"] = array("name" => "IP", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["url"] = array("name" => "URL", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["useragent"] = array("name" => "User agent", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["qs"] = array("name" => "Message", "fieldtype" => "text");
			//dit is tekst en geen ref naar loginusers
			$DS[$table]["fields"]["user"] = array("name" => $lang["username"], "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "mailqueue";
		$DS[$table]["primary"] = "queueid";
		$DS[$table]["openlink"] = "";
		$DS[$table]["addlink"] = "";
		$DS[$table]["addlinktext"] = "";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = false;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Mail queue message";
		$DS[$table]["plural"] = "Mail queue messages";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
		$DS[$table]["ShowInlineDuplicateLink"] = false;
		$DS[$table]["ShowInlineDeleteLink"] = true;
			$DS[$table]["fields"]["queueid"] = array("name" => $lang["id"], "fieldtype" => "id");

			$relation_to = "users";
			$DS[$table]["fields"]["user"] = array("name" => "User", "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "ExtraSelectConditionForSelectingNewRecords" => $ExtraSelectConditionForSelectingNewRecords_for_user_refs, "filter" => true);

			$DS[$table]["fields"]["template"] = array("name" => "Template", "fieldtype" => "text");
			$DS[$table]["fields"]["from"] = array("name" => "From", "fieldtype" => "text", "fieldsubtype" => "mail", "filter" => true);
			$DS[$table]["fields"]["fromname"] = array("name" => "From name", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["to"] = array("name" => "To", "fieldtype" => "text", "fieldsubtype" => "mail", "filter" => true);
			$DS[$table]["fields"]["PDF"] = array("name" => "PDF", "fieldtype" => "text", "options" => array("not", "used"), "filter" => true);
			$DS[$table]["fields"]["subject"] = array("name" => "Subject", "fieldtype" => "text");
			$DS[$table]["fields"]["attach_to_dossier"] = array("name" => "Attach to record", "fieldtype" => "text", "options" => array("both", "entity", "customer"), "filter" => true);
			$DS[$table]["fields"]["attach_as_filename"] = array("name" => "Attach as filename", "fieldtype" => "text");
			//dit is een text field
			$DS[$table]["fields"]["report_attach"] = array("name" => "Report attach", "fieldtype" => "text");
			//isrelationtodo
			$DS[$table]["fields"]["flextableid"] = array("name" => "Table", "fieldtype" => "id", "filter" => true);

			$DS[$table]["fields"]["flextablerecord"] = array("name" => "Record " . $lang["id"], "fieldtype" => "id");
			//$DS[$table]["fields"]["entity"] = array("name" => "Entity " . $lang["id"], "fieldtype" => "id");
			//$DS[$table]["fields"]["customer"] = array("name" => "Customer " . $lang["id"], "fieldtype" => "id");

			$DS[$table]["fields"]["status"] = array("name" => "Status", "fieldtype" => "text", "options" => array("unsent", "sent", "error"), "filter" => true);
			$DS[$table]["fields"]["date_queued"] = array("name" => "Date/time queued", "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);
			$DS[$table]["fields"]["date_sent"] = array("name" => "Date/time sent", "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);
			$DS[$table]["fields"]["worker_hash"] = array("name" => "Worker hash", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["table"] = array("name" => "Table", "fieldtype" => "text", "filter" => true);

			//zou eigenlijk subtabel moeten zijn met ref naar files
			$DS[$table]["fields"]["files_to_attach"] = array("name" => "Files to attach", "fieldtype" => "text", "serialized" => true);
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "todo";
		$DS[$table]["primary"] = "todoid";
		$DS[$table]["openlink"] = "";
		$DS[$table]["addlink"] = "";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = false;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Todo";
		$DS[$table]["plural"] = "Todo's";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
			$DS[$table]["fields"]["todoid"] = array("name" => $lang["id"], "fieldtype" => "id");
			$DS[$table]["fields"]["onchange"] = array("name" => "On change", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["from_value"] = array("name" => "From value", "fieldtype" => "text");
			$DS[$table]["fields"]["to_value"] = array("name" => "To value", "fieldtype" => "text");
			$DS[$table]["fields"]["eid"] = array("name" => "Record " . $lang["id"], "fieldtype" => "id");

			$relation_to = "users";
			$DS[$table]["fields"]["user"] = array("name" => "User", "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "ExtraSelectConditionForSelectingNewRecords" => $ExtraSelectConditionForSelectingNewRecords_for_user_refs, "filter" => true);

			$DS[$table]["fields"]["worker_hash"] = array("name" => "Worker hash", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "cache";
		$DS[$table]["primary"] = "stashid";
		$DS[$table]["openlink"] = "";
		$DS[$table]["addlink"] = "";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = false;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Cache record";
		$DS[$table]["plural"] = "Cache records";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
			$DS[$table]["fields"]["stashid"] = array("name" => $lang["id"], "fieldtype" => "id");
			$DS[$table]["fields"]["epoch"] = array("name" => "Date/time", "fieldtype" => "date", "dateformat_in_database" => "epoch", "filter" => true);
			//dit is een ref naar sessions
			$DS[$table]["fields"]["session"] = array("name" => "Session id", "fieldtype" => "text", "filter" => true);
			$DS[$table]["fields"]["value"] = array("name" => "Value", "fieldtype" => "text", "sometimes_serialized" => true);
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "stats";
		$DS[$table]["primary"] = "id";
		$DS[$table]["openlink"] = "";
		$DS[$table]["addlink"] = "";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = false;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Statistic record";
		$DS[$table]["plural"] = "Statistic records";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
			$DS[$table]["fields"]["id"] = array("name" => $lang["id"], "fieldtype" => "id");
			$DS[$table]["fields"]["item"] = array("name" => "Item", "fieldtype" => "text");
			$DS[$table]["fields"]["value"] = array("name" => "Value", "fieldtype" => "text");
			$DS[$table]["fields"]["count"] = array("name" => "Count", "fieldtype" => "number");
			$DS[$table]["fields"]["sum"] = array("name" => "Sum", "fieldtype" => "number", "decimals" => 2);
			$DS[$table]["fields"]["avg"] = array("name" => "Average", "fieldtype" => "number", "decimals" => 2);
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "datastructures";
		$DS[$table]["primary"] = "userid";
		$DS[$table]["openlink"] = "";
		$DS[$table]["addlink"] = "";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = false;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Data structure";
		$DS[$table]["plural"] = "Data structures";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$DS[$table]["open_records_in_popup_mainlist"] = 0;
		$DS[$table]["open_records_in_popup_sublist"] = 0;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
//			$relation_to = "users";
			//geen ref naar users voor veiligheid dat hij NOOIT gejoined wordt in een query
			//isrelationtodo
//			$DS[$table]["fields"]["userid"] = array("name" => "User", "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "ExtraSelectConditionForSelectingNewRecords" => $ExtraSelectConditionForSelectingNewRecords_for_user_refs, "filter" => true);
			$DS[$table]["fields"]["userid"] = array("name" => "User", "fieldtype" => "id");
			$DS[$table]["fields"]["datastructure"] = array("name" => "Data structure", "fieldtype" => "text", "serialized" => true);
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

		$table = "extendedprivsgroup";
		$DS[$table]["primary"] = "id";
		$DS[$table]["openlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlink"] = "record.php?table=" . $table . "&record=";
		$DS[$table]["addlinktext"] = "Add";
		$DS[$table]["allowed_to_write"] = true;
		$DS[$table]["allowed_to_add"] = true;
		$DS[$table]["allowed_to_delete"] = true;
		$DS[$table]["allow_direct_view"] = true;
		$DS[$table]["record_access_restrictions"]["read"] = array();
		$DS[$table]["record_access_restrictions"]["write"] = array();
		$DS[$table]["default_sortby"] = $table . "-groupid";
		$DS[$table]["default_sortdirection"] = "asc";
		$DS[$table]["IncludeInSystemWideSearches"] = false;
		$DS[$table]["singular"] = "Record access rule";
		$DS[$table]["plural"] = "Record access rules";
		$DS[$table]["layout"] = "@" . $table . "-" . $DS[$table]["primary"] . "@";
		$DS[$table]["perpage"] = 50;
		$DS[$table]["download_allowed"] = $download_allowed;
		$DS[$table]["download_limit"] = 0;
		$DS[$table]["search_parents"] = true;
		$DS[$table]["search_childs"] = false;
		$DS[$table]["HasFileAttachments"] = false;
		$DS[$table]["search_files"] = false;
		$popupsize = getpopupsize($table);
		$DS[$table]["open_records_in_popup_mainlist"] = $popupsize;
		$DS[$table]["open_records_in_popup_sublist"] = $popupsize;
		$DS[$table]["MaySelectColumns"] = $MaySelectColumns;
		$DS[$table]["ShowInlineDuplicateLink"] = true;
		$DS[$table]["ShowInlineDeleteLink"] = true;
		$DS[$table]["dqs_enabled"] = true;
			$DS[$table]["fields"]["id"] = array("name" => $lang['id'], "fieldtype" => "id");

			$relation_to = "groups";
			$DS[$table]["fields"]["groupid"] = array("name" => $DS[$relation_to]["singular"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $DS[$relation_to]["layout"], "ExtraSelectConditionForSelectingNewRecords" => "`" . $relation_to . "`.`active`='yes'", "filter" => true);

			$relation_to = "fields";
			$DS[$table]["fields"]["field"] = array("name" => $DS[$relation_to]["singular"], "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $layout_of_field_table, "ExtraSelectConditionForSelectingNewRecords" => "`" . $relation_to . "`.`deleted`='n'", "filter" => true, "showsearchbox" => true);

			$DS[$table]["fields"]["operand"] = array("name" => "Operand", "fieldtype" => "text", "options" => array("IS", "IS NOT"), "filter" => true, "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes");

			$DS[$table]["fields"]["value"] = array("name" => "Value", "fieldtype" => "text", "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes", "filter" => true);
			$DS[$table]["fields"]["rule"] = array("name" => "Rule", "fieldtype" => "text", "filter" => true, "options" => array("readonly", "invisible"), "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes");

			$relation_to = "fields";
			$DS[$table]["fields"]["only_apply_for_field"] = array("name" => "Only apply for field", "fieldtype" => "id", "relation_to" => $relation_to, "layout" => $layout_of_field_table, "ExtraSelectConditionForSelectingNewRecords" => "`" . $relation_to . "`.`deleted`='n'", "filter" => true, "showsearchbox" => true, "FormComment" => "Set to 0 to apply this privilege for the complete record");

			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);

			$DS[$table]["fields"]["visual_only"] = array("name" => "Visual only", "fieldtype" => "text", "filter" => true, "options" => array("n", "y"), "formelementtype" => "checkbox", "fieldtype_in_extrafields_table" => "checkbox", "formelementtype" => "checkbox", "defaultvalue" => "n", "FormComment" => "When checked, this rule will NOT be a security rule; javascript will be generated to hide/show this field on forms based on the condition. Any field you want to evaluate needs to be present on the form.", "interactive" => true, "interactive-alwaysinline" => true, "InteractiveDontReloadListOnChange" => "Yes");
			$DS[$table]["fields"]["visual_only"]["optionclass"]["n"] = "errmsg";
			$DS[$table]["fields"]["visual_only"]["optionclass"]["y"] = "okmsg";
	}

	$OpenAllTablesWithFormId = GetAttribute("group", "OpenAllTablesWithFormId", $GLOBALS['GROUPID']);
	if (!is_positive_integer($OpenAllTablesWithFormId))
	{
		$OpenAllTablesWithFormId = 0;
	}

	foreach (db_GetArray("SELECT `recordid`, `tablename`, `refer_field_layout`, `tableheaderrepeat`, `addlinktext`, `headerhtml`, `add_in_popup`, `sort_on`, `sort_direction`, `maxrowsperpage`, `formid`, `stayinformaftersave` FROM `CRMflextabledefs` ORDER BY `ordering`, `recordid`") as $flextable)
	{
		$table = "t" . $flextable["recordid"];
		if ($tables_allowed_to_read[$table])
		{
			$DS[$table]["layout"] = "@" . $table . "-recordid@";
			if ($flextable["refer_field_layout"] != "")
			{
				$DS[$table]["layout"] = $flextable["refer_field_layout"];
			}

			//deze support zit NIET op veld specifieke refer layouts (BlindReferenceFieldLayout); ReferLayoutLanguageENGLISH
			$language_specific_layout = GetAttribute("flextable", "ReferLayoutLanguage" . $GLOBALS["LANGUAGE"], $flextable["recordid"]);
			if ($language_specific_layout != "")
			{
				$DS[$table]["layout"] = $language_specific_layout;
			}

			$name_in_database = "flextable" . $flextable["recordid"];
			$DS[$table]["primary"] = "recordid";
			$DS[$table]["name_in_database"] = $name_in_database;

			$logical_delete = true;
			if (GetAttribute("flextable", "UsePhysicalDelete", $flextable["recordid"]) == "Yes")
			{
				$logical_delete = false;
			}
			$DS[$table]["logical_delete"] = $logical_delete;
			$DS[$table]["datatable"] = true;

			$MainTable = false;
			if (GetAttribute("flextable", "MainTable", $flextable["recordid"]) == "Yes")
			{
				$MainTable = true;
			}
			$DS[$table]["MainTable"] = $MainTable;

			$IncludeThisTableInSearchesFromParentTable = true;
			if (GetAttribute("flextable", "IncludeThisTableInSearchesFromParentTable", $flextable["recordid"]) == "No")
			{
				$IncludeThisTableInSearchesFromParentTable = false;
			}
			$DS[$table]["IncludeThisTableInSearchesFromParentTable"] = $IncludeThisTableInSearchesFromParentTable;

			$AddRowEnabledOnSublistsOnAutoForms = false;

			if (GetAttribute("flextable", "AddRowEnabledOnSublistsOnAutoForms", $flextable["recordid"]) == "Yes")
			{
				$AddRowEnabledOnSublistsOnAutoForms = true;
			}
			$DS[$table]["AddRowEnabledOnSublistsOnAutoForms"] = $AddRowEnabledOnSublistsOnAutoForms;

			if ($OpenAllTablesWithFormId > 0)
			{
				$DS[$table]["form-template-id"] = $OpenAllTablesWithFormId;
			}
			else
			{
				$DS[$table]["form-template-id"] = $flextable["formid"];
			}
			$DS[$table]["openlink"] = "record.php?table=" . $table . "&record=";
			$DS[$table]["addlink"] = "record.php?table=" . $table . "&record=";

			//interessant om hier ook op ft niveau onderscheid tussen users te kunnen maken
			$DS[$table]["allow_direct_view"] = true;

			$DS[$table]["record_access_restrictions"]["read"] = $record_access_restrictions["read"][$table];
			$DS[$table]["record_access_restrictions"]["write"] = $record_access_restrictions["write"][$table];
			$DS[$table]["default_sortby"] = $table . "-" . $DS[$table]["primary"];
			$DS[$table]["default_sortdirection"] = "asc";
			$DS[$table]["IncludeInSystemWideSearches"] = true;

			$DS[$table]["singular"] = $flextable["tablename"];
			//flextable1singular
			if ($lang[$name_in_database . 'singular'] != "")
			{
				$DS[$table]["singular"] = $lang[$name_in_database . 'singular'];
			}
			$DS[$table]["plural"] = $flextable["tablename"];
			if ($lang[$name_in_database . 'plural'] != "")
			{
				$DS[$table]["plural"] = $lang[$name_in_database . 'plural'];
			}
			$DS[$table]["headerrepeat"] = $flextable["tableheaderrepeat"];

			$DS[$table]["addlinktext"] = $flextable['addlinktext'];
			if ($lang[$name_in_database . 'add'] != "")
			{
				$DS[$table]["addlinktext"] = $lang[$name_in_database . 'add'];
			}

			$DS[$table]["allowed_to_write"] = $tables_allowed_to_write[$table];
			$DS[$table]["allowed_to_add"] = $tables_allowed_to_add[$table];
			$DS[$table]["allowed_to_delete"] = $tables_allowed_to_delete[$table];

			$DS[$table]["headerhtml"] = unescape_interleave_characters(ParseTemplateLanguageTags($flextable["headerhtml"]));

			$dqs_enabled = false;
			$dqs_fixedset = false;
			$DQSMethod = GetAttribute("flextable", "DQSMethod", $flextable["recordid"]);
			if ($DQSMethod != "Disabled")
			{
				$dqs_enabled = true;
				if ($DQSMethod == "Fixed set")
				{
					$dqs_fixedset = false;
				}
			}
			$DS[$table]["dqs_enabled"] = $dqs_enabled;
			$DS[$table]["dqs_fixedset"] = $dqs_fixedset;

			$DS[$table]["perpage"] = $flextable["maxrowsperpage"];

			$DS[$table]["default_sortby"] = $table . "-recordid";
			if (is_positive_integer($flextable['sort_on']))
			{
				$DS[$table]["default_sortby"] = $table . "-f" . $flextable['sort_on'];
			}

			if ($flextable['sort_direction'] == "Descending")
			{
				$DS[$table]["default_sortdirection"] = "desc";
			}
			else
			{
				$DS[$table]["default_sortdirection"] = "asc";
			}

			if (GetAttribute("flextable", "IncludeInSystemWideSearches", $flextable["recordid"]) == "No")
			{
				$DS[$table]["IncludeInSystemWideSearches"] = false;
			}

			$ShowInlineDuplicateLink = false;
			if (($DS[$table]["allowed_to_add"]) && (GetAttribute("flextable", "ShowInlineDuplicateLink", $flextable["recordid"]) == "Yes"))
			{
				$ShowInlineDuplicateLink = true;
			}
			$DS[$table]["ShowInlineDuplicateLink"] = $ShowInlineDuplicateLink;

			$ShowInlineDeleteLink = false;
			//check of iemand een ft mag editten; zo niet dan mag hij ook niet deleten
			if (GetAttribute("flextable", "ShowInlineDeleteLink", $flextable["recordid"]) == "Yes")
			{
				$ShowInlineDeleteLink = true;
			}
			$DS[$table]["ShowInlineDeleteLink"] = $ShowInlineDeleteLink;

			$ExtraSelectCondition = GetAttribute("flextable", "ExtraSelectCondition", $flextable["recordid"]);
			if ($ExtraSelectCondition != "")
			{
				$DS[$table]["ExtraSelectCondition"] = $ExtraSelectCondition;
				$DS[$table]["RemoveExtraSelectCondionFromQueryLinkText"] = GetAttribute("flextable", "RemoveExtraSelectCondionFromQueryLinkText", $flextable["recordid"]);
				$DS[$table]["AddExtraSelectCondionToQueryLinkText"] = GetAttribute("flextable", "AddExtraSelectCondionToQueryLinkText", $flextable["recordid"]);
			}

			$popupsize = 0;
			//dit attribuut is "misbruikt" voor de main list; zowel add als edit
			if (GetAttribute("flextable", "UsePopupAlsoWhenViewingPlainList", $flextable["recordid"]) == "Yes")
			{
				$popupsize = getpopupsize($table);
			}
			$DS[$table]["open_records_in_popup_mainlist"] = $popupsize;

			$popupsize = 0;
			//deze setting is "misbruikt" voor alle sub lijsten; zowel add als edit
			if ($flextable["add_in_popup"] == "y")
			{
				$popupsize = getpopupsize($table);
			}
			$DS[$table]["open_records_in_popup_sublist"] = $popupsize;

			$download_allowed_this_table = true;
			$download_limit = 0;

			if (!$is_admin)
			{
				if ((GetAttribute("flextable", "DenyDownloads", $flextable["recordid"]) == "Yes") || (!$download_allowed))
				{
					$download_allowed_this_table = false;
				}
				else
				{
					//er zijn downloadrules VOOR DEZE TABEL; dus niemand heeft rechten om te downloaden tenzij er een record in deze tabel staat voor deze group
					if (db_GetValue("SELECT COUNT(*) FROM `CRMdownloadrules` WHERE `table`=" . $flextable["recordid"]) > 0)
					{
						$download_allowed_this_table = false;
						$row = db_GetRow("SELECT * FROM `CRMdownloadrules` WHERE `table`=" . $flextable["recordid"] . " AND `group`=" . $GLOBALS['GROUPID']);
						//er is voor deze specifieke group een downloadrule, downloaden is dus toegestaan
						if ($row['id'] > 0)
						{
							$download_allowed_this_table = true;
							//deze kan ook 0 zijn voor alle rows
							$download_limit = $row['max_rows'];
						}
					}
				}
			}

			$DS[$table]["download_allowed"] = $download_allowed_this_table;
			$DS[$table]["download_limit"] = $download_limit;

			$search_parents = true;
			if (GetAttribute("flextable", "IncludeParentTableInSearches", $flextable["recordid"]) == "No")
			{
				$search_parents = false;
			}
			$DS[$table]["search_parents"] = $search_parents;

			$search_childs = true;
			if (GetAttribute("flextable", "IncludeChildTablesInSearches", $flextable["recordid"]) == "No")
			{
				$search_childs = false;
			}
			$DS[$table]["search_childs"] = $search_childs;

			$DS[$table]["addrow_on_bottom"] = false;
			if (GetAttribute("flextable", "AddRowLocation", $flextable["recordid"]) == "bottom")
			{
				$DS[$table]["addrow_on_bottom"] = true;
			}

			$HasFileAttachments = true;
			$search_files = true;
			if (GetAttribute("flextable", "HasFileAttachments", $flextable["recordid"]) == "No")
			{
				$HasFileAttachments = false;
				$search_files = false;
			}
			else
			{
				if (GetAttribute("flextable", "WhenSearchingAlsoSearchThroughBinaryAttachments", $flextable["recordid"]) == "No")
				{
					$search_files = false;
				}
			}
			$DS[$table]["HasFileAttachments"] = $HasFileAttachments;
			$DS[$table]["search_files"] = $search_files;

			$PartOfConfiguration = false;
			if (GetAttribute("flextable", "TableIsPartOfConfiguration", $flextable["recordid"]) == "Yes")
			{
				$PartOfConfiguration = true;

			}
			$DS[$table]["PartOfConfiguration"] = $PartOfConfiguration;

			$DoNotLock = false;
			if (GetAttribute("flextable", "DoNotLock", $flextable["recordid"]) == "Yes")
			{
				$DoNotLock = true;
			}
			$DS[$table]["DoNotLock"] = $DoNotLock;

			$MaySelectColumns_this_table = false;
			//de global var kan op ft niveau worden overrulet
			if (($is_admin) || ($MaySelectColumns))
			{
				$MaySelectColumns_this_table = true;
			}
			$DS[$table]["MaySelectColumns"] = $MaySelectColumns_this_table;

			$DS[$table]["stayinform"] = false;
			if ($flextable["stayinformaftersave"] == "y")
			{
				$DS[$table]["stayinform"] = true;
			}

			$PutCellDataInOwnDivInLists = false;
			if (GetAttribute("flextable", "PutCellDataInOwnDivInLists", $flextable["recordid"]) == "Yes")
			{
				$PutCellDataInOwnDivInLists = true;
			}
			$DS[$table]["PutCellDataInOwnDivInLists"] = $PutCellDataInOwnDivInLists;

			$DS[$table]["fields"]["recordid"] = array("name" => $lang["id"], "fieldtype" => "id");

			//deze hebben we nooit nodig in lijsten
//			$DS[$table]["fields"]["deleted"] = array("name" => "Deleted");
			$DS[$table]["fields"]["timestamp_last_change"] = array("name" => $lang["lastupdate"], "fieldtype" => "date", "dateformat_in_database" => "yyyy-mm-dd hh:mm:ss", "filter" => true);
		}
	}

	foreach (db_GetArray("SELECT `id`, `fieldtype`, `options`, `name`, `displaylistname`, `sum_column`, `number_format`, `excludefromfilters`, `israwhtml`, `underwaterfield`, `forcing`, `optioncolors`, `defaultval`, `calculation`, `calculation_type`, `showsearchbox`, `size`, `showasradio`, `sort` FROM `CRMextrafields` WHERE `deleted`='n' ORDER BY `ordering`, `id`") as $extrafield)
	{
		$table = get_ds_tablename_for_field($extrafield['id']);

		$field = "f" . $extrafield["id"];
		$name_in_database = "EFID" . $extrafield["id"];

		if (!is_array($DS[$table]))
		{
			$reasons[] = "user has NO access to table " . $table . ", so also NO access to field " . $field;
		}
		else
		{
			$allowed_to_read = true;
			$allowed_to_write = true;

			if (!$is_admin)
			{
				if (db_GetValue("SELECT COUNT(*) FROM `CRMfieldaccessrules` WHERE `field`=" . $extrafield['id']) > 0)
				{
					$allowed_to_read = false;
					$allowed_to_write = false;
					//warning triggeren op meer dan 1 record
					$row = db_GetRow("SELECT * FROM `CRMfieldaccessrules` WHERE `field`=" . $extrafield['id'] . " AND `group`='" . $GLOBALS['GROUPID'] . "' ORDER BY `id`");
					if ($row['read'] == 1)
					{
						$allowed_to_read = true;
					}
					if ($row['write'] == 1)
					{
						$allowed_to_write = true;
					}
				}
			}

			$calculation = "";
			$calculation_type = "";
			if ($extrafield["calculation"] != "")
			{
				$calculation = $extrafield["calculation"];
				//een calculation is ALTIJD readonly, zelfs voor admin
				$allowed_to_write = false;
				$calculation_type = $extrafield["calculation_type"];
				if ($calculation_type == "calculated on display")
				{
					$calculation = unescape_interleave_characters(ParseTemplateUserAndGroupFields($calculation));
				}
			}

			$reasons[] = "user access to field " . $field . " in table " . $table . " read/write is " . ($allowed_to_read ? "1" : "0") . "/" . ($allowed_to_write ? "1" : "0");

			$fieldtype = "";
			$relation_to = "";
			$ExtraSelectConditionForSelectingNewRecords = "";
			$InheritWriteAccessFromParent = true;
			//we bepalen het fieldtype en relation_to al BUITEN de lus of iemand dit veld wel mag zien vanwege de warnings die we willen geven hieronder
			if (($extrafield["fieldtype"] == "Reference to FlexTable") || ($extrafield["fieldtype"] == "Reference to FlexTable (multiselect)"))
			{
				if (is_positive_integer($extrafield["options"]))
				{
					$relation_to = "t" . $extrafield["options"];
					$fieldtype = "id";
					if (GetAttribute("extrafield", "InheritWriteAccessFromParent", $extrafield["id"]) == "No")
					{
						$InheritWriteAccessFromParent = false;
					}
				}
			}
			else if ($extrafield["fieldtype"] == "User-list of all CRM-CTT users")
			{
				$relation_to = "users";
				$fieldtype = "id";
				$ExtraSelectConditionForSelectingNewRecords = $ExtraSelectConditionForSelectingNewRecords_for_user_refs;
				$InheritWriteAccessFromParent = false;
			}
			else if (preg_match("/^Users of profile (\d+)$/", $extrafield["fieldtype"], $match))
			{
				$profile_id = $match[1];
				$relation_to = "users";
				$fieldtype = "id";
				if (is_positive_integer($profile_id))
				{
					$ExtraSelectConditionForSelectingNewRecords = "`active`='yes' AND `HIDEFROMASSIGNEEANDOWNERLISTS`='n' AND `PROFILE`=" . $profile_id;
				}
				else
				{
					$ExtraSelectConditionForSelectingNewRecords = "`active`='yes' AND `administrator`='yes'";
				}
				$InheritWriteAccessFromParent = false;
			}
			else if ($extrafield["fieldtype"] == "List of all groups")
			{
				$relation_to = "groups";
				$fieldtype = "id";
				$ExtraSelectConditionForSelectingNewRecords = "`active`='yes'";
				$InheritWriteAccessFromParent = false;
			}

/*
			if (!$allowed_to_read)
			{
				if ($relation_to != "")
				{
					$errortext = "SECURITY WARNING/MISCONFIGURATION: this user has no access to field " . $field . " which is a relation to " . $relation_to . ". *IF* there are access rules on " . $relation_to . " (or its parents) these rules are NOT applied for this user, which could mean this user sees records he/she shouldn't see!";
					log_msg("ERROR: " . $errortext);
					$reasons[] = $errortext;
				}
				//was handiger geweest als table eerste element was en read/write het tweede element
				if ((is_array($record_access_restrictions["read"][$table][$field])) || (is_array($record_access_restrictions["write"][$table][$field])))
				{
					$errortext = "SECURITY WARNING/MISCONFIGURATION: field " . $field . " has access rules defined, but these are NOT applied because this user has no access to this field. This probably means this user sees records he/she shouldn't see!";
					log_msg("ERROR: " . $errortext);
					$reasons[] = $errortext;
				}
			}
*/

			if ($allowed_to_read)
			{
				$filter = false;
				if ($extrafield['excludefromfilters'] == "n")
				{
					$filter = true;
				}

				$name = $extrafield["name"];
				if ($lang[$name] != "")
				{
					$name = $lang[$name];
				}
				//JH 2017-06-17 THIS MUST GO; LANG TAGS EFID123 ARE DEPRECATED
				else if ($lang[$name_in_database] != "")
				{
					$name = $lang[$name_in_database];
				}

				$ishtml = false;
				if (($extrafield['israwhtml'] == "y") || ($extrafield["fieldtype"] == "text area (rich text)"))
				{
					$ishtml = true;
				}

				$displaylistname = $extrafield["displaylistname"];
				if ($lang[$name . "DLN"] != "")
				{
					$displaylistname = $lang[$name . "DLN"];
				}
				//2017-06-16 THIS MUST GO EFID123 LANG TAGS ARE DEPRECATED; ONLY f123 IS SUPPORTED
				else if ($lang[$name_in_database . "DLN"] != "")
				{
					$displaylistname = $lang[$name_in_database . "DLN"];
				}

				$dateformat_in_database = "";
				$decimals = 0;
				$cast_number_to_decimal = false;
				$add_up_total = false;
				$multiselect = false;
				$hide_from_list = false;
				$hide_from_form = false;
				$serialized = false;

				//voegt css "f23" toe aan de dt/dd
				$AddIdentifierClass = false;
				if (GetAttribute("extrafield", "AddIdentifierClass", $extrafield["id"]) == "Yes")
				{
					$AddIdentifierClass = true;
				}

				$InteractiveAlwaysInline = false;
				if (GetAttribute("extrafield", "InteractiveAlwaysInline", $extrafield["id"]) == "Yes")
				{
					$InteractiveAlwaysInline = true;
				}

				$ExcludeFromSearches = false;
				if (GetAttribute("extrafield", "ExcludeFromSearches", $extrafield["id"]) == "Yes")
				{
					$ExcludeFromSearches = true;
				}

				$PerFieldSearch = false;
				if (GetAttribute("extrafield", "PerFieldSearchEnable", $extrafield["id"]) == "Yes")
				{
					$PerFieldSearch = true;
				}
				if (preg_match("/multiselect\)$/", $extrafield["fieldtype"]))
				{
					$multiselect = true;
					$serialized = true;
				}

				if ($extrafield["fieldtype"] == "diary")
				{
					$serialized = true;
				}

				$required = false;
				//2016-09-05 for admins extra fields are never required; this allows you to correct records without being bothered to set (conditional) required fields. Be careful! ValidateFieldInput must be changed as well!
				//if ((!$is_admin) && ($extrafield["forcing"] == "y"))
				if ($extrafield["forcing"] == "y")
				{
					$required = true;
				}

				$interactive = false;
				if ((in_array($extrafield["id"], $INTERACTIVEFIELDS)) || ($table == "files") || GetAttribute("extrafield", "Interactive", $extrafield["id"]) == "Yes")
				{
					if ($allowed_to_write)
					{
						$interactive = true;
					}
				}

				$RecalculateDaily = false;

				if ($extrafield["calculation"] != "")
				{
					if (GetAttribute("extrafield", "AlwaysRecalculateWhenNightlyCronIsRan", $extrafield["id"]) == "Yes")
					{
						$RecalculateDaily = true;
					}
				}

//"SQL Query","SQL Query (multiselect)" -> dit zijn eigenlijk ook id's

				$defaultvalue = $extrafield["defaultval"];

				$options = array();
				if (($extrafield["fieldtype"] == "drop-down") || ($extrafield["fieldtype"] == "drop-down (multiselect)"))
				{
					$options = unserialize($extrafield["options"]);
				}
				else if ($extrafield["fieldtype"] == "checkbox")
				{
					//in defaultval staat UNCHECKED, in options staat CHECKED; volgorde is belangrijk want make_checkbox() gaat van deze volgorde uit
					$options = array($defaultvalue, $extrafield["options"]);
					if (GetAttribute("extrafield", "CheckBoxCheckedByDefault", $extrafield["id"]) == "Yes")
					{
						$defaultvalue = $extrafield["options"];
					}
				}

				//alleen in gebruik voor numeric
				$prefix = "";
				//een fieldpostfix, zoals "kg" of "m" of "m3"; nog niet in gebruik
				$postfix = "";

				$multiline = false;
				if (($extrafield["fieldtype"] == "text area") || ($extrafield["fieldtype"] == "text area (rich text)"))
				{
					$multiline = true;
				}

				$fieldsubtype = "";
				if ($extrafield["fieldtype"] == "mail")
				{
					$fieldsubtype = "mail";
				}
				else if ($extrafield["fieldtype"] == "hyperlink")
				{
					$fieldsubtype = "link";
				}
				else if ($extrafield["fieldtype"] == "phone number")
				{
					$fieldsubtype = "tel";
				}

				if (($extrafield["fieldtype"] == "textbox") || ($extrafield["fieldtype"] == "text area") || ($extrafield["fieldtype"] == "drop-down") || ($extrafield["fieldtype"] == "mail") || ($extrafield["fieldtype"] == "hyperlink") || ($extrafield["fieldtype"] == "diary") || ($extrafield["fieldtype"] == "drop-down (populate by code)") || ($extrafield["fieldtype"] == "text area (rich text)") || ($extrafield["fieldtype"] == "SQL Query") || ($extrafield["fieldtype"] == "color") || ($extrafield["fieldtype"] == "phone number") || ($extrafield["fieldtype"] == "checkbox"))
				{
					//een diary is niet echt een text field maar een serialized array; op deze manier wordt hij in ieder geval doorzocht
					$fieldtype = "text";
				}
				else if ($extrafield["fieldtype"] == "numeric")
				{
					if ($extrafield["number_format"] == "don't format") //als ie op don't format staat dan blijft cast_number_to_decimal op false waardoor hij niet gecast wordt
					{
						$fieldtype = "text";
					}
					else
					{
						$fieldtype = "number";
						if (GetAttribute("extrafield", "NumericDatabaseField", $extrafield["id"]) != "Yes")
						{
							$cast_number_to_decimal = true;
						}
						//no decimals
						if ((preg_match("/^currency/", $extrafield["number_format"])) || ($extrafield["number_format"] == "normal"))
						{
							$decimals = 2;
						}
						else if (preg_match("/^(\d+) decimal/", $extrafield["number_format"], $match))
						{
							$decimals = $match[1];
						}
						if ($extrafield["sum_column"] == "yes")
						{
							$add_up_total = true;
						}

						if ($extrafield["number_format"] == "currency EUR")
						{
							$prefix = "€";
						}
						else if ($extrafield["number_format"] == "currency DOL")
						{
							$prefix = "$";
						}
						else if ($extrafield["number_format"] == "currency BRP")
						{
							$prefix = "£";
						}
						else if ($extrafield["number_format"] == "currency YEN")
						{
							$prefix = "¥";
						}
					}
				}
				else if ($extrafield["fieldtype"] == "date")
				{
					$fieldtype = "date";
					$dateformat_in_database = "dd-mm-yyyy";
				}
				else if ($extrafield["fieldtype"] == "date (iso)")
				{
					$fieldtype = "date";
					$dateformat_in_database = "yyyy-mm-dd";
				}
				else if ($extrafield["fieldtype"] == "date/time")
				{
					$fieldtype = "date";
					$dateformat_in_database = "yyyy-mm-dd hh:mm:ss";
				}
				else if ($extrafield["fieldtype"] == "Button")
				{
					$hide_from_list = true;
				}
				else if ($extrafield["fieldtype"] == "comment")
				{
					$hide_from_list = true;
				}

				if ($extrafield["underwaterfield"] == "y")
				{
					//2014-12-07
					if (!$is_admin)
					{
						$hide_from_list = true;
						$hide_from_form = true;
					}
				}

				$IgnoreCellClick = false;
				if (GetAttribute("extrafield", "IgnoreCellClick", $extrafield["id"]) == "Yes")
				{
					$IgnoreCellClick = true;
				}

				$DS[$table]["fields"][$field]["name"] = $name;
				$DS[$table]["fields"][$field]["name_in_database"] = $name_in_database;
				$DS[$table]["fields"][$field]["allowed_to_write"] = $allowed_to_write;

				//alleen voor selects gebruiken. nog check op maken?
				$showsearchbox = false;
				if ($extrafield["showsearchbox"] == "y")
				{
					$showsearchbox = true;
				}

				//overrulet de dynamisch opgebouwde listname
				$DS[$table]["fields"][$field]["displaylistname"] = $displaylistname;
				$DS[$table]["fields"][$field]["fieldtype"] = $fieldtype;
				$DS[$table]["fields"][$field]["fieldsubtype"] = $fieldsubtype;
				$DS[$table]["fields"][$field]["fieldtype_in_extrafields_table"] = $extrafield["fieldtype"];
				$DS[$table]["fields"][$field]["IgnoreCellClick"] = $IgnoreCellClick;
				$DS[$table]["fields"][$field]["filter"] = $filter;
				$DS[$table]["fields"][$field]["ishtml"] = $ishtml;
				$DS[$table]["fields"][$field]["multiselect"] = $multiselect;
				$DS[$table]["fields"][$field]["serialized"] = $serialized;
				$DS[$table]["fields"][$field]["hide_from_list"] = $hide_from_list;
				$DS[$table]["fields"][$field]["hide_from_form"] = $hide_from_form;
				$DS[$table]["fields"][$field]["PerFieldSearch"] = $PerFieldSearch;
				$DS[$table]["fields"][$field]["required"] = $required;
				$DS[$table]["fields"][$field]["defaultvalue"] = $defaultvalue;
				$DS[$table]["fields"][$field]["options"] = $options;
				$DS[$table]["fields"][$field]["interactive"] = $interactive;
				$DS[$table]["fields"][$field]["interactive-alwaysinline"] = $InteractiveAlwaysInline;
				$DS[$table]["fields"][$field]["ExcludeFromSearches"] = $ExcludeFromSearches;
				$DS[$table]["fields"][$field]["AddIdentifierClass"] = $AddIdentifierClass;
				$DS[$table]["fields"][$field]["multiline"] = $multiline;
				$DS[$table]["fields"][$field]["calculation"] = $calculation;
				$DS[$table]["fields"][$field]["calculation_type"] = $calculation_type;
				$DS[$table]["fields"][$field]["RecalculateDaily"] = $RecalculateDaily;
				$DS[$table]["fields"][$field]["FormComment"] = unescape_interleave_characters(ParseTemplateLanguageTags(GetAttribute("extrafield", "FormComment", $extrafield["id"])));
				$DS[$table]["fields"][$field]["showsearchbox"] = $showsearchbox;
				$DS[$table]["fields"][$field]["CustomValidationFunctionPHP"] = GetAttribute("extrafield", "CustomValidationFunctionPHP", $extrafield["id"]);
				$DS[$table]["fields"][$field]["field_access_restrictions"]["read"] = array();
				$DS[$table]["fields"][$field]["field_access_restrictions"]["write"] = array();


				//2017-08-21; these fields are NOT hidden; so you can view user fields from other tables (if you have access)
/*				if ($table == "users")
				{
					$DS[$table]["fields"][$field]["hide_field_when_in_child_table"] = true;
				}*/
				if (GetAttribute("extrafield", "allowed_to_change_once_set", $extrafield["id"]) == "No") {

					if (!$is_admin)
					{
						$DS[$table]["fields"][$field]["field_access_restrictions"]["write"][$field]["IS"][] = "";
					}
				}

				// 2015-10-26 extra velden en attributen zodat record en form-functions zelf niet meer hoeft te querien
				if (($extrafield["fieldtype"] == "textbox") || ($extrafield["fieldtype"] == "mail") || ($extrafield["fieldtype"] == "hyperlink") || ($extrafield["fieldtype"] == "phone number") || ($extrafield["fieldtype"] == "numeric") || ($extrafield["fieldtype"] == "date") || ($extrafield["fieldtype"] == "date (iso)") || ($extrafield["fieldtype"] == "date/time"))
				{
					//dit zou ook met css kunnen
					$DS[$table]["fields"][$field]["size_width"] = $extrafield["size"];
				}
				else if (($extrafield["fieldtype"] == "text area") || ($extrafield["fieldtype"] == "text area (rich text)"))
				{
					//the size and with are stored in the field options syntax "columns:rows" ("40:4")
					//dit zou ook met css kunnen
					list($columns, $rows) = explode(":", $extrafield['options']);
					$DS[$table]["fields"][$field]["size_width"] = $columns;
					$DS[$table]["fields"][$field]["size_height"] = $rows;
				}

				//we slaan het formelementtype alleen op voor velden die een beperkt aantal keuzes hebben, niet voor textvelden
				if (($extrafield["fieldtype"] == "drop-down") || ($extrafield["fieldtype"] == "drop-down (multiselect)"))
				{
					$DS[$table]["fields"][$field]["formelementtype"] = "select";
				}
				else if ($extrafield["fieldtype"] == "checkbox")
				{
					$DS[$table]["fields"][$field]["formelementtype"] = "checkbox";
				}
				else if ($relation_to != "")
				{
					$DS[$table]["fields"][$field]["formelementtype"] = "select";

					//'display as text, not as dropdown' voor refs, dit veld is misbruikt
					if ($extrafield["sort"] == "y")
					{
						$DS[$table]["fields"][$field]["formelementtype"] = "list";
					}
				}
				else if ($extrafield["fieldtype"] == "date/time")
				{
					$DS[$table]["fields"][$field]["ShowSeconds"] = false;
					if (GetAttribute("extrafield", "ShowSeconds", $extrafield["id"]) == "Yes")
					{
						$DS[$table]["fields"][$field]["ShowSeconds"] = true;
					}
				}

				if (($DS[$table]["fields"][$field]["formelementtype"] == "select") && ($extrafield['showasradio'] == "y"))
				{
					//hiermee zou een ref-to-flex ook als list of checkboxes (of radio buttons) weergegeven kunnen worden, daar is nu nog geen support voor in form-functions maar dat zou wel gemaakt kunnen worden als we het ooit nodig hebben
					$DS[$table]["fields"][$field]["formelementtype"] = "checkbox";
				}

				if ($extrafield["fieldtype"] == "comment")
				{
					$DS[$table]["fields"][$field]["comment_template_id"] = $extrafield['options'];
				}


				//boolean attributen
				//InteractiveDontReloadListOnChange zouden we ook moeten kunnen beredeneren; als er geen calculations en geen access rules van dit veld afhankelijk zijn dan hoeven we geen listreload te doen. voor nu is het een handmatig te zetten attribuut
				foreach (array("InteractiveDontReloadListOnChange", "AutoSizeOnInput", "UsePlanningCalendar", "ShowFieldHistory", "AutoUpdateDiaryField") as $attrib)
				{
					if (GetAttribute("extrafield", $attrib, $extrafield["id"]) == "Yes")
					{
						$DS[$table]["fields"][$field][$attrib] = true;
					}
				}

				//string attributen
				foreach (array("Placeholder", "MaximumLength", "ListOptionsForChildTable", "InlineAddIconURLExtras", "ButtonClickConfirmMessage", "MaximumValue", "MinimumValue", "RefreshByAjaxOnChangeOfField", "FieldsOnFormWhichMustHaveAValueCSV", "FieldsToAskInJSPopupCSV") as $attribute)
				{
					$str = GetAttribute("extrafield", $attribute, $extrafield["id"]);
					if ($str != "")
					{
						//DIT IS HEEL LOMP; MAAR ListOptionsForChildTable is een misnomer, het zou ListOptionsForParentTable moeten zijn
						if ($attribute == "ListOptionsForChildTable")
						{
							$attribute = "ListOptionsForParentTable";
						}
						$DS[$table]["fields"][$field][$attribute] = $str;
					}
				}
				if ($DS[$table]["fields"][$field]["ButtonClickConfirmMessage"] != "")
				{
					$DS[$table]["fields"][$field]["ButtonClickConfirmMessage"] = unescape_interleave_characters(ParseTemplateLanguageTags($DS[$table]["fields"][$field]["ButtonClickConfirmMessage"]));
				}

				// 2015-10-26 EIND


				$optioncolors = array();
				if ($extrafield['optioncolors'] != '')
				{
					$optioncolors = unserialize($extrafield['optioncolors']);
				}

				if (count($optioncolors) > 0)
				{
					foreach (array_keys($optioncolors) as $key)
					{
						$color = "";
						$operand = "";
						$value = "";
						if (is_array($optioncolors[$key]))
						{
							$color = $optioncolors[$key]["color"];
							$operand = $optioncolors[$key]["operand"]; //operand is LT, GT of HAS
							$value = $optioncolors[$key]["select"];
							if ($extrafield["fieldtype"] == "date")
							{
								$value = NLDate2Epoch($value);
							}
						}
						else
						{
							$color = $optioncolors[$key];
							$operand = "IS";
							$value = $key;
						}

						if (substr($color,0,1) == "#")
						{
							$DS[$table]["fields"][$field]["colors"][$operand][$value]["background"] = strtolower($color);

							$classname = ReturnClassnameForTextColorBasedOnBackgroundColor($color);
							if ($classname == "white")
							{
								$DS[$table]["fields"][$field]["colors"][$operand][$value]["foreground"] = "#ffffff";
							}
							else
							{
								$DS[$table]["fields"][$field]["colors"][$operand][$value]["foreground"] = "#000000";
							}
						}
						else
						{
							//2016-06-06 ontwikkeld; optionclasses
							//nadelen
							//- dit werkt NIET in Excel's
							//- geen support voor $operand
							//- geen support voor "hele hele hele vieze uitzondering om alleen kleur te leggen op het eigen ref veld als die wijst naar hetzelfde doelveld, om "M" wel te kleuren en "de heer" niet in dezelfde flextable"
							//voordelen
							//- je 1 class name kan gebruiken voor dezelfde kleur op meerdere plekken
							//- je kan afwijken van de standaard witte of zwarte tekst (dat is soms nodig)
							//- je kan meer styling meegeven, font-weight, font-size, borders, etc.
							$DS[$table]["fields"][$field]["optionclass"][$value] = $color;
						}
					}
				}

				if ($fieldtype == "date")
				{
					$DS[$table]["fields"][$field]["dateformat_in_database"] = $dateformat_in_database;
				}
				if ($fieldtype == "number")
				{
					$DS[$table]["fields"][$field]["decimals"] = $decimals;
					$DS[$table]["fields"][$field]["cast_number_to_decimal"] = $cast_number_to_decimal;
					$DS[$table]["fields"][$field]["add_up_total"] = $add_up_total;
					$DS[$table]["fields"][$field]["prefix"] = $prefix;
					$DS[$table]["fields"][$field]["postfix"] = $postfix;
				}

				if ($relation_to != "")
				{
					$DS[$table]["fields"][$field]["relation_to"] = $relation_to;

					//we nemen de standaard layout voor deze tabel
					$DS[$table]["fields"][$field]["layout"] = $DS[$relation_to]["layout"];

					//tenzij er een BlindReferenceFieldLayout is
					$BlindReferenceFieldLayout = GetAttribute("extrafield", "BlindReferenceFieldLayout", $extrafield["id"]);
					if ($BlindReferenceFieldLayout != "")
					{
						$DS[$table]["fields"][$field]["layout"] = $BlindReferenceFieldLayout;
					}

					if ($ExtraSelectConditionForSelectingNewRecords != "")
					{
						//deze wordt gebruikt voor het selecteren van nieuwe waardes in nieuwe records; hij kan zo gelimiteerd worden op "alleen mensen in dienst", oude records die verwijzen naar iemand die inmiddels "uit dienst is", houden hun referentie. deze mag NOOIT verwijderd worden
						$DS[$table]["fields"][$field]["ExtraSelectConditionForSelectingNewRecords"] = $ExtraSelectConditionForSelectingNewRecords;
					}

					$DS[$table]["fields"][$field]["InheritWriteAccessFromParent"] = $InheritWriteAccessFromParent;
				}
			}
		}
	}

	//coditionele field access
	if (!$is_admin)
	{
		if (is_positive_integer($GLOBALS['GROUPID']))
		{
			//zelfde code staat 2x in lf, moet nog samenvoegen
			foreach (db_GetArray("SELECT `field`, `operand`, `value`, `rule`, `only_apply_for_field` FROM `CRMextendedprivsgroup` WHERE `groupid`=" . $GLOBALS['GROUPID'] . " AND `only_apply_for_field`>0 AND `visual_only`='n'") as $row)
			{
				$table = get_ds_tablename_for_field($row['field']);
				$field = "f" . $row['field'];
				$value = $row['value'];
				$operand = $row['operand'];

				$type_of_access = "read"; //invisible
				if ($row['rule'] == "readonly")
				{
					$type_of_access = "write";
				}

				//LET OP: precies tegenovergesteld opgeslagen in record_access_restrictions want die houdt bij wat je WEL mag zien
				if ($operand == "IS NOT")
				{
					$operand = "IS";
				}
				else
				{
					$operand = "NOT";
				}

				$value = unescape_interleave_characters(ParseTemplateUserAndGroupFields($value));

				//2015-08-31
				if ($value == "")
				{
					if ($DS[$table]["fields"][$field]["fieldtype"] == "date")
					{
						//dit matcht ook op een date/time veld in mysql waar eigenlijk 0000-00-00 00:00:00 in staat
						$value = "0000-00-00";
					}
					else if ($DS[$table]["fields"][$field]["fieldtype"] == "number")
					{
						$value = 0;
					}
				}

				$only_apply_for_field = "f" . $row['only_apply_for_field'];
				$only_apply_for_field_table = get_ds_tablename_for_field($row['only_apply_for_field']);

				//field en only_apply_for_field moeten in dezelfde tabel zitten
				if ($table == $only_apply_for_field_table)
				{
					if (is_array($DS[$table]["fields"][$field]))
					{
						if (is_array($DS[$only_apply_for_field_table]["fields"][$only_apply_for_field]))
						{
							$DS[$only_apply_for_field_table]["fields"][$only_apply_for_field]["field_access_restrictions"][$type_of_access][$field][$operand][] = $value;
						}
						else
						{
							$errortext = "ERROR: user has no access to only_apply_for_field " . $only_apply_for_field . " therefore this security setting is NOT applied. Either give this user access to this field, or remove this rule.";
							log_msg($errortext);
						}
					}
					else
					{
						$errortext = "ERROR: user has no access to field " . $field . " AND THE TABLE IT RESIDES IN (" . $table . ") therefore this security setting is NOT applied. Either give this user access to this field and table, or remove this rule.";
						log_msg($errortext);
					}
				}
				else
				{
					//record_access_restrictions defineren altijd velden IN DE EIGEN TABEL. zou het ook mogelijk moeten zijn om te zeggen dat je een Order niet mag zien wanneer de bovenliggende Klant een bepaalde status heeft? dat kan nu niet
					$errortext = "ERROR: no support for access rules on table A (" . $table . ") defined by a field in table B (" . $only_apply_for_field . " in table " . $only_apply_for_field_table . ")";
					log_msg($errortext);
				}
			}
		}
	}

	foreach (array_keys($DS) as $table)
	{
		foreach (array_keys($DS[$table]["fields"]) as $field)
		{
			if (($DS[$table]["fields"][$field]["relation_to"] != "") && (!array_key_exists("InheritWriteAccessFromParent", $DS[$table]["fields"][$field])))
			{
				$DS[$table]["fields"][$field]["InheritWriteAccessFromParent"] = true;
			}
		}
	}

	//HIER DOEN WE OVERERVENDE TABLE ACCESS read EN write
	//hier lussen we door alle velden van alle tabellen heen; als een relation veld verwijst naar een tabel die niet bestaat dan betekent dat waarschijnlijk dat deze user de parent tabel niet mag zien. in dat geval mag hij de child ook niet zien
	//het kan ook zijn dat er nog een oud relation veld bestaat; daardoor raak je de onderliggende tabellen kwijt, maar dat is terecht
	//we moeten blijven lussen zolang er changes zijn; een kleinkind zou de eerste lus namelijk niet verwijderd worden als je wel toegang tot het kind hebt maar niet tot de grootouder
	$changes_made = true;
	while ($changes_made)
	{
		$changes_made = false;
		foreach (array_keys($DS) as $table)
		{
//deze bestaat hier nog niet omdat we de volgorde hebben omgedraaid, we moeten hier dus de fields doorlopen
//volgorde omgedraaid van bepalen van OVERERVENDE TABLE ACCESS (eerst) en opbouwen van cache van childs/parents van een tabel (daarna). Anders zouden er relics achter kunnen blijven in de parents/childs cache wat tot errors zou kunnen leiden. DIT IS POTENTIEEL EEN GEVAARLIJKE CHANGE
//			foreach (array_keys($DS[$table]["parents"]) as $parent_table)
			foreach (array_keys($DS[$table]["fields"]) as $field)
			{
				$parent_table = $DS[$table]["fields"][$field]["relation_to"];
				if ($parent_table != "")
				{
					if (!is_array($DS[$parent_table]))
					{
						$reasons[] = "User has NO access to parent table " . $parent_table . " so also NO access to child table " . $table;
						unset($DS[$table]);
						$changes_made = true;
						continue; //we skippen de rest van de fields lus; die bestaat nu immers toch niet meer
					}
					else if ((!$DS[$parent_table]["allowed_to_write"]) && ($DS[$table]["fields"][$field]["InheritWriteAccessFromParent"]) && ($DS[$table]["fields"][$field]["required"]))
					{
						if ($DS[$table]["allowed_to_write"])
						{
							$reasons[] = "User has readonly access to parent table " . $parent_table . ", InheritWriteAccessFromParent is true and " . $table . "-" . $field . " is required so also readonly access to child table " . $table;
							$DS[$table]["allowed_to_write"] = false;
							$changes_made = true;
						}
						if ($DS[$table]["allowed_to_add"])
						{
							$reasons[] = "User has readonly access to parent table " . $parent_table . ", InheritWriteAccessFromParent is true and " . $table . "-" . $field . " is required so also not allowed to add records to child table " . $table;
							$DS[$table]["allowed_to_add"] = false;
							$changes_made = true;
						}
					}
				}
			}
		}
	}

	$perpage_options = array();
	foreach (explode(",", GetSetting("PAGINATEOPTIONSLIST")) as $option)
	{
		if (is_positive_integer($option))
		{
			$perpage_options[] = $option;
		}
	}

	foreach (array_keys($DS) as $table)
	{
		if (!$DS[$table]["allowed_to_add"])
		{
			$DS[$table]["ShowInlineDuplicateLink"] = false;
		}
		if (!$DS[$table]["allowed_to_delete"])
		{
			$DS[$table]["ShowInlineDeleteLink"] = false;
		}

		//deze 2 velden zijn bedoeld als cache, om sneller relaties op te kunnen halen ipv door alle velden heen te lussen
		//we gebruiken deze vars pas de lus hierop, maar we moeten hem hier al initialiseren zodat ie ook voor de parent al bestaat
		$DS[$table]["parents"] = array(); //from_this_table
		$DS[$table]["childs"] = array(); //to_this_table

		//we houden per tabel de per options bij en zetten hem in de ds voor meer snelheid in make_list
		$perpage_options_this_table = $perpage_options;

		//de perpage setting van deze tabel hoeft niet per se in de PAGINATEOPTIONSLIST te staan; het zou beter zijn om dit wel te forceren
		$perpage_options_this_table[] = $DS[$table]["perpage"];
		$perpage_options_this_table = array_unique($perpage_options_this_table);
		asort($perpage_options_this_table);
		$DS[$table]["perpage_options"] = $perpage_options_this_table;

		//safety: dit zou nooit voor mogen komen
		if (!array_key_exists("allowed_to_write", $DS[$table]))
		{
			$DS[$table]["allowed_to_write"] = false;
			$errortext = "ERROR: allowed_to_write not defined for table " . $table;
			log_msg($errortext);
		}

		//dit is eigenlijk een calculation en is dus nooit *rechtstreeks* beschrijfbaar voor een user
		$DS[$table]["fields"][$DS[$table]["primary"]]["allowed_to_write"] = false;

		//deze bestaat altijd toch?
		if (is_array($DS[$table]["fields"]["timestamp_last_change"]))
		{
			//dit is eigenlijk een calculation en is dus nooit *rechtstreeks* beschrijfbaar voor een user
			$DS[$table]["fields"]["timestamp_last_change"]["allowed_to_write"] = false;
		}

		foreach (array_keys($DS[$table]["fields"]) as $field)
		{
			//indien de key niet bestaat overerft hij hem uit de tabel
			//OF als de tabel niet writeable is dan overruled die het veld
			if ((!array_key_exists("allowed_to_write", $DS[$table]["fields"][$field])) || (!$DS[$table]["allowed_to_write"]))
			{
				$DS[$table]["fields"][$field]["allowed_to_write"] = $DS[$table]["allowed_to_write"];
				$reasons[] = "field " . $field . " in table " . $table . " inherited write access from table: " . ($DS[$table]["fields"][$field]["allowed_to_write"] ? "1" : "0");
			}

			if (!$DS[$table]["fields"][$field]["allowed_to_write"])
			{
				//als je niet mag writen mag je hem ook niet verwijderen
				$DS[$table]["fields"][$field]["allowed_to_delete"] = false;
			}

			if ((!$DS[$table]["fields"][$field]["allowed_to_write"]) && ($DS[$table]["fields"][$field]["interactive"]))
			{
				$DS[$table]["fields"][$field]["interactive"] = false;
				$reasons[] = "field " . $field . " in table " . $table . " interactive is set to FALSE because not allowed to write this field";
			}

			//buttons bevatten geen informatie. als je NOOIT op een button mag klikken dan kunnen we hem dus net zo goed verbergen (van de forms) om het cleaner te houden en geen verwarring te zaaien (mensen die misschien denken dat de security niet op orde is omdat ze een gegrijsde Save button zie)
			if ((!$DS[$table]["fields"][$field]["allowed_to_write"]) && ($DS[$table]["fields"][$field]["fieldtype_in_extrafields_table"] == "Button"))
			{
				$reasons[] = "field " . $field . " in table " . $table . " is a button, this button is never clickable for this user, so we remove it from its data structure";
				unset($DS[$table]["fields"][$field]);
			}
		}
	}

	$autosave = false;
	$access_dependent_on = array();
	if (GetSetting("USE_AUTOSAVE") == "Yes")
	{
		$autosave = true;
		if (!$is_admin)
		{
			if (is_positive_integer($GLOBALS['GROUPID']))
			{
				//heel slordig om hier weer te moeten querien op `CRMextendedprivsgroup`, maar het is complex om het resultaat van eerdere queries op deze tabel te hergebruiken (omdat die zowel voor een tabel als ook voor een enkel veld kan gelden)
				$access_dependent_on = db_GetFlatArray("SELECT DISTINCT(`field`) FROM `CRMextendedprivsgroup` WHERE `groupid`=" . $GLOBALS['GROUPID'] . " AND `visual_only`='n'");
			}
		}
	}

	//hier vullen we altijd de name_in_database var
	//EN cachen we alle relaties. dat maakt een hoop code sneller
	foreach (array_keys($DS) as $table)
	{
		//DIT IS wat slordig; maar hierdoor is in ieder geval altijd de var name_in_database gevuld en dat scheelt een hoop IF's elders in de code; dit is alleen nodig vanwege de vertaling flextable4 -> t4 en EFID29 -> f29
		if ($DS[$table]["name_in_database"] == "")
		{
			$DS[$table]["name_in_database"] = $table;
		}

		//default 2016-02-25
		if (!array_key_exists("IncludeThisTableInSearchesFromParentTable", $DS[$table]))
		{
			$DS[$table]["IncludeThisTableInSearchesFromParentTable"] = true;
		}

		foreach (array_keys($DS[$table]["fields"]) as $field)
		{
			if ($DS[$table]["fields"][$field]["name_in_database"] == "")
			{
				$DS[$table]["fields"][$field]["name_in_database"] = $field;
			}

			$autosave_this_field = $autosave;
			if ($autosave_this_field)
			{
				if (!$DS[$table]["fields"][$field]["allowed_to_write"])
				{
					$autosave_this_field = false;
				}
				else if (preg_match("/^f(\d+)$/", $field, $match))
				{
					if (GetAttribute("extrafield", "DisableAutoSave", $match[1]) == "Yes")
					{
						$autosave_this_field = false;
					}
					else if (in_array($match[1], $access_dependent_on))
					{
						//deze warning is tijdelijk voor debugging
						//log_msg("WARNING: autosave disabled for f" . $match[1] . " because access is dependent on this field");
						$autosave_this_field = false;
					}
				}
			}
			//autosave wordt op alle velden gezet; dus ook op primary id velden waar helemaal niks te autosaven valt. is niet erg.
			$DS[$table]["fields"][$field]["autosave"] = $autosave_this_field;

			$parent_table = $DS[$table]["fields"][$field]["relation_to"];

			if ($parent_table != "")
			{
				//field is ALTIJD opgeslagen in de child
				//er kunnen meerdere links vanuit 1 tabel naar de andere tabel liggen
				$DS[$table]["parents"][$parent_table][] = $field;

				//deze persoon heeft ALTIJD toegang tot de parent, want anders was de child er eerder al uitgegooid, deze conditie is dus niet per se nodig
				if (is_array($DS[$parent_table]))
				{
					//er kunnen meerdere links vanuit 1 tabel naar de andere tabel liggen
					$DS[$parent_table]["childs"][$table][] = $field;
				}
			}
		}
	}

	if (GetSetting("UseURLBeautification") == "Yes")
	{
		foreach (array_keys($DS) as $table)
		{
			if (preg_match("/^record\.php/", $DS[$table]["openlink"]))
			{
				$DS[$table]["openlink"] = $DS[$table]["plural"] . "/";
			}
			if (preg_match("/^record\.php/", $DS[$table]["addlink"]))
			{
				$DS[$table]["addlink"] = $DS[$table]["plural"] . "/0?";
			}
		}
	}

	return array($DS, $reasons);
}
function make_query($table, $search = "", $searchperfield = array(), $start = 0, $perpage = 0, $sortby = "", $sortdirection = "asc", $filtercolumn = "", $filter = array(), $filtertype = array(), $listlayout = array(), $RemoveExtraSelectCondition = false, $connecting_field = "", $use_pre_select_records = false, $pre_select_records = array(), $order_by_pre_select = false, $search_parents = true, $search_childs = true, $search_files = true, $return_countquery = false, $type_of_list = "mainlist", $bypass_record_access_by_sql = false, $tabletype = "normal", $childtable = "", $childfield = "", $childrecordid = 0, $already_joined = array(), $hiddenfilter = array(), $hiddenfiltertype = array(), $listid = "", $do_not_check_write_access = false)
{
	//listid kan nu ook worden meegegeven; dit is binnen make_query() nergens in gebruik, maar kan gebruikt worden door de eval() code om wel of niet te runnen

	$sortby_db_arr = array();

	//de function call is in string context; hier omzetten naar array; niet zo sjiek, maar anders moet hoop code aangepast worden
	$sortby_arr = explode(",", $sortby);
	$sortdirection_arr = explode(",", $sortdirection);

	if (!is_positive_integer($start))
	{
		$start = 0;
	}
	if (!is_positive_integer($perpage))
	{
		$perpage = 0;
	}

	$table_alias = "";
	//dit zijn velden die we afbeelden of nodig hebben in de html
	$fields_to_select = array();
	//dit zijn velden die impact hebben op de resultset, bijvoorbeeld filters (die niet afgebeeld worden) of sorts
	$fields_to_search_or_filter = array();

	if (!$return_countquery)
	{
		//we selecteren ALTIJD de primary key; ook als die niet wordt afgebeeld in de lijst, moeten we hem hebben om het record te kunnen openen/selecteren. Indien het primary id al genoemd was in de fields_to_select dan doet deze code niks
		$fields_to_select[$table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"]] = 1;
	}

	if ($tabletype == "select")
	{
		//hiddenfilters in ListOptionsForParentTable are enforced as security; other listoptions are NOT security; we need to change this by moving the hiddenfilters to a separate add-child-security-table
		$ListOptionsForParentTable = $GLOBALS['UC']['PersonalDataStructure'][$childtable]["fields"][$childfield]["ListOptionsForParentTable"];
		//dit hele blok code staat zowel in make_query als ook in make_list, maar wel met net andere logica; het doel is hetzelfde; om hiddenfilters die in ListOptionsForParentTable staan ofwel uit de GET ofwel server side te halen. Hierdoor is het ook security.
		if ($ListOptionsForParentTable != "")
		{
			//LET OP: het parsen van ListOptionsForParentTable is heel exotisch, we callen ParseTemplateRecordData() zonder recordid maar MET recordid_to_preload. Het gevolg is dat de huidige database waarde worden ingeladen van de recordid_to_preload (== het record) en dat die worden aangevuld/overschreven met waarden uit de GET. Hierdoor kan een ListOptionsForParentTable zowel data uit de db gebruiken als ook een nieuw gekozen waarde op het formulier wat je aan het invullen bent (bijvoorbeeld afhankelijk van de keuze van contactpersonen die werken bij het bedrijf wat tevens geselecteerd staat) N.B.: dit zou beter omgezet kunnen worden tot een selectvia constructie

			parse_str($ListOptionsForParentTable, $ListOptionsForParentTable_unparsed_array);
			$ListOptionsForParentTable = unescape_interleave_characters(ParseTemplateRecordData($ListOptionsForParentTable, $childtable, 0, $childrecordid, true)); //do_not_check_write_access=true
			parse_str($ListOptionsForParentTable, $ListOptionsForParentTable_array);

			foreach (array_keys($ListOptionsForParentTable_unparsed_array['hiddenfilter']) as $key)
			{
				$tags_in_filter = false;
				foreach ($ListOptionsForParentTable_unparsed_array['hiddenfilter'][$key] as $tmp_filter)
				{
					//dit is extreem slordig; er zit minimaal 1 @ of % tag in een filter; in dat geval gebruiken we de waarde zoals de client die doorstuurt
					if (count(get_tags($tmp_filter)) > 0)
					{
						$tags_in_filter = true;
						break;
					}
				}

				//als er tags in de filters staan, dan moeten we de waarde gebruiken zoals die gepost zijn door de client, zo niet dan MOETEN we de server variabelen gebruiken want anders zou de client een LOFPT security setting kunnen overschrijven
				//als er niks van de client ontvangen is, dan worden sowieso de server variabelen gebruikt
				if (($tags_in_filter) && (array_key_exists($key, $hiddenfilter)))
				{
					//we gebruiken de waarde zoals de client die doorstuurt; die staat reeds al in het hiddenfilter; bijvoorbeeld een hiddenfilter op een bepaald bedrijf bij de selectie van een contactpersoon zodat je alleen contactpersonen ziet die bij dat bedrijf werken
//					print $key . " CLIENT<br>";
				}
				else
				{
					//we overschrijven het hiddenfilter met de server variabele, want er zitten geen tags in dit filter. Dit is bijvoorbeeld een filter dat er voor zorgt dat er alleen maar orders onder actieve klanten kunnen worden toegevoegd. we willen niet dat iemand met de query fiddelt en een hiddenfilter op INACTIEVE klanten meestuurt zodat zij daar dan wel orders voor kan aanmaken
//					print $key . " SERVER<br>";

					$hiddenfilter[$key] = array();
					$hiddenfiltertype[$key] = "IS";

					if ($ListOptionsForParentTable_array['hiddenfiltertype'][$key] != "")
					{
						$hiddenfiltertype[$key] = $ListOptionsForParentTable_array['hiddenfiltertype'][$key];
					}

					foreach ($ListOptionsForParentTable_array['hiddenfilter'][$key] as $tmp_filter)
					{
						$hiddenfilter[$key][] = $tmp_filter;
					}
				}
			}
		}
	}

	//we moeten de kolom waarop we willen sorteren altijd selecteren; hiermee is het mogelijk om te sorteren op een kolom die je niet afbeeldt. niet nodig voor count query.
	if ((!$return_countquery) && (count($sortby_arr) > 0))
	{
		for ($x = 0; $x < count($sortby_arr); $x++)
		{
			$sortby = $sortby_arr[$x];
			$table_parts = explode("-", $sortby);
			$last_field = $table_parts[count($table_parts)-1];
			$last_table = $table_parts[count($table_parts)-2];
			$relation_to = $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["relation_to"];
			//we sorteren op een relatie veld; dan nemen we de default sort van de tabel waar we naar verwijzen en NIET de alfanumerieke geparsde waarde. Hiermee is het mogelijk om een volgorde veld te definieren in de bovenliggende kolom terwijl de refer layout een tekstuele waarde bevat
			//2017-05-22 multiselect fields we sort on the serialized value; this is not very useful, but better than not sorting at all
			if (($relation_to != "") && (!$GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["multiselect"]))
			{
				//de default sortby bevat altijd de fully qualified name in de eigen tebale. het is niet mogelijk om een tabel te sorteren by default op een veld in een andere tabel
				$sortby .= "-" . $GLOBALS['UC']['PersonalDataStructure'][$relation_to]["default_sortby"];
				$sortby_arr[$x] = $sortby;
			}

			$fields_to_search_or_filter[$sortby] = 1;
		}
	}

	//0 = normaal filter, 1 = hiddenfilter
	//omdat er zowel een normaal filter als ook een hidden filter op hetzelfde veld kan liggen, moeten we ze los behandelen.
	//deze situatie kan nu: een hiddenfilter op status != 3 en een filter op status = 1
	$allfilter = array();
	$allfilter[0] = $filter;
	$allfilter[1] = $hiddenfilter;
	$allfiltertype = array();
	$allfiltertype[0] = $filtertype;
	$allfiltertype[1] = $hiddenfiltertype;

	//deze velden moeten geselecteerd worden want anders kan er verderop niet op gefilterd worden
	foreach (array(0, 1) as $normal_or_hidden)
	{
		foreach (array_keys($allfilter[$normal_or_hidden]) as $key)
		{
			foreach ($allfilter[$normal_or_hidden][$key] as $key2 => $filterval)
			{
				if (is_numeric($key2))
				{
					$fields_to_search_or_filter[$key] = 1;
				}
				//3x = want key2 kan ook numeriek zijn en dat matcht anders ook. waarom?
				else if (($key2 === "from") || ($key2 === "to"))
				{
					if ($filterval != "")
					{
						$fields_to_search_or_filter[$key] = 1;
					}
				}
			}
		}
	}

	foreach ($searchperfield as $searchperfieldkey => $searchperfieldvalue)
	{
		if ($searchperfieldvalue != "")
		{
			$fields_to_search_or_filter[$searchperfieldkey] = 1;
		}
	}

	if (!$bypass_record_access_by_sql)
	{
		foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["allfields"] as $allfield)
		{
			$table_parts = explode("-", $allfield);
			$last_field = $table_parts[count($table_parts)-1];
			$last_table = $table_parts[count($table_parts)-2];

			$first_field = $table_parts[1];
			$first_table = $table_parts[0];

			$addfield = false;

			//er ligt security op dit veld; we moeten hem daarom selecteren
			if (is_array($GLOBALS['UC']['PersonalDataStructure'][$last_table]["record_access_restrictions"]["read"][$last_field]))
			{
				$addfield = true;
			}

			//er ligt misschien write security op dit veld; maar we zijn niet geinteresseerd in write access
			else if (!$do_not_check_write_access)
			{
				//TODO: Hier zouden we nog kunnen checken voor write access of de hele tabel readonly is, dan hoeven we geen velden te adden waar security op ligt; ik heb dit eerder geprobeerd maar dat leidde tot query errors; die heeft ook te maken met de rules op [recordid][] = 0 => wat nooit waar kan zijn

				if (is_array($GLOBALS['UC']['PersonalDataStructure'][$last_table]["record_access_restrictions"]["write"][$last_field]))
				{
					$addfield = true;
					//het is een veld in een hogergelegen tabel; write access zou wel eens niet van toepassing kunnen zijn
					if (count($table_parts) > 2)
					{
						//van 0 t/m count -2 zijn het ref velden; het laatste veld is geen ref veld en die checken we dus ook niet op InheritWriteAccessFromParent
						for ($x = 0; $x < count($table_parts) - 2; $x += 2)
						{
							$last_table2 = $table_parts[$x];
							$last_field2 = $table_parts[$x+1];
							//write access overerft tenzij InheritWriteAccessFromParent == false voor 1 van de tussengelegen relaties
							if (!$GLOBALS['UC']['PersonalDataStructure'][$last_table2]["fields"][$last_field2]["InheritWriteAccessFromParent"])
							{
								$addfield = false;
								break;
							}
						}
					}
				}
			}

			//dit is lomp met die addfield maar wel lekker leesbaar
			if ($addfield)
			{
				//already joined is ALLEEN gevuld in make_query() wanneer we een child search doen, in dat geval willen we geen security opleggen die in already_joined staat, want daar komen we vandaan en die tabel wordt niet opnieuw gejoined om circulaire joins te voorkomen
				if ($already_joined[$first_table][$first_field] == 0)
				{
					//we moeten deze velden hier toevoegen zodat alle tabellen naar boven gejoind worden
					$fields_to_search_or_filter[$allfield] = 1;
				}
			}
		}
	}

	//dit zijn alleen de velden die we daadwerkelijk afbeelden in de lijst en die door de gebruiker opgevraagd worden. ALS daar een relation veld bij zit dan moeten we ook de velden ophalen die deel zijn van zijn refer field layout
	foreach ($listlayout as $key)
	{
		//in geval van een reference layout selecteren we alle velden die daarin voorkomen ook mee
		//TODO: eigenlijk hadden deze velden al meegegeven moeten worden aan de listlayout array, want nu selecteren we ook wel eens te veel velden die we misschien helemaal niet nodig hebben; dit kan dus sneller
		$table_parts = explode("-", $key);
		$last_field = $table_parts[count($table_parts)-1];
		$last_table = $table_parts[count($table_parts)-2];

		//buttons en comments bevatten geen data en hoeven dus ook niet uit de database geselecteerd te worden
		//2015-09-20 ze worden nu wel geselecteerd omdat de conditionele write access van deze "velden" ook geselecteerd moet worden *t1-f2 dus. als we ze hier niet selecteren dan zou dat namelijk niet gebeuren. in de toekomst willen we de db velden weggooien (want ze bevatten toch geen data), maar willen we wel de * velden selecteren
//		if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype_in_extrafields_table"] != "Button") && ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype_in_extrafields_table"] != "comment"))
		{
			//20150726 DIT IS DE sum in sql van alle pages, staat nu uit, omdat dit lang duurt top grote lijsten. TODO: op een countonly list tag hoef je geen add up total te doen
			if ($return_countquery)
			{
				/*
				if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["add_up_total"])
				{
					//TODO check moet dit associatief blijven?
					$fields_to_select[$key] = 1;
				}
				*/
			}
			else
			{
				//TODO check moet dit associatief blijven?
				$fields_to_select[$key] = 1;

				if ($already_joined[$last_table][$last_field] == 0)
				{
					if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["layout"] != "")
					{
						$key_zonder_huidige_tabel_en_field = preg_replace("/" . $last_table . "\-" . $last_field . "$/", "", $key);

						foreach ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["layout_tags"] as $tag)
						{
							//we moeten hier het volledige path voorzetten want layout kent niet de lagerliggende tabel waarin wij nu (eventueel) zitten
							$fullkey = $key_zonder_huidige_tabel_en_field . $tag;
							$fields_to_select[$fullkey] = 1;
						}
					}
				}
			}
		}
	}

	if ($return_countquery)
	{
		//2016-01-08 dit is een dirty hack; omdat er anders geen 1 veld geselecteerd wordt op een countquery wanneer je een ongefilterde lijst opvraagt, waardoor make_query_parts() ALLE velden gaat selecteren en de count query onnodig traag wordt
		$fields_to_select[$table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"]] = 1;
	}

	//dit is een beetje lomp; dit zijn velden waarop alleen gezocht, gefilterd of gesorteerd wordt maar die niet worden weergegeven
	$fields_NOT_to_select = array();
	foreach (array_keys($fields_to_search_or_filter) as $field)
	{
		if (!array_key_exists($field, $fields_to_select))
		{
			//nodig voor make_query_parts
			$fields_to_select[$field] = 1;
			//nodig voor de lus daaronder om dat veld niet daadwerkelijk te selecteren maar wel op te kunnen filters of zoeken
			$fields_NOT_to_select[$field] = 1;
		}
	}

	//safety mysql kan maximaal 61 tabellen joinen in een enkele query
	$max_tables_to_join = 61;

	//connecting_field bevindt zich altijd in de table die wordt aangeroepen in make_query_parts
	//2013-11-02 already_joined wordt nu ook meegegeven aan make_query_parts() die is alleen gevuld wanneer we een child search doen vanuit make_query() lager in deze functie

	//fields_to_select bepaald welke velden er geselcteerd worden
	//die retourneert in $query_parts["select_field"] en DAAR wordt de security er op gelegd. als een veld dus niet in fields_to_select staat dan wordt security NIET toegepast. dat is potentieel gevaarlijk
	$query_parts = make_query_parts($table, $table_alias, $already_joined, $fields_to_select, $search, $max_tables_to_join);

	//hier worden ALLE where rules ingezet; die helemaal onderaan aan $query worden toegevoegd
	$wheres = array();

	//list modifier
	$module_to_run = GetSetting("ModuleToEvaluateOnEveryListBuild");
	if (is_positive_integer($module_to_run))
	{
		//kan niet met runmodule draaien; want hij moet in zelfde scope draaien
//		RunModule($module_to_run);
		if (CheckModuleAccess($module_to_run) == "ok")
		{
			//deze kan velden toevoegen die beginnen met ShowThisField
			eval("?>" . db_GetValue("SELECT `module_code` FROM `CRMmodules` WHERE `mid`=" . $module_to_run));
		}
	}

	//2017-12-05 we also check if this is a flextable because otherwise deleted extrafields can never be shown
	if ((preg_match("/^t\d+$/", $table)) && ($GLOBALS['UC']['PersonalDataStructure'][$table]["logical_delete"]))
	{
		$wheres[] = "`" . $table . "`.`deleted`='n'";
	}

	$fields_to_select_counter = 0;

	$query = "SELECT ";
	if ($return_countquery)
	{
		$query .= "COUNT(*) AS `number_of_records`";
		$fields_to_select_counter++;
	}

	//wanneer een child gejoined wordt (connecting_field != "") dan willen we * selecteren voor de EXISTS () query
	if ($connecting_field != "")
	{
		$query .= "*";
		$connecting_table = $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$connecting_field]["relation_to"];
		$wheres[] = return_database_fieldname($table, $connecting_field, $table) . "=`" . $connecting_table . "`.`" . $GLOBALS['UC']['PersonalDataStructure'][$connecting_table]["primary"] . "`";
	}
	else
	{
		$from_to_done = array();
		for ($x = 0; $x < count($query_parts["select_field"]); $x++)
		{
			foreach ($sortby_arr as $sortby)
			{
				if ($query_parts["select_alias"][$x] == $sortby)
				{
					//sortby_db is gelijk aan sortby, maar is nu gevalideerd; verderop waar we de daadwerkelijke ORDER BY opmaken zetten we de alias weer om naar database naam, omdat we zouden kunnen sorteren op een veld wat we niet selecteren
					$sortby_db_arr[$sortby] = $query_parts["select_alias"][$x];
				}
			}

			if ((!$return_countquery) && (!array_key_exists($query_parts["select_alias"][$x], $fields_NOT_to_select)))
			{
				if (array_key_exists($query_parts["select_alias"][$x], $fields_to_select))
				{
					if ($fields_to_select_counter > 0)
					{
						$query .= ",";
					}

					if ($return_countquery)
					{
						$query .= "SUM(";
					}
					$query .= $query_parts["select_field"][$x];
					if ($return_countquery)
					{
						$query .= ")";
					}
					$query .= " AS `" . $query_parts["select_alias"][$x] . "`";

					$fields_to_select_counter++;
				}
			}
			$filters_to_apply = array();

			//voor filter queries nemen we de andere filter mee maar de EIGEN filter niet; dan zou namelijk de multiselect checkboxen geen functie meer hebben
			if ($filtercolumn == $query_parts["select_alias"][$x])
			{
				$filtercolumn = $query_parts["select_field"][$x];
				$query .= ",COUNT(*) AS `number_of_records`";
			}
			else
			{
				$filters_to_apply[] = 0;
			}
			//het hiddenfilter moet er altijd opgelegd worden ook als we de NORMALE filter query voor deze kolom maken
			$filters_to_apply[] = 1;

			foreach ($filters_to_apply as $normal_or_hidden)
			{
				if (count($allfilter[$normal_or_hidden][$query_parts["select_alias"][$x]]) > 0)
				{
					$table_parts = explode("-", $query_parts["select_alias"][$x]);
					$last_field = $table_parts[count($table_parts)-1];
					$last_table = $table_parts[count($table_parts)-2];

					//CODE DIE HIER VEEL OP LIJKT STAAT OP 5 PLEKKEN; number_of_filters_set, HIGHLIGHT TH EN HIGHLIGHT TD, HUMAN_READABLE_FILTER EN DE OPBOUW VAN SQL QUERIES OP BASIS VAN FILTERS
					//in onderstaande code lopen we de VELDEN door en checken we of er filters op gezet zijn. op de andere plekken lopen we de FILTERS door. net andersom dus

					if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["multiselect"])
					{
						foreach ($allfilter[$normal_or_hidden][$query_parts["select_alias"][$x]] as $filterval)
						{
							$where = $query_parts["select_field"][$x];
							if ($allfiltertype[$normal_or_hidden][$query_parts["select_alias"][$x]] == "NOT")
							{
								$where .= " NOT";
							}

							$searchstr = $filterval;
							//TODO checken of dit goed is; in json moet double quote en backslash geescapet worden; deze chars komen nooit voor in base64 maar we moeten eigenlijk van die base64 af, vandaar deze codering
		//					$searchstr = preg_replace("/\\/", "\\\\", $searchstr);
		//					$searchstr = preg_replace("/\"/", "\\\"", $searchstr);
							$where .= " LIKE '%\"" . mres($searchstr) . "\"%'";
							$wheres[] = $where;
						}
					}
					else
					{
						if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "date")
						{
							//is er niet een global in interleave waarin de datum staat? TODO
							$current_date = date("Y-m-d");
							list($current_year, $current_month, $current_day) = explode("-", $current_date);

							$this_query_part = "";
							foreach ($allfilter[$normal_or_hidden][$query_parts["select_alias"][$x]] as $key2 => $filterval)
							{
								$from = "";
								$to = "";

								//dit is een regulier filter TODAY, PAST, FUTURE etc. de bijzonder from -> to filters staan hieronder
								if (is_numeric($key2))
								{
									if ($filterval == "PAST")
									{
										$to = date("Y-m-d", strtotime($current_date . " -1 day"));
									}
									else if ($filterval == "FUTURE")
									{
										$from = date("Y-m-d", strtotime($current_date . " +1 day"));
									}
									else if (preg_match("/^(\d{4})$/", $filterval, $match))
									{
										$from = $match[1] . "-01-01";
										$to = $match[1] . "-12-31";
									}
									else if ($filterval == "")
									{
										$from = "0000-00-00";
										$to = "0000-00-00";
									}
									else if (preg_match("/^(L|T|N)(D|W|M|Q|Y)((W|M|Q)(\d+))?$/", $filterval, $match))
									{
										$this_last_or_next = $match[1];
										$day_week_month_quarter_or_year = $match[2];

										//alleen in combinatie met day_week_month_quarter_or_year == Y
										$week_month_or_quarter = $match[4];
										$which_number = $match[5];

										if ($day_week_month_quarter_or_year == "D")
										{
											$from = $current_date;
											//dit is slordig dat we de strtotime functie nodig hebben; voor maand, kwartaal en jaar berekening doen we het namelijk zelf
											if ($this_last_or_next == "L")
											{
												$from = date("Y-m-d", strtotime($from . " -1 day"));
											}
											else if ($this_last_or_next == "N")
											{
												$from = date("Y-m-d", strtotime($from . " +1 day"));
											}
											$to = $from;
										}
										else if ($day_week_month_quarter_or_year == "W")
										{
											//we laten de rest door de standaard jaar functie afhandelen; zie comment bij month
											$week_month_or_quarter = $day_week_month_quarter_or_year;
											$day_week_month_quarter_or_year = "Y";

											$day_of_week = date("N", strtotime($current_date));
											$mondaydate = date("Y-m-d", strtotime($current_date . " -" . ($day_of_week - 1) . " day"));

											if ($this_last_or_next == "L")
											{
												$mondaydate = date("Y-m-d", strtotime($mondaydate . " -7 day"));
											}
											else if ($this_last_or_next == "N")
											{
												$mondaydate = date("Y-m-d", strtotime($mondaydate . " +7 day"));
											}

											//de week kan in het vorige of het volgende jaar vallen; dit kan OOK wanneer de huidge week wordt opgevraagd
											$this_last_or_next = "T";
											$which_number = date("W", strtotime($mondaydate));
											if (date("o", strtotime($mondaydate)) < $current_year)
											{
												$this_last_or_next = "L";
											}
											else if (date("o", strtotime($mondaydate)) > $current_year)
											{
												$this_last_or_next = "N";
											}
										}
										else if ($day_week_month_quarter_or_year == "M")
										{
											//we laten de rest door de standaard jaar functie afhandelen; TM (IN APRIL) is hetzelfe als TYM4, NM (IN december) is hetzelfde als NYM1, LM (in januari) is hetzelfe als LYM12
											$week_month_or_quarter = $day_week_month_quarter_or_year;
											$day_week_month_quarter_or_year = "Y";
											$which_number = $current_month;
											if ($this_last_or_next == "L")
											{
												$which_number--;
												$this_last_or_next = "T";
												if ($which_number == 0)
												{
													$which_number = 12;
													$this_last_or_next = "L";
												}
											}
											else if ($this_last_or_next == "N")
											{
												$which_number++;
												$this_last_or_next = "T";
												if ($which_number == 13)
												{
													$which_number = 1;
													$this_last_or_next = "N";
												}
											}
										}
										else if ($day_week_month_quarter_or_year == "Q")
										{
											//we laten de rest door de standaard jaar functie afhandelen; zie comment bij month
											$week_month_or_quarter = $day_week_month_quarter_or_year;
											$day_week_month_quarter_or_year = "Y";
											$which_number = 1;
											if ($current_month > 9)
											{
												$which_number = 4;
											}
											else if ($current_month > 6)
											{
												$which_number = 3;
											}
											else if ($current_month > 3)
											{
												$which_number = 2;
											}

											if ($this_last_or_next == "L")
											{
												$which_number--;
												$this_last_or_next = "T";
												if ($which_number == 0)
												{
													$which_number = 4;
													$this_last_or_next = "L";
												}
											}
											else if ($this_last_or_next == "N")
											{
												$which_number++;
												$this_last_or_next = "T";
												if ($which_number == 5)
												{
													$which_number = 1;
													$this_last_or_next = "N";
												}
											}
										}

										if ($day_week_month_quarter_or_year == "Y")
										{
											$start_month = 1;
											$end_month = 12;

											$year = $current_year;
											if ($this_last_or_next == "L")
											{
												$year--;
											}
											else if ($this_last_or_next == "N")
											{
												$year++;
											}

											if ($week_month_or_quarter == "M")
											{
												$start_month = $which_number;
												$end_month = $which_number;
											}
											else if ($week_month_or_quarter == "Q")
											{
												$start_month = 1;
												$end_month = 3;
												if ($which_number == 2)
												{
													$start_month = 4;
													$end_month = 6;
												}
												else if ($which_number == 3)
												{
													$start_month = 7;
													$end_month = 9;
												}
												else if ($which_number == 4)
												{
													$start_month = 10;
													$end_month = 12;
												}
											}

											if ($week_month_or_quarter == "W")
											{
												$from = year_weeknumber_to_mondaydate($year, $which_number);
												//het is inclusief, dus slechts 6 dagen optellen en niet 7
												$to = date("Y-m-d", strtotime($from . " +6 day"));
											}
											else
											{
												$from = sprintf("%04d-%02d-%02d", $year, $start_month, 1);
												$to = sprintf("%04d-%02d-%02d", $year, $end_month, cal_days_in_month(CAL_GREGORIAN, $end_month, $year));
											}
										}
									}
								}
								//3x = want key2 kan ook numeriek zijn en dat matcht anders ook. waarom?
								else if (($key2 === "from") || ($key2 === "to"))
								{
									//er zijn 2 filter keys voor numberic values; to en from; als we er niet 1 skippen dan worden deze waarden 2x afgebeeld als human filter
									if (!$from_to_done[$query_parts["select_alias"][$x]])
									{
										$from = $allfilter[$normal_or_hidden][$query_parts["select_alias"][$x]]["from"];
										//NL formaat; we moeten US formaat nog ondersteunen! TODO
										if (preg_match("/^(\d{2})\-(\d{2})\-(\d{4})$/", $from, $match))
										{
											$from = $match[3] . "-" . $match[2] . "-" . $match[1];
										}
										//TD -> This Day
										else if (preg_match("/^TD((\-|\+)(\d+)(DAY|WEEK|MONTH|YEAR))?$/", $from, $match))
										{
											$sign = $match[2];
											$amount = $match[3];
											$timetype = $match[4];

											$from = $current_date;
											if ($sign != "")
											{
												$tmpsql = "SELECT DATE_ADD('" . $from . "', INTERVAL ";
												if ($sign == "-")
												{
													$tmpsql .= "-";
												}
												$tmpsql .= $amount . " " . $timetype . ")";
												$from = db_GetValue($tmpsql);
											}
										}
										if (!preg_match("/^\d{4}\-\d{2}\-\d{2}( \d{2}\:\d{2}\:\d{2})?$/", $from))
										{
											$from = "";
										}

										$to = $allfilter[$normal_or_hidden][$query_parts["select_alias"][$x]]["to"];
										//NL formaat; we moeten US formaat nog ondersteunen! TODO
										if (preg_match("/^(\d{2})\-(\d{2})\-(\d{4})$/", $to, $match))
										{
											$to = $match[3] . "-" . $match[2] . "-" . $match[1];
										}
										//TD -> This Day
										else if (preg_match("/^TD((\-|\+)(\d+)(DAY|WEEK|MONTH|YEAR))?$/", $to, $match))
										{
											$sign = $match[2];
											$amount = $match[3];
											$timetype = $match[4];

											$to = $current_date;
											if ($sign != "")
											{
												$tmpsql = "SELECT DATE_ADD('" . $to . "', INTERVAL ";
												if ($sign == "-")
												{
													$tmpsql .= "-";
												}
												$tmpsql .= $amount . " " . $timetype . ")";
												$to = db_GetValue($tmpsql);
											}
										}
										if (!preg_match("/^\d{4}\-\d{2}\-\d{2}( \d{2}\:\d{2}\:\d{2})?$/", $to))
										{
											$to = "";
										}

										$from_to_done[$query_parts["select_alias"][$x]] = true;
									}
								}

								if (($from != "") || ($to != ""))
								{
									if ($this_query_part != "")
									{
										$this_query_part .= " OR ";
									}

									if (($from != "") && (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["dateformat_in_database"] == "yyyy-mm-dd hh:mm:ss") || ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["dateformat_in_database"] == "epoch")))
									{
										if (strlen($from) == 10)
										{
											$from .= " 00:00:00";
										}
									}
									if (($to != "") && (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["dateformat_in_database"] == "yyyy-mm-dd hh:mm:ss") || ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["dateformat_in_database"] == "epoch")))
									{
										if (strlen($to) == 10)
										{
											$to .= " 23:59:59";
										}
									}

									$this_query_part .= "(";

									if ($from == $to)
									{
										$this_query_part .= $query_parts["select_field"][$x] . "='" . $from . "'";
									}
									else if (($from != "") && ($to != ""))
									{
										$this_query_part .= $query_parts["select_field"][$x] . ">='" . $from . "' AND " . $query_parts["select_field"][$x] . "<='" . $to . "'";
									}
									else if ($from != "")
									{
										$this_query_part .= $query_parts["select_field"][$x] . ">='" . $from . "'";
									}
									else if ($to != "")
									{
										//om 0000-00-00 en 0000-00-00 00:00:00 te excluden uit de lijst; anders zie je ook alle records met een niet ingevulde datum als je bv PAST selecteert
 //dit ging mis met GetCalendarEvents
 //										$this_query_part .= $query_parts["select_field"][$x] . "<='" . $to . "' AND " . $query_parts["select_field"][$x] . "!=0";
										$this_query_part .= $query_parts["select_field"][$x] . "<='" . $to . "'";
									}

									$this_query_part .= ")";
								}
							}

							if ($this_query_part != "")
							{
								$where = "";
								if ($allfiltertype[$normal_or_hidden][$query_parts["select_alias"][$x]] == "NOT")
								{
									$where .= " NOT";
								}
								$where .= " (" . $this_query_part . ")";
								$wheres[] = $where;
							}
						}
						else if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "number") || ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "id"))
						{
							$this_query_part = "";
							foreach ($allfilter[$normal_or_hidden][$query_parts["select_alias"][$x]] as $key2 => $filterval)
							{
								//dit is een regulier filter, de bijzondere from -> to filters staan hieronder
								if (is_numeric($key2))
								{
									if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "id")
									{
										//filterval kan de waarde _new_ bevatten.
										//dit maakt de query (in m.): `t1`.`EFID2` IN ('_new_')
										//Mysql retourneert hier dan alle records met waarde 0. Geeft wel deze Warning (Code 1292): Truncated incorrect DOUBLE value: '_new_'
										//we willen helemaal geen enkel record terugzien, vandaar:
										if (!is_positive_integer_or_zero($filterval))
										{
											$filterval = -1;
										}
									}
//TODO: change to recordid IN (2, 4, 6, 12) -> this is much faster than repeated OR's
									//list($valid, $filterval_valid) = is_valid_number($filterval);
									//if ($valid)
									//{
										if ($this_query_part != "")
										{
											$this_query_part .= " OR ";
										}
										//$this_query_part .= " (" . $query_parts["select_field"][$x] . "=" . $filterval_valid . ")";
										$this_query_part .= " (" . $query_parts["select_field"][$x] . "='" . mres($filterval) . "')";
									//}
								}
								//3x = want key2 kan ook numeriek zijn en dat matcht anders ook. waarom?
								else if (($key2 === "from") || ($key2 === "to"))
								{
									//er zijn 2 filter keys voor numberic values; to en from; als we er niet 1 skippen dan worden deze waarden 2x afgebeeld als human filter
									if (!$from_to_done[$query_parts["select_alias"][$x]])
									{
										//deze condities worden herhaald voor de human readable filter
										list($fromvalid, $from) = is_valid_number($allfilter[$normal_or_hidden][$query_parts["select_alias"][$x]]["from"]);
										list($tovalid, $to) = is_valid_number($allfilter[$normal_or_hidden][$query_parts["select_alias"][$x]]["to"]);

										if (($fromvalid) || ($tovalid))
										{
											if ($this_query_part != "")
											{
												$this_query_part .= " OR ";
											}

											$this_query_part .= " (";

											if (($fromvalid) && ($tovalid))
											{
												if ($from == $to)
												{
													$this_query_part .= $query_parts["select_field"][$x] . "=" . $from;
												}
												else
												{
													$this_query_part .= $query_parts["select_field"][$x] . ">=" . $from . " AND " . $query_parts["select_field"][$x] . "<=" . $to;
												}
											}
											else if ($fromvalid)
											{
												$this_query_part .= $query_parts["select_field"][$x] . ">=" . $from;
											}
											else if ($tovalid)
											{
												$this_query_part .= $query_parts["select_field"][$x] . "<=" . $to;
											}
											$this_query_part .= ")";

											$from_to_done[$query_parts["select_alias"][$x]] = true;
										}
									}
								}
							}

							if ($this_query_part != "")
							{
								$where = "";
								if ($allfiltertype[$normal_or_hidden][$query_parts["select_alias"][$x]] == "NOT")
								{
									$where .= " NOT";
								}
								$where .= " (" . $this_query_part . ")";
								$wheres[] = $where;
							}

						}
						else
						{
							$select_null = false;

							$tmp_filter_query = "";

							//code herschrijven dat wanneer er maar 1 waarde is dan we = gebruiken ipv in(), dat lijkt namelijk *iets* sneller te zijn
							foreach ($allfilter[$normal_or_hidden][$query_parts["select_alias"][$x]] as $filterval)
							{
								if ($tmp_filter_query != "")
								{
									$tmp_filter_query .= ",";
								}
								//filterval kan ook leeg zijn

								$tmp_filter_query .= "'" . mres($filterval) . "'";
								if ($filterval == "")
								{
									$select_null = true;
								}
							}
							if ($tmp_filter_query != "")
							{
								$where = "";
								if ($allfiltertype[$normal_or_hidden][$query_parts["select_alias"][$x]] == "NOT")
								{
									$where .= " NOT";
								}
								$where .= " (" . $query_parts["select_field"][$x] . " IN (" . $tmp_filter_query . ")";
								if ($select_null)
								{
									$where .= " OR " . $query_parts["select_field"][$x] . " IS NULL";
								}
								$where .= ")";

								$wheres[] = $where;
							}
						}
					}
				}
			}
		}
	}

	$lockquery = "";
	$type_of_access_query = "";
	//voor een select table kijken we niet naar locks; want je mag er wel een nieuw record onderhangen als child
	//files kunnen nooit gelockt zijn. als je met record.php?table=files&record=100 een file opent dan wordt er wel een lock record geinsert (denk ik) maar we negeren deze hier
	//2016-12-01: we kijken alleen voor tabellen waar je in mag schrijven of het record gelockt is; als je niet mag schrijven dan interesseert het je namelijk niet of hij gelockt is; en het gaat je ook niks aan
	if ((GetSetting("EnableEntityLocking") == "Yes") && ($tabletype != "select") && ($table != "files") && ($GLOBALS['UC']['PersonalDataStructure'][$table]["allowed_to_write"]) && (!$GLOBALS['UC']['PersonalDataStructure'][$table]["DoNotLock"]))
	{
		//2015-05-12; vanaf nu staat de DS naam in locktable; "t4" dus ipv "4"
		$lockquery = "(SELECT `CRMentitylocks`.`lockby` FROM `CRMentitylocks` WHERE `CRMentitylocks`.`lockon`=`" . $table . "`.`" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"] . "` AND `CRMentitylocks`.`lockby`!='" . mres($GLOBALS['USERID']) . "' AND `CRMentitylocks`.`locktable`='" . mres($table) . "' LIMIT 1)";
		$type_of_access_query .= "," . $lockquery . " AS `record_locked_by`";
	}

	$write_restrictions = array();

	if (!$bypass_record_access_by_sql)
	{
		//ok voor read/write access
		//readonly voor read access
		$type_of_access_query .= ",";

		//we hoeven in dit geval ook de lockquery niet nogmaals te runnen, we WETEN immers al dat alle records in deze tabel ALTIJD readonly zijn
		if (!$GLOBALS['UC']['PersonalDataStructure'][$table]["allowed_to_write"])
		{
			$type_of_access_query .= "'readonly'";
			$write_restrictions[] = "FALSE";
		}
		else
		{
//2017-04-26 in case "type_of_access" == "ok" we also need to check if "record_locked_by" == 0
/*			if ($lockquery != "")
			{
				$write_restrictions[] = $lockquery . " IS NULL";
			}*/

			$restrictions = get_recursive_access_querypart($table, "", "write", $already_joined);
			if ($restrictions != "")
			{
				$write_restrictions[] = $restrictions;
			}

			if (count($write_restrictions) > 0)
			{
				$type_of_access_query .= "IF(" . implode(" AND ", $write_restrictions) . ",'ok','readonly')";
			}
			else
			{
				$type_of_access_query .= "'ok'";
			}
		}

//2017-04-26 a better name would be "recordwritable" with 0 or 1 as possible values
		$type_of_access_query .= " AS `type_of_access`";

		for ($x = 0; $x < count($query_parts["select_field"]); $x++)
		{
			$table_parts = explode("-", $query_parts["select_alias"][$x]);
			//we kijken alleen naar de velden in de eigen tabel; een make_query() zegt dus NOOIT iets over de writability van parent records, die zouden zo maar readonly kunnen zijn terwijl het child record writable is
			if (count($table_parts) == 2)
			{
				$last_table = $table_parts[0];
				$last_field = $table_parts[1];

				//als het hele record readonly is, dan zijn alle velden dat ook, dus moeten we al die condities per veld opnieuw runnen. dit is relatief duur; daarom doen we dat niet anders wordt ook de lockquery voor ieder veld uit de eigen tabel opnieuw gedraaid. mysql zou intelligent genoeg moeten zijn om hem maar 1x uit te voeren, maar het levert enorm lange queries op. mooier nog zou zijn als we het derived field (calc) `type_of_access` zouden kunnen gebruiken IN deze select, maar dat kan niet in mysql. de correcte werking bij het gebruik van een variabel @type_of_access IN DE SELECT is niet gegarandeerd door mysql. we doen daarom iets meer in php:
				//type_of_access -> ok/readonly, t1-f2 -> "abc def", *t1-f2 -> 0/1
				//om te weten of t1-f2 writable is moet type_of_access == ok, record_locked_by == 0 en *t1-f2 == 1 zijn

				$field_access_restrictions = array();

				//dit veld is NOOIT writable
				if (!$GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["allowed_to_write"])
				{
					$field_access_restrictions[] = "FALSE";
				}
				else
				{
//MERGEN IN apply_conditional_read_access_to_field() !!
					//we moeten zowel read als write access controleren. als een veld conditioneel niet leesbaar is, dan is ie namelijk ook niet schrijfbaar
					foreach (array("read", "write") as $type_of_access)
					{
						if (count($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["field_access_restrictions"][$type_of_access]) > 0)
						{
							foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["field_access_restrictions"][$type_of_access]) as $restricting_field)
							{
								foreach (array("IS", "NOT") as $operand)
								{
									$select_field = return_database_fieldname($last_table, $restricting_field, $last_table);
									$select_field = convert_database_date_and_number_column($last_table, $restricting_field, $select_field);

									$field_access_restriction = $select_field;

									if ($operand == "NOT")
									{
										$field_access_restriction .= " NOT";
									}
									$field_access_restriction .= " IN (";
									$tel = 0;
									foreach ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["field_access_restrictions"][$type_of_access][$restricting_field][$operand] as $value)
									{
										if ($value == "")
										{
											if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$restricting_field]["fieldtype"] == "date")
											{
												//dit matcht ook op een date/time veld in mysql waar eigenlijk 0000-00-00 00:00:00 in staat
												$value = "0000-00-00";
											}
											else if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$restricting_field]["fieldtype"] == "number")
											{
												$value = 0;
											}
										}

										if ($tel > 0)
										{
											$field_access_restriction .= ",";
										}
										if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$restricting_field]["fieldtype"] == "id")
										{
											//om dan maar geen results te hebben
											if (!is_positive_integer_or_zero($value))
											{
												//IS NULL?
												$value = -1;
											}
											//hoeft niet door mres() en hoeven ook geen quotes om heen
											$field_access_restriction .= $value;
										}
										else
										{
											$field_access_restriction .= "'" . mres($value) . "'";
										}
										$tel++;
									}
									$field_access_restriction .= ")";
									if ($tel > 0)
									{
										$field_access_restrictions[] = $field_access_restriction;
									}
								}
							}
							$select_field = "IF(" . implode(" AND ", $field_access_restrictions) . "," . $select_field . ",'')";
						}
					}
				}

				$type_of_access_query .= ",";
				if (count($field_access_restrictions) > 0)
				{
					$type_of_access_query .= "IF(" . implode(" AND ", $field_access_restrictions) . ",1,0)";
				}
				else
				{
					$type_of_access_query .= "1";
				}
				//"*" om onderscheid te maken tussen reguliere tags; "t1-f2" vs "*t1-f2" (of die writable is 0 of 1)
				$type_of_access_query .= " AS `*" . $query_parts["select_alias"][$x] . "`";
			}
		}
	}
	//dit hele blok stond voorheen in een if (!$do_not_check_write_access), maar omdat we nu ook records weg kunnen later die readonly zijn (een soort filter dus) moeten we deze code altijd uitvoeren

	//op een select table willen we geen records tonen waarop niet gewrite mag worden; want daar mag je immers ook geen nieuwe records onderhangen
	//we zouden dit ook nog configurabel kunnen maken met een listoption hide_readonly_records maar dat is het nu niet

	//TODO on a <select> dropdown we do not need the IF(`t2`.`EFID10` NOT IN ('No'),'ok','readonly') AS `type_of_access` query part. We could leave that off.
	if (($tabletype == "select") && ($GLOBALS['UC']['PersonalDataStructure'][$childtable]["fields"][$childfield]["InheritWriteAccessFromParent"]))
	{
		foreach ($write_restrictions as $write_restriction)
		{
			$wheres[] = $write_restriction;
		}
	}

	if (!$do_not_check_write_access)
	{
		$query .= $type_of_access_query;
	}

	$query .= " FROM";

	for ($x = 0; $x < count($query_parts["table"]); $x++)
	{
		$query .= " " . $query_parts["table"][$x];
	}

	if (!$bypass_record_access_by_sql)
	{
		$restrictions = get_recursive_access_querypart($table, "", "read", $already_joined);
		if ($restrictions != "")
		{
			$wheres[] = $restrictions;
		}
	}

	if ($use_pre_select_records)
	{
		if (count($pre_select_records) > 0)
		{
			$wheres[] = "`" . $table . "`.`" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"] . "` IN (" . implode(",", $pre_select_records) . ")";
		}
		else
		{
			$wheres[] = "FALSE";
		}
	}

	//er wordt een extra conditie op deze select lijst gelegd op basis van wat het referende veld in de child table *WIL* zien; bijvoorbeeld alleen de "actieve" klanten, of de werknemers die nog "in dienst" zijn
	if ($tabletype == "select")
	{
//TODO deze code zou helemaal weg moeten; want we kunnen dit nu met hiddenfilters in ListOptionsForParentTable oplossen
		if ($GLOBALS['UC']['PersonalDataStructure'][$childtable]["fields"][$childfield]["ExtraSelectConditionForSelectingNewRecords"] != "")
		{
			$wheres[] = "(" . $GLOBALS['UC']['PersonalDataStructure'][$childtable]["fields"][$childfield]["ExtraSelectConditionForSelectingNewRecords"] . ")";
		}


//we selecteren een parent uit t1. we zitten zelf in child t2. we willen alleen maar records uit t1 laten zien die, na opslaan van t2, nog steeds een beschrijfbaar record in t2 overhoudt
//daarom leggen we de security uit de child op de parent als restrictie voor de selecteren records
//stel; je mag Nederland en Frankrijk zien in een parent tabel, maar je mag alleen medewerkers die aan Nederland gekoppeld zijn muteren. Bij het toevoegen van een nieuwe medewerker moet je dan dus niet Frankrijk kunnen kiezen.
//deze code moet daar voor gaan zorgen
//kan er ook nog sql security via die veld gelegd worden? zo ja dan kunnen we die hier niet toepassen, toch? dit blijft dan dus wel een halve maatregel

		foreach (array("read", "write") as $childtable_read_or_write)
		{
			foreach (array("IS", "NOT") as $childtable_is_or_not)
			{
				if (is_array($GLOBALS['UC']['PersonalDataStructure'][$childtable]["record_access_restrictions"][$childtable_read_or_write][$childfield][$childtable_is_or_not]))
				{
					$sqlpart = "`" . $table . "`.`" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"] . "`";
					if ($childtable_is_or_not == "NOT")
					{
						$sqlpart .= " NOT";
					}
					//geen mres() want dit zijn ALTIJD numerieke waardes, het is immers een referentie naar een andere tabel
					$sqlpart .= " IN (" . implode(",", $GLOBALS['UC']['PersonalDataStructure'][$childtable]["record_access_restrictions"][$childtable_read_or_write][$childfield][$childtable_is_or_not]) . ")";
					$wheres[] = $sqlpart;
				}
			}
		}

	}

	if (!$RemoveExtraSelectCondition)
	{
		if ($GLOBALS['UC']['PersonalDataStructure'][$table]["ExtraSelectCondition"] != "")
		{
			//haken omdat er eventueel EFID2='abc' OR EFID5='def' in de ExtraSelectCondition staat
			$wheres[] = "(" . $GLOBALS['UC']['PersonalDataStructure'][$table]["ExtraSelectCondition"] . ")";
		}
	}

	//een deel van deze code is ook gedupliceerd in de lijst functie voor de highlighting -> dat is nogal lomp
	foreach ($searchperfield as $searchperfieldkey => $searchperfieldvalue)
	{
		if ($searchperfieldvalue != "")
		{
			//TODO deze lus is lomp; hij vertaald t1-f69-t2-f29 NAAR `t1-f69-t2`.`EFID29`
			//we zouden ook een associatieve array kunnen maken van alias -> dbfield
			for ($x = 0; $x < count($query_parts["select_field"]); $x++)
			{
				if ($query_parts["select_alias"][$x] == $searchperfieldkey)
				{
					list($searchtype, $searchparts) = searchstring_to_searchparts($searchperfieldvalue);
					$searchtype_sql = "AND";
					if ($searchtype == "ANY")
					{
						$searchtype_sql = "OR";
					}
					if (count($searchparts) > 0)
					{
						$where = "(";
						$tel = 0;
//						foreach (array_keys($searchparts) as $searchpart)
						for ($searchpart_counter = 0; $searchpart_counter < count($searchparts); $searchpart_counter++)
						{
							$searchpart = $searchparts[$searchpart_counter]["str"];

							$searchpart_escaped = $searchpart;
							//escapen want deze karakters zijn operator in de Mysql LIKE functie
							$searchpart_escaped = str_replace("%", "\\%", $searchpart_escaped);
							$searchpart_escaped = str_replace("_", "\\_", $searchpart_escaped);
							$searchpart_escaped = mres($searchpart_escaped);
							if ($tel > 0)
							{
								$where .= " " . $searchtype_sql . " ";
							}
							$where .= $query_parts["select_field"][$x];
							if (!$searchparts[$searchpart_counter]["is_positive"])
							{
								$where .= " NOT";
							}
							$where .= " LIKE '%" . $searchpart_escaped . "%'";
							//check: moet hier ook nog die OR IS NULL in??
							$tel++;
						}
						$where .= ")";
						$wheres[] = $where;
					}
				}
			}
		}
	}

	if ($search != "")
	{
		list($searchtype, $searchparts) = searchstring_to_searchparts($search);

		//checken of het sneller is alle velden te concatten en dan 1x een like te doen
		if (count($searchparts) > 0)
		{
			//dit gaat om de buitenste AND; ieder woord in een zoekopdracht moet matchen...
			//(EFID1 LIKE '%bla%' OR EFID2 LIKE '%bla%') AND (EFID1 LIKE '%boe%' OR EFID2 LIKE '%boe%')
			$match_all_searchparts = "AND";
			if ($searchtype == "ANY")
			{
				//...tenzij je op OR zoekt
				$match_all_searchparts = "OR";
			}

			//alle searches staan in hun eigen haak omdat searches OR kunnnen zijn terwijl de filter dan nog AND moeten zijn
			$where = "(";

			//we zouden moeten beginnen met de langste search string, want daarvan is de kans het grootst dat die niet matcht
//			foreach (array_keys($searchparts) as $searchpart)
			for ($searchpart_counter = 0; $searchpart_counter < count($searchparts); $searchpart_counter++)
			{
				$searchpart = $searchparts[$searchpart_counter]["str"];

				$searchpart_escaped = $searchpart;
				//escapen want deze karakters zijn operator in de Mysql LIKE functie
				$searchpart_escaped = str_replace("%", "\\%", $searchpart_escaped);
				$searchpart_escaped = str_replace("_", "\\_", $searchpart_escaped);
				$searchpart_escaped = mres($searchpart_escaped);

				$this_searchpart_wheres = array();

				//alle velden CONCAT-en IS 0.08 sec tegenover 0.11 sec voor multiple likes. probleem is dat je dan dingen zou kunnen matchen die aan elkaar geplakt worden. concat_ws doet er 0.10 sec over. de winst is dan minimaal. maar is er nog steeds. ik vind het alleen minder mooi

				for ($x = 0; $x < count($query_parts["select_field"]); $x++)
				{
					$table_parts = explode("-", $query_parts["select_alias"][$x]);
					$last_field = $table_parts[count($table_parts)-1];
					$last_table = $table_parts[count($table_parts)-2];

					$search_this_field = false;
					if ($last_table == $table)
					{
						$search_this_field = true;
					}
					else if ($search_parents)
					{
						$search_this_field = true;
						//2015-11-20 in deze lus werden ALLE gejoinde velden gesearcht. als je velden uit parents weergeeft of op filters werden die ook doorzcoth terwijl search_parents uit kon staan
						//-2 want de laatste tabel wordt altijd doorzocht als de voorgaanse search_parents op true staat
						for ($zz = 0; $zz < count($table_parts) - 2; $zz += 2)
						{
							if (!$GLOBALS['UC']['PersonalDataStructure'][$table_parts[$zz]]["search_parents"])
							{
//								print '<h1>not searching ' . $query_parts["select_alias"][$x] . ' because search parents is false for ' . $table_parts[$zz] . '</h1>';
								$search_this_field = false;
								break;
							}
						}
					}

					if ($search_this_field)
					{
						$search_fors = array($searchpart_escaped);

						//beetje lomp; 2 step search_this_field
						$search_this_field = false;
						$exact_match = false;

						$wildcard_start = true;
						$wildcard_end = true;

						if (!$GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["ExcludeFromSearches"])
						{
							if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "text")
							{
								$search_this_field = true;
							}
							else if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "date") && ($searchparts[$searchpart_counter]["strtype"] == "date"))
							{
								$search_this_field = true;
								$wildcard_start = false;
							}
							else if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "date") && ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["dateformat_in_database"] == "yyyy-mm-dd hh:mm:ss") && ($searchparts[$searchpart_counter]["strtype"] == "time"))
							{
								$search_this_field = true;
								//LET OP: geen seconden (zie 3 plekken in lf.php)
								//$wildcard_end = false;
							}
							//TODO controle op aantal decimalen
							else if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "number") && ($searchparts[$searchpart_counter]["strtype"] == "number"))
							{
								$search_this_field = true;
							}
							//we doorzoeken alleen het primary id veld van de tabel waar we nu in zitten
							else if (($connecting_field == "") && ($query_parts["select_alias"][$x] == $table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"]) && (is_positive_integer($searchpart_escaped)))
							{
								$search_this_field = true;
								$exact_match = true;
								$wildcard_start = false;
								$wildcard_end = false;
							}
							else if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["multiselect"])
							{
								$multiselectparent_table = $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["relation_to"];
								if ($multiselectparent_table != "")
								{
									$search_fors = array();

									if ($already_joined[$last_table][$last_field] == 0)
									{
										$multiselectparent_primary = $multiselectparent_table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$multiselectparent_table]["primary"];

										//voor iedere searchpart moet deze where RECORDID IN (SELECT ...) query opnieuw worden uitgevoerd
										$multiselectparent_search = $searchpart;
										//omdat er gezocht kan zijn op een letterlijke string; moeten we deze searchpart tussen quotes zetten; eigenlijk zouden we de searchpart parser maar 1x moeten runnen, BUITEN make_query. maar voor nu werkt dit goed
										$multiselectparent_search = preg_replace("/\"/", "\\\"", $multiselectparent_search);
										$multiselectparent_search = '"' . $multiselectparent_search . '"';

										$multiselectparent_searchperfield = array();
										$multiselectparent_start = 0;
										$multiselectparent_perpage = 0;
										$multiselectparent_sortby = "";
										$multiselectparent_sortdirection = "asc";
										$multiselectparent_filtercolumn = "";
										$multiselectparent_filter = array();
										$multiselectparent_filtertype = array();
										$multiselectparent_listlayout = array($multiselectparent_primary);
										$multiselectparent_RemoveExtraSelectCondition = false;
										$multiselectparent_connecting_field = "";

										//pre select zijn altijd de records in de parent table; die moeten dus niet meegegeven worden naar de child sub query
										$multiselectparent_use_pre_select_records = false;
										$multiselectparent_pre_select_records = array();
										$multiselectparent_order_by_pre_select = false;
										$multiselectparent_return_countquery = false;

										$multiselectparent_type_of_list = "mainlist";
										$multiselectparent_tabletype = "normal";
										$multiselectparent_childtable = "";
										$multiselectparent_childfield = "";
										$multiselectparent_childrecordid = 0;

										$multiselectparent_already_joined = $already_joined;
										//belangrijk om circulaire joins van t1 -> t2 -> t1 te voorkomen; de $multiselectparent_table moet weten dat de parent tabel al gejoined is om een nieuwe join hiernaar toe te voorkomen
										//TODO: are last_table and last_field the correct fields to use in this context? check also with search throguth child tables
										$multiselectparent_already_joined[$last_table][$last_field]++;

										$multiselectparent_hiddenfilter = array();
										$multiselectparent_hiddenfiltertype = array();
										$multiselectparent_listid = "";
										$multiselectparent_do_not_check_write_access = true;

										//pre_select_records is altijd leeg; anders kom je hier niet

										$multiselectparent_query = make_query($multiselectparent_table, $multiselectparent_search, $multiselectparent_searchperfield, $multiselectparent_start, $multiselectparent_perpage, $multiselectparent_sortby, $multiselectparent_sortdirection, $multiselectparent_filtercolumn, $multiselectparent_filter, $multiselectparent_filtertype, $multiselectparent_listlayout, $multiselectparent_RemoveExtraSelectCondition, $multiselectparent_connecting_field, $multiselectparent_use_pre_select_records, $multiselectparent_pre_select_records, $multiselectparent_order_by_pre_select, $search_parents, $search_childs, $search_files, $multiselectparent_return_countquery, $multiselectparent_type_of_list, $bypass_record_access_by_sql, $multiselectparent_tabletype, $multiselectparent_childtable, $multiselectparent_childfield, $multiselectparent_childrecordid, $multiselectparent_already_joined, $multiselectparent_hiddenfilter, $multiselectparent_hiddenfiltertype, $multiselectparent_listid, $multiselectparent_do_not_check_write_access);

//inside make_query() we cannot add debug to $output
/*
										if ($GLOBALS['SesMem']['debug'])
										{
											print '<h1>multi selectparent_query (' . $listid . ')</h1>';
											print '<div>' . htme($multiselectparent_query) . '</div>';
										}
*/

										//we could probably use db_GetFlatArray IF we are sure only a single field is returned?
										foreach (db_GetArray($multiselectparent_query) as $multiselectparent_row)
										{
											//no mres is needed because this is always a int() database field
											$search_fors[] = "\"" . $multiselectparent_row[$multiselectparent_primary] . "\"";
											$search_this_field = true;
										}
									}
								}
								else if (count($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["options"]) > 0)
								{
									//we search this multiselect field in a normal fashion. but we would need to escape quotes because serialize() also adds quotes (TODO)
									$search_this_field = true;
								}
							}
							//hiermee worden alle id velden doorzocht; dit is snel verwarrend; zeker als die id's niet worden weergegeven in lijsten
	/*						else if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "id") && (is_positive_integer($searchpart_escaped)))
							{
								$search_this_field = true;
								$exact_match = true;
							}*/
						}

						if ($search_this_field)
						{
							//only multiple select fields can have more than 1 search_fors
							foreach ($search_fors as $search_for)
							{
								$this_where = "";
								if (!$searchparts[$searchpart_counter]["is_positive"])
								{
									$this_where .= "(";
								}
								$this_where .= $query_parts["select_field"][$x];
								if (!$searchparts[$searchpart_counter]["is_positive"])
								{
									if ($exact_match)
									{
										$this_where .= " !";
									}
									else
									{
										$this_where .= " NOT";
									}
								}
								if ($exact_match)
								{
									$this_where .= "='" . $search_for . "'";
								}
								else
								{
									$this_where .= " LIKE '";
									if ($wildcard_start)
									{
										$this_where .= "%";
									}
									$this_where .= $search_for;
									if ($wildcard_end)
									{
										$this_where .= "%";
									}
									$this_where .= "'";
								}
								//zie http://dev.mysql.com/doc/refman/5.0/en/string-comparison-functions.html#operator_not-like
								if (!$searchparts[$searchpart_counter]["is_positive"])
								{
									$this_where .= " OR " . $query_parts["select_field"][$x] . " IS NULL)";
								}

								$this_searchpart_wheres[] = $this_where;
							}
						}
					}
				}

				if (($search_childs) && ($GLOBALS['UC']['PersonalDataStructure'][$table]["search_childs"]))
				{
					foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["childs"]) as $child_table)
					{
						if ($GLOBALS['UC']['PersonalDataStructure'][$child_table]["IncludeThisTableInSearchesFromParentTable"])
						{
							foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["childs"][$child_table] as $child_field)
							{
								//multiselects worden niet doorzocht; hiervoor moet je een koppeltabel maken
								if (!$GLOBALS['UC']['PersonalDataStructure'][$child_table]["fields"][$child_field]["multiselect"])
								{
									//om circulaire joins te voorkomen
									if ($already_joined[$child_table][$child_field] == 0)
									{
										//voor iedere searchpart moet deze where RECORDID IN (SELECT ...) query opnieuw worden uitgevoerd
										$child_search = $searchpart;
										//omdat er gezocht kan zijn op een letterlijke string; moeten we deze searchpart tussen quotes zetten; eigenlijk zouden we de searchpart parser maar 1x moeten runnen, BUITEN make_query. maar voor nu werkt dit goed
										$child_search = preg_replace("/\"/", "\\\"", $child_search);
										$child_search = '"' . $child_search . '"';

										$child_searchperfield = array();
										$child_start = 0;
										$child_perpage = 0;
										$child_sortby = "";
										$child_sortdirection = "asc";
										$child_filtercolumn = "";
										$child_filter = array();
										$child_filtertype = array();
										$child_listlayout = array($child_table . "-" . $child_field);
										$child_RemoveExtraSelectCondition = false;
										$child_connecting_field = $child_field;

										//pre select zijn altijd de records in de parent table; die moeten dus niet meegegeven worden naar de child sub query
										$child_use_pre_select_records = false;
										$child_pre_select_records = array();
										$child_order_by_pre_select = false;
										$child_return_countquery = false;

										$child_type_of_list = "mainlist";
										$child_tabletype = "normal";
										$child_childtable = "";
										$child_childfield = "";
										$child_childrecordid = 0;

										$child_already_joined = $already_joined;
										//belangrijk om circulaire joins van t1 -> t2 -> t1 te voorkomen; de child tabel moet weten dat de parent tabel al gejoined is om een nieuwe join hiernaar toe te voorkomen
										$child_already_joined[$child_table][$child_field]++;

										$child_hiddenfilter = array();
										$child_hiddenfiltertype = array();
										$child_listid = "";
										$child_do_not_check_write_access = true;

										//pre_select_records is altijd leeg; anders kom je hier niet

										$child_query = make_query($child_table, $child_search, $child_searchperfield, $child_start, $child_perpage, $child_sortby, $child_sortdirection, $child_filtercolumn, $child_filter, $child_filtertype, $child_listlayout, $child_RemoveExtraSelectCondition, $child_connecting_field, $child_use_pre_select_records, $child_pre_select_records, $child_order_by_pre_select, $search_parents, $search_childs, $search_files, $child_return_countquery, $child_type_of_list, $bypass_record_access_by_sql, $child_tabletype, $child_childtable, $child_childfield, $child_childrecordid, $child_already_joined, $child_hiddenfilter, $child_hiddenfiltertype, $child_listid, $child_do_not_check_write_access);

										$this_where = "";
										if (!$searchparts[$searchpart_counter]["is_positive"])
										{
											$this_where .= "NOT ";
										}
										$this_where .= "EXISTS (" . $child_query . ")";

										$this_searchpart_wheres[] = $this_where;
									}
								}
							}
						}
					}
				}

				if (($search_files) && ($GLOBALS['UC']['PersonalDataStructure'][$table]["search_files"]))
				{
					$this_where = "";

					if (!$searchparts[$searchpart_counter]["is_positive"])
					{
						$this_where .= "NOT ";
					}
					$this_where .= "EXISTS (SELECT * FROM `CRMbinfiles` WHERE `CRMbinfiles`.`type`='" . $GLOBALS['UC']['PersonalDataStructure'][$table]["name_in_database"] . "' AND `CRMbinfiles`.`koppelid`=`" . $table . "`.`" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"] . "` AND (`CRMbinfiles`.`filename` LIKE '%" . $searchpart_escaped . "%' OR `CRMbinfiles`.`extractedascii` LIKE '%" . $searchpart_escaped . "%'))";

					$this_searchpart_wheres[] = $this_where;
				}

				//er is geen 1 text field wat doorzocht kan worden (bijvoorbeeld in een tabel met alleen maar numerieke waarden), maar er is wel gezocht op iets, dus deze tabel moet dan 0 results teruggeven
				if (count($this_searchpart_wheres) == 0)
				{
					$this_searchpart_wheres[] = "FALSE";
//					$where .= "FALSE";
				}

				//dit gaat om de binnenste OR; er hoeft maar 1 van de velden te matchen op het zoekwoord...
				//(EFID1 LIKE '%bla%' OR EFID2 LIKE '%bla%') AND (EFID1 LIKE '%boe%' OR EFID2 LIKE '%boe%')
				$match_all_fields = "OR";
				if (!$searchparts[$searchpart_counter]["is_positive"])
				{
					//...tenzij je iets NIET wilt vinden; dan moet de zoekwoord in GEEN van de velden voorkomen
					$match_all_fields = "AND";
				}
				if ($searchpart_counter > 0)
				{
					$where .= " " . $match_all_searchparts . " ";
				}

				$where .= "(" . implode(" " . $match_all_fields . " ", $this_searchpart_wheres) . ")";
			}
			$where .= ")";

			$wheres[] = $where;
		}
	}

	if (count($wheres) > 0)
	{
		$query .= " WHERE " . implode(" AND ", $wheres);
	}

	//geen group by op count query
	if (!$return_countquery)
	{
		if ($filtercolumn != "")
		{
			$query .= " GROUP BY ";
			if ($filtercolumn != "")
			{
				$query .= $filtercolumn;
			}
		}
	}

	//geen order by op count query
	if (!$return_countquery)
	{
		if ((count($sortby_db_arr) > 0) || (($order_by_pre_select) && (count($pre_select_records) > 0)))
		{
			$sort_depth = 0;
			$query .= " ORDER BY ";
			if (($order_by_pre_select) && (count($pre_select_records) > 0))
			{
				//sort_depth is hier ALTIJD 0, voor consistentie en eventuele toekomstige aanpassingen
				if ($sort_depth > 0)
				{
					$query .= ",";
				}
				$query .= "FIELD(`" . $table . "`.`" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"] . "`," . implode(",", $pre_select_records) . ")";
				//geen sortdirection (?)
				$sort_depth++;
			}
			//dit is lomp maar er kan gesorteerd worden op een custom extra veld wat er via de module eval is ingezet (postcode); wanneer we de dqs maken dan kunnen we niet op het alias sorteren want die wordt dan niet geselecteerd. de volgorde van records (postcode) is wel van belang ivm de next en prev buttons
			else
			{
				for ($sortcounter = 0; $sortcounter < count($sortby_arr); $sortcounter++)
				{
					$sortby = $sortby_arr[$sortcounter];
					$sortdirection = $sortdirection_arr[$sortcounter];

					$sortby_db = $sortby_db_arr[$sortby];
					if (preg_match("/^ShowThisField\-(.*)$/", $sortby_db))
					{
						for ($x = 0; $x < count($query_parts["select_field"]); $x++)
						{
							if ($query_parts["select_alias"][$x] == $sortby_db)
							{
								$sortby_db = $query_parts["select_field"][$x];
								//als een numeriek veld selecteren wat wordt afgerond, dan willen we op de niet afgeronde waarde sorteren omdat de order anders undefined is. Dit is een slordige uitzondering
								//TODO als de kolom waarop je sorteert dezelfde waardes bevat moet is de sort order dan undefined en moet je dan dus altijd nog als 2e sort optie record id toevoegen?
								//2015-06-17 zat deze conditie er nou alleen in voor de list modifier van a65? volgens mij is die niet meer nodig; het gaat sowieso niet goed met een list modifier die round(KOLOM,2) doet
		/*						if (preg_match("/^ROUND\((.*)\)$/", $sortby_db, $match))
								{
									$sortby_db = $match[1];
								}*/
								break;
							}
						}


						if ($sort_depth > 0)
						{
							$query .= ",";
						}
						$query .= $sortby_db;
						if ($sortdirection == "desc")
						{
							$query .= " DESC";
						}
						$sort_depth++;
					}
					else if ($sortby_db != "")
					{
						$field_to_sort_on_added = false;

						$table_parts = explode("-", $sortby_db);
						$last_field = $table_parts[count($table_parts)-1];
						$last_table = $table_parts[count($table_parts)-2];

						$table_parts_minus_field = $table_parts;
						array_pop($table_parts_minus_field);

						//we sorteren verderop op `t1-f69-t2`.`EFID37`, vandaar deze lichtelijk lompe code
						$table_alias = implode("-", $table_parts_minus_field);

						$select_field = return_database_fieldname($last_table, $last_field, $table_alias);

/*
//JH 2017-02-14 blijkbaar niet meer nodig deze code. dit mag naar verloop van tijd weg
						//alleen voor ModuleToEvaluateOnEveryListBuild
						if (preg_match("/^ShowThisField\-/", $sortby_db))
						{
							$select_field = "`" . $sortby_db . "`";
						}
*/


						//dit kan volgens mij niet meer voorkomen omdat we nu sorteren op de default sort van de bovenliggende tabel, zie bovenaan make_query()
						if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["layout"] != "")
						{
							//we kunnen klakkeloos op deze velden sorten want we WETEN dat ze geselect zijn door make_query_parts()
							foreach ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["layout_tags"] as $tag)
							{
								//dit is een beetje lomp maar we weten niet per se het path naar deze tags, want het kan een aantal tabellen hoger liggen.
								//daarom lussen we door alle velden die we selecten en nemen we degene die op de tag eindigt. dat is hem namelijk altijd
								//todo dit is traag
								foreach ($query_parts["select_alias"] as $select_field_loop)
								{
									if (preg_match("/" . $tag . "$/", $select_field_loop))
									{
										$table_parts = explode("-", $select_field_loop);
										$last_field = $table_parts[count($table_parts)-1];
										$last_table = $table_parts[count($table_parts)-2];

										$table_parts_minus_field = $table_parts;
										array_pop($table_parts_minus_field);

										//we sorteren verderop op `t1-f69-t2`.`EFID37`, vandaar deze lichtelijk lompe code
										$table_alias = implode("-", $table_parts_minus_field);

										$select_field = return_database_fieldname($last_table, $last_field, $table_alias);
										$select_field = convert_database_date_and_number_column($last_table, $last_field, $select_field);
										$select_field = apply_conditional_read_access_to_field($last_table, $last_field, $table_alias, $select_field);

										if ($sort_depth > 0)
										{
											$query .= ",";
										}
										$query .= $select_field;
										if ($sortdirection == "desc")
										{
											$query .= " DESC";
										}
										$sort_depth++;
										$field_to_sort_on_added = true;
										break;
									}
								}
							}
						}
						//het veld is NIET een relation naar een andere tabel OF de refer field layout heeft geen enkele tag in zich; in dat geval sorteren we op het numerieke veld wat naar de andere tabel verwijst om een query error te voorkomen
						//TODO dit is slordig; samenvoegen met de lus hierboven
						if (!$field_to_sort_on_added)
						{
							$select_field = convert_database_date_and_number_column($last_table, $last_field, $select_field);
							$select_field = apply_conditional_read_access_to_field($last_table, $last_field, $table_alias, $select_field);

							if ($sort_depth > 0)
							{
								$query .= ",";
							}
							$query .= $select_field;
							if ($sortdirection == "desc")
							{
								$query .= " DESC";
							}
							$sort_depth++;
						}
					}
				}
			}

			//to prevent ERROR 1055 (42000): Expression #2 of ORDER BY clause is not in GROUP BY clause and contains nonaggregated column which is not functionally dependent on columns in GROUP BY clause; this is incompatible with sql_mode=only_full_group_by
			if ($filtercolumn == "")
			{
				if (!in_array($table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"], $sortby_arr))
				{
					//sort recordid wordt toegevoegd (als die nog niet bestaat) om een onveranderlijke volgorde te forceren wanneer je op een niet uniek veld sorteert
					if ($sort_depth > 0)
					{
						$query .= ",";
					}
					$query .= "`" . $table . "`.`" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"] . "`";
					//if the first column on which sort is applied is descending; the recordid is also descending (this is neccessary to create a valid opposite search)
					if ($sortdirection_arr[0] == "desc")
					{
						$query .= " DESC";
					}
					$sort_depth++;
				}
			}
		}
	}

	//geen limit op count query
	if (!$return_countquery)
	{
		//0 == alle records; bij filtercolumn queries is perpage 0; je wilt dan immers alles zien
		if ($perpage > 0)
		{
			$query .= " LIMIT " . $start . "," . $perpage;
		}
	}

	return $query;
}
function make_query_parts($table, $table_alias = "", $already_joined = array(), $fields_to_select = array(), $search = "", $max_tables_to_join = 0)
{
	//$UseInnerJoinInsteadOfLeftJoinForRequiredReferenceFields = false;

	//max_tables_to_join is standaard 0; dit kan gebruikt worden om alle referenties in een tabel op te halen zonder de query daadwerkelijk uit te voeren
	//voor de main lijst query moet max_tables_to_join op 61 staan want dat is het limiet van mysql

	//we houden bij of een tabel al gejoind is of niet; we joinen in 1 path van make_query_parts() een tabel maar 1 keer; als een tabel nogmaals voorkomt dan IS dit een een circulaire join. we tellen lussen en zouden de vergelijking hier verderop (already_joined < 1) ook op kleiner dan 2 kunnen zetten, zodat je in een Personen tabel die naar zichzelf verwijst ook de Vader en Moeder van iemand doorzoekt

	$query_parts = array();
	$query_parts["table"] = array();

	$query_parts["select_field"] = array();
	$query_parts["select_alias"] = array();

	//dit is de main table
	if ($table_alias == "")
	{
		$table_alias = $table;
		$table_in_database = $GLOBALS['UC']['PersonalDataStructure'][$table]["name_in_database"];
		$query_parts["table"][] = "`CRM" . $table_in_database . "` `" . $table_alias . "`";
	}

	if (is_array($GLOBALS['UC']['PersonalDataStructure'][$table]))
	{
		//2 losse lussen; eerst de eigen velden dan de parents. dit is niet mega efficient
		for ($lus = 1; $lus <= 2; $lus++)
		{
			foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]) as $field)
			{
				//2017-08-21; fields in the user table can no longer be displayed in other tables; they ARE still displayed when viewing the user table directly. this code is duplicated in make_query_parts() and get_recursive_access_querypart()
				if ((!$GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["hide_field_when_in_child_table"]) || ($table_alias == $table))
				{
					$select_field = return_database_fieldname($table, $field, $table_alias);
					$select_alias = $table_alias . "-" . $field;

					//in lus == 1 voegen we de velden in de eigen tabel toe zodat die bovenaan staan in de allfields array. deze array wordt gebruikt door list layout en zo staan de eigen velden dus bovenaan
					if ($lus == 1)
					{
						//check of we dit veld wel willen afbeelden; zo niet, dan selecten we hem ook niet
						//bij een search worden alle velden geselect omdat er later op gezocht wordt. we zouden hiervoor eigenlijk een andere lijst van moeten aanleggen: search fields
						if (($fields_to_select[$select_alias] == 1) || ($search != "") || (count($fields_to_select) == 0))
						{
							//pay attention: format_empty_numbers == false
							$select_field = convert_database_date_and_number_column($table, $field, $select_field, false);

							if ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["fieldtype_in_extrafields_table"] == "comment")
							{
								//2016-02-16 een comment field bevat nooit data, we selecteren hem als letterlijke string "CommentReadAccess" uit de database. als er conditionele read access op dit veld ligt dan komt hij als "" uit de database en daarop acteren we in ptrd()
								$select_field = "'CommentReadAccess'";
							}
							$select_field = apply_conditional_read_access_to_field($table, $field, $table_alias, $select_field);

							$query_parts["select_field"][] = $select_field;
							$query_parts["select_alias"][] = $select_alias;
						}
					}


//fields in the user table are not joined upwards from child tables; to prevent bloated (and potentially circular) datastructures. this code is duplicated in make_query_parts() and get_recursive_access_querypart()
if (($table == "users") && ($table_alias != "users") && (preg_match("/^f\d+$/", $field)))
{
}
else
{
					//in lus == 2 joinen we de parent velden
					if ($lus == 2)
					{
						$parent_table = $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["relation_to"];

						//multiselects kunnen we niet joinen omdat daar een serialized array in staat. multiselects moeten niet meer gebruikt worden; in plaats daarvan een koppeltabel
						if (($parent_table != "") && (!$GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["multiselect"]))
						{
							//om circulaire joins te voorkomen. we zouden ook 1 of 2x een circulaire join kunnen doorlopen en zo ook iemand zijn vader en moeder te vinden op een circulaire persoonstabel op een search
							if ($already_joined[$table][$field] == 0)
							{
								$table_alias_parent = $table_alias . "-" . $field . "-" . $parent_table;

								$make_join = false;
								if (count($fields_to_select) == 0)
								{
									//alle velden worden gejoined
									$make_join = true;
								}
								else if (($search != "") && ($GLOBALS['UC']['PersonalDataStructure'][$table]["search_parents"]))
								{
									//alle velden worden gejoined
									$make_join = true;
								}
								else
								{
									foreach (array_keys($fields_to_select) as $field_to_select)
									{
										if (preg_match("/^" . $table_alias_parent . "/", $field_to_select))
										{
											$make_join = true;
											break;
										}
									}
								}

								if ($make_join)
								{
									//safety mysql kan maximaal 61 tabellen joinen in een enkele query
									$tables_in_join = count($query_parts["table"]);
									if (($max_tables_to_join == 0) || ($tables_in_join <= $max_tables_to_join))
									{
/*										$joinquery = "";
										//verplichte velden zouden ge inner joint kunnen worden (de referentie MOET dan bestaan anders zie je het record niet meer). dit is SOMS sneller, maar soms ook veel trager. Daarom uitgezet
										if (($UseInnerJoinInsteadOfLeftJoinForRequiredReferenceFields) && ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["required"]))
										{
											$joinquery = "INNER";
										}
										else
										{
											$joinquery = "LEFT";
										}
										$joinquery .= " JOIN `CRM" . $GLOBALS['UC']['PersonalDataStructure'][$parent_table]["name_in_database"] . "` `" . $table_alias_parent . "` ON ";
*/
										$joinquery = "LEFT JOIN `CRM" . $GLOBALS['UC']['PersonalDataStructure'][$parent_table]["name_in_database"] . "` `" . $table_alias_parent . "` ON ";

										if ((preg_match("/^t\d+$/", $parent_table)) && ($GLOBALS['UC']['PersonalDataStructure'][$parent_table]["logical_delete"]))
										{
											$joinquery .= "`" . $table_alias_parent . "`.`deleted`='n' AND ";
										}
										$joinquery .= "`" . $table_alias_parent . "`.`" . $GLOBALS['UC']['PersonalDataStructure'][$parent_table]["primary"] . "`=" . return_database_fieldname($table, $field, $table_alias);
										$query_parts["table"][] = $joinquery;

										$parent_already_joined = $already_joined;
										$parent_already_joined[$table][$field]++;

										$query_parts_parent = make_query_parts($parent_table, $table_alias_parent, $parent_already_joined, $fields_to_select, $search, $max_tables_to_join);

										//safety mysql kan maximaal 61 tabellen joinen in een enkele query
										$tables_in_join = count($query_parts["table"]) + count($query_parts_parent["table"]);
										if (($max_tables_to_join == 0) || ($tables_in_join <= $max_tables_to_join))
										{
											//kan dit sjieker dan met array_merge?
											$query_parts["table"] = array_merge($query_parts["table"], $query_parts_parent["table"]);
											$query_parts["select_field"] = array_merge($query_parts["select_field"], $query_parts_parent["select_field"]);
											$query_parts["select_alias"] = array_merge($query_parts["select_alias"], $query_parts_parent["select_alias"]);
										}
										else
										{
											$errortext = "WARNING: Too many tables in join (1): " . $tables_in_join . " (" . $table_alias . "). No sql error is generated, because these tables are not joined at all. However, this could affect security or user experience.";
											log_msg($errortext);
										}
									}
									else
									{
										$errortext = "WARNING: Too many tables in join (2): " . $tables_in_join . " (" . $table_alias . "). No sql error is generated, because these tables are not joined at all. However, this could affect security or user experience.";
										log_msg($errortext);
									}
								}
							}
						}
}
					}
				}
			}
		}
	}

	return $query_parts;
}
function parse_old_template_to_new_template($template = "", $currentpath = "", $tags_already_parsed = array())
{
	//this function is a misnomer; it should be called pre_parse_record_tags()

	//update 2014-11-25 in 1 regexp, ook support voor ^ tags voor urlcodering. backreference gebruikt in pattern om geen #EFID12% meer te matchen
	$count = preg_match_all("/(\@|\#|\%|\^)([a-zA-Z0-9\-\_]+)\\1/", $template, $matches);
	for ($x = 0; $x < $count; $x++)
	{
		$open_char = $matches[1][$x];
		//deze kan namelijk gechanged worden
		$orig_open_char = $open_char;
		$oldtag = $matches[2][$x];

		$table_parts = explode("-", $oldtag);
		$last_field = $table_parts[count($table_parts)-1];

		$target = "";
		//deze tag is al geparsed; want hij bevat de nieuwe layout: t1-f2
		if (count($table_parts) > 1)
		{
			$target = $oldtag;
		}

		$matchkey = "";
		//niks gevonden
		//JH 2017-10-17 do we still need this block of code below now that old tags are no longer supported?
		if ($target != "")
		{
			$table_parts_target = explode("-", $target);
			$first_table = $table_parts_target[0];

			//eventueel het currentpath als voorvoegsel gebruiken
			if (($currentpath != "") && (!preg_match("/^" . preg_quote($currentpath . "-", "/") . "/", $target)))
			{
				$table_parts = explode("-", $currentpath);
				$last_field = $table_parts[count($table_parts)-1];
				$last_table = $table_parts[count($table_parts)-2];

				$first_table = $table_parts[0];

				//dit is wat lomp; maar door de volgorde van omzetten van oude naar nieuwe tags; kan dit al een nieuwe zijn; als we dit niet doen dan staat de naam van de tabel/veld er 2x in t17-f188-t4-f54-t4-f54-t5-f46
				//ik weet niet 100% zeker of dit wel goed gaat op 3, 4 of N laags tabellen
				if (($table_parts_target[0] == $last_table) && ($table_parts_target[1] == $last_field))
				{
					$target = implode("-", array_slice($table_parts_target, 2));
				}

				$target2 = $currentpath . "-" . $target;
				//hij matcht na het toevoegen van het voorvoegsel (currentpath)
				if (in_array($target2, $GLOBALS['UC']['PersonalDataStructure'][$first_table]["allfields"]))
				{
					$matchkey = $target2;
				}
			}
			//hij matcht volledig
			else if (in_array($target, $GLOBALS['UC']['PersonalDataStructure'][$first_table]["allfields"]))
			{
				$matchkey = $target;
			}
		}

		//de target begint niet met het currentpath. bijvoorbeeld bij een circulaire referentie.
		//target: t1-f4
		//t1-f2 verwijst naar t1
		//currentpath is dan: t1-f2
		//target wordt dan: t1-f2-t1-f4

		$newtag = "";

		if ($matchkey != "")
		{
			$table_parts = explode("-", $matchkey);
			$last_field = $table_parts[count($table_parts)-1];
			$last_table = $table_parts[count($table_parts)-2];

			//wanneer layout gezet is dan betreft het een relation_to, we parsen die met hun layout zodat de reference wordt opgedeeld in de component parts. BEHALVE wanneer het een multiselect is, daar kunnen we niks mee
			//we willen alleen @ en ^ tags recursief parsen en uit elkaar plukken als het een referentie veld is, # en % niet!
			//TODO het nadeel is dat we op deze plek niet weten of de referentie gezet is (of 0 is). In lijsten is de refer layout van een niet gezette referentie "", op forms blijven vaste teksten in de refer layout zichtbaar. Als een refer layout bijvoorbeeld "@t1-f1@ (@t1-f2@)" is dan is op een form zichtbaar " ()"
			if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["layout"] != "") && (!$GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["multiselect"] != "") && (($open_char == "@") || ($open_char == "^")))
			{
				//om circulaire referenties te voorkomen
				if (!$tags_already_parsed[$matchkey])
				{
					$tags_already_parsed_loop = $tags_already_parsed;
					$tags_already_parsed_loop[$matchkey] = true;
					$newtag = parse_old_template_to_new_template($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["layout"], $matchkey, $tags_already_parsed_loop);
				}
			}
			else
			{
				$newtag = $open_char . $matchkey . $open_char;
			}
		}

		//als de key niet gevonden is blijft newtag leeg en zou hij verdwijnen hij uit het template, daarom zetten we de originele tag weer terug; behalve als de oude tag een efid tag was, waar je blijkbaar geen toegang tot had
		if ($newtag == "")
		{
			//daarom check of die eindigt op een f872, hiermee worden niet t15-recordid en t16-timestamp_last_change gevangen, terwijl deze user daar misschien ook geen toegang tot heeft; geen groot probleem want de data zelf wordt nooit gereplaced maar de tags blijven dan staan
			if (!preg_match("/\-f\d+$/", $oldtag))
			{
				//dit kunnen tags zijn zoals GROUPONLY, XONLY en ARROWS die we onveranderd willen laten
				$newtag = $orig_open_char . $oldtag . $orig_open_char;
			}
		}

		$template = str_replace($orig_open_char . $oldtag . $orig_open_char, $newtag, $template);
	}
	return $template;
}
//dit is de inverse functie van onderstaande; om van searchparts weer een searchstring te maken
//TODO kan ik deze ook gebruiken in de child search? daar bouw ik toch ook iets op met de searchparts
function searchparts_to_searchstring($searchtype, $searchparts, $skip_element = -1)
{
	$searchstring = "";
	for ($x = 0; $x < count($searchparts); $x++)
	{
		if ($x != $skip_element)
		{
			$searchpart = $searchparts[$x]["str"];
			if ($searchstring != "")
			{
				if ($searchtype == "ANY")
				{
					$searchstring .= " or";
				}
				$searchstring .= " ";
			}
			if (!$searchparts[$x]["is_positive"])
			{
				$searchstring .= "-";
			}
			if ($searchparts[$x]["is_quoted"])
			{
				$searchstring .= "\"";
				$searchpart = preg_replace("/\"/", "\\\"", $searchpart);
			}
			$searchstring .= $searchpart;
			if ($searchparts[$x]["is_quoted"])
			{
				$searchstring .= "\"";
			}
		}
	}
	return $searchstring;
}
function searchstring_to_searchparts($searchstring)
{
	//safety om niet te veel searchparts te krijgen
	$max_search_parts = 6;

	//als je wilt zoeken naar 2 spaties zet ze dan tussen quotes: ["  "]
	//als je wilt zoeken naar een leading of trailing space zet hem dan tussen quotes: [" jeroen"] of ["jeroen "]
	//als je wilt zoeken naar een double of single quote escape hem dan of zet hem tussen de andere quote soort: ["\""] of ['"']
	//gebruik [+amsterdam -bier] om te zoeken naar voorkomens van amsterdam ZONDER bier; de plus doet niks die kan ook weg
	//de plus en min moet direct na een spatie staan zoeken naar [amsterdam-bier] zoekt naar letterlijk dat
	//plus min kan in combin met quotes dus [-"jeroen van den hoed" amsterdam]
	//zoeken naar [-  jeroen  hidde] zoekt als [+jeroen +hidde] de string moet direct volgen op de min (of plus)
	//als er een pipe |	ERGENS in zoeksting staat dan wordt er gezocht met OR anders met AND

	//we zoeken altijd lowercase
	$searchstring = strtolower($searchstring);

	//tabs en enters naar spaties want daar matchen we op als separator
	$searchstring = preg_replace("/\t|\r|\n/", " ", $searchstring);

	$parts = array();
	$searchpart_counter = -1;
	$charcounter = 0;
	$quote_open = false;
	$next_char_escaped = false;
	$is_positive = true;
	$quote_char = "";

	for ($x = 0; $x < strlen($searchstring); $x++)
	{
		$char = substr($searchstring, $x, 1);
		$add_char_to_parts = false;

		if ($next_char_escaped)
		{
			$next_char_escaped = false;
			$add_char_to_parts = true;
		}
		else if (($quote_open) && ($char == "\\"))
		{
			$next_char_escaped = true;
		}
		//spatie is het split char
		else if ((!$quote_open) && ($char == " "))
		{
			//default
			$is_positive = true;
			$charcounter = 0;
		}
		else if ((!$quote_open) && ($charcounter == 0) && ($char == "+"))
		{
			//een plus is default en veranderd hem nooit
		}
		else if ((!$quote_open) && ($charcounter == 0) && ($char == "-"))
		{
			//een min draait hem om; 1 min is negatief; 2 minnen is positief
			$is_positive = !$is_positive;
		}
		//quote is niet open, en er is 1 van de toegestande quotes chars gebruikt
		else if ((!$quote_open) && (($char == "\"") || ($char == "'")))
		{
			$quote_open = true;
			$quote_char = $char;
		}
		//hij moet weer met dezelfde soort quote char beeindigd worden
		else if (($quote_open) && ($char == $quote_char))
		{
			$quote_open = false;
			$quote_char = "";
		}
		else
		{
			$add_char_to_parts = true;
		}

		if ($add_char_to_parts)
		{
			//het is het eerste char; dus aanmaken
			if ($charcounter == 0)
			{
				if ($searchpart_counter == $max_search_parts)
				{
					break;
				}
				$searchpart_counter++;
				$parts[$searchpart_counter]["is_positive"] = $is_positive;
				$parts[$searchpart_counter]["is_quoted"] = $quote_open;
				$parts[$searchpart_counter]["str"] = "";
				$parts[$searchpart_counter]["strtype"] = "text";
			}
			$parts[$searchpart_counter]["str"] .= $char;

			$charcounter++;
		}
	}

	if ($quote_open)
	{
		//WARNING: MALFORMED SEARCH STRING
	}

	$searchtype = "ALL";

	//hier voegen we meerdere gelijke strings samen
	$parts_assoc = array();
	$assoc_count = 0;
	for ($x = 0; $x <= $searchpart_counter; $x++)
	{
		/*
		zoeken naar		results		methode
		toyota			226
		lagos			1211
		toyota lagos	31			AND
		toyota or lagos	1406		OR				226 + 1211 - 31 = 1406
		*/

		//er moeten minimaal 3 searchparts zijn (searchpart_counter is 0 based dus >= 2); update: hoeven we niet meer op te checken doordat ($x > 0) && ($x < $searchpart_counter)
		//de operator moet ERGENS tussen de andere tags staan

		//zoeken naar "or" zoekt naar "or"
		//zoeken naar "mercedes or" zoekt naar "mercedes" && "or" (minimaal 2 andere zoektermen)
		//zoeken naar "mercedes or nigeria" zoekt "mercedes" || "nigeria"
		//zoeken naar "or mercedes nigeria" zoekt "or" && "mercedes" && "nigeria" -> de operator moet tussen de andere in staan
		//zoeken naar "mercedes or nigeria or lagos" zoekt "mercedes" || "nigeria" || "lagos"
		//zoeken naar "mercedes or nigeria lagos" zoekt "mercedes" || "nigeria" || "lagos"
		//zoeken naar "mercedes or nigeria and lagos" zoekt "mercedes" || "nigeria" || "lagos" -> deze laatste zouden we nog moeten ondersteunen, maar dan moeten we ook met haken gaan werken; dat is iets voor de toekomst

		if (!$parts[$x]["is_quoted"])
		{
			//zet +1200,50 om naar 1200.50
			//rfc 1402, als onderstaande code wordt uitgezet worden "00123" niet meer omgezet naar "123" maar dan wordt ook nooit meer numerieke velden doorzocht
			$is_valid_number = is_valid_number($parts[$x]["str"]);
			if ($is_valid_number[0])
			{
				$parts[$x]["str"] = $is_valid_number[1];
				$parts[$x]["strtype"] = "number";
			}

			$checkdate = $parts[$x]["str"];
			$checkdate_iso = "";
			if (preg_match("/^(\d{4})\-(\d{1,2})\-(\d{1,2})$/", $checkdate, $match))
			{
				$checkdate_iso = sprintf("%04d-%02d-%02d", $match[1], $match[2], $match[3]);
			}
			else if (preg_match("/^(\d{4})\-(\d{1,2})$/", $checkdate, $match))
			{
				$checkdate_iso = sprintf("%04d-%02d", $match[1], $match[2]);
			}
			else if (preg_match("/^(\d{4})$/", $checkdate, $match))
			{
				$checkdate_iso = sprintf("%04d", $match[1]);
			}
			else if (preg_match("/^(\d{1,2})\-(\d{1,2})\-(\d{4})$/", $checkdate, $match))
			{
				if ($GLOBALS['UC']['DateFormat'] == "mm-dd-yyyy")
				{
					$checkdate_iso = sprintf("%04d-%02d-%02d", $match[3], $match[1], $match[2]);
				}
				//defaulteert naar dd-mm-yyyy
				else
				{
					$checkdate_iso = sprintf("%04d-%02d-%02d", $match[3], $match[2], $match[1]);
				}
			}
			else if (preg_match("/^(\d{1,2})\-(\d{4})$/", $checkdate, $match))
			{
				//"5-2016" is interpreted as month-year, irrespective of DateFormat
				$checkdate_iso = sprintf("%04d-%02d", $match[2], $match[1]);
			}

			if ($checkdate_iso != "")
			{
//				pre("checkdate_iso: " . $checkdate_iso);
				//TODO check of dit een valide datum is!
				$parts[$x]["str"] = $checkdate_iso;
				$parts[$x]["strtype"] = "date";
			}


			$checktime = $parts[$x]["str"];
			$checktime_iso = "";
			if (preg_match("/^(\d{1,2})\:(\d{2})\:(\d{2})$/", $checktime, $match))
			{
				//LET OP: geen seconden (zie 3 plekken in lf.php)
//				$checktime_iso = sprintf("%02d:%02d:%02d", $match[1], $match[2], $match[3]);
				$checktime_iso = sprintf("%02d:%02d", $match[1], $match[2]);
			}
			else if (preg_match("/^(\d{1,2})\:(\d{2})$/", $checktime, $match))
			{
				//LET OP: geen seconden (zie 3 plekken in lf.php)
//				$checktime_iso = sprintf("%02d:%02d:00", $match[1], $match[2], $match[3]);
				$checktime_iso = sprintf("%02d:%02d", $match[1], $match[2]);
			}

			if ($checktime_iso != "")
			{
				$parts[$x]["str"] = $checktime_iso;
				$parts[$x]["strtype"] = "time";
			}
		}

		$is_operator = false;
		if ((!$parts[$x]["is_quoted"]) && ($x > 0) && ($x < $searchpart_counter))
		{
			if (($parts[$x]["str"] == "or") || ($parts[$x]["str"] == "|") || ($parts[$x]["str"] == "||"))
			{
				$searchtype = "ANY";
				$is_operator = true;
			}
			else if (($parts[$x]["str"] == "and") || ($parts[$x]["str"] == "&") || ($parts[$x]["str"] == "&&"))
			{
				//dit is default
				//TODO we hoeven van "and" daarom eigenlijk geen operator te maken
	//			$searchtype = "ALL";
				$is_operator = true;
			}
		}

		if (!$is_operator)
		{
			$add = true;
			for ($y = 0; $y <= $searchpart_counter; $y++)
			{
				//deze string komt ook voor in een andere string
				//er is bijvoorbeeld gezocht op "dam" en "amsterdam" -> we hoeven dan alleen naar "amsterdam" te zoeken; want "dam" zit daar in
				//er moet wel beiden positief of beiden negatief gezocht worden

				//TODO als je op overlappende zoekwoorden zoekt "amste erdam" dan highlight alleen de eerste. niet makkelijk omheen te werken dit. ook niet een mega groot issue
				if (($x != $y) && (preg_match("/" . preg_quote($parts[$x]["str"]) . "/", $parts[$y]["str"])) && ($parts[$x]["is_positive"] == $parts[$y]["is_positive"]))
				{
					//als je zoekt naar "werk werk" dan heeft alleen de eeste searchpart effect, de tweede wordt genegeerd
					if (($parts[$x]["str"] != $parts[$y]["str"]) || ($x < $y))
					{
						$add = false;
						break;
					}
				}
			}

			if ($add)
			{
				$parts_assoc[$assoc_count++] = $parts[$x];
			}
		}
	}

	if ($GLOBALS['SesMem']['debug'])
	{
		print '<h1>search for "' . $searchtype . '" matches</h1>';
		pre($parts_assoc);
	}

	return array($searchtype, $parts_assoc);
}
function str_to_bool($str = "", $default_value = false)
{
	$bool = $default_value;
	if (isset($str))
	{
		if (($str == "1") || ($str == true))
		{
			$bool = true;
		}
		else if (($str == "0") || ($str == "") || ($str == false))
		{
			$bool = false;
		}
	}
	return $bool;
}
function make_list($listoptions)
{
	//********************************************************************************************
	//LET OP get_list_options_and_descriptions() BIJWERKEN WANNEER HIER OPTIES WORDEN TOEGEVOEGD!!
	//********************************************************************************************

	//qlog(INFO, "Start make_list " . string_r($listoptions));

	global $lang;

	$output = '';

	//werkt nog niet omdat Excel klaagt dat er iets mis is; dit komt omdat de xsd download url beveiligd is; en Excel niet met een passwd prompt komt; we zouden veel liever de xsd informatie embedden in de xml maar dit wordt ook niet ondersteund (door Excel)
	$xml_with_xsd = false;
	if ($_GET['xml_with_xsd'] == 1)
	{
		$xml_with_xsd = true;
	}
	$xml_ms_excel = false;
	if ($_GET['xml_ms_excel'] == 1)
	{
		$xml_ms_excel = true;
	}

	//altijd leeg; is nodig als argument voor make_query()
	$already_joined = array();

	//deze MAG alleen gezet zijn door list calls vanuit templates parse, list_filebox() en GetCalendarEvents()
	//ook voor de plain list in pdf
	//deze setting staat perpage=0 toe
	//via de GET staan we dat niet toe
	//LET OP dat je geen lijsten maakt die gefresht worden door de client (bijvoorbeeld een sorteer link). want DAN komen de vars wel via GET binnen en is de perpage setting dus niet langer 0
	$internallistcall = str_to_bool($listoptions["internallistcall"], false);

	//het is sneller om eerst de search te isolaten, wannneer er veel filter in de lijst staan; want anders moet de volledige search opnieuw gedaan worden voor ieder filter
	//voor een lijst zonder filter is het sneller om het in 1x te doen; we zouden dat nog variabel kunnnen maken
	//bij ARS met 4000 records met ongeveer 10 filter is het 0.3s tegenover 5.8s. Deze var MOET dus op true!
	//dit is alleen functioneel wanneer er met adaptieve filter gewerkt wordt (dat is op dit moment de ENIGE mogelijkheid). zo niet; dan zijn de filter queries simpel en hoeft er geen pre_select meer gedraaid te worden bij een search (het is dan dus ook sneller om het NIET te doen)
	//$run_pre_select_query_when_search = true;

	//json output uitgezet; staat wel helemaal klaar om te gebruiken
	//$list_outputtypes = array("table", "tile", "array", "template", "json");
	$list_outputtypes = array("table", "tile", "array", "template");
	//deze array is gedupliceerd in listlayout.php en in make_list in listfunctions.php dat is slordig
	$list_outputtypes_download = array("tsv", "xml", "excel");

	$t1_make_list = microtime(true);

	$human_readable_filter = '';

	$table = isset($listoptions["table"]) ? $listoptions["table"] : "";

	//tabletype, ALLEEN in combinatie met outputtype == table of tile
	//normal => de volledige lijst met alle opties
	//select => voor het selecteren van een record in een popup
	//plain => voor de plain list in een pdf
	$tabletype = isset($listoptions["tabletype"]) ? $listoptions["tabletype"] : "normal";

	//wordt afgeleid op basis van selectvia
	$childtable = "";
	$childfield = "";
	//alleen voor select lists; het id van het record in childtable die deze list callt
	//kan ook _new_ zijn
	$childrecordid = is_positive_integer($listoptions["childrecordid"]) ? $listoptions["childrecordid"] : 0;

	//naar boven gehaald ivm selectvia
	$hiddenfilter = is_array($listoptions["hiddenfilter"]) ? $listoptions["hiddenfilter"] : array();
	$hiddenfiltertype = is_array($listoptions["hiddenfiltertype"]) ? $listoptions["hiddenfiltertype"] : array();
	$type_of_list = isset($listoptions["type_of_list"]) ? $listoptions["type_of_list"] : "mainlist";
	$listid = isset($listoptions["listid"]) ? $listoptions["listid"] : "";
	$listid_div_exists = str_to_bool($listoptions["listid_div_exists"], false);

	//this var can only be set via #LIST# tags (internal calls) and not via the url
	$do_not_check_write_access = str_to_bool($listoptions["do_not_check_write_access"], false);

	//highlight row, bevat recordid van record wat gehighlight moet worden (dit kan het laatst geopende record zijn)
	//$highlight = $listoptions["hl"];
	//dit is slordig; er wordt in de GET gekeken. zou niet moeten
	$highlight = is_positive_integer($_GET["hl"]) ? $_GET["hl"] : 0;


	//wordt verderop toegevoegd nadat de <div> met de listid er in is gezet; we MOETEN dit herstructureren
	$selectvia_output = '';

	$selectvia = isset($listoptions["selectvia"]) ? $listoptions["selectvia"] : "";
	//dit zijn de selectvia filters
	$selectviafilter = isset($listoptions["selectviafilter"]) ? $listoptions["selectviafilter"] : "";
	$this_is_the_last_selectvia_step = false;

	$selectvia_steps = 0;

	if ($selectvia != "")
	{
		$table = "";
		$tabletype = "select";

		//dit is onprettig; we gebruiken een globaal gefixeerd id waarin ook nog eens verschillende tables worden geladen. ONDERZOEKEN OF DIT TOT ISSUES LEIDT
		$listid = "selectlist";

		$table_parts_selectvia = explode("-", $selectvia);
		$selectvia_steps = count($table_parts_selectvia) / 2 - 1;

		//hoeft niet meer los doorgegeven te worden
		$childtable = $table_parts_selectvia[0];
		$childfield = $table_parts_selectvia[1];

		$type_of_list = "sublist"; //zodat op de contactpersonen lijst het bedrijf verborgen wordt

		//the currently selected value in the form; we get it from the GET (REQUEST) vars. this should be changed
		$selectviacurrentvalue = $_REQUEST[$childtable . "-" . $childfield];

		if (($selectviafilter == "") && (is_positive_integer($selectviacurrentvalue)) && (count($table_parts_selectvia) > 4))
		{
			//there is a select via path of 2 steps or more; we need to set selectviafilters so list can start at the lowest level and the breadcrumb trails to (grand)parents work
			$selectvia_fields = array();
			$selectvia_field = $table_parts_selectvia[2] . "-" . $table_parts_selectvia[3];
			for ($x = 4; $x < count($table_parts_selectvia); $x += 2)
			{
				$table_name_selectvia = $table_parts_selectvia[$x];
				$field_name_selectvia = $table_parts_selectvia[$x+1];
				$selectvia_field .= "-";
				//we only select primary id's; we do not care about refer layouts here
				$selectvia_fields[] = $selectvia_field . $table_name_selectvia . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table_name_selectvia]["primary"];
				$selectvia_field .= $table_name_selectvia . "-" . $field_name_selectvia;
			}

			$selectvia_record = getrecord($table_parts_selectvia[2], $selectviacurrentvalue, false, false, $selectvia_fields, true); //do_not_check_write_access=true

			//format: table1-id1,table2-id2,table3-id3 etc.
			for ($selectvia_field_counter = 0; $selectvia_field_counter < count($selectvia_fields); $selectvia_field_counter++)
			{
				$selectvia_field = $selectvia_fields[$selectvia_field_counter];
				$parts = explode("-", $selectvia_field);
				$selectviafilter_table = $parts[count($parts)-2];
				$selectviafilter_val = $selectvia_record[$selectvia_field];
				if ($selectvia_field_counter > 0)
				{
					$selectviafilter .= ",";
				}
				$selectviafilter .= $selectviafilter_table . "-" . $selectviafilter_val;
			}
		}


		//example: selectvia: t3-f33-f2-f22-t1-recordid t1 = companies, t2 = contactpersons, t3 = invoice. an invoice is linked to a contactperson of a company. With this selectvia we select a company first and then drilldown to choose a contactperson who works at the selected company
		//selectvia: t3-f33-f2-f22-	t1-recordid
		//t1 = bedrijven
		//t2 = contactpersonen, verwijst naar bedrijven (t1) via f22
		//t3 = vacatures, verwijst naar contactpersonen (t2) via t33
		//we kiezen eerst een bedrijf en vervolgens een contactpersoon die bij dat bedrijf werkt.
		//de eerste list heeft een tweede list tot gevolg, de tweede list stopt het recordid van de contactpersoon terug in het callende formulier

		//GEEN SUPPORT voor circulaire joins

		//opbouw filter str array t3-2,t2-192,t1-25
		$selectviafilter_array = array();
		foreach (explode(",", $selectviafilter) as $selectviafilter_tmp)
		{
			if ($selectviafilter_tmp != "")
			{
				list($selectviafilter_key, $selectviafilter_val) = explode("-", $selectviafilter_tmp);
				$selectviafilter_array[$selectviafilter_key] = $selectviafilter_val;
			}
		}
		$selectviafilter_parts = '';

		$last_hiddenfilter_name = '';
		$last_hiddenfilter_value = '';

		$breadcrumb = array();
		//we lopen hem achterstevoren af; we beginnen hoog gelegen en lopen naar laag
		for ($x = count($table_parts_selectvia) - 2; $x >= 0; $x -= 2)
		{
			$table_name_selectvia = $table_parts_selectvia[$x];
			$field_name_selectvia = $table_parts_selectvia[$x+1];
			if (!is_array($GLOBALS['UC']['PersonalDataStructure'][$table_name_selectvia]["fields"][$field_name_selectvia]))
			{
				$errortext = "error in selectvia element " . $table_name_selectvia . "-" . $field_name_selectvia;
				return array(0, $errortext);
			}
			$relation_to_selectvia = $GLOBALS['UC']['PersonalDataStructure'][$table_name_selectvia]["fields"][$field_name_selectvia]["relation_to"];
			$this_selectviafilter_value = $selectviafilter_array[$relation_to_selectvia];


			if ($x == count($table_parts_selectvia) - 2)
			{
				$breadcrumb[] = array($selectviafilter_parts, $GLOBALS['UC']['PersonalDataStructure'][$table_name_selectvia]["plural"]);
			}

			if (is_positive_integer($this_selectviafilter_value))
			{
				$selectviafilter_parts .= $relation_to_selectvia . '-' . $this_selectviafilter_value . ',';

				$breadcrumb[] = array($selectviafilter_parts, get_refer_layout_for_record_field($table_name_selectvia, $field_name_selectvia, $this_selectviafilter_value));
				$table = $table_name_selectvia;

				$last_hiddenfilter_name = $table_name_selectvia . "-" . $field_name_selectvia;
				$last_hiddenfilter_value = $this_selectviafilter_value;
				if ($x == 2)
				{
					$this_is_the_last_selectvia_step = true;
				}
			}
		}
		//beetje lomp dit; kan dit beter?
		if ($selectvia_steps == 1)
		{
			$this_is_the_last_selectvia_step = true;
		}
		if ($last_hiddenfilter_name != '')
		{
			//eerst resetten zodat er geen dubbele filters worden gelegd
			$hiddenfilter[$last_hiddenfilter_name] = array();
			$hiddenfiltertype[$last_hiddenfilter_name] = "IS";

			$hiddenfilter[$last_hiddenfilter_name][] = $last_hiddenfilter_value;
		}

		//het eerste element
//		$selectvia_output .= htme($GLOBALS['UC']['PersonalDataStructure'][$table_parts_selectvia[0]]["fields"][$table_parts_selectvia[1]]["name"]) . ': ';
		for ($x = 0; $x < count($breadcrumb); $x++)
		{
			if ($x > 0)
			{
				$selectvia_output .= ' / ';
			}
			$selectviafilter_parts = $breadcrumb[$x][0];
			$name = $breadcrumb[$x][1];
			if ($x < count($breadcrumb) - 1)
			{
				$selectvia_output .= '<a onclick="refresh_contents_of_div(\'' . $listid . 'div\', \'list.php?childrecordid=' . $childrecordid . '&amp;nonavbar=1&amp;selectvia=' . inline_jsencode($selectvia) . '&amp;selectviafilter=' . inline_jsencode($selectviafilter_parts) . '\', \'\', true, \'GET\', \'\', true)">';
			}
			$selectvia_output .= htme($name);
			if ($x < count($breadcrumb) - 1)
			{
				$selectvia_output .= '</a>';
			}
		}
		if ($table == "")
		{
			//we beginnen achteraan
			$table = $table_parts_selectvia[count($table_parts_selectvia) - 2];
		}
		if (count($breadcrumb) > 1)
		{
			$selectvia_output .= ' ' . htme($GLOBALS['UC']['PersonalDataStructure'][$table]["plural"]);
		}

		$ListOptionsForParentTable = $GLOBALS['UC']['PersonalDataStructure'][$childtable]["fields"][$childfield]["ListOptionsForParentTable"];
		//dit hele blok code staat zowel in make_query als ook in make_list, maar wel met net andere logica; het doel is hetzelfde; om hiddenfilters die in ListOptionsForParentTable staan ofwel uit de GET ofwel server side te halen. Hierdoor is het ook security.
		if ($ListOptionsForParentTable != "")
		{
			//LET OP: het parsen van ListOptionsForParentTable is heel exotisch, we callen ParseTemplateRecordData() zonder recordid maar MET recordid_to_preload. Het gevolg is dat de huidige database waarde worden ingeladen van de recordid_to_preload (== het record) en dat die worden aangevuld/overschreven met waarden uit de GET. Hierdoor kan een ListOptionsForParentTable zowel data uit de db gebruiken als ook een nieuw gekozen waarde op het formulier wat je aan het invullen bent (bijvoorbeeld afhankelijk van de keuze van contactpersonen die werken bij het bedrijf wat tevens geselecteerd staat) N.B.: dit zou beter omgezet kunnen worden tot een selectvia constructie

			parse_str($ListOptionsForParentTable, $ListOptionsForParentTable_unparsed_array);
			$ListOptionsForParentTable = unescape_interleave_characters(ParseTemplateRecordData($ListOptionsForParentTable, $childtable, 0, $childrecordid, true)); //do_not_check_write_access=true
			parse_str($ListOptionsForParentTable, $ListOptionsForParentTable_array);

			foreach (array_keys($ListOptionsForParentTable_unparsed_array['hiddenfilter']) as $key)
			{
				$tags_in_filter = false;
				foreach ($ListOptionsForParentTable_unparsed_array['hiddenfilter'][$key] as $tmp_filter)
				{
					//dit is extreem slordig; er zit minimaal 1 @ of % tag in een filter; in dat geval gebruiken we de waarde zoals de client die doorstuurt
					if (count(get_tags($tmp_filter)) > 0)
					{
						$tags_in_filter = true;
						break;
					}
				}

				//als er tags in de filters staan, dan moeten we de waarde gebruiken zoals die gepost zijn door de client, zo niet dan MOETEN we de server variabelen gebruiken want anders zou de client een LOFPT security setting kunnen overschrijven
				//als er niks van de client ontvangen is, dan worden sowieso de server variabelen gebruikt
				if (($tags_in_filter) && (array_key_exists($key, $hiddenfilter)))
				{
					//we gebruiken de waarde zoals de client die doorstuurt; die staat reeds al in het hiddenfilter; bijvoorbeeld een hiddenfilter op een bepaald bedrijf bij de selectie van een contactpersoon zodat je alleen contactpersonen ziet die bij dat bedrijf werken

					//- stel er is een hiddenfilter op t1-f1 met de waarde %t3-f3%
					//- in de eerste load wordt die door ParseTemplateRecordData() erin gezet vanuit de GET variabele t3-f3=12 die door form-functions wordt meegestuurd
					//- stel dat je die lijst dan anders sorteert; dan is de GET variabele t3-f3=12 niet meer beschikbaar, maar is deze al gezet in het hiddenfilter die via list.php binnenkomt. we nemen dan deze waarde
					//- als er GEEN tags in het filter staan dan kijken we niet naar wat de client meestuurt
					$ListOptionsForParentTable_array['hiddenfilter'][$key] = $hiddenfilter[$key];
					$ListOptionsForParentTable_array['hiddenfiltertype'][$key] = $hiddenfiltertype[$key];
				}
			}

			/*
				selectvia: cp -> bedrijf -> land

				hiddenfilter op bedrijf = actief, we leggen die op de cp zodat die ook gevalidered wordt bij opslaan van uiteindelijke ref;
					hiddenfilter[t48-f567-t47-f587][]=Ja

				bij stap omhoog voor selectie van bedrijf moet de childtable en field eerst van het hiddenfilter afgehaald worden; zodat je alleen bedrijven kan kiezen die actief zijn
					hiddenfilter[t47-f587][]=Ja
			*/

			foreach (array_keys($ListOptionsForParentTable_array['hiddenfilter']) as $key)
			{
				$LOFPT_usable_part = "";

				$LOFPT_PARTS = explode("-", $key);
				for ($x = 0; $x < count($LOFPT_PARTS); $x += 2)
				{
					if ($LOFPT_PARTS[$x] == $table)
					{
						$LOFPT_usable_part = implode("-", array_slice($LOFPT_PARTS, $x, count($LOFPT_PARTS) - $x));
						break;
					}
				}
				if ($LOFPT_usable_part != "")
				{
					//eerst resetten zodat er geen dubbele filters worden gelegd
					$hiddenfilter[$LOFPT_usable_part] = array();
					$hiddenfiltertype[$LOFPT_usable_part] = "IS";

					foreach ($ListOptionsForParentTable_array['hiddenfilter'][$key] as $val)
					{
						$hiddenfilter[$LOFPT_usable_part][] = $val;
					}
					if ($ListOptionsForParentTable_array['hiddenfiltertype'][$key] != "")
					{
						$hiddenfiltertype[$LOFPT_usable_part] = $ListOptionsForParentTable_array['hiddenfiltertype'][$key];
					}
				}
			}

			foreach ($ListOptionsForParentTable_array as $key => $val)
			{
				//dit is heel lomp; we zouden ALLE listoptions die in de LOFPT zijn gedefineerd om deze list moeten leggen; maar we doen er slechts een heel paar waarvan ik weet dat we ze in gebruik hebben (TODO)
				if (($key == "hide_filters") || ($key == "hide_search") || ($key == "hide_listlayoutselector") || ($key == "search_parents") || ($key == "search_childs") || ($key == "search_files") || ($key == "use_adaptive_filters"))
				{
					//we doen ALSOF deze opties zijn meegegeven in de call naar make_list
					$listoptions[$key] = $val;
				}
			}
		}

//		$output .= 'hiddenfilter<pre>' . print_r($hiddenfilter, true) . '</pre>';
//		$output .= 'hiddenfiltertype<pre>' . print_r($hiddenfiltertype, true) . '</pre>';
	}

	if ((!is_array($GLOBALS['UC']['PersonalDataStructure'][$table])) || (!$GLOBALS['UC']['PersonalDataStructure'][$table]["allow_direct_view"]))
	{
		$errortext = "";

		if (GetSetting("DevelopmentOrProduction") == "Development")
		{
			$errortext = "Table \"" . $table . "\" does not exist or no access";
			if ($GLOBALS['SesMem']['ShowReasons'])
			{
				$errortext .= GetBackTrace();
			}
			//via buitenom opgevraagd dus /list.php?table=t299, dit willen we weten want iemand zit te fiddelen met de url
			if (!$internallistcall)
			{
				PrintAD($errortext);
				log_msg("ERROR: " . $errortext);
			}

			//verbergen, dit is vooral voor formulieren van belang; alleen op development
			$errortext = "<!-- " . $errortext . " -->";
		}

		//geen max_one_concurrent_load_finish() want max_one_concurrent_load_start() heeft nog niet gedraaid

		return array(0, $errortext);
	}


	//niet meer random; dat betekent wel dat als er meerdere lijsten uit dezelfde tabel op deze page staan, er een listid moet worden meegegeven
	//een listid moet globaal uniek zijn in een applicatie
	//als je meerdere views aanmaakt op dezelfde tabel dan moeten ze beide hun eigen listid hebben zodat ze ook hun eigen filters, sort en column selectie hebben
	if (!preg_match("/^[a-zA-Z][a-zA-Z0-9]+$/", $listid))
	{
		$listid = $table;
		if ($type_of_list == "sublist")
		{
			$listid .= $type_of_list;
		}
		$listid_div_exists = false;
	}


/*	$unique_command_id = "user_" . $GLOBALS['USERID'] . "_list_" . $table;
	if (!max_one_concurrent_load_start($unique_command_id))
	{
		return array(0, '<div class="errmsg">This list is already loading. Please be patient.</div>');
	}*/

	$hide_addlink = str_to_bool($listoptions["hide_addlink"], false);

	$ignorelastview = str_to_bool($listoptions["ignorelastview"], false);

	$countonly = str_to_bool($listoptions["countonly"], false);

	//als deze var gezet is moet iemand op EXACT dezelfde plek terugkomen als waar hij de lijst de laatste keer achterliet; dus inclusief pagina nummer, search en RemoveExtraSelectCondition
//	$sameresultandpage = str_to_bool($listoptions["sameresultandpage"], false);
	//dit is slordig; er wordt in de GET gekeken. zou niet moeten, maar anders moet dit; $listoptions["sameresultandpage"] = $_GET["sameresultandpage"]; $listoptions["hl"] = $_GET["hl"]; worden toegevoegd aan IEDERE make_list() call
	$sameresultandpage = str_to_bool($_GET["sameresultandpage"], false);

	if (($type_of_list == "sublist") || ($tabletype == "select") || ($countonly))
	{
		//we onthouden de last view niet op sublist's (ook niet op global searches -> in dat geval zet list.php de ignorelastview var zelf) en ook niet op select tables
		//om die reden beelden we ook niet de list layout selector af want dat wordt toch niet onthouden. TODO?

		//sinds 2014-09-03 kan dit ook voor de selectvia select list gebruikt worden
		if (!$sameresultandpage)
		{
			$ignorelastview = true;
		}
	}
	else
	{
		$type_of_list = "mainlist";
	}

	$default_search = "";
	$default_start = 0;
	$default_RemoveExtraSelectCondition = false;

	$default_listname = "";
	$lastview_attribute = "";

	$lastview = array();
	$lastview_current = array();
	//JH 2015-10-12 we willen NIKS onthouden
	if (($listoptions["outputtype"] != "array") && ($listoptions["outputtype"] != "json"))
	//if ($listoptions["outputtype"] != "array")
	{
		//zodat lastview altijd gevuld is, ook wanneer we een sublist bekijken; TODO dit zou netter en compacter kunnen
		$lastview["searchperfield"] = array();
		$lastview["filter"] = array();
		$lastview["filtertype"] = array();
		$lastview["sortby"] = "";
		$lastview["sortdirection"] = "asc";
		$lastview["perpage"] = $GLOBALS['UC']['PersonalDataStructure'][$table]["perpage"];
		$lastview["outputtype"] = "table";
		$lastview["selectionnew"] = "";
		$lastview["search"] = "";
		$lastview["start"] = 0;
		$lastview["RemoveExtraSelectCondition"] = false;
		$lastview["hidetableheader"] = false;
		$lastview["hidetableheaderwhennoresults"] = false;
		$lastview["hide_tablewhennoresults"] = false;

		$default_listname = "system-";
		if ($type_of_list == "sublist")
		{
			//dit is niet een fraaie conditie, maar hiermee krijgen sublists BY DEFAULT een andere listname
			$default_listname .= "sublist";
		}

		$lastview["listname"] = $default_listname;

		//we nemen de lastview alleen mee als we op de mainlist zitten, anders negeren we hem om te voorkomen dat op sublist's filters gezet worden.

		//de laatste geopende list view (filters, searches etc) staat in een los attribuut per table EN per listid
		//de lastview is opgebouwd op basis van listid
		//eigenlijk zou hier de table var niet in moeten staan; de listid zelf is immers al uniek
		$lastview_attribute = "List-lastview-" . $table . "-" . $listid;

		$lastview = GetAttribute("user", $lastview_attribute, $GLOBALS['USERID']);
		$lastview_current = $lastview;

		//defaults voor de lastview
		//TODO moeten hierin nog meer vars worden opgeslagen (ALLE listoptions bijvoorbeeld?)
		//hiddenfilter worden hierin niet bewaard

		//DEZE VARS WORDEN ALTIJD ONTHOUDEN, DUS OOK VOOR SUBLISTS
		//HUH? JH 2015-10-12 deze code doet niks toch?
		$lastview["sortby"] = isset($lastview["sortby"]) ? $lastview["sortby"] : "";
		$lastview["sortdirection"] = isset($lastview["sortdirection"]) ? $lastview["sortdirection"] : "asc";
		$lastview["perpage"] = is_positive_integer_or_zero($lastview["perpage"]) ? $lastview["perpage"] : $GLOBALS['UC']['PersonalDataStructure'][$table]["perpage"];
		$lastview["outputtype"] = isset($lastview["outputtype"]) ? $lastview["outputtype"] : "table";
		//dit is slordig; maar om te voorkomen dat een excel of xml als outputtype vast wordt gezet
		if (!in_array($lastview["outputtype"], $list_outputtypes))
		{
			$lastview["outputtype"] = "table";
		}
		$lastview["selectionnew"] = isset($lastview["selectionnew"]) ? $lastview["selectionnew"] : "";

		$lastview["listname"] = isset($lastview["listname"]) ? $lastview["listname"] : $default_listname;

		//DEZE VARS WORDEN ALLEEN VOOR DE MAINLIST ONTHOUDEN
		if (!$ignorelastview)
		{
			$lastview["searchperfield"] = is_array($lastview["searchperfield"]) ? $lastview["searchperfield"] : array();
			$lastview["filter"] = is_array($lastview["filter"]) ? $lastview["filter"] : array();
			$lastview["filtertype"] = is_array($lastview["filtertype"]) ? $lastview["filtertype"] : array();

			$lastview["search"] = isset($lastview["search"]) ? $lastview["search"] : "";
			$lastview["start"] = is_positive_integer_or_zero($lastview["start"]) ? $lastview["start"] : 0;
			$lastview["RemoveExtraSelectCondition"] = str_to_bool($lastview["RemoveExtraSelectCondition"], false);
			$lastview["hidetableheader"] = str_to_bool($lastview["hidetableheader"], false);
			$lastview["hidetableheaderwhennoresults"] = str_to_bool($lastview["hidetableheaderwhennoresults"], false);
			$lastview["hide_tablewhennoresults"] = str_to_bool($lastview["hide_tablewhennoresults"], false);

			if ($sameresultandpage)
			{
				//deze vals worden doorgaans niet overgenomen
				$default_search = $lastview["search"];
				$default_start = $lastview["start"];
				$default_RemoveExtraSelectCondition = $lastview["RemoveExtraSelectCondition"];
			}
		}
		else
		{
			//dit is wat lomp
			$lastview["searchperfield"] = array();
			$lastview["filter"] = array();
			$lastview["filtertype"] = array();
		}
	}

	if ($GLOBALS['SesMem']['debug'])
	{
		$output .= '<h1>listoptions (' . htme($listid) . ')</h1><pre>' . htme(print_r($listoptions, true)) . '</pre>';
		$output .= '<h1>lastview (' . htme($listid) . ')</h1><pre>' . htme(print_r($lastview, true)) . '</pre>';
	}

	$search_highlight_only = str_to_bool($listoptions["search_highlight_only"], false);

	$search = isset($listoptions["search"]) ? $listoptions["search"] : $default_search;
	$start = is_positive_integer_or_zero($listoptions["start"]) ? $listoptions["start"] : $default_start;
	$RemoveExtraSelectCondition = false;

	//hij kan alleen verwijderd worden indien hij bestaat en de remove text is gevuld. anders KAN en MAG het niet
	if (($GLOBALS['UC']['PersonalDataStructure'][$table]["ExtraSelectCondition"] != "") && ($GLOBALS['UC']['PersonalDataStructure'][$table]["RemoveExtraSelectCondionFromQueryLinkText"] != ""))
	{
		$RemoveExtraSelectCondition = str_to_bool($listoptions["RemoveExtraSelectCondition"], $default_RemoveExtraSelectCondition);
	}

	$searchperfield = is_array($listoptions["searchperfield"]) ? $listoptions["searchperfield"] : $lastview["searchperfield"];
	$perpage = is_positive_integer_or_zero($listoptions["perpage"]) ? $listoptions["perpage"] : $lastview["perpage"];

	//een gezette sort overrulet de laatst gezette sortdirection; m.a.w. een nieuwe sort is altijd dus asc wanneer je geen direction meegeeft
	if ($listoptions["sortby"] != "")
	{
		$lastview["sortdirection"] = "";
	}
	$sortby = isset($listoptions["sortby"]) ? $listoptions["sortby"] : $lastview["sortby"];
	$sortdirection = isset($listoptions["sortdirection"]) ? $listoptions["sortdirection"] : $lastview["sortdirection"];


	$outputtype = isset($listoptions["outputtype"]) ? $listoptions["outputtype"] : $lastview["outputtype"];
	$filter = is_array($listoptions["filter"]) ? $listoptions["filter"] : $lastview["filter"];
	$filtertype = is_array($listoptions["filtertype"]) ? $listoptions["filtertype"] : $lastview["filtertype"];
	$selectionnew = isset($listoptions["selectionnew"]) ? $listoptions["selectionnew"] : $lastview["selectionnew"];

	$hidetableheader = str_to_bool($listoptions["hidetableheader"], false);
	$hidetableheaderwhennoresults = str_to_bool($listoptions["hidetableheaderwhennoresults"], false);
	$hide_tablewhennoresults = str_to_bool($listoptions["hide_tablewhennoresults"], false);

	$listname = isset($listoptions["listname"]) ? $listoptions["listname"] : $lastview["listname"];

	$hiddensearchperfield = is_array($listoptions["hiddensearchperfield"]) ? $listoptions["hiddensearchperfield"] : array();
	$hiddensearch = isset($listoptions["hiddensearch"]) ? $listoptions["hiddensearch"] : "";

	$search_parents = str_to_bool($listoptions["search_parents"], true);
	$search_childs = str_to_bool($listoptions["search_childs"], true);
	$search_files = str_to_bool($listoptions["search_files"], true);
	$recent = is_positive_integer_or_zero($listoptions["recent"]) ? $listoptions["recent"] : 0;

	$disable_click_record_to_open = str_to_bool($listoptions["disable_click_record_to_open"], false);

	$use_adaptive_filters = str_to_bool($listoptions["use_adaptive_filters"], true);

	$hide_header = str_to_bool($listoptions["hide_header"], false);
	$customheader = isset($listoptions["customheader"]) ? $listoptions["customheader"] : "";

	$group = isset($listoptions["group"]) ? $listoptions["group"] : "";
	$hide_groupedcolumn = str_to_bool($listoptions["hide_groupedcolumn"], true);

	$lastgroup = "";

	$hide_empty_cells = str_to_bool($listoptions["hide_empty_cells"], false);

	$hide_pagination_if_only_1_page = str_to_bool($listoptions["hide_pagination_if_only_1_page"], false);
	//change 2014-01-16 standaard hiden
	$hide_bottom_pagination = str_to_bool($listoptions["hide_bottom_pagination"], true);
	$hide_perpage = str_to_bool($listoptions["hide_perpage"], false);
	$hide_recordcounter = str_to_bool($listoptions["hide_recordcounter"], false);
	$hide_search = str_to_bool($listoptions["hide_search"], false);
	$hide_selections = str_to_bool($listoptions["hide_selections"], false);
	$hide_addrow = str_to_bool($listoptions["hide_addrow"], true);

	$hide_id_input = str_to_bool($listoptions["hide_id_input"], false);
	$hide_listlayout = str_to_bool($listoptions["hide_listlayout"], false);
	//aanpassing 2014-06-17 worden bij default gehide. er zit sowieso een bug in de allfilters; de eerste keer doen ze het goed, de tweede keer (na ajax) werken l labels niet meer goed en komen ongezette filters weer terug
	$hide_listlayoutselector = str_to_bool($listoptions["hide_listlayoutselector"], false);


	//default hiden
	$hide_outputtype = array();
	$hide_outputtype['table'] = str_to_bool($listoptions["hide_outputtype_table"], true);
	$hide_outputtype['tile'] = str_to_bool($listoptions["hide_outputtype_tile"], true);
	$hide_outputtype['tsv'] = str_to_bool($listoptions["hide_outputtype_tsv"], true);
	$hide_outputtype['xml'] = str_to_bool($listoptions["hide_outputtype_xml"], true);

	//ALTIJD hiden
	$hide_outputtype['array'] = true;
	$hide_outputtype['json'] = true;
	$hide_outputtype['template'] = true;

	//default showen
	$hide_outputtype['excel'] = str_to_bool($listoptions["hide_outputtype_excel"], false);

	//default hiden
	$hide_recent = str_to_bool($listoptions["hide_recent"], true);
	$hide_modules = str_to_bool($listoptions["hide_modules"], false);

	//dit is feitelijk DISABLE sort
	$hide_sortby = str_to_bool($listoptions["hide_sortby"], false);
	$hide_filters = str_to_bool($listoptions["hide_filters"], false);
	$hide_add_up_total = str_to_bool($listoptions["hide_add_up_total"], false);

	$hide_must_be_must_not_be_filter = str_to_bool($listoptions["hide_must_be_must_not_be_filter"], false);

	$hide_human_readable_filter = str_to_bool($listoptions["hide_human_readable_filter"], false);

	$hide_all = str_to_bool($listoptions["hide_all"], false);

	$hide_massupdate = str_to_bool($listoptions["hide_massupdate"], false);
	$add_class_to_each_cell = str_to_bool($listoptions["add_class_to_each_cell"], false);

	$listlayoutcolumns = isset($listoptions["listlayoutcolumns"]) ? $listoptions["listlayoutcolumns"] : "";

	$continuous_scroll = str_to_bool($listoptions["continuous_scroll"], false);

	//this is the list template id
	$template_id = is_positive_integer($listoptions["template_id"]) ? $listoptions["template_id"] : 0;
	//this is the template id to open the record
	$template_to_use_open = is_positive_integer($listoptions["template_to_use_open"]) ? $listoptions["template_to_use_open"] : 0;


	$watch = is_positive_integer($listoptions["watch"]) ? $listoptions["watch"] : 0;
	if ($watch > 0)
	{
		if ($watch < 5)
		{
			$watch = 5;
		}
		//er MOET gesorteerd worden op het primary id
		$sortby = $table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"];
		//zowel asc als desc is toegestaan
		$hide_sortby = true;
	}



	$templated_list_html = '';
	$templated_list_html_tags = array();
	if ($outputtype == "template")
	{
		if (is_positive_integer($template_id))
		{
			$templated_list_html = GetTemplate($template_id);
			$templated_list_html = ParseTemplateLanguageTags($templated_list_html);
			if ($templated_list_html != "")
			{
				$templated_list_html = parse_old_template_to_new_template($templated_list_html, $table);
				$templated_list_html_tags = get_tags($templated_list_html);
				//beetje slordig dit
				$listlayoutcolumns = implode(",", $templated_list_html_tags);
			}
		}
	}

	//2015-06-15 deze vars worden gebruikt om bij te houden welke regulier (dus NIET-hidden) filters en searches per field geprint zijn als html checkboxen, zodat ze niet onderaan de lijst er opnieuw in gezet worden
	$filter_printed = array();
	$filter_from_to_printed = array();
	$filtertype_printed = array();
	$searchperfield_printed = array();

	//om direct een record te openen op numerieke input; ipv id_input
	$id_input_in_search = true;

	if ($tabletype == "select")
	{
		//TODO dit zou wel aan kunnen staan om hem direct te selecteren; daar is nog geen support voor (todo)
		$hide_id_input = true;
		//we willen in dit geval gewoon zoeken naar dat nummer en niet het record openen
		$id_input_in_search = false;
	}
	$search = trim($search);
	if ($id_input_in_search)
	{
		$hide_id_input = true;
	}

	if ($hide_all)
	{
		if ($customheader == "")
		{
			$hide_header = true;
		}
		$hide_pagination_if_only_1_page = true;
		$hide_bottom_pagination = true;
		$hide_perpage = true;
		$hide_recordcounter = true;
		$hide_search = true;
		$hide_selections = true;
		$hide_addlink = true;
		$hide_id_input = true;
		$hide_listlayout = true;

		$hide_outputtype['table'] = true;
		$hide_outputtype['tile'] = true;
		$hide_outputtype['tsv'] = true;
		$hide_outputtype['xml'] = true;
		$hide_outputtype['excel'] = true;

		$hide_recent = true;
		$hide_modules = true;

		$hide_sortby = true;
		$hide_filters = true;
//		$hide_add_up_total = true;
		$hide_must_be_must_not_be_filter = true;

		$hide_human_readable_filter = true;

		$hide_listlayoutselector = true;

		$hide_massupdate = true;
	}

	$hide_totalrecords = false;
	if ((($GLOBALS['UC']['PersonalDataStructure'][$table]["headerhtml"] == "") || ($customheader != "")) && (!$hide_header))
	{
		//het aantal records staat al in de header tussen haken, er wordt dan geen 1 - 100 / 400 afgebeeld maar die info kan gemist worden (dit maakt de lijst compacter en makkelijker leesbaar)
		$hide_totalrecords = true;
		//dit verbergt de << en >> buttons en de [1] - 10 telling; een lijst kan beginnen op meerdere pages, maar door het zetten van een filter kan het aantal results op 1 page passen waardoor deze buttons aan en uit schakelen. dat is misschien wat onrustig; maar op korte lijsten met minder results dan de perpage var geeft dit een mooie compacte weergave
		$hide_pagination_if_only_1_page = true;
	}

	if (count($GLOBALS['UC']['PersonalDataStructure'][$table]["perpage_options"]) == 1)
	{
		$hide_perpage = true;
	}

	//vaak nodig dus lokaal opslaan helpt
	$is_admin = is_administrator();

	$html_output = false;
	if (($outputtype == "table") || ($outputtype == "tile") || ($outputtype == "template"))
	{
		$html_output = true;
	}
	$only_data = str_to_bool($listoptions["only_data"], false);

	if ((($html_output) && (!$only_data)) || ($selectionnew != ""))
	{
		//html output: we need to print a dropdown of selection (unless only_data is true)
		//selectionnew is set; we need to load that selection to make this list
	}
	else
	{
		//otherwise; do not query attributes for selections at all
		$hide_selections = true;
	}

	if (!$is_admin)
	{
		if (!$hide_selections)
		{
			if (CheckFunctionAccess("HideSelectionsSelectBox") != "nok")
			{
				$hide_selections = true;
			}
		}

		if (!$hide_selections)
		{
			if (preg_match("/^t(\d+)$/", $table, $match))
			{
				$flextable_id = $match[1];
				if ($type_of_list == "mainlist")
				{
					if (GetAttribute("flextable", "ShowSelectionsWhenNotInline", $flextable_id) == "No")
					{
						$hide_selections = true;
					}
				}
				else
				{
					if (GetAttribute("flextable", "ShowSelectionsWhenInline", $flextable_id) == "No")
					{
						$hide_selections = true;
					}
				}
			}
		}
	}

	if (!$GLOBALS['UC']['PersonalDataStructure'][$table]["MaySelectColumns"])
	{
		$hide_listlayout = true;
	}

	//2014-08-22 deze 2 zijn verplaats van onder
	if ((!in_array($outputtype, $list_outputtypes)) && (!in_array($outputtype, $list_outputtypes_download)))
	{
		$outputtype = "table";
	}

	//!= array en !$countonly is lomp dit moet beter! we kunnen die niet doen op basis van ignorelastview, want die wordt ook gebruikt voor sublists, en daarvan wil je oa de sortorder onthouden
	if (($lastview_attribute != "") && ($outputtype != "array") && ($outputtype != "json") && (!$countonly))
	{
		$lastview["searchperfield"] = $searchperfield;
		$lastview["filter"] = $filter;
		$lastview["filtertype"] = $filtertype;
		$lastview["sortby"] = $sortby;
		$lastview["sortdirection"] = $sortdirection;
		$lastview["perpage"] = $perpage;
		$lastview["outputtype"] = $outputtype;
		$lastview["selectionnew"] = $selectionnew;
		$lastview["search"] = $search;
		$lastview["start"] = $start;
		$lastview["RemoveExtraSelectCondition"] = $RemoveExtraSelectCondition;
		$lastview["hidetableheader"] = $hidetableheader;
		$lastview["hidetableheaderwhennoresults"] = $hidetableheaderwhennoresults;
		$lastview["hide_tablewhennoresults"] = $hide_tablewhennoresults;

		//deze is one-time-only
		if ($listname != "user-DOWNLOAD")
		{
			$lastview["listname"] = $listname;
		}

		if ($lastview != $lastview_current)
		{
			SetAttribute("user", $lastview_attribute, $lastview, $GLOBALS['USERID']);
		}

//JH 2017-12-15 nodig als we het lastview attribute willen gaan gebruiken als vervanger voor dqs
//		SetAttribute("user", $lastview_attribute, $listoptions, $GLOBALS['USERID']);
	}

	//deze var kan nog niet als listoption worden meegegeven; kleine moeite om dat te doen
	$filters_as_checkbox = true;

	//deze wordt gesaved omdat die in het divform gezet wordt; anders wordt LASTVIEW alsnog overschreven wanneeer je naar de next of prev page gaat
	$orig_perpage = $perpage;

	$set_tilelistoptionsandfilters_div = false;

	//2015-10-25 dit is default voor tile
	if ($outputtype == "tile")
	{
		$hide_empty_cells = true;
	}
	//de cellen van een tabel staan niet altijd meer op dezelfde plek; de tableheader heeft dus geen functie
	//hide_empty_cells in table context leidt tot een tabel waarbij niet iedere row evenveel cellen heeft. dit is dus invalide html. bovendien kan de add up total niet meer goed gepositioneerd worden, dus die moet niet gebruik worden in dit soort tabellen
	if ($hide_empty_cells)
	{
		$hidetableheader = true;
	}


	$primary_key = $table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"];

	if ($sortby == "")
	{
		$sortby = $GLOBALS['UC']['PersonalDataStructure'][$table]["default_sortby"];
		$sortdirection = $GLOBALS['UC']['PersonalDataStructure'][$table]["default_sortdirection"];
	}

	//de function call is in string context; hier omzetten naar array; niet zo sjiek, maar anders moet hoop code aangepast worden
	$sortby_arr = explode(",", $sortby);
	$sortdirection_arr = explode(",", $sortdirection);

	if (($group != "") && (!in_array($group, $GLOBALS['UC']['PersonalDataStructure'][$table]["allfields"])))
	{
		$group = "";
		$output .= '<div class="errmsg">group cannot be set, key &quot;' . htme($group) . '&quot; does not exists</div>';
	}

	for ($sortcounter = 0; $sortcounter < count($sortby_arr); $sortcounter++)
	{
		if ($sortdirection_arr[$sortcounter] != "desc")
		{
			$sortdirection_arr[$sortcounter] = "asc";
		}
	}

	if (GetSetting("DevelopmentOrProduction") == "Development")
	{
		foreach (array("hiddenfilter", "filter", "filtertype", "hiddenfiltertype", "searchperfield", "hiddensearchperfield") as $var)
		{
			foreach (array_keys($$var) as $key)
			{
				if (!in_array($key, $GLOBALS['UC']['PersonalDataStructure'][$table]["allfields"]))
				{
					$output .= '<div class="errmsg">' . $var . ' cannot be set, key &quot;' . htme($key) . '&quot; does not exists</div>';
				}
			}
		}
		foreach (array("filtertype", "hiddenfiltertype") as $var)
		{
			foreach ($$var as $key => $val)
			{
				if (($val != "") && ($val != "IS") && ($val != "NOT"))
				{
					$output .= '<div class="errmsg">' . $var . ' &quot;' . htme($key) . '&quot; value is incorrect, should be either &quot;IS&quot; or &quot;NOT&quot;, value is &quot;' . htme($val) . '&quot;</div>';
				}
			}
		}
		for ($sortcounter = 0; $sortcounter < count($sortby_arr); $sortcounter++)
		{
			$sortby = $sortby_arr[$sortcounter];
			$sortdirection = $sortdirection_arr[$sortcounter];

			if (($sortby != "") && (!in_array($sortby, $GLOBALS['UC']['PersonalDataStructure'][$table]["allfields"])))
			{
				$output .= '<div class="errmsg">sortby cannot be set, key &quot;' . htme($sortby) . '&quot; does not exists</div>';
			}
			if (($sortdirection != "") && ($sortdirection != "asc") && ($sortdirection != "desc"))
			{
				$output .= '<div class="errmsg">sortdirection value is incorrect, should be either &quot;asc&quot; or &quot;desc&quot;, value is &quot;' . htme($sortdirection) . '&quot;</div>';
			}
		}
	}

	$close_listid_div = false;
	if ((!$listid_div_exists) && ($html_output) && ($tabletype != "plain"))
	{
		$close_listid_div = true;
		//deze container wordt hergebruikt voor list reloads
		$output .= '<div id="' . $listid . 'div" class="listcontainer">';
	}

	//we slaan dit op een in var; omdat ik niet 100% ben of onderstaande code in zijn geheel verplaatst kan worden naar de correcte plek
	//deze wordt op het eind van listoptionsleft ingevoegd
	$extra_html_above_list = '';

	$AllowFoldersInFilelists = false;
	//wordt gebruikt om de inline delete te hiden
	$all_records_in_this_list_readonly = false;

	if ($html_output)
	{
		//geen browse button op de revert pagina
		if (($table == "files") && ($hiddenfilter["files-version_belonging_to"][0] == 0))
		{
			if (is_array($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]["folder"]))
			{
				//caching voor later de lus door de records
				$AllowFoldersInFilelists = true;
			}

			//deze vars worden gezet door list_filebox()
			$files_relation_to = $hiddenfilter["files-type"][0];
			$files_for_koppelid = $hiddenfilter["files-koppelid"][0];
			$files_for_folder = $hiddenfilter["files-folder"][0];

			$files_relation_to_ds = $files_relation_to;
			if (preg_match("/^flextable(\d+)$/", $files_relation_to_ds, $match))
			{
				$files_relation_to_ds = "t" . $match[1];
			}

			//er kunnen alleen bestanden worden toegevoegd wanneer we naar een file list kijken ONDER 1 specifiek record
			$access_of_record_to_which_file_is_attached = CheckTableRecordAccess($files_relation_to_ds, $files_for_koppelid);

			if (is_array($GLOBALS['UC']['PersonalDataStructure'][$files_relation_to_ds]))
			{
				if (is_positive_integer($files_for_koppelid))
				{
					if ($access_of_record_to_which_file_is_attached != "nok")
					{
						if ($access_of_record_to_which_file_is_attached == "readonly")
						{
							//wordt nog niet gebruikt; is voor consistentie
							$all_records_in_this_list_readonly = true;
						}

						//if (is_positive_integer($files_for_folder))
						//om ook de root folder in z'n eentje weer te geven
						if ($AllowFoldersInFilelists)
						{
							$extra_html_above_list .= '<div class="folderpath">';

							$steps = array();

							$target_folder = $files_for_folder;
							//maximaal 16 levels folder diepte om eindeloze lussen te voorkomen
							while (($target_folder > 0) && (count($steps) < 16))
							{
								//voor security worden hier alleen paden geselecteerd waar het type en recordid gelijk zijn. Dit zou altijd zo moeten zijn, MAAR er kan een willekeurig id in files_for_folder worden geparsed uit de REQUEST. zonder deze condities in de query zou je daarmee bestandsnamen kunnen zien die je niet mag zien
								//wanneer je 1 file mag zien onder een record, dan mag je ALLE files zien onder dat record
								$row = db_GetRow("SELECT `filename`, `folder` FROM `CRMbinfiles` WHERE `fileid`='" . mres($target_folder) . "' AND `type`='" . mres($files_relation_to) . "' AND `koppelid`='" . mres($files_for_koppelid) . "'");
								$target_filename = $row['filename'];
								$steps[] = array($target_folder, $target_filename);
								$target_folder = $row['folder'];
							}

							$steps[] = array(0, $GLOBALS['UC']['PersonalDataStructure'][$files_relation_to_ds]["singular"] . " " . $files_for_koppelid);

							for ($x = count($steps) - 1; $x >= 0; $x--)
							{
								$target_folder = $steps[$x][0];
								$target_filename = $steps[$x][1];
								$extra_html_above_list .= '<a onclick="ListSetFileFolder(\'' . $listid . '\',' . $target_folder . ')">' . htme($target_filename) . '</a> / ';
							}

							$extra_html_above_list .= '</div>';

							if ($access_of_record_to_which_file_is_attached == "ok")
							{
								//let op: geen support voor meerdere file lists op 1 pagina; omdat upload_files() dat niet ondersteund, zou wel kunnen op basis van listid
								//suffix van div wordt meegegeven aan upload_files(), dat zou vervangen moeten worden door listid

								/* the button for adding a folder */
								$extra_html_above_list .= '

								<input type="hidden" name="newfolder" id="JS_newfolder" value="">

								<button type="button" onclick="

									$(\'#JS_newfolder\').val(prompt(\'' . inline_jsencode($lang['addfolder']) . '\', \'\'));
									if ($(\'#JS_newfolder\').val() != \'\') {
										upload_files(\'-' . $files_relation_to_ds . '\', true)
									}
								"
								title="' . htme($lang['addfolder']) . '"

								 class="uploadbuttonfilelist addbutton">
								 <span class="fa fa-plus-square-o"></span>
								 <span class="fa fa-folder-o"></span>

								 </button>
								';

							}
						}
						if ($access_of_record_to_which_file_is_attached == "ok")
						{
							/* the button for adding a new, empty file */
							$extra_html_above_list .= '

							<input type="hidden" name="newdocument" id="JS_newdocument" value="">

							<button type="button" onclick="

								$(\'#JS_newdocument\').val(prompt(\'' . inline_jsencode($lang['createdocument']) . '\',\'\'));
								if ($(\'#JS_newdocument\').val() != \'\') {
									upload_files(\'-' . $files_relation_to_ds . '\', true)
								}
							"
							title="' . htme($lang['createdocument']) . '"

							 class="uploadbuttonfilelist addbutton">
							 <span class="fa fa-plus-square-o"></span>
							 <span class="fa fa-file-text-o"></span></button>';
						 }
					}
				}
			}
		}
	}

//JH 2016-10-21 uitgezet
//	$inline_select_table = false;

	if (!is_positive_integer_or_zero($start))
	{
		$start = 0;
	}
	if (!is_positive_integer_or_zero($perpage))
	{
		$perpage = $GLOBALS['UC']['PersonalDataStructure'][$table]["perpage"];
	}

	$buffering = true;
	$run_countquery = true;
	if (($outputtype == "array") || ($outputtype == "json"))
	{
		$run_countquery = false;
	}
	else if (($outputtype == "xml") || ($outputtype == "tsv") || ($outputtype == "excel"))
	{
		if ($GLOBALS['UC']['PersonalDataStructure'][$table]["download_allowed"])
		{
			$run_countquery = false;
			$start = 0;
			$perpage = 0;
			//directe output door print omdat de output var anders te groot wordt
			//voor excel is dit niet nodig
			$buffering = false;

			//we do not need to know if the user has write access to these records
			$do_not_check_write_access = true;

			//verderop wordt nog gecheckt of de download niet meer records retourneert dan dat je mag zien
		}
		else
		{
			$errortext = "download not allowed";
			PrintAD($errortext);
			log_msg("ERROR: " . $errortext);
//			max_one_concurrent_load_finish($unique_command_id);
			return array(0, $errortext);
		}
	}
	//voor in pdf's
	if ($tabletype == "plain")
	{
		$start = 0;
		$perpage = 0; //dit werkt alleen wanneer die via internallistcall binnenkomt
		$hide_sortby = true;
		$hide_filters = true;
		$disable_click_record_to_open = true;
	}
	if ($tabletype == "array")
	{
		$perpage = 0; //dit werkt alleen wanneer die via internallistcall binnenkomt
	}

	if ($GLOBALS["webapp"]) //wordt gezet in webapp.php of initiate.php
	{
		//overrides voor webapp
		$hide_filters = true;
		//filters are disabled; but IF we would decide to re-enable filters we wish to show them as <select multiple> instead of a list of checkboxes
		//$filters_as_checkbox = false;

		if ($watch == 0)
		{
			$continuous_scroll = true;
		}

		$hide_perpage = true;
		$hide_listlayout = true;
		$hide_listlayoutselector = true;
		$hide_outputtype['excel'] = true;

		//alleen als die gezet is op een nummer (en dus niet voor array calls e.d.)...
		/*if ($perpage > 0)
		{
			//...dan zetten we hem om naar de minimaal toegestane waaarde
			$perpage = $GLOBALS['UC']['PersonalDataStructure'][$table]["perpage_options"][0];
		}
		*/

		//the select list is already displayed in the stackedcontent; we do not have an extra level left to put an add window in, while retaining the content of the underlying form
		if ($tabletype == "select")
		{
			$hide_addlink = true;
		}
	}

	if (!$internallistcall)
	{
		if ($outputtype == "array")
		{
			$errortext = "array not allowed";
			PrintAD($errortext);
			log_msg("ERROR: " . $errortext);
//			max_one_concurrent_load_finish($unique_command_id);
			return array(0, $errortext);
		}
		if ($html_output)
		{
			if (!in_array($perpage, $GLOBALS['UC']['PersonalDataStructure'][$table]["perpage_options"]))
			{
				$use_perpage = $GLOBALS['UC']['PersonalDataStructure'][$table]["perpage_options"][0];
				if ($perpage == 0)
				{
					$last_element = count($GLOBALS['UC']['PersonalDataStructure'][$table]["perpage_options"]) - 1;
					$use_perpage = $GLOBALS['UC']['PersonalDataStructure'][$table]["perpage_options"][$last_element];
				}

				$perpage = $use_perpage;
/*				$errortext = "perpage " . $perpage . " incorrect, using value " . $use_perpage;
				log_msg("ERROR: " . $errortext);*/
			}
		}
	}

	$listlayout = array();

	if ($listlayoutcolumns != "")
	{
		$hide_listlayout = true;
		//beetje vies dit
		if ($listlayoutcolumns == "allfieldsthistable")
		{
			//code duplicatie van hieronder
			foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]) as $field)
			{
				if (!$GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["hide_from_list"])
				{
					$key = $table . "-" . $field;
					$listlayout[] = $key;
				}
			}
		}
		else
		{
			$listlayout = explode(",", $listlayoutcolumns);
		}
	}
	else
	{
		//in listname staat of het een system of een user var list betreft gevolgd door de listname
		$identifier = "system";
		$entity = 1;
		$listname_without_type = "";

		if (preg_match("/^(system|user)-(.*)$/", $listname, $match))
		{
			$identifier = $match[1];
			$listname_without_type = $match[2];
		}

		if ($identifier == "user")
		{
			$entity = $GLOBALS['USERID'];
		}

		$attribute = "List-layout-" . $table . "-" . $listname_without_type;

		$listlayout = GetAttribute($identifier, $attribute, $entity);
	}

	//hij moet geselecteerd worden; maar we beelden hem niet af IN de lijst, alleen erboven
	if (($group != "") && (!in_array($group, $listlayout)))
	{
		$listlayout[] = $group;
	}

	if (count($listlayout) == 0)
	{
		//defaulteren naar de standaard list
		$listname = $default_listname;

		//code duplicatie; zie hierboven
		$identifier = "system";
		$entity = 1;
		$listname_without_type = "";

		if (preg_match("/^(system|user)-(.*)$/", $listname, $match))
		{
			$identifier = $match[1];
			$listname_without_type = $match[2];
		}

		if ($identifier == "user")
		{
			$entity = $GLOBALS['USERID'];
		}

		$attribute = "List-layout-" . $table . "-" . $listname_without_type;

		$listlayout = GetAttribute($identifier, $attribute, $entity);
		//code duplicatie; zie hierboven
	}

	if ((count($listlayout) == 0) && (preg_match("/sublist$/", $default_listname, $match)))
	{
		//defaulteren naar de standaard list main list, zodat een sub list niet een eigen layout HOEFT te hebben
		$listname = $match[0];

		//code duplicatie; zie hierboven
		$identifier = "system";
		$entity = 1;
		$listname_without_type = "";

		if (preg_match("/^(system|user)-(.*)$/", $listname, $match))
		{
			$identifier = $match[1];
			$listname_without_type = $match[2];
		}

		if ($identifier == "user")
		{
			$entity = $GLOBALS['USERID'];
		}

		$attribute = "List-layout-" . $table . "-" . $listname_without_type;

		$listlayout = GetAttribute($identifier, $attribute, $entity);
		//code duplicatie; zie hierboven
	}

	if (count($listlayout) == 0)
	{
		//default alle velden in de huidige tabel
		foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]) as $field)
		{
			if (!$GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["hide_from_list"])
			{
				$key = $table . "-" . $field;
				$listlayout[] = $key;
			}
		}
	}

	//omdat we in de h2 tag van de tile de refer layout afbeelden zoals deze tabel zichzelf wil presenteren moeten we hier alle kolommen selecteren die we daar voor nodig hebben
	//met een hack beelden we alle velden die in de layout zitten vervolgens NIET af in de tile view
	if ($outputtype == "tile")
	{
		foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["layout_tags"] as $tag)
		{
			$listlayout[] = $tag;
		}
	}

	//wordt gebruikt om lijsten met calculations te herladen op een update in een child window
	$reload_list_on_child_update = 0;

	$listlayout_validated = array();
	foreach ($listlayout as $key)
	{
		$table_parts = explode("-", $key);
		$last_field = $table_parts[count($table_parts)-1];
		$last_table = $table_parts[count($table_parts)-2];

		//de eerste conditie gooit opgevraagde velden weg die niet bestaan in de datastructure van deze user
		//de tweede conditie gooit dubbele velden weg; ieder veld kan maar 1x worden weergegeven in een lijst
		if ((in_array($key, $GLOBALS['UC']['PersonalDataStructure'][$table]["allfields"])) && (!in_array($key, $listlayout_validated)) && (!$GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["hide_from_list"]))
		{
			$listlayout_validated[] = $key;
			if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["calculation"] != "")
			{
				$reload_list_on_child_update = 1;
			}
		}
	}
	$listlayout = $listlayout_validated;

	//listlayout wordt ook gedisplayed; maar we zouden hieronder meer velden kunnen toevoegen aan de select die we niet willen displayen
	$listlayout_for_select = $listlayout;

	if ($type_of_list == "sublist")
	{
		//in sublisten op een edit formulier worden vaak de onderliggende records weergegeven; we willen dan niet de kolom tonen waar het hiddenfilter op gezet is (het referfield) daarom gooien we die kolom hier uit de layout
		//update 2014-04-07 we doen dit alleen op de kolommen waar slechts 1 IS filter op zit. Als er meer dan 1 IS filter op zit dan beelden we hem wel af, ook als er een NOT filter op zit beelden we hem af, want in die gevallen bevat de kolom nog onderscheidende informatie
		$columns_to_remove_from_listlayout = array();
		foreach (array_keys($hiddenfilter) as $key)
		{
			//slechts 1 IS filter
			if ((count($hiddenfilter[$key]) == 1) && ($hiddenfiltertype[$key] != "NOT"))
			{
				$columns_to_remove_from_listlayout[] = $key;
			}
		}

		if (count($columns_to_remove_from_listlayout) > 0)
		{
			//http://stackoverflow.com/questions/1883421/removing-array-item-by-value
			$listlayout = array_diff($listlayout, $columns_to_remove_from_listlayout);
		}
	}

	if ($tabletype == "select")
	{
		foreach ($GLOBALS['UC']['PersonalDataStructure'][$childtable]["fields"][$childfield]["layout_tags"] as $tag)
		{
			//omdat we 1 tabel hoger selecteren moeten we de huidige tabel en field name van de tags afhalen want die gaat er vanuit dat er vanaf 1 niveau LAGER wordt geparsed. DEZE CODE STAAT MEERDERE KEREN IN DEZE FILE
			$tag_zonder_parent_referentie = preg_replace("/^" . $childtable . "-" . $childfield . "-/", "", $tag);
			//we voegen alle velden in het refer field die we nodig hebben voor de parent table toe. er zouden in het refer field velden kunnen staan die we niet displayen. daarom MOETEN we ze hier hiertoevoegen anders blijft dat deel van het refer field leeg
			$listlayout_for_select[] = $tag_zonder_parent_referentie;
		}
	}

	$filtercolumn = "";
	$connecting_field = "";

	$use_pre_select_records = false;
	$pre_select_records = array();
	$order_by_pre_select = false;

	//tijdelijk in gebruik voor listindexedsearch
	if ($listoptions["use_pre_select_records"])
	{
		$use_pre_select_records = true;
		$pre_select_records = $listoptions["pre_select_records"];
	}

	if ($recent > 0)
	{
		//TODO moeten we niet selecteren op 1 van onderstaande waardes? nu voegen we wellicht journal regels toe voor een record via een module zonder dat iemand die per se geopend heeft
		//Flextable record viewed/edited

		$t1 = microtime(true);
		$pre_select_records = db_GetFlatArray("SELECT DISTINCT(`eid`) FROM `CRMjournal` WHERE `type`='" . $GLOBALS['UC']['PersonalDataStructure'][$table]["name_in_database"] . "' AND `user`='" . mres($GLOBALS['USERID']) . "' AND `eid`>0 ORDER BY `id` DESC LIMIT " . $recent);

		$t2 = microtime(true);
		$t = $t2 - $t1;

		if ($GLOBALS['SesMem']['debug'])
		{
			$output .= '<h1>recent query (' . $listid . ')</h1>';
			$output .= '<div>runtime: ' . $t . '</div>';
		}

		$order_by_pre_select = true;
		$use_pre_select_records = true;
		$hide_sortby = true; //de sort is immers gefixeerd op recentness

		//kan niet in combi met selection werken
		$hide_selections = true;
	}

	$available_selectionsnew = array();
	if (!$hide_selections)
	{
		$available_selectionsnew["system"] = GetAttribute("system", "SelectionNew" . $table, 1);
		$available_selectionsnew["user"] = GetAttribute("user", "SelectionNew" . $table, $GLOBALS['USERID']);

		if ($selectionnew != "")
		{
			//dit gaat verkeerd wanneer er een system en een user selection is met dezelfde naam
			$selection_name = preg_replace("/^(system|user)\:/", "", $selectionnew);
			$selectionnew = "";

			foreach (array("system", "user") as $type)
			{
				if (is_array($available_selectionsnew[$type][$selection_name]))
				{
					$filter = array_merge($filter, $available_selectionsnew[$type][$selection_name]["filter"]);
					$filtertype = array_merge($filtertype, $available_selectionsnew[$type][$selection_name]["filtertype"]);
					//dit is heel vies; om verdubbeling van zoektermen te voorkomen
					if ($search != $available_selectionsnew[$type][$selection_name]["search"])
					{
						$search = trim($search . " " . $available_selectionsnew[$type][$selection_name]["search"]);
					}
					$searchperfield = array_merge($searchperfield, $available_selectionsnew[$type][$selection_name]["searchperfield"]);

					//om zeker te weten dat we de goeie type hebben; ook handig voor de eerste conversie van old2new selection
					$selectionnew = $type . ":" . $selection_name;

					break;
				}
			}
		}
	}

//update filter en hidden filter worden nu los doorgegeven aan make_query en dus niet meer gemerged
//dit geldt nog niet voor search per field en all search
	$allsearchperfield = array_merge($searchperfield, $hiddensearchperfield);
	//trim is wat lomp maar wel makkelijk
	$allsearch = trim($search . " " . $hiddensearch);

	//afhankelijk van preselect
	$allsearch_not_preselect = $allsearch;

	//deze MOET op false staan
	$bypass_record_access_by_sql = false;

	//alleen de pre_select_records wanneer er een search is die ALLE parents en CHILDS joint en dus traag is
	//omdat dit zo traag is voeren we het slechts 1x uit en gebruiken we voor vervolgqueries (voor filter) deze lijst van primary id's
	if (($allsearch != "") && (!$GLOBALS['UC']['PersonalDataStructure'][$table]["compoundprimarykey"]))
	{
		$tmp_start = 0;
		$tmp_perpage = 0;
		$tmp_sortby = "";
		$tmp_sortdirection = "";
		$tmp_listlayout = array($primary_key);
		$tmp_return_countquery = false;
		$tmp_do_not_check_write_access = true;

		$t1 = microtime(true);

		$query = make_query($table, $allsearch, $allsearchperfield, $tmp_start, $tmp_perpage, $tmp_sortby, $tmp_sortdirection, $filtercolumn, $filter, $filtertype, $tmp_listlayout, $RemoveExtraSelectCondition, $connecting_field, $use_pre_select_records, $pre_select_records, $order_by_pre_select, $search_parents, $search_childs, $search_files, $tmp_return_countquery, $type_of_list, $bypass_record_access_by_sql, $tabletype, $childtable, $childfield, $childrecordid, $already_joined, $hiddenfilter, $hiddenfiltertype, $listid, $tmp_do_not_check_write_access);

		//pre_select_records kan al gevuld zijn wanneer er een saved selection sql gebruikt wordt. Die pre_select_records zijn HIERBOVEN al aan make_query() meegegeven, en alleen die records waarbij OOK de allsearch matcht, moeten dus in pre_select_records blijven staan. Daarom gooien we hier de array leeg.
		$pre_select_records = array();
		//in het geval van een selection was pre_select_records al gevuld; maar die gaan we nu overschrijven met de (mogelijk) kleinere resultset uit de search
		//we zouden hier misschien ook db_GetFlatArray kunnen gebruiken, indien primary_key altijd het eerste (en enige) element is wat geselecteerd wordt
		foreach (db_GetArray($query) as $record)
		{
			$pre_select_records[] = $record[$primary_key];
		}

		$t2 = microtime(true);
		$t = $t2 - $t1;

		if ($GLOBALS['SesMem']['debug'])
		{
			$output .= '<h1>preselect search query (' . $listid . ')</h1>';
			$output .= '<div>' . htme($query) . '</div>';
			$output .= '<div>runtime: ' . $t . '</div>';
		}

		$use_pre_select_records = true;

		//we limiteren de lijst niet in werkelijkheid, maar highlighten alleen de gevonden woorden, terwijl de rest van de records zichtbaar blijven. momenteel alleen in gebruik voor de filelist
		//de gevonden records worden in het geheel gehighlight op basis van dat het recordid in $pre_select_records() zit
		if ($search_highlight_only)
		{
			$use_pre_select_records = false;
		}

		//search is al uitgevoerd als preselect; dat gaan we dus niet opnieuw doen
		$allsearch_not_preselect = "";
	}


	//losse var omdat we deze sort niet als zodanig willen weergeven in de UI, omdat de call naar make_query weer in string context is implode-en we hem hier. dit is niet zo fraai
	$sortby_including_group = implode(",",$sortby_arr);
	$sortdirection_including_group = implode(",",$sortdirection_arr);
	if (($group != "") && ($group != $sortby_arr[0]))
	{
		$sortby_including_group = $group . "," . $sortby_including_group;
		//group wordt altijd ASC gesort, als er descending gesort moet worden neem dan de group column als eerste sortby op (inclusief sortbydirection=desc) zodat bovenstaande conditie ($group != $sortby_arr[0]) hier niet terecht komt
		$sortdirection_including_group = "asc," . $sortdirection_including_group;
	}

	$number_of_records = 0;
	if ($run_countquery)
	{
		$return_countquery = true;
		//we willen geen 1 veld selecteren zodat make_query() geen onnodige left joins hoeft te leggen die de count(*) query flink kunnen vertragen EN die toch geen invloed hebben op het aantal records
		//we moeten hier minimaal 1 veld in zetten anders retourneert make_query_parts ALLE velden; TODO kunnen we dat niet omschrijven?
		//20150726
		//2016-01-08 niet meer nodig gebeurt ook al in make_query()
		$countquery_listlayout = array($primary_key);
		$tmp_do_not_check_write_access = true;

		$t1 = microtime(true);

		$countquery = make_query($table, $allsearch_not_preselect, $allsearchperfield, $start, $perpage, $sortby_including_group, $sortdirection_including_group, $filtercolumn, $filter, $filtertype, $countquery_listlayout, $RemoveExtraSelectCondition, $connecting_field, $use_pre_select_records, $pre_select_records, $order_by_pre_select, $search_parents, $search_childs, $search_files, $return_countquery, $type_of_list, $bypass_record_access_by_sql, $tabletype, $childtable, $childfield, $childrecordid, $already_joined, $hiddenfilter, $hiddenfiltertype, $listid, $tmp_do_not_check_write_access);
		$count_and_sums = db_GetRow($countquery);
		$number_of_records = $count_and_sums['number_of_records'];

		$t2 = microtime(true);
		$t = $t2 - $t1;

		if ($GLOBALS['SesMem']['debug'])
		{
			$output .= '<h1>count query (' . $listid . ')</h1>';
			$output .= '<div>' . htme($countquery) . '</div>';
			$output .= '<div>number of records: ' . $number_of_records . '</div>';
			$output .= '<div>runtime: ' . $t . '</div>';
		}
	}

	$max_sort_records = 25000;
	if (($number_of_records > $max_sort_records) && (!$is_admin))
	{
		if ($sortby_including_group != $table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"])
		{
//			log_msg("ERROR: " . $number_of_records . " results in list " . $listid . ", sort adjusted to primary key");
			$sortby = $table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"];
			$sortby_including_group = $sortby;
			$sortby_arr = array($sortby);
			//take first element
			$sortdirection = $sortdirection_arr[0];
			$sortdirection_including_group = $sortdirection;
			$sortdirection_arr = array($sortdirection);
			if ($html_output)
			{
				$output .= '<div class="warnmsg">&gt; ' . $max_sort_records . ' ' . htme($GLOBALS['UC']['PersonalDataStructure'][$table]["plural"] . ", " . $lang["sort_" . $sortdirection_arr[0]] . ' = ' . $lang["id"]) . '</div>';
			}
		}
	}

	//2 dimensionale array; alleen in gebruik voor outputtype == excel
	$exceldata = array();
	//beetje lomp; alleen in gebruik voor outputtype == array
	$arraydata = array();

	if (!$countonly)
	{
		//deze staat in het divform en moet daarom buiten onderstaande lus gedeclareerd worden
		$div_all_filters = '';

		if (!$GLOBALS['UC']['PersonalDataStructure'][$table]["allowed_to_add"])
		{
			$hide_addlink = true;
			$hide_addrow = true;
		}

		//waardes te prefillen op een add form
		$prefill = array();
		if ((!$hide_addlink) || (!$hide_addrow))
		{
			//alleen de eigen tabel (en niet op bovenliggende tabellen)
			//alleen op IS filter (en niet op NOT)
			//alleen als er een enkel filter is gezet (en niet 2 of meer)
			//alleen op relation_to velden en text velden (dat zijn ook dropdowns) (en niet op bv. datum of numerieke (from-to) filters)

			//TODO: ook support voor meerdere filters op 1 multiselect field
			foreach (array_keys($hiddenfilter) as $key)
			{
				$table_parts = explode("-", $key);
				$last_field = $table_parts[count($table_parts)-1];
				$last_table = $table_parts[count($table_parts)-2];
				if ((count($table_parts) == 2) && ($hiddenfiltertype[$key] != "NOT") && (count($hiddenfilter[$key]) == 1) && (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["relation_to"] != "") || ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "text")))
				{
					$prefill[$key] = $hiddenfilter[$key][0];
				}
			}
//JH 2016-11-15 reguliere filters worden niet meer meegenomen naar een add form en de add row
/*
			foreach (array_keys($filter) as $key)
			{
				$table_parts = explode("-", $key);
				$last_field = $table_parts[count($table_parts)-1];
				$last_table = $table_parts[count($table_parts)-2];
				if ((count($table_parts) == 2) && ($filtertype[$key] != "NOT") && (count($filter[$key]) == 1) && (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["relation_to"] != "") || ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "text")))
				{
					$prefill[$key] = $filter[$key][0];
				}
			}
*/
		}

		$addlink_html = '';
		//2015-07-24 zodat all_records_in_this_list_readonly ook op true gaat wanneer hide_all aan staat zodat de inline delete verborgen wordt
//		if (!$hide_addlink)
//		{
			if ($html_output)
			{
				if ($GLOBALS['UC']['PersonalDataStructure'][$table]["allowed_to_add"])
				{
					$addlink_starttable = "";
					$addlink_startrecordid = 0;

					//update 2014-03-24 select toegevoegd
					if (($type_of_list == "sublist") || ($tabletype == "select"))
					{
						foreach (array_keys($hiddenfilter) as $key)
						{
							$table_parts = explode("-", $key);
							$last_field = $table_parts[count($table_parts)-1];
							$last_table = $table_parts[count($table_parts)-2];
							if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["relation_to"] != "")
							{
								if (count($hiddenfilter[$key]) == 1)
								{
									//als er een hiddenfilter gezet is op de sublist die verwijst naar een andere tabel dan maken we nieuwe records in deze subtable aan met deze waarde geprefilled; alleen als er maar 1 hiddenfilter op dit veld is
									$addlink_starttable = $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["relation_to"];
									$addlink_startrecordid = $hiddenfilter[$key][0];

									if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["InheritWriteAccessFromParent"])
									{
										$parent_table = $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["relation_to"];
										$record_access_parent = CheckTableRecordAccess($parent_table, $addlink_startrecordid);
										//als de parent readonly is, en omdat we weten dat record access overerft (InheritWriteAccessFromParent) dan kan deze gebruiker geen records toevoegen in deze child tabel; dan kunnen we de addlink dus hiden (of misschien uitgrijzen in de toekomst)
										if ($record_access_parent == "readonly")
										{
											$hide_addlink = true;
											$hide_addrow = true;
											//wordt gebruikt om inline delete te hiden
											$all_records_in_this_list_readonly = true;
										}

										break;
									}
								}
							}
						}
					}

					//hide_addlink kan nu true zijn door de CheckTableRecordAccess op de parent
					if (!$hide_addlink)
					{
						$addlink_html .= '<div class="addlink">';

						$popupsize = 0;
						$addlink = $GLOBALS['UC']['PersonalDataStructure'][$table]["addlink"];

						//hiermee wordt direct het toegevoegde record in de child gezet en het parent add venster gesloten; we kunnen dat niet doen wanneer selectvia uit meerdere stappen bestaat, in dat geval openen add forms in popups
						if ($tabletype == "select")
						{
							if ($this_is_the_last_selectvia_step)
							{
								//geen popup wanneer uit select; maar wel nonavbar want we zitten dan al in een popup die we hergebruiken -> dit is niet fraai
								//put_id_in_child werkt met dom id's niet mooi
								$addlink .= '&nonavbar=1&put_id_in_child=' . $childtable . '-' . $childfield . '-';
								if ($childrecordid > 0)
								{
									$addlink .= $childrecordid;
								}
								else
								{
									$addlink .= '_new_';
								}
							}
							else
							{
								//we zijn bezig met het aflopen van een selectvia path; en moet nu een nieuwe lijst inladen om de volgende stap te zetten
								//dit is geen fraaie oplossing want in put_id_in_child staat het woord "selectvia" gevolgd door de base64 gecodeerde volgende url waar alleen nog het net toegevoegde recordid achter gezet moet worden
								$selectviafilter_new = $selectviafilter;
								if ($selectviafilter_new != '')
								{
									$selectviafilter_new .= ',';
								}
								$selectviafilter_new .= $table . '-'; //hier moet nog nieuw toegevoegd id achter
								$next_selectvia_url = 'list.php?childrecordid=' . $childrecordid . '&nonavbar=1&selectvia=' . $selectvia . '&selectviafilter=' . $selectviafilter_new;
								$addlink .= '&nonavbar=1&put_id_in_child=selectvia' . base64_encode($next_selectvia_url);
							}
						}
						else
						{
							$popupsize = $GLOBALS['UC']['PersonalDataStructure'][$table]["open_records_in_popup_" . $type_of_list];
						}

						//2015-09-01 om ALLE filters te prefillen; we doen dat toch maar niet omdat mensen hier per ongeluk een status op GEFACTUREERD kunnen zetten wat niet de bedoeling is. dit is waarschijnlijk in 95% van de gevallen behoorlijk handig, en in 5% verschrikkelijk onhandig
						foreach ($prefill as $prefill_field => $prefill_value)
						{
							$addlink .= '&' . rawurlencode($prefill_field) . '=' . rawurlencode($prefill_value);
						}

						if (($addlink_starttable != "") && ($addlink_startrecordid > 0))
						{
							$addlink .= '&starttable=' . $addlink_starttable . '&startrecordid=' . $addlink_startrecordid;
						}

						$addlink .= '&add=1';

						$addlink_html .= '<input type="button" class="addbutton" onclick="ilor(\'' . inline_jsencode($addlink) . '\',' . $popupsize . ',\'' . $listid . '\',0)" value="' . htme($GLOBALS['UC']['PersonalDataStructure'][$table]["addlinktext"]) . '">';
						$addlink_html .= '</div>';
					}
				}
			}
//		}


		if (($run_countquery) && ($number_of_records == 0) && ($hide_tablewhennoresults))
		{
			//alleen in gebruik voor files lists
			//de addlink wordt weergegeven zodat je nieuwe records kan toevoegen
			if ($tabletype != "plain")
			{
				$output .= $addlink_html;
			}
		}
		else
		{
			$return_countquery = false;

			$t1 = microtime(true);

			$query = make_query($table, $allsearch_not_preselect, $allsearchperfield, $start, $perpage, $sortby_including_group, $sortdirection_including_group, $filtercolumn, $filter, $filtertype, $listlayout_for_select, $RemoveExtraSelectCondition, $connecting_field, $use_pre_select_records, $pre_select_records, $order_by_pre_select, $search_parents, $search_childs, $search_files, $return_countquery, $type_of_list, $bypass_record_access_by_sql, $tabletype, $childtable, $childfield, $childrecordid, $already_joined, $hiddenfilter, $hiddenfiltertype, $listid, $do_not_check_write_access);

			$close_listinner_div = false;
			if (!$only_data)
			{
				if ($outputtype == "table")
				{
					//deze heeft een display: inline-block zodat we de excel icon rechts kunnen floaten gelijk met het einde van de tabel
					$output .= '<div class="listinner">';
					$close_listinner_div = true;
				}
			}

			$dqs_stash_id = 0;

			//dqs alleen opbouwen als er modules in de lijst staan
			//arrows opbouwen niet meer op basis van dqs maar op basis van listid en sameresultandpage. pas doen zodra iemand op de knop klikt en niet bij maken van formulier
			//2015-11-19 dqs staat aan voor alle admin tables

			//2016-10-13: niet voor downloads en array output
			if ($html_output)
			{
				if (($GLOBALS['UC']['PersonalDataStructure'][$table]["dqs_enabled"]) || (!$GLOBALS['UC']['PersonalDataStructure'][$table]["datatable"]))
				{
					$t1_dqs = microtime(true);

					//in de dqs query staat de query die gerunt moet worden om de ongelimiteerde lijst terug te krijgen met primary id's; deze is nodig voor inline module support en voor de browsearray voor de arrows
					$dqs_perpage = 0;
					//we selecteren alleen het primary field; da's snel
					$dqs_listlayout = array();
					$dqs_listlayout[] = $table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"];
					$tmp_do_not_check_write_access = true;

					$dqs_stash = make_query($table, $allsearch_not_preselect, $allsearchperfield, $start, $dqs_perpage, $sortby_including_group, $sortdirection_including_group, $filtercolumn, $filter, $filtertype, $dqs_listlayout, $RemoveExtraSelectCondition, $connecting_field, $use_pre_select_records, $pre_select_records, $order_by_pre_select, $search_parents, $search_childs, $search_files, $return_countquery, $type_of_list, $bypass_record_access_by_sql, $tabletype, $childtable, $childfield, $childrecordid, $already_joined, $hiddenfilter, $hiddenfiltertype, $listid, $tmp_do_not_check_write_access);

					if ($GLOBALS['UC']['PersonalDataStructure'][$table]["dqs_fixedset"])
					{
						$dqs_stash = db_GetArray($dqs_stash);
					}

					//hier wordt ofwel de query of wel een array met recordid's in gezet
					//de eerste is adaptief: als je resultset tussentijds wijzigt, dan zie je die wijzigingen terwijl je (bv) met je arrows door de records loopt
					//de tweede is fixed: behalve dan dat records in de tussentijd verwijderd kunnen zijn of niet meer zichtbaar voor de gebruiker. ook niet helemaal fixed dus.
					$dqs_stash_id = PushStashValue($dqs_stash);

					$t2_dqs = microtime(true);
					$t_dqs = $t2_dqs - $t1_dqs;

					if ($GLOBALS['SesMem']['debug'])
					{
						$output .= '<h1>dqs query (' . $listid . ')</h1>';
						$output .= '<div>dqs_stash_id: ' . $dqs_stash_id . '</div>';
						$output .= '<div>runtime make_query(): ' . $t_dqs . '</div>';
					}
				}
			}

			//vanwege memory gebruik (downloaden van grote tabel via xml) zouden we MYSQLI_USE_RESULT moeten gebruiken maar er worden nog queries uitgevoerd tijdens het afbeelden van de lijst en dan kan dat niet; MySQL Error: Commands out of sync; you can't run this command now
			$MYSQLI_USE_RESULT = false;
			if ($_GET['MYSQLI_USE_RESULT'] == 1)
			{
				$MYSQLI_USE_RESULT = true;
			}
			//2e false is voor $db die we niet meegeven
			$ResultSet = mcq($query, false, $MYSQLI_USE_RESULT);

			//$ResultSet = mcq($query);

			$t2 = microtime(true);
			$t = $t2 - $t1;

			if ($GLOBALS['SesMem']['debug'])
			{
				$output .= '<h1>main query (' . $listid . ')</h1>';
				$output .= '<div>' . htme($query) . '</div>';
				$output .= '<div>runtime (RESULT SET NOT FULLY LOADED YET): ' . $t . '</div>';
			}

			$number_of_records_on_this_page = $perpage;
			if ($number_of_records_on_this_page == 0)
			{
				$number_of_records_on_this_page = $number_of_records;
			}

			$first_record_on_this_page = $start;

			//voor de laatste pagina
			if ($first_record_on_this_page + $number_of_records_on_this_page > $number_of_records)
			{
				$number_of_records_on_this_page = $number_of_records - $first_record_on_this_page;
			}

			$last_record_on_this_page = $first_record_on_this_page + $number_of_records_on_this_page;

			//when start is larger than the result set
			if ($number_of_records_on_this_page < 0)
			{
				$number_of_records_on_this_page = 0;
			}

			$number_of_records_before_this_page = $first_record_on_this_page;
			$number_of_records_after_this_page = $number_of_records - $last_record_on_this_page;

			//before en after moet los uitgerekend worden omdat start 40 zou kunnen zijn en per page 50
			$number_of_pages_before_this_page = floor($number_of_records_before_this_page / $perpage);

			$number_of_pages_after_this_page = 0;

			/* dit levert een division by 0 op in PHP7, fix HF20160721 door if >0 eromheen te zetten TODO JEROEN is dit ok? */
			if ($perpage > 0)
			{
				if ($number_of_records_before_this_page % $perpage > 0)
				{
					$number_of_pages_before_this_page++;
				}

				$number_of_pages_after_this_page = floor($number_of_records_after_this_page / $perpage);
				if ($number_of_records_after_this_page % $perpage > 0)
				{
					$number_of_pages_after_this_page++;
				}
			}
			$number_of_pages = $number_of_pages_before_this_page + $number_of_pages_after_this_page + 1;

//			$current_page = $number_of_pages_before_this_page + 1;

			$startprint = $first_record_on_this_page;
			if ($number_of_records_on_this_page > 0)
			{
				$startprint++;
			}

			//als het aantal records op deze pagina groter is dan je download_limit dan mag je de pagina niet zien, dit geldt zowel voor downloads als voor onscreen lijsten
			if (($GLOBALS['UC']['PersonalDataStructure'][$table]["download_limit"] > 0) && ($number_of_records_on_this_page > $GLOBALS['UC']['PersonalDataStructure'][$table]["download_limit"]))
			{
				$errortext = "Too many records selected; access to download this amount was denied by your administrator. Select fewer records.";
				PrintAD($errortext);
				log_msg("WARNING: " . $errortext);
//				max_one_concurrent_load_finish($unique_command_id);
				return array(0, $errortext);
			}

			$full_fieldnames_displaylistname = array();
			$full_fieldnames = array();

			//rechtstreekse call mysqli
			while ($fieldinfo = mysqli_fetch_field($ResultSet))
			{
				$key = $fieldinfo->name;

				//voor custom toegevoegde velden via ModuleToEvaluateOnEveryListBuild
				if (preg_match("/^ShowThisField\-(.*)$/", $key, $match))
				{
					$full_fieldnames[$key] = $match[1];
					$full_fieldnames_displaylistname[$key] = $match[1];
					$listlayout[] = $key;
				}
				else
				{
					$full_fieldnames_displaylistname[$key] = fieldalias_to_fullname($key);
					$full_fieldnames[$key] = fieldalias_to_fullname($key, false);
				}
			}

			if ($outputtype == "xml")
			{
				//er is op dit moment nog geen enkele output (als het goed is)
				$output = '<?xml version="1.0" encoding="UTF-8"?>';
				if ($xml_ms_excel)
				{
					$output .= '<?mso-application progid="Excel.Sheet"?>';
					$output .= '<Workbook xmlns="urn:schemas-microsoft-com:office:spreadsheet" xmlns:x="urn:schemas-microsoft-com:office:excel" xmlns:ss="urn:schemas-microsoft-com:office:spreadsheet" xmlns:html="http://www.w3.org/TR/REC-html40">';
					$output .= '<Styles>';
					$output .= '<Style ss:ID="s62">';
					$output .= '<NumberFormat ss:Format="Short Date"/>';
					$output .= '</Style>';
					$output .= '<Style ss:ID="s63">';
					$output .= '<NumberFormat ss:Format="General Date"/>';
					$output .= '</Style>';
					$output .= '</Styles>';
					$output .= '<Worksheet ss:Name="' . htme($GLOBALS['UC']['PersonalDataStructure'][$table]["plural"]) . '">';
					$output .= '<Table>';
					$output .= '<Column ss:Index="1" ss:AutoFitWidth="0" ss:Width="110"/>';
				}
				else
				{
					$output .= '<interleave';
					if ($xml_with_xsd)
					{
						$output .= ' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="' . $GLOBALS['BASE'] . 'list.php?xsd=' . implode(",", $listlayout) . '"';
					}
					$output .= '>';
				}
			}

			$datefilteroptions = array();

			$datefilteroptions[] = array("PAST", $lang["inthepast"], 0);
			$datefilteroptions[] = array("FUTURE", $lang["inthefuture"], 0);
			$datefilteroptions[] = array("", "", -1);

			//NU: THIS
			$datefilteroptions[] = array("TD", $lang["today"], 0);
			$datefilteroptions[] = array("TW", $lang["thisweek"], 0);
			$datefilteroptions[] = array("TM", $lang["thismonth"], 0);
			$datefilteroptions[] = array("TQ", $lang["thisquarter"], 0);
			$datefilteroptions[] = array("TY", $lang["thisyear"], 0);
			$datefilteroptions[] = array("", "", -1);

			//TOEKOMST: NEXT
			$datefilteroptions[] = array("ND", $lang["tomorrow"], 0);
			$datefilteroptions[] = array("NW", $lang["nextweek"], 0);
			$datefilteroptions[] = array("NM", $lang["nextmonth"], 0);
			$datefilteroptions[] = array("NQ", $lang["nextquarter"], 0);
			$datefilteroptions[] = array("NY", $lang["nextyear"], 0);
			$datefilteroptions[] = array("", "", -1);

			//VERLEDEN: LAST
			$datefilteroptions[] = array("LD", $lang["yesterday"], 0);
			$datefilteroptions[] = array("LW", $lang["lastweek"], 0);
			$datefilteroptions[] = array("LM", $lang["lastmonth"], 0);
			$datefilteroptions[] = array("LQ", $lang["lastquarter"], 0);
			$datefilteroptions[] = array("LY", $lang["lastyear"], 0);
			$datefilteroptions[] = array("", "", -1);

			//dit is nogal lomp; zou adaptief moeten zijn
			$current_year = date("Y");
			for ($year = $current_year - 2; $year > $current_year - 5 ; $year--)
			{
				$datefilteroptions[] = array($year, $year, 0);
			}
			$datefilteroptions[] = array("", "", -1);

			for ($quarter = 1; $quarter <= 4; $quarter++)
			{
				$datefilteroptions[] = array("TYQ" . $quarter, $lang["thisyearq" . $quarter], 0);
			}
			$datefilteroptions[] = array("", "", -1);
			//eigenlijk hier ook toekomst!
			for ($quarter = 1; $quarter <= 4; $quarter++)
			{
				$datefilteroptions[] = array("LYQ" . $quarter, $lang["lastyearq" . $quarter], 0);
			}
			$datefilteroptions[] = array("", "", -1);

			for ($month = 1; $month <= 12; $month++)
			{
				$datefilteroptions[] = array("TYM" . $month, $lang["thisyear"] . " " . $lang["month" . $month], 0);
			}
			$datefilteroptions[] = array("", "", -1);
			//eigenlijk hier ook toekomst!
			for ($month = 1; $month <= 12; $month++)
			{
				$datefilteroptions[] = array("LYM" . $month, $lang["lastyear"] . " " . $lang["month" . $month], 0);
			}
			$datefilteroptions[] = array("", "", -1);

			for ($week = 1; $week <= 53; $week++)
			{
				$datefilteroptions[] = array("TYW" . $week, $lang["thisyear"] . " " . $lang["week"] . " " . $week, 0);
			}
			$datefilteroptions[] = array("", "", -1);
			//eigenlijk hier ook toekomst!
			for ($week = 1; $week <= 53; $week++)
			{
				$datefilteroptions[] = array("LYW" . $week, $lang["lastyear"] . " " . $lang["week"] . " " . $week, 0);
			}

			//ook als hide_selections aan staat moeten we dit runnen omdat we de naam van de selectie boven de lijst weer geven

			//PAGINATOR: deze wordt in losse var gezet omdat die nogmaals onder de lijst gezet kan worden
			$paginator = '';

			if (($html_output) && (!$only_data))
			{
				if (($tabletype == "normal") || ($tabletype == "select"))
				{
					if (!$hide_header)
					{
						$headerhtml = '';

						$headername = "";
						$headersize = "h1";
						if ($type_of_list == "sublist")
						{
							$headersize = "h2";
						}

						if ($tabletype == "select")
						{
							//we gebruiken de naam van het select veld als header, we zouden hier eigenlijk "Kies een ...." voor moeten zetten
							//$headername = $GLOBALS['UC']['PersonalDataStructure'][$childtable]["fields"][$childfield]["name"];
							$headerhtml = '<' . $headersize . '>' . $selectvia_output . ' (' . FormatNumberNEW($number_of_records) . ')</' . $headersize . '>';
						}
						//een custom header is gezet via listoptions
						else if ($customheader != "")
						{
							$headername = $customheader;
						}
						else if ($GLOBALS['UC']['PersonalDataStructure'][$table]["headerhtml"] != "")
						{
							$headerhtml = get_parsed_template_with_given_template("", 0, array(), array(), false, "", $GLOBALS['UC']['PersonalDataStructure'][$table]["headerhtml"], "DEFAULTVALUE");
						}
						//geen html header ingesteld, dus maken we hem zelf
						else
						{
							$headername = $GLOBALS['UC']['PersonalDataStructure'][$table]["plural"];
						}

						if ($headerhtml == "")
						{
							if ($selectionnew != "")
							{
								if ($headername != "")
								{
									$headername .= ": ";
								}
								$selection_name = preg_replace("/^(system|user)\:/", "", $selectionnew);
								$headername .= $selection_name;
							}
							$headerhtml .= '<' . $headersize . '>' . htme($headername) . ' (' . FormatNumberNEW($number_of_records) . ')</' . $headersize . '>';
						}

						$output .= $headerhtml;
					}

					$number_of_filters_set = 0;

					//deze HRF code draait ook wanneer de HRF verborgen wordt; dat is nodig omdat we de number_of_filters_set variabele hier optellen die later gebruikt wordt voor het weergeven van de clearallfilters
					if (($tabletype == "normal") || ($tabletype == "select"))
					{
						$human_readable_filter = '';

						if ($search != "")
						{
							list($searchtype, $searchparts) = searchstring_to_searchparts($search);
							for ($searchpart_counter = 0; $searchpart_counter < count($searchparts); $searchpart_counter++)
							{
								$searchpart = $searchparts[$searchpart_counter]["str"];
								$search_without_this_searchpart = searchparts_to_searchstring($searchtype, $searchparts, $searchpart_counter);

								//kijken naar date num en formateren
								$langtag = 'mustcontain';
								if (!$searchparts[$searchpart_counter]["is_positive"])
								{
									$langtag = 'mustnotcontain';
								}
								$human_readable_filter .= '<li>' . htme($lang[$langtag]) . ' <strong>' . htme($searchpart) . '</strong> <a><label onclick="ListSetSearch(\'' . $listid . '\',\'' . inline_jsencode($search_without_this_searchpart) . '\',\'' . inline_jsencode($GLOBALS['UC']['PersonalDataStructure'][$table]['openlink']) . '\',' . $GLOBALS['UC']['PersonalDataStructure'][$table]["open_records_in_popup_" . $type_of_list] . ')"><span class="fa fa-times"></span></label></a></li>';
								$number_of_filters_set++;
							}
						}
						//hidden wordt genegeerd
						foreach ($searchperfield as $searchperfieldkey => $searchperfieldvalue)
						{
							if ($searchperfieldvalue != "")
							{
								list($searchtype, $searchparts) = searchstring_to_searchparts($searchperfieldvalue);
								for ($searchpart_counter = 0; $searchpart_counter < count($searchparts); $searchpart_counter++)
								{
									$searchpart = $searchparts[$searchpart_counter]["str"];
									$search_without_this_searchpart = searchparts_to_searchstring($searchtype, $searchparts, $searchpart_counter);

									//kijken naar date num en formateren
									$langtag = 'mustcontain';
									if (!$searchparts[$searchpart_counter]["is_positive"])
									{
										$langtag = 'mustnotcontain';
									}
									$human_readable_filter .= '<li>' . htme(fieldalias_to_fullname($searchperfieldkey)) . ' ' . htme($lang[$langtag]) . ' <strong>' . htme($searchpart) . '</strong> <a><label onclick="ListSetSearchperfield(\'' . $listid . '\',\'' . inline_jsencode($searchperfieldkey) . '\',\'' . inline_jsencode($search_without_this_searchpart) . '\')"><span class="fa fa-times"></span></label></a></li>';
									$number_of_filters_set++;
								}
							}
						}

						$from_to_done = array();

						//let op: we lussen alleen door $filter en niet door $hiddenfilter
						foreach (array_keys($filter) as $key)
						{
							//CODE DIE HIER VEEL OP LIJKT STAAT OP 5 PLEKKEN; number_of_filters_set, HIGHLIGHT TH EN HIGHLIGHT TD, HUMAN_READABLE_FILTER EN DE OPBOUW VAN SQL QUERIES OP BASIS VAN FILTERS
							foreach ($filter[$key] as $key2 => $val)
							{
								//key2 is lomp; zie comment bij numeric filter
								$table_parts = explode("-", $key);
								$last_field = $table_parts[count($table_parts)-1];
								$last_table = $table_parts[count($table_parts)-2];

								//deze conditie moet VOOR de from/to condities staan omdat als key2 "0" is dit evalueeert op key2 == "to" of "from"
								if (is_numeric($key2))
								{
									$displayval = $val;
									$show_fieldname = true;

									//TODO deze code is haast identiek aan het toevoegen van gezette filter die niet meer in de array staan
									if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["layout"] != "")
									{
										//stom om get_refer_layout_for_record_field re runnen die getrecord runt voor ieder filter, terwijl die zelfde filters verderop opnieuw gequeried worden (behalve als er gefilterd wordt op een veld waar het filter uit staat)
										$displayval = get_refer_layout_for_record_field($last_table, $last_field, $displayval);
									}
									else if ((count($table_parts) == 2) && ($last_field == $GLOBALS['UC']['PersonalDataStructure'][$last_table]["primary"]))
									{
										//to display "must be Jeroen" instead of "id must be Jeroen"
										$show_fieldname = false;
										$displayval = get_refer_layout_for_record($last_table, $displayval);
									}
									else if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "date")
									{
										for ($x = 0; $x < count($datefilteroptions); $x++)
										{
											if ($datefilteroptions[$x][0] == $displayval)
											{
												$displayval = $datefilteroptions[$x][1];
												continue;
											}
										}
									}

									$langtag = 'mustbe';
									if ($filtertype[$key] == "NOT")
									{
										$langtag = 'mustnotbe';
									}

									$human_readable_filter .= '<li>';
									if ($show_fieldname)
									{
										$human_readable_filter .= htme(fieldalias_to_fullname($key)) . ' ';
									}

									if (!$GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["ishtml"])
									{
									}
									$human_readable_filter .= htme($lang[$langtag]) . ' <strong>' . $displayval . '</strong> <a><label onclick="ListSetFilter(\'' . $listid . '\',\'' . htme($key) . '\',\'' . htme(whitespace_to_underscore($val)) . '\',false)"><span class="fa fa-times"></span></label></a></li>';
									$number_of_filters_set++;
								}
								//3x = want key2 kan ook numeriek zijn en dat matcht anders ook. waarom?
								else if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "number") && (($key2 === "from") || ($key2 === "to")))
								{
									//er zijn 2 filter keys voor numberic values; to en from; als we er niet 1 skippen dan worden deze waarden 2x afgebeeld als human filter
									if (!$from_to_done[$key])
									{
										list($fromvalid, $from) = is_valid_number($filter[$key]["from"]);
										list($tovalid, $to) = is_valid_number($filter[$key]["to"]);

										if (($fromvalid) || ($tovalid))
										{
											$displayval = "";
											$decimals = $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["decimals"];
											$prefix = $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["prefix"];
											$postfix = $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["postfix"];

											if (($fromvalid) && ($tovalid))
											{
												if ($from == $to)
												{
													$displayval = FormatNumberNEW($from, $decimals, $prefix, $postfix);
												}
												else
												{
													$displayval = $lang['between1'] . ' ' . FormatNumberNEW($from, $decimals, $prefix, $postfix) . ' ' . $lang['between2'] . ' ' . FormatNumberNEW($to, $decimals, $prefix, $postfix);
												}
											}
											else if ($fromvalid)
											{
												$displayval = $lang['higherthanorequal'] . ' ' . FormatNumberNEW($from, $decimals, $prefix, $postfix);
											}
											else if ($tovalid)
											{
												$displayval = $lang['lowerthanorequal'] . ' ' . FormatNumberNEW($to, $decimals, $prefix, $postfix);
											}

											$human_readable_filter .= '<li>' . htme(fieldalias_to_fullname($key)) . ' <strong>' . htme($displayval) . '</strong> <a><label onclick="ListSetFromToFilter(\'' . $listid . '\', \'' . inline_jsencode($key) . '\', \'\', \'\')"><span class="fa fa-times"></span></label></a></li>';
											$number_of_filters_set++;
										}
										$from_to_done[$key] = true;
									}
								}
								//3x = want key2 kan ook numeriek zijn en dat matcht anders ook. waarom?
								else if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "date") && (($key2 === "from") || ($key2 === "to")))
								{
									//er zijn 2 filter keys voor numberic values; to en from; als we er niet 1 skippen dan worden deze waarden 2x afgebeeld als human filter
									if (!$from_to_done[$key])
									{
										$fromval = $filter[$key]["from"];
										$toval = $filter[$key]["to"];

										//waardes valideren?
										if (($fromval != "") || ($toval != ""))
										{
											$displayval = "";
											if (($fromval != "") && ($toval != ""))
											{
												if ($fromval == $toval)
												{
													$displayval = htme($fromval);
												}
												else
												{
													$displayval = $lang['between1'] . ' ' . htme($fromval) . ' ' . $lang['between2'] . ' ' . htme($toval);
												}
											}
											else if ($fromval != "")
											{
												$displayval = $lang['higherthanorequal'] . ' ' . htme($fromval);
											}
											else if ($toval)
											{
												$displayval = $lang['lowerthanorequal'] . ' ' . htme($toval);
											}

											$human_readable_filter .= '<li>' . htme(fieldalias_to_fullname($key)) . ' <strong>' . htme($displayval) . '</strong> <a><label onclick="ListSetFromToFilter(\'' . $listid . '\', \'' . inline_jsencode($key) . '\', \'\', \'\')"><span class="fa fa-times"></span></label></a></li>';
											$number_of_filters_set++;
										}
										$from_to_done[$key] = true;
									}
								}
							}
						}

						if ((!$hide_human_readable_filter) && ($human_readable_filter != ''))
						{
							//human readable filters moeten aanstaan als je selecties op wilt kunnen slaan
							if (!$hide_selections)
							{
								if (($is_admin) || (CheckFunctionAccess("AddEditSelections") == "ok"))
								{
									$selection_name = preg_replace("/^(system|user)\:/", "", $selectionnew);

									$human_readable_filter .= '<li class="savethesefilters">' . htme($lang['savethesefilters']) . ' <a onclick="var selectionname = prompt(\'' . inline_jsencode($lang['savethesefilters'] . ". " . $lang['name'] . ":") . '\',\'' . inline_jsencode($selection_name) . '\'); if ((selectionname != null) && (selectionname != \'\')) { ListSetSaveSelectionNew(\'' . $listid . '\',\'user:\' + selectionname) };"><span class="fa fa-floppy-o"></span></a>';

									if (($is_admin) || (CheckFunctionAccess("AllowedToEditSystemWideSelections") == "ok"))
									{
										$human_readable_filter .= ' <a onclick="var selectionname = prompt(\'' . inline_jsencode($lang['savethesefilters'] . ". " . $lang['name'] . " (system):") . '\',\'' . inline_jsencode($selection_name) . '\'); if ((selectionname != null) && (selectionname != \'\')) { ListSetSaveSelectionNew(\'' . $listid . '\',\'system:\' + selectionname) };"><span class="fa fa-floppy-o"></span></a>';
									}
									$human_readable_filter .= '</li>';
								}
							}

							//als er slechts 1 filter is, dan hoeft de clear all er niet te staan
							if ($number_of_filters_set > 1)
							{
								$human_readable_filter .= '<li class="clearallfilters">' . htme($lang['clearallfilters']) . ' <a><label onclick="ListSetClearfilter(\'' . $listid . '\')"><span class="fa fa-times"></span></label></a></li>';
							}
							$output .= '<div class="human_readable_filter"><ul>' . $human_readable_filter . '</ul></div>';
						}
					}


					//PAGINATOR BEGIN
						if (!$continuous_scroll)
						{
							if (($number_of_pages > 1) || (!$hide_pagination_if_only_1_page))
							{
								//PREVIOUS PAGE
								$paginator .= '<div class="prevpage">';
								$link_active = false;
								if ($first_record_on_this_page > 0)
								{
									$link_active = true;
									$prev = $first_record_on_this_page - $perpage;
									$paginator .= '<a onclick="ListSetStart(\'' . $listid . '\',' . $prev . ')">';
								}
								$paginator .= '<span class="fa fa-arrow-left';
								if (!$link_active)
								{
									$paginator .= ' disabled';
								}
								$paginator .= '"></span>';
								if ($link_active)
								{
									$paginator .= '</a>';
								}
								$paginator .= '</div>';

								//NEXT PAGE
								$paginator .= '<div class="nextpage">';
								$link_active = false;
								if ($last_record_on_this_page < $number_of_records)
								{
									$link_active = true;
									$next = $first_record_on_this_page + $perpage;
									$paginator .= '<a onclick="ListSetStart(\'' . $listid . '\',' . $next . ')">';
								}
								$paginator .= '<span class="fa fa-arrow-right';
								if (!$link_active)
								{
									$paginator .= ' disabled';
								}
								$paginator .= '"></span>';
								if ($link_active)
								{
									$paginator .= '</a>';
								}
								$paginator .= '</div>';
							}
						}
					//PAGINATOR EIND

					if ($set_tilelistoptionsandfilters_div)
					{
						//deze lang tag is niet helemaal kloppend, maar goed genoeg
						$output .= '<label for="' . $listid . 'tileall">' . htme($lang['showallfilters']) . '</label> <input type="checkbox" class="tilelistoptionsandfiltersshowall" id="' . $listid . 'tileall">';
						$output .= '<div class="tilelistoptionsandfilters">';
					}

					//BEGIN listoptions
					$output .= '<div class="listoptions">';
					$output .= '<div class="listoptionsleft">';

					//PAGINATOR
					//deze kan nogmaals onder de lijst staan
						$output .= $paginator;

					//START AND END RECORD, TOTAL NUMBER OF RECORDS + NAME OF RECORDS
					if (!$continuous_scroll)
					{
						if ((!$hide_recordcounter) && (($number_of_pages > 1) || (!$hide_pagination_if_only_1_page)))
						{
							$output .= '<div class="recordcounter">';
							$output .= '<input type="text" id="' . $listid . 'startrecord" value="' . $startprint . '" onchange="this.value=trim(this.value); if(is_positive_integer(this.value)){$(this).removeClass(\'errorinfield\'); ListSetStart(\'' . $listid . '\',(this.value - 1))}else{$(this).addClass(\'errorinfield\');}" style="width: ' . (strlen($startprint)*0.66) . 'em;"> - ' . FormatNumberNEW($last_record_on_this_page);
							if (!$hide_totalrecords)
							{
								$output .= ' / ' . FormatNumberNEW($number_of_records);
							}
							$output .= '</div>';
						}
					}

					//ADD RECORD
						$output .= $addlink_html;

					//SEARCH BOX
						if (!$hide_search)
						{
							//id "searchinputbox" om search field leeg te maken bij direct id input
							//JH 2017-11-08 the onkeydown is there to be able to reload a list by focusing the searchbox and pressing enter
							//this is a workaround for when you press BACK in your browser; you will see the list without the search term applied, but the search term is still filled out in the searchbox
							//the REAL solution to this problem is to cache html locally (final ajax result) or use the webapp where back navigation is properly taken care of
							$output .= '<div class="searchbox"><span class="fa fa-search search_img"></span><input type="search" id="' . $listid . 'searchinputbox" value="' . htme($search) . '" class="RecordSearch enter2blur" onchange="ListSetSearch(\'' . $listid . '\',this.value,\'' . inline_jsencode($GLOBALS['UC']['PersonalDataStructure'][$table]['openlink']) . '\',' . $GLOBALS['UC']['PersonalDataStructure'][$table]["open_records_in_popup_" . $type_of_list] . ')" onkeydown="if (event.keyCode == 13) { this.onchange(); }" placeholder="' . htme($lang["search"]) . '"></div>';
						}

					//DIRECT INPUT RECORDID
						if ((!$hide_id_input) && ($GLOBALS['UC']['PersonalDataStructure'][$table]["openlink"] != ""))
						{
							$popupsize = $GLOBALS['UC']['PersonalDataStructure'][$table]["open_records_in_popup_" . $type_of_list];

							$output .= '<div class="id_input"><input type="text" id="' . $listid . 'idinput" onchange="this.value=trim(this.value); if(is_positive_integer(this.value)){$(this).removeClass(\'errorinfield\');ilor(\'' . inline_jsencode($GLOBALS['UC']['PersonalDataStructure'][$table]["openlink"]) . '\'+this.value,' . $popupsize . ',\'' . $listid . '\',0);this.value=\'\';}else{$(this).addClass(\'errorinfield\');}" class="NumericRecordInput" placeholder="' . htme($lang["id"]) . '"></div>';
						}

					//EXTRASELECTCONDITION
					//de extra select conditie kan alleen verwijderd worden als de remove text gevuld, anders wordt de link niet getoond en MAG de selectie ook niet verwijderd worden
						if (($GLOBALS['UC']['PersonalDataStructure'][$table]["ExtraSelectCondition"] != "") && ($GLOBALS['UC']['PersonalDataStructure'][$table]["RemoveExtraSelectCondionFromQueryLinkText"] != ""))
						{
							//invers; als de extraselectcondition erop zit kan ie eraf, zo niet dan kan ie erop
							$output .= '<div class="extraselectcondition"><input type="button" onclick="ListSetRemoveextraselectcondition(\'' . $listid . '\',' . var_export(!$RemoveExtraSelectCondition, true) . ')" value="';
							if ($RemoveExtraSelectCondition)
							{
								$output .= htme($GLOBALS['UC']['PersonalDataStructure'][$table]["AddExtraSelectCondionToQueryLinkText"]);
							}
							else
							{
								$output .= htme($GLOBALS['UC']['PersonalDataStructure'][$table]["RemoveExtraSelectCondionFromQueryLinkText"]);
							}
							$output .= '"></div>';
						}

					//SELECTIONS
						if (!$hide_selections)
						{
							$output .= '<div class="selections">';
							if (count($available_selectionsnew["system"]) + count($available_selectionsnew["user"]) > 0)
							{
								$output .= htme($lang['savedfilters']) . ' <select onchange="ListLoadSelectionNew(\'' . $listid . '\',this.value)">';
								$output .= '<option></option>';
								foreach (array("system", "user") as $type)
								{
									$sorted = array_keys($available_selectionsnew[$type]);
									sort($sorted);
									foreach ($sorted as $option)
									{
										$value = $type . ":" . $option;
										$output .= '<option value="' . htme($value) . '"';
										if ($value == $selectionnew)
										{
											$output .= ' selected';
										}
										$output .= '>' . htme($option) .'</option>';
									}
								}
								$output .= '</select>';
							}
							if ($selectionnew != "")
							{
								$selection_can_be_deleted = false;
								if ($is_admin)
								{
									$selection_can_be_deleted = true;
								}
								else if (CheckFunctionAccess("AddEditSelections") == "ok")
								{
									if (preg_match("/^user\:/", $selectionnew))
									{
										$selection_can_be_deleted = true;
									}
									else if (CheckFunctionAccess("AllowedToEditSystemWideSelections") == "ok")
									{
										$selection_can_be_deleted = true;
									}
								}

								if ($selection_can_be_deleted)
								{
									$output .= ' <a onclick="if (confirm(\'' . inline_jsencode($lang["areyousure"]) . '\')) { ListDeleteSelectionNew(\'' . $listid . '\',\'' . inline_jsencode($selectionnew) . '\'); }"><span class="fa fa-times"></span></a>';
								}
							}
							$output .= '</div>';
						}


					//SHOW RECENTLY ACCESSED RECORDS
						if (!$hide_recent)
						{
							$output .= '<div class="recentrecords">' . htme($lang['showrecentlyaccessed']) . '<select onchange="ListSetRecent(\'' . $listid . '\',this.value)"><option value="0"></option>';
							foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["perpage_options"] as $option)
							{
								$output .= '<option value="' . htme($option) . '"';
								if ($recent == $option)
								{
									$output .= ' selected';
								}
								$output .= '>' . htme($option) . '</option>';
							}
							$output .= '</select></div>';
						}

					//CLEAR FILTER; deze staat standaard achteraan in de human readable filters, tenzij die uitstaan
						if (($number_of_filters_set > 0) && ($hide_human_readable_filter))
						{
							$output .= '<div class="clearfilter"><input type="button" onclick="ListSetClearfilter(\'' . $listid . '\')" value="' . htme($lang['clearallfilters']) . '" class="ClearFilter"></div>';
						}

					$output .= $extra_html_above_list;

					$output .= '</div>';
					$output .= '<div class="listoptionsright">';

					//NUMBER OF RECORDS PER PAGE SELECTOR
						if (!$hide_perpage)
						{
							$output .= '<div class="perpage"><select onchange="ListSetPerpage(\'' . $listid . '\',this.value)" title="' . htme($lang["rowsperpage"]) . '">';
							foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["perpage_options"] as $option)
							{
								$output .= '<option value="' . htme($option) . '"';
								if ($perpage == $option)
								{
									$output .= ' selected';
								}
								$output .= '>' . FormatNumberNEW($option) . '</option>';
							}
							$output .= '</select></div>';
						}

					//LIST LAYOUT: COLUMN SELECTOR
						if (!$hide_listlayout)
						{
							//LIST LAYOUT SELECTOR
							if (!$hide_listlayoutselector)
							{
								$available_listnames = array();
								foreach (db_GetArray("SELECT `attribute` FROM `CRMattributes` WHERE `entity`=1 AND `identifier`='system' AND `attribute` LIKE 'List-layout-" . $table . "-%'") as $row)
								{
									$parts = explode("-", $row['attribute']);
									$listname_this = $parts[3];
									$available_listnames[] = array("system", $listname_this);
								}
								//we laten de download select columns layout hier weg
								foreach (db_GetArray("SELECT `attribute` FROM `CRMattributes` WHERE `entity`='" . mres($GLOBALS['USERID']) . "' AND `identifier`='user' AND `attribute` LIKE 'List-layout-" . $table . "-%' AND `attribute`!='List-layout-" . $table . "-DOWNLOAD'") as $row)
								{
									$parts = explode("-", $row['attribute']);
									$listname_this = $parts[3];
									$available_listnames[] = array("user", $listname_this);
								}

								if (count($available_listnames) > 1)
								{
									$output .= '<div class="listlayoutselector"><select onchange="ListSetListname(\'' . $listid . '\',this.value)" title="' . htme($lang["selectsavedcolumnlayout"]) . '">';
									for ($x = 0; $x < count($available_listnames); $x++)
									{
										$listname_type = $available_listnames[$x][0];
										$listname_this = $available_listnames[$x][1];

										$listname_this_display = $lang["listlayout-" . $listname_type] . " ";
										if ($listname_this != "")
										{
											$listname_this_display .= $listname_this;
										}
										else
										{
											$listname_this_display .= $lang["listlayout-default"];
										}
										$val = $listname_type . "-" . $listname_this;
										$output .= '<option value="' . htme($val) . '"';
										if ($val == $listname)
										{
											$output .= ' selected';
										}
										$output .= '>' . htme($listname_this_display) . '</option>';
									}
									$output .= '</select></div>';
								}
							}

							if ($GLOBALS['UC']['PersonalDataStructure'][$table]["MaySelectColumns"])
							{
								$output .= '<div class="listlayout"><a href="listlayout.php?table=' . rawurlencode($table) . '&amp;listname=' . rawurlencode($listname) . '&amp;nonavbar=1&amp;listid=' . rawurlencode($listid) . '" onclick="PopFancyBoxLarge(\'\',this.href); return false;" title="' . htme($lang["selectcolumns"]) . '"><span class="fa fa-cog"></span></a></div>';
							}

							//het zetten van een advanced filter hangt niet per se samen met het opslaan van selections, maar we hangen hem daar maar even aan op
							if (($is_admin) || (CheckFunctionAccess("AddEditSelections") == "ok"))
							{
								$output .= '<div class="listlayout"><a onclick="PopFancyBoxLarge(\'\',\'list.php?advancedfilter=1&amp;nonavbar=1&amp;filtertable=' . $table . '&amp;listid=' . $listid . '\')" title="' . htme($lang['advancedfilter']) . '"><span class="fa fa-filter"></span></a></div>';
							}
						}

					//TABLE PROPERTIES
						if (($is_admin) && (GetAttribute("user", "OnlineEditMode-ShowInlineLinks", $GLOBALS['USERID']) == "Yes"))
						{
							if (preg_match("/^t(\d+)$/", $table, $match))
							{
								$flextable_id = $match[1];
								$output .= '<div class="tableproperties"><a href="flextable.php?EditFlexTable=' . $flextable_id . '&amp;fromlist=' . base64_encode($_SERVER['REQUEST_URI']) . '"><span class="fa fa-cogs" title="' . htme($lang["tableproperties"]) . '"></span></a></div>';
							}
							if ($number_of_records > 0)
							{
								$output .= '<div class="tableproperties"><a href="record.php?mass_del=' . $dqs_stash_id . '&amp;nonavbar=1&amp;table=' . $table . '&amp;listid=' . rawurlencode($listid) . '" onclick="PopFancyBoxLarge(\'\',this.href); return false;"><span class="fa fa-times"></span></a></div>';
							}
						}

					//MODULE SUPPORT
						if (!$hide_modules)
						{
							foreach (db_GetArray("SELECT `mid`, `module_list_html` FROM `CRMmodules` WHERE `version_belonging_to`=0 AND `module_list_html`!='' AND `for_table`='" . mres($GLOBALS['UC']['PersonalDataStructure'][$table]["name_in_database"]) . "'") as $mid)
							{
								if (CheckModuleAccess($mid['mid']) == "ok")
								{
									$popupsize = 0;
									//dezelfde setting als MassUpdatePopup wordt hiervoor gebruikt
									$popupsize_txt = GetAttribute("module", "MassUpdatePopup", $mid['mid']);
									if ($popupsize_txt == "Small")
									{
										$popupsize = 1;
									}
									else if ($popupsize_txt == "Medium")
									{
										$popupsize = 2;
									}
									else if ($popupsize_txt == "Large")
									{
										$popupsize = 3;
									}
									else if ($popupsize_txt == "Extra large")
									{
										$popupsize = 4;
									}
									//"Full screen (no popup)"

									$output .= '<div class="module_list"><a onclick="module_list(' . $mid['mid'] . ',' . $popupsize . ',\'' . $listid . '\',' . $dqs_stash_id . ')">' . $mid['module_list_html'] . '</a></div>';
								}
							}
						}

                        //if ($table == "files")
						//{
                            // HF20160501 Jeroen check s.v.p.
                            // TODO: i18n
                        //    $output .= '<div class="module_list"><a onclick="PopFancyBoxDimensioned(\'\',\'file.php?move=' . $dqs_stash_id . '\', 300, 500);" title="Moves files/folders"><span class="fa fa-files-o"></span></a></div>';

                            /*
                             * $output .= '<div class="tableproperties"><a href="record.php?mass_del=' . $dqs_stash_id . '&amp;nonavbar=1&amp;table=' . $table . '&amp;listid=' . rawurlencode($listid) . '" onclick="PopFancyBoxLarge(\'\',this.href); return false;"><span class="fa fa-times"></span></a></div>';
                             *
                             **/
                      //      $output .= '<div class="module_list"><a onclick="" title="' . htme($lang['delete']) . '"><a href="record.php?mass_del=' . $dqs_stash_id . '&amp;nonavbar=1&amp;table=' . $table . '&amp;listid=' . rawurlencode($listid) . '" onclick="PopFancyBoxLarge(\'\',this.href); return false;"><span class="fa fa-times"></span></a></a></div>';
                        //}

					//TABLE/TILE SWITCHER
						foreach ($list_outputtypes as $option)
						{
							if (!$hide_outputtype[$option])
							{
								$output .= '<div class="outputtype' . $option . '">';
								if ($option != $outputtype)
								{
									$output .= '<a onclick="ListSetOutputtype(\'' . $listid . '\',\'' . inline_jsencode($option) . '\')">';
								}
								$output .= $option;
								if ($option != $outputtype)
								{
									$output .= '</a>';
								}
								$output .= '</div>';
							}
						}

					//DOWNLOAD
//2015-09-18 op verzoek van Hidde uitgezet
//						if (($GLOBALS['UC']['PersonalDataStructure'][$table]["download_allowed"]) && (($GLOBALS['UC']['PersonalDataStructure'][$table]["download_limit"] == 0) || ($number_of_records <= $GLOBALS['UC']['PersonalDataStructure'][$table]["download_limit"])))
						if ($GLOBALS['UC']['PersonalDataStructure'][$table]["download_allowed"])
						{
							$fontawesome = array("tsv" => "fa-file-text-o", "xml" => "fa-file-code-o", "excel" => "fa-file-excel-o");

							foreach ($list_outputtypes_download as $option)
							{
								if (!$hide_outputtype[$option])
								{
									$txt = $lang['downloadexport'] . ' ' . $lang[$option];
									$output .= '<div class="outputtype' . $option . '"><a onclick="ListSetOutputtypeDownload(\'' . $listid . '\',\'' . inline_jsencode($option) . '\',\'\');"><span class="fa ' . $fontawesome[$option] . '" title="' . htme($txt) . '"></span></a></div>';

									$txt = $lang['downloadexport'] . ' ' . $lang[$option] . ' (' . $lang["selectcolumns"] . ')';
									$listname_download = "user-DOWNLOAD";
									$output .= '<div class="outputtype' . $option . ' downloadwithcolumnselector"><a href="listlayout.php?table=' . rawurlencode($table) . '&amp;listname=' . rawurlencode($listname_download) . '&amp;nonavbar=1&amp;listid=' . rawurlencode($listid) . '&amp;download=' . $option . '" onclick="PopFancyBoxLarge(\'\',this.href); return false;"><span class="fa ' . $fontawesome[$option] . '" title="' . htme($txt) . '"></span><span class="fa fa-cog"></span></a></div>';
								}
							}
						}

//switched off, not in use
/*
						if ($watch > 0)
						{
							//TODO change class, we misuse thee paginator class
							$output .= '<div class="paginator"><span id="' . $listid . 'watchicon" class="fa fa-spinner"></span></div>';
						}
*/

					$output .= '</div>';
					//eind listoptions
					$output .= '</div>';
				}
			}

			//20150726
			//voor de add up numeric
			$totals = array();
			$totals_per_group = array();

			//TODO
			//van alle multi select velden die we weergeven in de lijst halen we voorafgaand ALLE data op; dit is wellicht onverstandig; neem ARS transport orders; die verwijst naar Cargo; dat zijn 1000-en records
			//hij cachet nu 2014-01-06
			$multiselect_data_array = array();

			if ($GLOBALS['UC']['PersonalDataStructure'][$table]["openlink"] == "")
			{
				$disable_click_record_to_open = true;
			}

			if (($number_of_records == 0) && ($hidetableheaderwhennoresults))
			{
				$hidetableheader = true;
			}

			if (!$only_data)
			{
				if (($outputtype == "table") || ($outputtype == "template"))
				{
					$output .= '<div class="horscrollouter clearboth">';
					$output .= '<div class="horscroll">';
				}

				//ook voor tile wordt de table afgebeeld, maar dan alleen voor de header row om te filteren en sorteren
				if ($outputtype == "table")
				{
					$output .= '<table class="crm';
					if ($disable_click_record_to_open)
					{
						$output .= ' noclick';
					}
					else
					{
						$output .= ' crmclick';
					}
					$output .= '">';

					if (!$hidetableheader)
					{
						$output .= '<thead>';
						$output .= '<tr class="headerrow">';
					}
				}
				else if ($outputtype == "tile")
				{
					//deze staat er om hem te kunnen hiden
					$output .= '<div class="tilefilters">';

					if (!$hide_sortby)
					{
						//dit is niet echt een tileFILTER maar hij staat in dezelfde lijst
						$output .= '<div class="tilesort">';

						$output .= '<select onchange="ListSetSort(\'' . $listid . '\',document.getElementById(\'' . $listid . 'tilesort\').value,this.value)" id="' . $listid . 'tilesortdirection">';
						foreach (array("asc", "desc") as $sortdirection_option)
						{
							$output .= '<option value="' . $sortdirection_option . '"';
							//alleen de eerste sort wordt bewaard via de UI
							if ($sortdirection_arr[0] == $sortdirection_option)
							{
								$output .= ' selected';
							}
							$output .= '>' . htme($lang["sort_" . $sortdirection_option]) . '</option>';
						}
						$output .= '</select> ';

						$output .= '<select onchange="ListSetSort(\'' . $listid . '\',this.value,document.getElementById(\'' . $listid . 'tilesortdirection\').value)" id="' . $listid . 'tilesort">';
						$output .= '<option></option>';
						foreach ($listlayout as $key)
						{
							$output .= '<option value="' . $key . '"';
							//alleen de eerste sort wordt bewaard via de UI
							if ($sortby_arr[0] == $key)
							{
								$output .= ' selected';
							}
							$output .= '>' . htme($full_fieldnames[$key]) . '</option>';
						}
						$output .= '</select>';

						$output .= '</div>';
					}
				}
			}

			//bevat het aantal kolommen (standaard velden en custom_extra_columns) wat is WEERGEGEVEN
			$columncounter = 0;
			foreach ($listlayout as $key)
			{
				//het veld waarop gegroupt wordt beelden we niet af in de lijst
				if (($key == $group) && ($hide_groupedcolumn))
				{
					continue;
				}
				$table_parts = explode("-", $key);

				$table_name = "";
				$field_name = "";

				// hierboven doorlopen we hem ook al voor de full_fieldnames KAN DAT SNELLER?
				//TODO count -1 of -2?
				for ($x = 0; $x < count($table_parts); $x += 2)
				{
					$table_name = $table_parts[$x];
					$field_name = $table_parts[$x+1];
				}

				$this_field_filterable = false;
				if ((!$hide_filters) && (($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["filter"]) || ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["PerFieldSearch"])))
				{
					$this_field_filterable = true;
				}

				if (($outputtype == "tile") && (!$this_field_filterable))
				{
					//we beelden dit veld niet af
					continue;
				}

				$columncounter++;

				$table_refer = $GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["relation_to"];

				if ($outputtype == "tsv")
				{
					if ($columncounter > 1)
					{
						$output .= "\t";
					}
					//deze preg staat ook lager! dubbelop!
					$output .= preg_replace("/[\t|\r\n]/", " ", $full_fieldnames[$key]);
				}

				if ($outputtype == "excel")
				{
					//voor Excel is de displaylistname in gebruik
					$exceldata[0][] = $full_fieldnames_displaylistname[$key];
				}

				//filter
				if ((($outputtype == "table") || ($outputtype == "tile")) && (!$hidetableheader) && (!$only_data))
				{
					$classes = array();

					$output .= '<';
					if ($outputtype == "table")
					{
						$output .= 'th';
					}
					else if ($outputtype == "tile")
					{
						$output .= 'div';
						$classes[] = 'tilefilter';
					}

					$this_field_filter = false;
					//CODE DIE HIER VEEL OP LIJKT STAAT OP 5 PLEKKEN; number_of_filters_set, HIGHLIGHT TH EN HIGHLIGHT TD, HUMAN_READABLE_FILTER EN DE OPBOUW VAN SQL QUERIES OP BASIS VAN FILTERS
					foreach ($filter[$key] as $key2 => $val)
					{
						//het een regulier filter; hiervan kan val ook leeg zijn, maar dan is hij nog steeds gezet. deze conditie moet VOOR de from/to condities staan omdat als key2 "0" is dit evalueeert op key2 == "to" of "from"
						if (is_numeric($key2))
						{
							$this_field_filter = true;
							break;
						}
						//het is een from-to filter op een numeriek of een datum veld
						//3x = want key2 kan ook numeriek zijn en dat matcht anders ook. waarom?
						else if (($key2 === "from") || ($key2 === "to"))
						{
							//voor deze filters MOET de value gevuld zijn, want hij wordt leeg doorgestuurd en DAT is geen valide filter
							if ($val != "")
							{
								$this_field_filter = true;
								break;
							}
						}
					}

					if ($this_field_filter)
					{
						$classes[] .= 'filter';
					}

					if ($table_refer == "")
					{
						if ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["fieldtype"] == "number")
						{
							$classes[] = 'num';
						}
						if ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["fieldtype"] == "id")
						{
							$classes[] = 'recordid';
						}
					}

					//voegt class="f23" toe aan iedere th, makkelijk voor css
					//als add_class_to_each_cell aan staat dan verderop ook aan td en dt en dd, hoop extra data maar makkelijk om met css bepaalde kolommen op te maken
					//let op staat 2x in deze code
					if (preg_match("/^f\d+$/", $field_name))
					{
						$classes[] = $field_name;
					}

					if ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["PerFieldSearch"])
					{
						$classes[] = "hasPerFieldSearch";
					}

					if (count($classes) > 0)
					{
						$output .= ' class="' . implode(" ", $classes) . '"';
					}
					$output .= '>';

					$output .= '<div>';

					if (($tabletype == "normal") || ($tabletype == "select"))
					{
						if (($is_admin) && (GetAttribute("user", "OnlineEditMode-ShowInlineLinks", $GLOBALS['USERID']) == "Yes"))
						{
							//alleen efid's online edit
							if (preg_match("/^f(\d+)$/", $field_name, $match))
							{
								$efid = $match[1];
								$output .= '<div class="InlineEditModeDivContainer"><div class="InlineEditModeDiv"><a href="extrafields.php?editextrafield=' . $efid . '&amp;fromlist=' . base64_encode($_SERVER['REQUEST_URI']) . '" title="' . htme($table . "-f" . $efid . " " . $GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["fieldtype_in_extrafields_table"]) . '"><span class="fa fa-edit"></span></a></div></div>';
							}
						}
					}

					if ((!$hide_sortby) && ($outputtype == "table"))
					{
						for ($sortcounter = 0; $sortcounter < count($sortby_arr); $sortcounter++)
						{
							$sortby = $sortby_arr[$sortcounter];
							$sortdirection = $sortdirection_arr[$sortcounter];
							$sortdirection_option = 'asc';
							if (($sortby == $key) && ($sortdirection != "desc"))
							{
								$sortdirection_option = 'desc';
							}
							$output .= '<a onclick="ListSetSort(\'' . $listid . '\',\'' . inline_jsencode($key) . '\',\'' . inline_jsencode($sortdirection_option) . '\')" title="';
							if ($full_fieldnames[$key] != $full_fieldnames_displaylistname[$key])
							{
								$output .= htme($full_fieldnames[$key]) . ' ';
							}
							$output .= htme($lang["sort_" . $sortdirection_option]) . '">';

							if ($sortby == $key)
							{
								if ($sortdirection == "desc")
								{
									$output .= '&#9660;';
								}
								//asc is de default, ook "" komt hier
								else
								{
									$output .= '&#9650;';
								}
								if (count($sortby_arr) > 1)
								{
									$output .= ($sortcounter + 1);
								}
								$output .= ' ';
	//							$output .= '<span class="fa fa-sort-' . $sortdirection_option . '"></span> ';
							}
						}

						$output .= htme($full_fieldnames_displaylistname[$key]) . '</a>';

						if ($GLOBALS['SesMem']['debug'])
						{
							$output .= '<div>' . htme($key) . '</div>';
						}
					}
					else
					{
						if ($full_fieldnames[$key] != $full_fieldnames_displaylistname[$key])
						{
							$output .= '<span title="' . htme($full_fieldnames[$key]) . '">';
						}
						$output .= htme($full_fieldnames_displaylistname[$key]);
						if ($full_fieldnames[$key] != $full_fieldnames_displaylistname[$key])
						{
							$output .= '</span>';
						}
					}

					//20150726
					/*
					if ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["add_up_total"])
					{
						$output .= '<div class="num">' . FormatNumberNEW($count_and_sums[$key], $GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["decimals"], $GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["prefix"], $GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["postfix"]) . '</div>';
					}
					*/

					$output .= '</div>';

					if ($this_field_filterable)
					{
						$filteroptions = array();

						if ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["filter"])
						{
							$t1 = microtime(true);
							$filterquery = "";

							if ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["fieldtype"] == "date")
							{
								$filteroptions = $datefilteroptions;
							}
							else if ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["fieldtype"] == "number")
							{
								//deze heeft geen filteroptions; alleen maar from en to boxen
							}
							else
							{
								$relation_to = $GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["relation_to"];

								//wanneer use_adaptive_filters=0 gebruikt wordt kan een gebruiker alle users in de user tabel zien, ook degene die niet gekoppeld zijn aan records die hij mag zien, daarom deze check op allow_direct_view. Die staat op false voor de user tabellen wanneer je geen user edit rechten hebt
								if ((!$use_adaptive_filters) && (($relation_to == "") || ($GLOBALS['UC']['PersonalDataStructure'][$relation_to]["allow_direct_view"])))
								{
									if ($relation_to != "")
									{
										//dit laat alleen maar de waarden in het filter zien die OP DIT MOMENT gekozen mogen worden voor nieuwe records. dat is niet per se goed, maar anders worden ook deleted users afgebeeld
										foreach (select_refer_field_options($table_name, $field_name) as $filterkey => $filterval)
										{
											$filteroptions[] = array($filterkey, $filterval, 0);
										}
									}
									else if (count($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["options"]) > 0)
									{
										//het is een dropdown
										//TODO populate by code en drop-down sql

										foreach ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["options"] as $filterval)
										{
											$filteroptions[] = array($filterval, $filterval, 0);
										}
									}
									else
									{
										//het is een text box; hier zou je helemaal niet op moeten willen filters; zou dit trouwens niet beter met een distinct en order by kunnen?
										$filterquery = "SELECT " . return_database_fieldname($table_name, $field_name, $table_name) . " AS `" . $field_name . "` FROM `CRM" . $GLOBALS['UC']['PersonalDataStructure'][$table_name]["name_in_database"] . "` " . $table_name . " GROUP BY `" . $field_name . "`";
										foreach (db_GetArray($filterquery) as $row)
										{
											$filteroptions[] = array($row[$field_name], $row[$field_name], 0);
										}
									}
								}
								//ADAPTIEF MULTISELECT FILTER
								else if ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["multiselect"])
								{
									if ($relation_to != "")
									{
										$tmp_fields_to_select = array();
										$tmp_fields_to_select[] = $relation_to . "-" . $GLOBALS['UC']['PersonalDataStructure'][$relation_to]["primary"];

										foreach ($GLOBALS['UC']['PersonalDataStructure'][$relation_to]["layout_tags"] as $tag)
										{
											$tmp_fields_to_select[] = $tag;
										}

										$tmp_search = "";
										$tmp_searchperfield = array();
										$tmp_start = 0;
										$tmp_perpage = 0;
										$tmp_sortby = $GLOBALS['UC']['PersonalDataStructure'][$relation_to]["default_sortby"];
										$tmp_sortdirection = $GLOBALS['UC']['PersonalDataStructure'][$relation_to]["default_sortdirection"];
										$tmp_filtercolumn = "";
										$tmp_filter = array();
										$tmp_filtertype = array();
										$tmp_RemoveExtraSelectCondition = false;
										$tmp_connecting_field = "";
										$tmp_use_pre_select_records = false;
										$tmp_pre_select_records = array();
										$tmp_order_by_pre_select = false;
										$tmp_return_countquery = false;
										$tmp_childtable = "";
										$tmp_childfield = "";
										$tmp_childrecordid = 0;
										$tmp_hiddenfilter = array();
										$tmp_hiddenfiltertype = array();
										$tmp_listid = "";
										$tmp_do_not_check_write_access = true;

										$filterquery = make_query($relation_to, $tmp_search, $tmp_searchperfield, $tmp_start, $tmp_perpage, $tmp_sortby, $tmp_sortdirection, $tmp_filtercolumn, $tmp_filter, $tmp_filtertype, $tmp_fields_to_select, $tmp_RemoveExtraSelectCondition, $tmp_connecting_field, $tmp_use_pre_select_records, $tmp_pre_select_records, $tmp_order_by_pre_select, $search_parents, $search_childs, $search_files, $tmp_return_countquery, $type_of_list, $bypass_record_access_by_sql, $tabletype, $tmp_childtable, $tmp_childfield, $tmp_childrecordid, $already_joined, $tmp_hiddenfilter, $tmp_hiddenfiltertype, $tmp_listid, $tmp_do_not_check_write_access);

										foreach (db_GetArray($filterquery) as $filterrecord)
										{
											$filteroptions[] = array($filterrecord[$relation_to . "-" . $GLOBALS['UC']['PersonalDataStructure'][$relation_to]["primary"]], parse_refer_layout_for_table($relation_to, $filterrecord), 0);
										}
									}
									else if (count($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["options"]) > 0)
									{
										foreach ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["options"] as $filterval)
										{
											$filteroptions[] = array($filterval, $filterval, 0);
										}
									}
								}
								//ADAPTIEF NIET-MULTISELECT FILTER
								else
								{
									$tmp_start = 0;
									$tmp_perpage = 0;

									//we sorteren op de alfanumerieke filter waarde
									$tmp_sortby = $key;
									$tmp_sortdirection = "asc";

									$relation_to = $GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["relation_to"];
									if ($relation_to != "")
									{
										//ref to flex sorteren we op de standaard sort van die tabel
										$tmp_sortby = $key . "-" . $GLOBALS['UC']['PersonalDataStructure'][$relation_to]["default_sortby"];
										$tmp_sortdirection = $GLOBALS['UC']['PersonalDataStructure'][$relation_to]["default_sortdirection"];
									}

									$tmp_connecting_field = "";

									$key_zonder_huidige_tabel_en_field = preg_replace("/" . $table_name . "\-" . $field_name . "$/", "", $key);
									$value_layout = $GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["layout"];

									$tmp_filtercolumn = $key;
									$tmp_listlayout = array($key); //deze wordt door make_query ge-expand zodat alle velden in de ref layout ook geselecteerd worden

									$tmp_return_countquery = false;

									$tmp_childtable = "";
									$tmp_childfield = "";
									$tmp_childrecordid = 0;

									$tmp_do_not_check_write_access = true;

									$filterquery = make_query($table, $allsearch_not_preselect, $searchperfield, $tmp_start, $tmp_perpage, $tmp_sortby, $tmp_sortdirection, $tmp_filtercolumn, $filter, $filtertype, $tmp_listlayout, $RemoveExtraSelectCondition, $tmp_connecting_field, $use_pre_select_records, $pre_select_records, $order_by_pre_select, $search_parents, $search_childs, $search_files, $tmp_return_countquery, $type_of_list, $bypass_record_access_by_sql, $tabletype, $tmp_childtable, $tmp_childfield, $tmp_childrecordid, $already_joined, $hiddenfilter, $hiddenfiltertype, $listid, $tmp_do_not_check_write_access);

									$setfilter = array();
									foreach ($filter[$key] as $cur)
									{
										$setfilter[$cur] = true;
									}

									foreach (db_GetArray($filterquery) as $filterrecord)
									{
										$value = $filterrecord[$tmp_filtercolumn];
										if ($value_layout != "")
										{
											if ($value == 0)
											{
												$value = "";
											}
											else
											{
												//TODO PARSING CENTRALISEREN
												$value = $value_layout;
												foreach ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["layout_tags"] as $tag)
												{
													$replacetag = $key_zonder_huidige_tabel_en_field . $tag;
													$replace = $filterrecord[$replacetag];
													$value = str_replace("@" . $tag . "@", $replace, $value);
												}
											}
										}
										$setfilter[$filterrecord[$tmp_filtercolumn]] = false;

										$filteroptions[] = array($filterrecord[$tmp_filtercolumn], $value, $filterrecord['number_of_records']);
									}

									foreach ($setfilter as $filter_on => $must_be_added)
									{
										//het is een filter op een waarde die op dit moment niet in de result set zit; we moeten hem dan wel toevoegen als filter optie
										if ($must_be_added)
										{
											$value = $filter_on;
											if ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["layout"] != "")
											{
												$value = get_refer_layout_for_record_field($table_name, $field_name, $value);
											}
											$filteroptions[] = array($filter_on, $value, 0);
										}
									}
								}
							}

							$t2 = microtime(true);
							$t = $t2 - $t1;
							if ($GLOBALS['SesMem']['debug'])
							{
								$output .= '<h1>filter query ' . htme($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["name"]) . ' (' . $listid . ')</h1>';
								$output .= '<div>' . htme($filterquery) . '</div>';
								$output .= '<div>number of filter options: ' . count($filteroptions) . '</div>';
								$output .= '<div>runtime: ' . $t . '</div>';
							}
						}

						$output .= '<div class="filtercheckboxes">';
						$output .= '<div class="filtercheckboxesinner">';

						if ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["PerFieldSearch"])
						{
							$searchperfield_printed[$key] = true;

							$output .= '<div class="searchperfield"><span class="fa fa-search search_img"></span><input type="search" value="' . htme($searchperfield[$key]) . '" class="RecordSearch enter2blur" onchange="ListSetSearchperfield(\'' . $listid . '\',\'' . $key . '\',this.value)" placeholder="' . htme($lang["search"]) . '"></div>';

							$div_all_filters .= '<input type="search" name="searchperfield[' . $key . ']" id="' . $listid . 'searchperfield[' . $key . ']" value="' . htme($searchperfield[$key]) . '">';
						}

						if ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["filter"])
						{
							if ((count($filteroptions) > 0) || ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["fieldtype"] == "number"))
							{
								if (!$hide_must_be_must_not_be_filter)
								{
									$filtertype_printed[$key] = true;

									if ($filtertype[$key] != "NOT")
									{
										$filtertype[$key] = "IS";
									}
									$output .= '<div class="must_be_must_not_be_filter">';
									foreach (array("IS", "NOT") as $filtertype_option)
									{
										$id = $listid . 'Pfiltertype[' . $key . ']=' . $filtertype_option;
										$output .= '<input type="radio" value="' . $filtertype_option . '" id="' . $id . '" onclick="ListSetFiltertype(\'' . $listid . '\',\'' . inline_jsencode($key) . '\',this.value)"';
										if ($filtertype[$key] == $filtertype_option)
										{
											$output .= ' checked';
										}
										$langtag = "mustbe";
										if ($filtertype_option == "NOT")
										{
											$langtag = "mustnotbe";
										}
										$output .= '> <label for="' . $id . '">' . htme($lang[$langtag]) . '</label> ';
									}
									$output .= '</div>';

									foreach (array("IS", "NOT") as $filtertype_option)
									{
										$id = $listid . 'filtertype[' . $key . ']=' . $filtertype_option;
										$div_all_filters .= '<input type="radio" name="filtertype[' . $key . ']" id="' . $id . '" value="' . $filtertype_option . '"';
										if ($filtertype[$key] == $filtertype_option)
										{
											$div_all_filters .= ' checked';
										}
										$div_all_filters .= '>';
									}
								}
							}
							else
							{
								$output .= '<div>' . htme($lang["nofilteroptions"]) . '</div>';
							}
						}

						//dit is lompe code
						$efid = 0;
						if (preg_match("/^f(\d+)$/", $field_name, $match))
						{
							$efid = $match[1];
						}

						if ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["filter"])
						{
							$filter_printed[$key] = true;

							if (($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["fieldtype"] == "number") || (($efid > 0) && ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["fieldtype"] == "date") && (GetAttribute("extrafield", "ShowFromToFieldsInListFilters", $efid) == "Yes")))
							{
								$filter_from_to_printed[$key] = true;
								//er moet hier een button bij want met een onchange kan je niet from EN to zetten zonder een bloody irritante reload
								//moet nog een class op om ze smaller te maken

								$classes_from = array();
								$classes_from[] = "enter2blur";
								if ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["fieldtype"] == "date")
								{
									$classes_from[] = 'datepicker';
								}

								$classes_to = array();
								$classes_to[] = "enter2blur";
								if ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["fieldtype"] == "date")
								{
									$classes_to[] = 'datepicker';
								}

								$idfrom = $listid . 'Pfilter[' . $key . '][from]';
								$output .= '<div><label for="' . $idfrom . '">' . htme($lang["higherthanorequal"]) . '</label> <input type="text" id="' . $idfrom . '" value="' . htme($filter[$key]['from']) . '" class="' . implode(" ", $classes_from) . '"></div>';

								$idto = $listid . 'Pfilter[' . $key . '][to]';
								$output .= '<div><label for="' . $idto . '">' . htme($lang["lowerthanorequal"]) . '</label> <input type="text" id="' . $idto . '" value="' . htme($filter[$key]['to']) . '" class="' . implode(" ", $classes_to) . '"></div>';

								$output .= '<div><input type="button" value="' . htme($lang["go"]) . '" onclick="ListSetFromToFilter(\'' . $listid . '\',\'' . inline_jsencode($key) . '\',document.getElementById(\'' . inline_jsencode($idfrom) . '\').value,document.getElementById(\'' . inline_jsencode($idto) . '\').value)"></div>';

								$idfrom = $listid . 'filter[' . $key . '][from]';
								$div_all_filters .= '<input type="text" name="filter[' . $key . '][from]" id="' . $idfrom . '" value="' . htme($filter[$key]['from']) . '">';

								$idto = $listid . 'filter[' . $key . '][to]';
								$div_all_filters .= '<input type="text" name="filter[' . $key . '][to]" id="' . $idto . '" value="' . htme($filter[$key]['to']) . '">';
							}
						}

						if (!$filters_as_checkbox)
						{
							$output .= '<select multiple id="' . $listid . 'Pfilter[' . $key . '][]" onchange="ListSetFilterMultipleSelect(\'' . $listid . '\', \'' . $key . '\')">';
							$div_all_filters .= '<select multiple id="' . $listid . 'filter[' . $key . '][]" name="filter[' . $key . '][]">';
						}

						for ($x = 0; $x < count($filteroptions); $x++)
						{
							$filter_on = $filteroptions[$x][0];
							$filter_text = $filteroptions[$x][1];
							$filter_results = $filteroptions[$x][2];

							if ($filters_as_checkbox)
							{
								$output .= '<div>';
							}

							if ($filter_results == -1)
							{
								//separator
								if ($filters_as_checkbox)
								{
									$output .= '&nbsp;';
								}
							}
							else
							{
								$styles = array();
								$optionclass = $GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["optionclass"][$filter_on];
								if ($optionclass == "")
								{
									list($bgcolor, $fgcolor) = getfieldcolors($table_name, $field_name, $filter_on);
									if ($bgcolor != "")
									{
										$styles[] = 'background-color:' . $bgcolor . ';';
										if ($fgcolor != "")
										{
											$styles[] = 'color:' . $fgcolor . ';';
										}
									}
								}

								//voor empty en reference not found
								if ($filter_text == "")
								{
									$styles[] = 'font-style:italic;';
								}

								$css = '';
								if ($optionclass != "")
								{
									$css .= ' class="' . $optionclass . '"';
								}
								if (count($styles) > 0)
								{
									$css .= ' style="' . join("", $styles) . '"';
								}

								$id = $listid . 'Pfilter[' . $key . '][]=' . whitespace_to_underscore($filter_on);

								if ($filters_as_checkbox)
								{
									$output .= '<input type="checkbox" value="' . htme($filter_on) . '" id="' . htme($id) . '" onclick="ListSetFilter(\'' . $listid . '\',\'' . inline_jsencode($key) . '\',\'' . inline_jsencode(whitespace_to_underscore($filter_on)) . '\',this.checked)"';
								}
								else
								{
									$output .= '<option value="' . htme($filter_on) . '" id="' . htme($id) . '"' . $css;
								}
								if ((array_key_exists($key, $filter)) && (in_array($filter_on, $filter[$key])))
								{
									if ($filters_as_checkbox)
									{
										$output .= ' checked';
									}
									else
									{
										$output .= ' selected';
									}
								}
								$output .= '>';
								if ($filters_as_checkbox)
								{
									$output .= ' <label for="' . htme($id) . '"' . $css . '>';
								}

								if ($filter_text != "")
								{
									if (!$GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["ishtml"])
									{
										$filter_text = htme($filter_text);
									}
									$output .= $filter_text;
								}
								//je kan nog steeds filteren op een referentie met als geparsde layout "0" omdat die hierboven al matcht filter_text is dan namelijk "0"
								else if (($filter_on != "") && ($filter_on != "0"))
								{
									$output .= 'Reference &quot;' . htme($filter_on) . '&quot; not found';
								}
								else
								{
									$output .= htme($lang["empty"]);
								}
								if ($filter_results > 0)
								{
									$output .= ' (' . FormatNumberNEW($filter_results) . ')';
								}

								if ($filters_as_checkbox)
								{
									$output .= '</label>';
								}
								else
								{
									$output .= '</option>';
								}

								//"HTML 5 is even more permissive, saying only that an id must contain at least one character and may not contain any space characters."
								$id = $listid . 'filter[' . $key . '][]=' . whitespace_to_underscore($filter_on);

								if ($filters_as_checkbox)
								{
									$div_all_filters .= '<input type="checkbox" name="filter[' . $key . '][]" id="' . htme($id) . '" value="' . htme($filter_on) . '"';
								}
								else
								{
									$div_all_filters .= '<option value="' . htme($filter_on) . '" id="' . htme($id) . '"';
								}
								if ((array_key_exists($key, $filter)) && (in_array($filter_on, $filter[$key])))
								{
									if ($filters_as_checkbox)
									{
										$div_all_filters .= ' checked';
									}
									else
									{
										$div_all_filters .= ' selected';
									}
								}
								$div_all_filters .= '>';

								if (!$filters_as_checkbox)
								{
									$div_all_filters .= '</option>';
								}
							}
							if ($filters_as_checkbox)
							{
								$output .= '</div>';
							}
						}
						if (!$filters_as_checkbox)
						{
							$output .= '</select>';
							$div_all_filters .= '</select>';
						}

						//filtercheckboxesinner
						$output .= '</div>';
						//filtercheckboxes
						$output .= '</div>';
					}

					if ($outputtype == "table")
					{
						$output .= '</th>';
					}
					else if ($outputtype == "tile")
					{
						//einde van tilefilters
						$output .= '</div>';
					}
				}
			}

			$massupdate_modules = array();

			$custom_extra_columns = array();
			if ((($outputtype == "table") || ($outputtype == "tile")) && ($tabletype == "normal"))
			{
				if ($table == "tables")
				{
					$custom_extra_columns[] = array("id" => "tables-recordcount", "header" => "Records", "cssclass" => "num", "noclick" => false);
					$custom_extra_columns[] = array("id" => "tables-showtable", "header" => "Show", "cssclass" => "", "noclick" => false);
					$custom_extra_columns[] = array("id" => "tables-noaccessrestrictionwarning", "header" => "Access", "cssclass" => "cntr", "noclick" => false);
				}
				else if ($table == "fields")
				{
					$custom_extra_columns[] = array("id" => "fields-noaccessrestrictionwarning", "header" => "Access", "cssclass" => "cntr", "noclick" => false);
				}
				else if ($table == "templates")
				{
					$custom_extra_columns[] = array("id" => "templates-noaccessrestrictionwarning", "header" => "Access", "cssclass" => "cntr", "noclick" => false);
				}
				else if ($table == "calendars")
				{
					$custom_extra_columns[] = array("id" => "calendars-noaccessrestrictionwarning", "header" => "Access", "cssclass" => "cntr", "noclick" => false);
				}
				else if ($table == "modules")
				{
					$custom_extra_columns[] = array("id" => "modules-size", "header" => "Size", "cssclass" => "num", "noclick" => false);
					$custom_extra_columns[] = array("id" => "modules-noaccessrestrictionwarning", "header" => "Access", "cssclass" => "cntr", "noclick" => false);
					$custom_extra_columns[] = array("id" => "modules-run", "header" => "Run", "cssclass" => "cntr", "noclick" => true);
				}
				else if ($table == "users" && (($is_admin) || (CheckFunctionAccess("UserAdmin") == "ok" && CheckFunctionAccess("UserAdminSudo") == "ok")))
				{
					$custom_extra_columns[] = array("id" => "users-bethisuser", "header" => "Be this user", "cssclass" => "cntr", "noclick" => true);
				}
				else if ($table == "files")
				{
					$custom_extra_columns[] = array("id" => "files-icon", "header" => htme($lang['open']), "cssclass" => "cntr", "noclick" => false);
					if (is_array($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]["version_no"]))
					{
						if ($hiddenfilter["files-version_belonging_to"][0] == 0)
						{
							$custom_extra_columns[] = array("id" => "files-version", "header" => htme($lang["version"]), "cssclass" => "cntr", "noclick" => true);
						}
						else
						{
							//TODO lang tag
							$custom_extra_columns[] = array("id" => "files-revert", "header" => "Revert", "noclick" => true);
						}
					}
				}
				else if ($table == "locks")
				{
					$custom_extra_columns[] = array("id" => "locks-delete", "header" => "Delete", "cssclass" => "cntr", "noclick" => true);
				}

				if (($GLOBALS['UC']['PersonalDataStructure'][$table]["ShowInlineDeleteLink"]) && (!$all_records_in_this_list_readonly))
				{
					$custom_extra_columns[] = array("id" => "ShowInlineDeleteLink", "header" => htme($lang['delete']), "cssclass" => "cntr inlinedelete", "noclick" => true);
				}
				if (($GLOBALS['UC']['PersonalDataStructure'][$table]["ShowInlineDuplicateLink"]) && ($GLOBALS['UC']['PersonalDataStructure'][$table]["allowed_to_add"]))
				{
					$custom_extra_columns[] = array("id" => "ShowInlineDuplicateLink", "header" => htme($lang['duplicate']), "cssclass" => "cntr", "noclick" => true);
				}

				if (!$hide_massupdate)
				{
					foreach (db_GetArray("SELECT `mid`, `module_name` FROM `CRMmodules` WHERE `version_belonging_to`=0 AND `for_table`='MU-" . mres($table) . "'") as $modules)
					{
						if (CheckModuleAccess($modules['mid']) == "ok")
						{
							$massupdate_modules[$modules['mid']] = $modules['module_name'];
						}
					}
					// HF20160501 Jeroen check s.v.p.  || $table == "files"
					if (count($massupdate_modules) > 0)
					{
						$custom_extra_columns[] = array("id" => "massupdate", "header" => '<input type="checkbox" onclick="massupdateallboxes(\'' . $listid . '\',this.checked)">', "cssclass" => "cntr", "noclick" => true);
					}
				}
			}

			if ($outputtype == "tsv")
			{
				$output .= "\n";
			}

			$addrow_html = '';
			if (!$hide_addrow)
			{
				//dit is lomp en kan beter gecentraliseerd worden
				$listlayout_for_add_row = array();
				foreach ($listlayout as $key)
				{
					//het veld waarop gegroupt wordt beelden we niet af in de lijst
					if (($key == $group) && ($hide_groupedcolumn))
					{
						continue;
					}
					$listlayout_for_add_row[] = $key;
				}
				$addrow_html = get_add_row_for_list_table($table, $listid, $listlayout_for_add_row, $prefill, $custom_extra_columns);
			}

			if (!$only_data)
			{
				if ($outputtype == "table")
				{
					if (!$hidetableheader)
					{
						foreach ($custom_extra_columns as $custom_extra_column)
						{
							$output .= '<th';
							if ($custom_extra_column['cssclass'] != "")
							{
								$output .= ' class="' . $custom_extra_column['cssclass'] . '"';
							}
							$output .= '>' . $custom_extra_column['header'] . '</th>';
						}
						$output .= '</tr>';
						$output .= '</thead>';
					}
					if ($outputtype == "table")
					{
						$output .= '<tbody>';
					}

					if (!$hide_addrow && !$GLOBALS['UC']['PersonalDataStructure'][$table]["addrow_on_bottom"])
					{
						$output .= $addrow_html;
					}
				}
				else if ($outputtype == "tile")
				{
					//einde van de hele header tabel voor sorteren en filteren wordt hier afgesloten
					$output .= '</div>';
				}

				if ($set_tilelistoptionsandfilters_div)
				{
					//eide van tilelistoptionsandfilters
					$output .= '</div>';
				}

				if ($outputtype == "tile")
				{
					$output .= '<div class="tiles">';
				}
				else if ($outputtype == "template")
				{
					if ($tabletype != "plain")
					{
						$output .= '<div class="templatecontainer">';
					}
				}
			}

			//TODO gebruiken we deze vars verderop nog wel?
			//dit is de search
			$searchtype = "";
			$searchparts = array();
			if ($search != "")
			{
				list($searchtype, $searchparts) = searchstring_to_searchparts($search);
			}
			//highlighting end

			//dit is duplicatie van de functie in make_query: dit is hier alleen nodig voor de highlighting; dat is nogal lomp
			$searchpartsperfield = array();
			foreach ($searchperfield as $searchperfieldkey => $searchperfieldvalue)
			{
				if ($searchperfieldvalue != "")
				{
					list($searchtype_pf, $searchparts_pf) = searchstring_to_searchparts($searchperfieldvalue);
					if (count($searchparts_pf) > 0)
					{
						$searchpartsperfield[$searchperfieldkey] = $searchparts_pf;
					}
				}
			}

			$recordcounter = 0;

			//alleen voor table view van belang
			$num_cols = count($listlayout) + count($custom_extra_columns);
			//de group column wordt zelf niet afgebeeld maar wel meegeselecteerd
			if (($group != "") && ($hide_groupedcolumn))
			{
				$num_cols--;
			}

			$access_array_for_records_to_which_files_are_attached = array();

			//rechtstreekse call mysqli
			while ($record = mysqli_fetch_array($ResultSet))
			{
				$recordcounter++;
				$primary_id = $record[$table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"]];

				if (($group != "") && ($record[$group] != $lastgroup))
				{
					if ($outputtype == "table")
					{
						if ($lastgroup != "")
						{
							//vergelijkbare code staat 3x in lf; 1x helemaal onderaan in de tfoot, 1x in de add-up-per-group BINNEN DE LUS, 1x in de add-up-per-group BUITEN DE LUS. dit kan beter
							if (!$hide_add_up_total)
							{
								$display_footer = false;

								$footer = '<tr class="liketfoot noclick">';
								foreach ($listlayout as $key)
								{
									//het veld waarop gegroupt wordt beelden we niet af in de lijst
									if (($key == $group) && ($hide_groupedcolumn))
									{
										continue;
									}

									$table_parts = explode("-", $key);
									$last_field = $table_parts[count($table_parts)-1];
									$last_table = $table_parts[count($table_parts)-2];

									$footer .= '<td';

									if (array_key_exists($key,$totals_per_group))
									{
										$footer .= ' class="num">';
										$display_footer = true;

										$value = FormatNumberNEW($totals_per_group[$key], $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["decimals"], $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["prefix"], $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["postfix"]);

										//de span zit er om alleen een lijn boven het getal te zetten
										$footer .= '<span>' . $value . '</span>';
									}
									else
									{
										$footer .= '>';
									}
									$footer .= '</td>';
								}
								$footer .= str_repeat('<td></td>', count($custom_extra_columns));

								$footer .= '</tr>';

								if ($display_footer)
								{
									$output .= $footer;
								}
							}
						}
						$output .= '<tr class="listgroupheader noclick"><td colspan="' . $num_cols . '">';
					}
					else if (($outputtype == "tile") || ($outputtype == "template"))
					{
						if ($lastgroup != "")
						{
							$output .= '</div>'; //listgroupdiv
						}
						$output .= '<div class="listgroupdiv">';
						$output .= '<div class="listgroupheader">';
					}
					if ($html_output)
					{
						$value = $record[$group];

						$table_parts = explode("-", $group);
						$last_field = $table_parts[count($table_parts)-1];
						$last_table = $table_parts[count($table_parts)-2];

						$table_refer = $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["relation_to"];

						//2015-10-23 TODO dit is veel code duplicatie kan ik hier niet parse_refer_layout_for_field gebruiken?
						if (($table_refer != "") && ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["layout"] != ""))
						{
							//er is geen referentie gezet
							if ($value == 0)
							{
								$value = "";
							}
							else
							{
								$value = $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["layout"];
								$key_zonder_huidige_tabel_en_field = preg_replace("/" . $last_table . "\-" . $last_field . "$/", "", $group);

								//TODO PARSING CENTRALISEREN
								foreach ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["layout_tags"] as $tag)
								{
									$replacetag = $key_zonder_huidige_tabel_en_field . $tag;
									$replace = $record[$replacetag];
									$value = str_replace("@" . $tag . "@", $replace, $value);
								}
							}
							$value = htme($value);
						}
						else
						{
							$value = formatvalue($value, $group);
						}

						$output .= '<h2>' . $value . '</h2>';
					}
					if ($outputtype == "table")
					{
						$output .= '</td></tr>';
					}
					else if (($outputtype == "tile") || ($outputtype == "template"))
					{
						$output .= '</div>';
					}
					$lastgroup = $record[$group];
					$totals_per_group = array();
				}

				$onclick_action = '';

				/* uitzonderingen specifiek voor de files list */
				$files_filetype = "";
				$files_isfolder = false;
				if ($table == "files")
				{
					//todo; al meeselecteren in de main query
					$files_tmp_record = db_GetRow("SELECT type, koppelid, filetype FROM CRMbinfiles WHERE fileid=" . $primary_id);
					$files_filetype = $files_tmp_record['filetype'];
					if (($AllowFoldersInFilelists) && ($files_filetype == "{{{folder}}}"))
					{
						$files_isfolder = true;
						$onclick_action = 'ListSetFileFolder(\'' . $listid . '\',' . $primary_id . ')';
					}

					//Vanuit sql weten we niet of een bestand writable (=verwijderbaar en efid's aanpasbaar) is. Als er record access rules zijn die een file readonly maken dan komt dat wel uit sql, maar als we kijken in sql niet of het record waaraan de file geattached is writable is, want dat zou erg complexe queries opleveren; de files list kan namelijk aan verschillende ft's hangen die ieder hun eigen readonly logica hebben. Dit zouden we eigenlijk ooit wel eens moeten doen (TOD0). Dan kan onderstaande code weer weg.
					//....
					//er is geen record access rule die dit bestand readonly maakt op basis van een efid in de files table...
					//...we weten nu echter nog niet of het record waaraan deze file hangt writable is (want die comlexe logica zit (nog) niet in de files access query)
					//daarom kijken we hier naar de access van het bovenliggende record; we weten al wel zeker dat het bovenliggende record readable is want anders dan had deze user dit bestand uberhaupt niet gezien
					if ($record["type_of_access"] == "ok")
					{
						if (!$is_admin)
						{
							$attached_to_table = $files_tmp_record['type'];
							//reguliere gebruikers mogen deze bestanden niet wissen; system files zien ze sowieso niet
							//deze logica zit gedupliceerd in CheckIfAttachingIsAllowed()
							if (($attached_to_table == "generalpurpose") || ($attached_to_table == "public"))
							{
								$record["type_of_access"] = "readonly";
							}
							else
							{
								if (preg_match("/^flextable(\d+)$/", $attached_to_table, $match))
								{
									$attached_to_table = "t" . $match[1];
								}
								$attached_to_record = $files_tmp_record['koppelid'];

								$access_of_record_to_which_file_is_attached = $access_array_for_records_to_which_files_are_attached[$attached_to_table][$attached_to_record];
								if ($access_of_record_to_which_file_is_attached == "")
								{
									$access_of_record_to_which_file_is_attached = CheckTableRecordAccess($attached_to_table, $attached_to_record);
									//we only query the access once per run; for improved speed
									$access_array_for_records_to_which_files_are_attached[$attached_to_table][$attached_to_record] = $access_of_record_to_which_file_is_attached;
								}

								if ($access_of_record_to_which_file_is_attached == "readonly")
								{
									$record["type_of_access"] = "readonly";
								}
								else if ($access_of_record_to_which_file_is_attached == "nok")
								{
									$record["type_of_access"] = "readonly";
									log_msg("ERROR: user has write access to file_id " . $primary_id . " which is attached to table " . $attached_to_table . ", record " . $attached_to_record . ", which this user is not allowed to see. THIS SHOULD NEVER HAPPEN.");
								}
							}
						}
					}
				}

				if (!$files_isfolder)
				{
					if ($tabletype == "select")
					{
/*
						if ($inline_select_table)
						{
							$onclick_action = 'document.getElementById(\'JS_' . $childfield_dom_id . '\').value = ' . $primary_id . '; this.style.backgroundColor = \'orange\';';
						}
*/
						if (($selectvia != "") && (!$this_is_the_last_selectvia_step))
						{
							$selectviafilter_new = $selectviafilter;
							if ($selectviafilter_new != '')
							{
								$selectviafilter_new .= ',';
							}
							$selectviafilter_new .= $table . '-' . $primary_id;
							$onclick_action = 'refresh_contents_of_div(\'' . $listid . 'div\', \'list.php?childrecordid=' . $childrecordid . '&amp;nonavbar=1&amp;selectvia=' . inline_jsencode($selectvia) . '&amp;selectviafilter=' . inline_jsencode($selectviafilter_new) . '\', \'\', true, \'GET\', \'\', true)';
						}
						else
						{
							//put_id_in_child werkt met dom id's niet mooi
							$onclick_action = 'put_id_in_child(\'' . $childtable . '-' . $childfield . '-';
							if ($childrecordid > 0)
							{
								$onclick_action .= $childrecordid;
							}
							else
							{
								$onclick_action .= '_new_';
							}
							$onclick_action .= '\', ' . $primary_id . ', \'' . inline_jsencode(parse_refer_layout_for_field($childtable, $childfield, $record, true)) . '\')';
						}
					}
					else if (!$disable_click_record_to_open)
					{
						$openlink = $GLOBALS['UC']['PersonalDataStructure'][$table]["openlink"] . $primary_id;
						$popupsize = $GLOBALS['UC']['PersonalDataStructure'][$table]["open_records_in_popup_" . $type_of_list];
						//de laatste "this" is om de row te highlighten met de class lastopenedrecord
						$onclick_action = 'ilor(\'' . inline_jsencode($openlink) . '\',' . $popupsize . ',\'' . $listid . '\',' . $dqs_stash_id . ',this,' . $primary_id . ')';
					}
				}

				$classes_record = array(); //classes for the whole record (not for specific fields in this record)
				if ($highlight == $primary_id)
				{
					$classes_record[] = 'highlight';
				}

				if ($record['type_of_access'] == "readonly")
				{
					$classes_record[] = 'RecordReadonly';
				}

//not needed at this time
/*				if (($watch > 0) && ($only_data))
				{
					$classes_record[] = 'watchnew';
				}*/

				if ($outputtype == "table")
				{
					$output .= '<tr';
					if ($watch > 0) //2017-02-19 misschien ALTIJD data-id er in zetten als we dit voor meer zaken kunnen gebruiken?
					{
						$output .= ' data-id="' . $primary_id . '"';
					}
					if ($onclick_action != "")
					{
						$output .= ' onclick="' . $onclick_action . '"';
					}
					if (count($classes_record) > 0)
					{
						$output .= ' class="' . implode(" ", $classes_record) . '"';
					}
					$output .= '>';
				}
				else if ($outputtype == "tile")
				{
					$output .= '<div';
					if ($watch > 0) //2017-02-19 misschien ALTIJD data-id er in zetten als we dit voor meer zaken kunnen gebruiken?
					{
						$output .= ' data-id="' . $primary_id . '"';
					}

					$classes_record[] = 'tile';
					if ($disable_click_record_to_open)
					{
						$classes_record[] = 'noclick';
					}
					else
					{
						$classes_record[] = 'tileclick';
					}
					$output .= ' class="' . implode(" ", $classes_record) . '"';

					//duplicatie
					if ($onclick_action != "")
					{
						$output .= ' onclick="' . $onclick_action . '"';
					}
					$output .= '>';

					$layout = htme(parse_refer_layout_for_table($table, $record));

					for ($searchpart_counter = 0; $searchpart_counter < count($searchparts); $searchpart_counter++)
					{
						//als deze false is dan is het een NOT search; die kan dan sowieso niet meer in value staan, maar we willen de preg_replace niet onnodig draaien
						if ($searchparts[$searchpart_counter]["is_positive"])
						{
							$layout = highlight($layout, $searchparts[$searchpart_counter]["str"]);
						}
					}

					$output .= '<h2>' . $layout . '</h2>';

					$output .= '<dl>';
				}
				else if ($outputtype == "xml")
				{
					if ($xml_ms_excel)
					{
						$output .= '<Row>';
					}
					else
					{
						$output .= '<record>';
					}
				}
				$templated_list_html_this = $templated_list_html;

				//bevat het aantal kolommen (standaard velden en custom_extra_columns) wat is WEERGEGEVEN
				$columncounter = 0;
				foreach ($listlayout as $key)
				{
					//het veld waarop gegroupt wordt beelden we niet af in de lijst
					if (($key == $group) && ($hide_groupedcolumn))
					{
						continue;
					}

					$table_parts = explode("-", $key);
					$last_field = $table_parts[count($table_parts)-1];
					$last_table = $table_parts[count($table_parts)-2];

					$table_refer = $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["relation_to"];

					$value = $record[$key];
					$value_array = array();
					if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["serialized"]) || ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["sometimes_serialized"]))
					{
						$value_array = unserialize($value);
						$do_print_r = true;
						if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["sometimes_serialized"])
						{
							if (!is_array($value_array))
							{
								$value_array = array();
								$do_print_r = false;
							}
						}

						//dit is een wat slordige conditie die er alleen voor admin tabellen in staat
						if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["multiselect"]) && ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype_in_extrafields_table"] == "diary"))
						{
							$do_print_r = false;
						}

						if ($do_print_r)
						{
							if (empty($value_array))
							{
								$value = "";
							}
							else
							{
								$value = print_r($value_array, true);
							}
						}
					}

					//hiermee kan op het recordid van een parent tabel geklikt worden om het parent record te openen van uit de child list. ALLEEN op priamry key van bovenliggende tabel. zo lang je die niet weergeeft in je list layout zie je deze functionaliteit dus niet
					$make_link = false;
					$openlink_other_table = $GLOBALS['UC']['PersonalDataStructure'][$last_table]["openlink"];
					if (($outputtype == "table") || ($outputtype == "tile"))
					{
						if (($last_table != $table) && ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["primary"] == $last_field) && ($value > 0) && ($openlink_other_table != ""))
						{
							$make_link = true;
							$openlink_other_table .= $value;
						}
					}

					$this_field_filter = false;
					//CODE DIE HIER VEEL OP LIJKT STAAT OP 5 PLEKKEN; number_of_filters_set, HIGHLIGHT TH EN HIGHLIGHT TD, HUMAN_READABLE_FILTER EN DE OPBOUW VAN SQL QUERIES OP BASIS VAN FILTERS
					foreach ($filter[$key] as $key2 => $val)
					{
						//het een regulier filter; hiervan kan val ook leeg zijn, maar dan is hij nog steeds gezet. deze conditie moet VOOR de from/to condities staan omdat als key2 "0" is dit evalueeert op key2 == "to" of "from"
						if (is_numeric($key2))
						{
							$this_field_filter = true;
							break;
						}
						//het is een from-to filter op een numeriek of een datum veld
						//3x = want key2 kan ook numeriek zijn en dat matcht anders ook. waarom?
						else if (($key2 === "from") || ($key2 === "to"))
						{
							//voor deze filters MOET de value gevuld zijn, want hij wordt leeg doorgestuurd en DAT is geen valide filter
							if ($val != "")
							{
								$this_field_filter = true;
								break;
							}
						}
					}

		//2014-01-22
		/*
					if ($searchperfield[$key] != "")
					{
						$this_field_filter = true;
					}
		*/
					$classes = array();
					$noclick_js = '';
					$interactive_field = false;
					if (($outputtype == "table") || ($outputtype == "tile"))
					{
						if ($tabletype == "normal")
						{
							//een field moet als interactive gedefineerd zijn, hij kan alleen in de EIGEN tabel geedit worden; we kunnen dus geen fields in de parent tabel editten en record moet writeable zijn
							//2015-09-18 EN de conditionele write access (uit sql) tot dit veld moet 1 opleveren
							if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["interactive"]) && ($key == $last_table . "-" . $last_field) && ($record['type_of_access'] == "ok") && ($record['record_locked_by'] == 0) && ($record['*' . $key] == 1))
							{
								$interactive_field = true;
							}

							if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["IgnoreCellClick"]) || ($interactive_field) || ($make_link))
							{
								$classes[] = 'noclick';
								//deze zetten we in een tile view ook op de dd
								$noclick_js = ' onclick="noclick(event)"';
							}
						}
					}

					if ($hide_empty_cells)
					{
						//interactieve velden beelden we altijd af, OOK als ze niet invulbaar zijn op dit moment. Dat zouden we misschien aan moeten passen. voor gnl komt dit nu wel goed uit want hierdoor staat de opm altijd in dezelfde kolom
						if (!$interactive_field)
						{
							if ($outputtype == "tile")
							{
								//dit veld is al afgebeeld in de h2 en hoeven we dus niet nogmaals weer te geven in de dl
								if (in_array($key, $GLOBALS['UC']['PersonalDataStructure'][$table]["layout_tags"]))
								{
									continue;
								}
								if (count($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["layout_tags"]) > 0)
								{
									$alltags_already_displayed = true;
									foreach ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["layout_tags"] as $tag)
									{
										if (!in_array($tag, $GLOBALS['UC']['PersonalDataStructure'][$table]["layout_tags"]))
										{
											$alltags_already_displayed = false;
										}
									}
									if ($alltags_already_displayed)
									{
										continue;
									}
								}
							}
							//helemaal niet afbeelden; dit kan beter
							if (($value == '') || ($value == '0000-00-00') || ($value == '0000-00-00 00:00:00'))
							{
								continue;
							}
						}
					}

					//we beelden dit veld af
					$columncounter++;


					if ($this_field_filter)
					{
						$classes[] .= 'filter';
					}

					//voegt class="f23" toe aan iedere td of dd, hoop extra data maar makkelijk om met css bepaalde kolommen op te maken
					if (($add_class_to_each_cell) || ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]['AddIdentifierClass']))
					{
						//let op staat 2x in deze code
						if (preg_match("/^f\d+$/", $last_field))
						{
							$classes[] = $last_field;
						}
					}

					//deze classes worden op het EERSTE veld gezet
					if (($columncounter == 1) && ($tabletype != "plain"))
					{
						if (($search_highlight_only) && (in_array($primary_id, $pre_select_records)))
						{
							$classes[] = 'SearchHighlight';
						}
					}

					if ($outputtype == "table")
					{
						$output .= '<td';
					}
					else if ($outputtype == "tile")
					{
						$output .= '<dt';
					}

					if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["calculation"] != "")
					{
						//dubbele streep omdat een enkele streep al gebruikt wordt in scheidingsteken in $key
						//let op dezelfde code voor "tile" staat lager omdat het id en de class dan op de <dd> moet zitten en niet op de <dt>
						if ($outputtype == "table")
						{
							$output .= ' id="' . $listid . '--' . $key . '--' . $primary_id . '"';
							$classes[] = 'calculation';
						}
					}

					$output .= $noclick_js;

					if ($outputtype == "xml")
					{
						if ($xml_ms_excel)
						{
							$excel_type = 'String';
							$excel_style = '';
							if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "date")
							{
								$excel_type = 'DateTime';
								$excel_style = 's62';
								if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["dateformat_in_database"] == "yyyy-mm-dd hh:mm:ss") || ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["dateformat_in_database"] == "epoch"))
								{
									$excel_style = 's63';
								}
							}
							else if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "number") || ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "id"))
							{
								$excel_type = 'Number';
							}
							$output .= '<Cell';
							if ($excel_style != "")
							{
								$output .= ' ss:StyleID="' . $excel_style . '"';
							}
							$output .= '><Data ss:Type="' . $excel_type . '"';
						}
						else
						{
							$output .= '<' . $key;
							if (!$xml_with_xsd)
							{
								$output .= ' name="' . htme($full_fieldnames[$key]) . '"';
							}
						}
					}

					$bgcolor = "";
					$fgcolor = "";
					//always interactive fields krijgen geen kleur, normale interactieve velden wel
					if ((!$interactive_field) || (!$GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["interactive-alwaysinline"]))
					{
						$optionclass = $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["optionclass"][$value];
						if ($optionclass != "")
						{
							$classes[] = $optionclass;
						}
						else
						{
							list($bgcolor, $fgcolor) = getfieldcolors($last_table, $last_field, $value);
						}
					}

					if ($table_refer == "")
					{
						if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "number") || ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "id"))
						{
							if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "id")
							{
								$classes[] = 'recordid';
							}
							else
							{
								$classes[] = 'num';
							}

							//id's formatten we niet
							if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "number")
							{
								//20150726
								if ((!$hide_add_up_total) && ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["add_up_total"]))
								{
									$totals[$key] += $value;
									$totals_per_group[$key] += $value;
								}

								//TODO voor alle downloads geen formatering
								if (($outputtype == "table") || ($outputtype == "tile"))
								{
									//slordige uitzondering maar kan nu niet anders
									if ($key == "files-filesize")
									{
										//als we files-filetype mee zouden selecteren dan wisten we hem hier al; dan kan echter niet want dan wordt ie ook afgebeekd
										//2016-01-29 de foldersize wordt nu weergegeven
/*										if ($files_isfolder)
										{
											$value = "";
										}
										else
										{*/
											$value = ceil($value / 1024) . " KB";
//										}
									}
									else
									{
										$value = FormatNumberNEW($value, $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["decimals"], $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["prefix"], $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["postfix"]);
									}
								}
							}
						}
					}

					if ($outputtype == "table")
					{
						if ($bgcolor != "")
						{
							$output .= ' style="background-color:' . $bgcolor . ';';
							if ($fgcolor != "")
							{
								$output .= 'color:' . $fgcolor . ';';
							}
							$output .= '"';
						}
					}
					if (($outputtype == "table") || ($outputtype == "tile"))
					{
						if (count($classes) > 0)
						{
							$output .= ' class="' . implode(" ", $classes) . '"';
						}
					}
					if (($outputtype == "table") || ($outputtype == "tile") || ($outputtype == "xml"))
					{
						$output .= '>';
						if (($outputtype == "table") && ($GLOBALS['UC']['PersonalDataStructure'][$table]["PutCellDataInOwnDivInLists"]))
						{
							$output .= '<div>';
						}
					}

					if ($outputtype == "tile")
					{
						$output .= htme($full_fieldnames_displaylistname[$key]) . '</dt>';
						$output .= '<dd' . $noclick_js;

						if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["calculation"] != "")
						{
							//let op dezelfde code voor "table" staat hoger omdat het id en de class dan op de <dd> moet zitten en niet op de <dt>
							if ($outputtype == "tile")
							{
								$output .= ' id="' . $listid . '--' . $key . '--' . $primary_id . '"';
								$classes[] = "calculation";
							}
						}

						//slordig een definition list heeft geen container voor een dd / dt element, daarom moet deze class op beide los gezet worden. specifieke styles zijn natuurlijk toe te voegen door dt.f23 of dd.f23 { font-weight: bold }
						if (count($classes) > 0)
						{
							$output .= ' class="' . implode(" ", $classes) . '"';
						}

						if ($bgcolor != "")
						{
							$output .= ' style="background-color:' . $bgcolor . ';';
							if ($fgcolor != "")
							{
								$output .= 'color:' . $fgcolor . ';';
							}
							$output .= '"';
						}
						$output .= '>';
					}

					if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["multiselect"])
					{
						$newvalue = '';
						foreach ($value_array as $val)
						{
							//multi dropdown en sql nog niet ondersteund
							if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["relation_to"] != "")
							{
								//TODO DIT IS NIET EFFICIENT: getrecord wordt voor ieder multiselect waarde gerunt. dat kan dus enkele 100-en keren zijn wanneer pagination hoog is ingesteld. Als je al records opvraagt uit een grote lijst dan duurt het extreem lang. een alternatief is eerst een array opbouwen met alle mogelijke refers. dat deed ik eerst met multiselect_data_array, nadeel daarvan is dat DAT heel erg duur is op korte lijsten
								//FIX 2014-01-06 we cachen eenmaal geparsde records; dit levert vooral veel op, bij weinig keuzes die vaak zijn geselecteerd

								if ($multiselect_data_array[$key][$val] != "")
								{
									$val = $multiselect_data_array[$key][$val];
								}
								else
								{
									$selectallfields = false;
									$bypass_record_access_by_sql = false;
									$selectfields = array();

									foreach ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["layout_tags"] as $tag)
									{
										$selectfields[] = $tag;
									}
									$one_record = getrecord($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["relation_to"], $val, $selectallfields, $bypass_record_access_by_sql, $selectfields, true); //do_not_check_write_access=true

									$newval = parse_refer_layout_for_field($last_table, $last_field, $one_record, false);

									$multiselect_data_array[$key][$val] = $newval;
									$val = $newval;
								}
							}

							if ($newvalue != '')
							{
								$newvalue .= ', ';
							}
							$newvalue .= $val;
						}
						$value = $newvalue;
					}
					else if (($table_refer != "") && ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["layout"] != ""))
					{
						//er is geen referentie gezet
						if ($value == 0)
						{
							$value = "";
						}
						else
						{
							$value = $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["layout"];
							$key_zonder_huidige_tabel_en_field = preg_replace("/" . $last_table . "\-" . $last_field . "$/", "", $key);

							//TODO PARSING CENTRALISEREN
							foreach ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["layout_tags"] as $tag)
							{
								$replacetag = $key_zonder_huidige_tabel_en_field . $tag;
								$replace = $record[$replacetag];
								$value = str_replace("@" . $tag . "@", $replace, $value);
							}
						}
					}

					if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "date")
					{
						if (($outputtype == "table") || ($outputtype == "tile") || ($outputtype == "template") || ($outputtype == "excel"))
						{
							$value = from_iso_datetime($value, $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["ShowSeconds"]);
						}
						//het is een datetime veld
						else if (($outputtype == "xml") && (($xml_with_xsd) || ($xml_ms_excel)) && (strlen($value) == 19))
						{
							//2016-06-03 00:24:54 -> 2016-06-03T00:24:54
							$value = str_replace(" ", "T", $value);
							/*if ($xml_ms_excel)
							{
								$value .= ".000";
							}*/
						}
						//voor tsv geen formattering
					}

					if (($outputtype == "table") || ($outputtype == "tile") || ($outputtype == "template") || ($outputtype == "xml"))
					{
						//slordige uitzondering diary
						if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype_in_extrafields_table"] != "diary") && (!$GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["ishtml"]))
						{
							//beetje slordig; htme wordt later gedaan
							if (!$interactive_field)
							{
								$value = htme($value);
							}
							if (($html_output) && ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["multiline"]))
							{
								$value = nl2br($value);
							}
						}
					}

					if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype_in_extrafields_table"] == "diary")
					{
						$value = return_diary($value_array, $outputtype);
						if ($outputtype == "xml")
						{
							$value = htme($value);
						}
					}

					if ($html_output)
					{
						if (($value != "") && (!$interactive_field))
						{
							//moet dit niet met urlencode?
							//noclick zodat het record zelf niet opent
							if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldsubtype"] == "mail")
							{
								//window.top.location vanwege google chrome die het anders niet doet vanuit een frame
								$value = '<a href="mailto:' . $value . '" onclick="window.top.location.href=this.href;noclick(event);return(false);">' . $value . '</a>';
							}
							else if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldsubtype"] == "link")
							{
								//value moet nog http voor gezet worden
								$value = '<a href="' . $value . '" onclick="window.open(this.href);noclick(event);return(false);">' . $value . '</a>';
							}
							else if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldsubtype"] == "tel")
							{
								//window.top.location vanwege google chrome die het anders niet doet vanuit een frame
								$value = '<a href="tel:' . $value . '" onclick="window.top.location.href=this.href;noclick(event);return(false);">' . $value . '</a>';
							}
						}

//TMP WEGHALEN VOOR CODE DEPLOY
/*						if (strstr($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["name"], "#LIST"))
						{
							$tmp = $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["name"];
							$tmp = ParseTemplateRecordData($tmp, $table, $primary_id);
							$value = unescape_interleave_characters(ParseTemplateListTags($tmp));
						}*/

					}
					else if ($outputtype == "tsv")
					{
						//safety
						$value = preg_replace("/[\t|\r|\n]/", " ", $value);

						if ($columncounter > 1)
						{
							$output .= "\t";
						}
					}

					if (($html_output) && (!$interactive_field))
					{
						for ($searchpart_counter = 0; $searchpart_counter < count($searchparts); $searchpart_counter++)
						{
							//als deze false is dan is het een NOT search; die kan dan sowieso niet meer in value staan, maar we willen de preg_replace niet onnodig draaien
							if ($searchparts[$searchpart_counter]["is_positive"])
							{
								$value = highlight($value, $searchparts[$searchpart_counter]["str"]);
							}
						}

						for ($searchpart_counter = 0; $searchpart_counter < count($searchpartsperfield[$key]); $searchpart_counter++)
						{
							//als deze false is dan is het een NOT search; die kan dan sowieso niet meer in value staan, maar we willen de preg_replace niet onnodig draaien
							if ($searchpartsperfield[$key][$searchpart_counter]["is_positive"])
							{
								$value = highlight($value, $searchpartsperfield[$key][$searchpart_counter]["str"]);
							}
						}

					//hiermee zouden subtabellen gehighlight kunnen worden ZONDER er daadwerkelijk in te zoeken (en ze dus te limiteren). ik ga het voor nu toch anders oplossen met een javascript highlight functie
					//NIET MEER NODIG DANKZIJ search_highlight_only
					/*
									if ($_GET['highlight'] != "")
									{
										$value = highlight($value, $_GET['highlight']);
									}
					*/
					}

					if ($make_link)
					{
						//we kijken naar de tabel waar we NU inzitten en niet de tabel die we openen. Dit is debatable.
						$popupsize = $GLOBALS['UC']['PersonalDataStructure'][$table]["open_records_in_popup_" . $type_of_list];
						//update 2014-04-08 dqs value hieruit want die bevat de id's van de tabel waarin we nu zitten en NIET van de tabel die we openen
						//geen this.href gebruiken want die maakt er een URL van en dat gaat mis bij beautification (https://admin.interleave.nl/admin/https://admin.interleave.nl/admin/Rules/0?&duplicate=2)
						$output .= '<a href="' . htme($openlink_other_table) . '" onclick="ilor($(this).attr(\'href\'),' . $popupsize . ',\'' . $listid . '\',0); return false;">';
					}

					if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["ishtml"])
					{
						if ($outputtype == "excel")
						{
							$value = html2plaintext($value);
						}
						else if ($outputtype == "xml")
						{
							$value = htme($value);
						}
					}


					if ($outputtype == "xml")
					{
						//remove all ascii control characters, except tab (09), carriage return (0D) and line feed (0A)
						//xml parsers usually don't except these characters
						$value = preg_replace('/[\x00-\x08\x0B\x0C\x0E-\x1F]/', ' ', $value);
					}

					if (($outputtype == "excel") || ($outputtype == "tsv"))
					{
						//https://www.owasp.org/index.php/CSV_Injection
						//prefix formula characters (=, +, - and @) with a space, to prevent formula's from being executed in Excel, Google Docs or other spreadsheet programs. Excel does not execute these formula's when opening an xml file, so we don't prefix this format.
						$value = preg_replace('/^([\=\+\-\@])/', ' \\1', $value);
					}

					if (($outputtype == "array") || ($outputtype == "json"))
					{
						$arraydata[$recordcounter-1][$key] = $value;
					}
					else if ($outputtype == "excel")
					{
						//2016-06-06 geen support voor optionclass
						if ($bgcolor != "")
						{
							$value = '[[bgcolor' . $bgcolor . ']]' . $value;
							if ($fgcolor != "")
							{
								$value = '[[fgcolor' . $fgcolor . ']]' . $value;
							}
						}
						$exceldata[$recordcounter][] = $value;
					}
					//om memory te saven op excel export; niet toevoegen aan output
					else if ($outputtype == "template")
					{
						//code duplication! we must probably change this code in the very near future to a call to get_parsed_template_with_given_template() so that ALL tags are supported (including list tags for sublists); get_parsed_template_with_given_template() and subsidiary functions MUST support passing the $record as a variable. Otherwise PTRD() will requery the database which is must too slow in list context
						//IF we do this; it could be argued that record.php must go, and that opening a single record is nothing more than opening a templated list with a filter on t1-recordid=8736

						if (strpos($templated_list_html_this, "%" . $key . "%") !== false)
						{
							//we use $record[$key] as unformatted value; I think this is correct
							$templated_list_html_this = str_replace("%" . $key . "%", escape_interleave_characters(htme($record[$key])), $templated_list_html_this);
						}

						if (strpos($templated_list_html_this, "^" . $key . "^") !== false)
						{
							$templated_list_html_this = str_replace("^" . $key . "^", escape_interleave_characters(rawurlencode($record[$key])), $templated_list_html_this);
						}

						if (strpos($templated_list_html_this, "@" . $key . "@") !== false)
						{
							$span = '';
							if ($bgcolor != "")
							{
								$span .= ' style="background-color:' . $bgcolor . ';';
								if ($fgcolor != "")
								{
									$span .= 'color:' . $fgcolor . ';';
								}
								$span .= '"';
							}
							if (count($classes) > 0)
							{
								$span .= ' class="' . implode(" ", $classes) . '"';
							}
							if ($span != '')
							{
								$value = '<span' . $span . '>' . $value . '</span>';
							}
							$templated_list_html_this = str_replace("@" . $key . "@", escape_interleave_characters($value), $templated_list_html_this);
						}

						if (strpos($templated_list_html_this, "#" . $key . "#") !== false)
						{
							//THIS IS A TEMPORARY SOLUTION, WE MUST CHANGE THIS WHOLE PARSING LOGIC TO PTRD() OR get_parsed_template_with_given_template()
							$replace = '';
							if (($record['type_of_access'] == "ok") && ($record['record_locked_by'] == 0) && ($record['*' . $key] == 1))
							{
								$replace = '<span onclick="noclick(event)">' . return_ifield($key, $primary_id, $record[$key], $listid, false, true) . '</span>';
							}
							else
							{
								//RAW VALUE LIKE %, THIS MUST BE CHANGED
								$replace = htme($record[$key]);
							}
							$templated_list_html_this = str_replace("#" . $key . "#", $replace, $templated_list_html_this);
						}

						if (strpos($templated_list_html, "#DELETEBUTTON#") !== false)
						{
							$replace = '';
							//this code is duplicated for the custom_extra_addon inlinedelete AND the #DELTEBUTTON# on a templated list
							if (($record['type_of_access'] == "ok") && ($record['record_locked_by'] == 0))
							{
								$replace = '<a onclick="PopFancyBoxDimensioned(\'\',\'record.php?table=' . $table . '&amp;record=' . $primary_id . '&amp;delete=1&amp;nonavbar=1&amp;ParentAjaxHandler=' . $listid . '\',300,500)"><span class="fa fa-times"></span></a>';
							}
							$templated_list_html_this = str_replace("#DELETEBUTTON#", $replace, $templated_list_html_this);
						}
					}
					else
					{
						if ($interactive_field)
						{
							//dit is ALTIJD in de eigen tabel, key bestaat dus uit 2 elementen; laatste conditie is true om security check te skippen; we WETEN immers al dat dit veld writable is voor deze user
							$output .= return_ifield($key, $primary_id, $record[$key], $listid, false, true);
						}
						else
						{
							$output .= $value;
						}
					}

					if ($make_link)
					{
						$output .= '</a>';
					}

					if ($outputtype == "table")
					{
						if ($columncounter == 1)
						{
							if (($record['type_of_access'] == "ok") && ($record['record_locked_by'] > 0))
							{
								$output .= ' <span class="fa fa-lock" title="' . htme($lang["recordislocked1"]) . ' ' . GetUserName($record['record_locked_by']) . '"></span>';
							}
						}
						//hiermee kan de tabel uitgeklapt worden op een hover
						if ($GLOBALS['UC']['PersonalDataStructure'][$table]["PutCellDataInOwnDivInLists"])
						{
							$output .= '</div>';
						}
						$output .= '</td>';
					}
					else if ($outputtype == "tile")
					{
						$output .= '</dd>';
					}
					else if ($outputtype == "xml")
					{
						if ($xml_ms_excel)
						{
							$output .= '</Data></Cell>';
						}
						else
						{
							$output .= '</' . $key . '>';
						}
					}
				}

				foreach ($custom_extra_columns as $custom_extra_column)
				{
					$columncounter++;

					$classes = array();
					if ($outputtype == "table")
					{
						$output .= '<td';
					}
					else if ($outputtype == "tile")
					{
						$output .= '<dt';
					}

					if ($custom_extra_column['noclick'])
					{
						$output .= ' onclick="noclick(event)"';
						$classes[] = 'noclick';
					}
					if ($custom_extra_column['cssclass'])
					{
						$classes[] = $custom_extra_column['cssclass'];
					}
					if (count($classes) > 0)
					{
						$output .= ' class="' . implode(" ", $classes) . '"';
					}
					$output .= '>';

					if ($outputtype == "tile")
					{
						$output .= $custom_extra_column['header'] . '</dt><dd';
						if ($custom_extra_column['noclick'])
						{
							$output .= ' onclick="noclick(event)"';
						}
						if (count($classes) > 0)
						{
							$output .= ' class="' . implode(" ", $classes) . '"';
						}
						$output .= '>';
					}

					if ($custom_extra_column['id'] == "ShowInlineDeleteLink")
					{
						if (($record['type_of_access'] == "ok") && ($record['record_locked_by'] == 0))
						{
							//this code is duplicated for the custom_extra_addon inlinedelete AND the #DELTEBUTTON# on a templated list
							$output .= '<a onclick="PopFancyBoxDimensioned(\'\',\'record.php?table=' . $table . '&amp;record=' . $primary_id . '&amp;delete=1&amp;nonavbar=1&amp;ParentAjaxHandler=' . $listid . '\',300,500)"><span class="fa fa-times"></span></a>';
						}
					}
					else if ($custom_extra_column['id'] == "ShowInlineDuplicateLink")
					{
						$addlink = $GLOBALS['UC']['PersonalDataStructure'][$table]["addlink"] . '&duplicate=' . $primary_id;
						$popupsize = 0;

						if (($tabletype == "select") && ($selectvia_steps == 1))
						{
							//geen popup wanneer uit select; maar wel nonavbar want we zitten dan al in een popup die we hergebruiken -> dit is niet fraai
							//put_id_in_child werkt met dom id's niet mooi
							$addlink .= '&nonavbar=1&put_id_in_child=' . $childtable . '-' . $childfield . '-';
							if ($childrecordid > 0)
							{
								$addlink .= $childrecordid;
							}
							else
							{
								$addlink .= '_new_';
							}
						}
						else
						{
							$popupsize = $GLOBALS['UC']['PersonalDataStructure'][$table]["open_records_in_popup_" . $type_of_list];
						}
						//geen this.href gebruiken want die maakt er een URL van en dat gaat mis bij beautification (https://admin.interleave.nl/admin/https://admin.interleave.nl/admin/Rules/0?&duplicate=2)
						$output .= '<a href="' . htme($addlink) . '" onclick="ilor($(this).attr(\'href\'),' . $popupsize . ',\'' . $listid . '\',0); return false;"><span class="fa fa-files-o"></span></a>';
					}
					else if ($custom_extra_column['id'] == "tables-recordcount")
					{
						$tmpsql = "SELECT COUNT(*) FROM `CRMflextable" . $primary_id . "`";
						if ($GLOBALS['UC']['PersonalDataStructure'][$table]["logical_delete"])
						{
							$tmpsql .= " WHERE `deleted`='n'";
						}
						$output .= FormatNumberNEW(db_GetValue($tmpsql));
					}
					else if ($custom_extra_column['id'] == "tables-showtable")
					{
						$output .= '<a href="list.php?table=t' . $primary_id . '"><button type="button">Show</button></a>';
					}
					else if ($custom_extra_column['id'] == "modules-size")
					{
						$output .= FormatNumberNEW(db_GetValue("SELECT LENGTH(`module_code`) FROM `CRMmodules` WHERE mid=" . $primary_id));
					}
					else if ($custom_extra_column['id'] == "massupdate")
					{
						$output .= '<input type="checkbox" name="massupdate[]" value="' . $primary_id . '">';
					}
					else if ($custom_extra_column['id'] == "modules-noaccessrestrictionwarning")
					{
						if (db_GetValue("SELECT COUNT(*) FROM `CRMmoduleaccessrules` WHERE `module`=" . $primary_id) == 0)
						{
							$output .= '<span class="fa fa-exclamation-triangle" title="No access configured"></span>';
						}
					}
					else if ($custom_extra_column['id'] == "tables-noaccessrestrictionwarning")
					{
						if (db_GetValue("SELECT COUNT(*) FROM `CRMtableaccessrules` WHERE `table`=" . $primary_id) == 0)
						{
							$output .= '<span class="fa fa-exclamation-triangle" title="No access configured"></span>';
						}
					}
					else if ($custom_extra_column['id'] == "fields-noaccessrestrictionwarning")
					{
						if (db_GetValue("SELECT COUNT(*) FROM `CRMfieldaccessrules` WHERE `field`=" . $primary_id) == 0)
						{
							$output .= '<span class="fa fa-exclamation-triangle" title="No access configured"></span>';

						}
						if (db_GetValue("SELECT COUNT(*) FROM `CRMfieldaccessrules` WHERE `field`=" . $primary_id . " AND `group` = 0") > 0 && db_GetValue("SELECT COUNT(*) FROM `CRMfieldaccessrules` WHERE `field`=" . $primary_id . " AND `group` != 0") == 0) {
						 	$output .= '<span class="fa fa-exclamation-triangle" title="Only administrators can access this field!"></span><span class="fa fa-exclamation-triangle" title="Only administrators can access this field!"></span>';
						}
					}
					else if ($custom_extra_column['id'] == "templates-noaccessrestrictionwarning")
					{
						if (db_GetValue("SELECT COUNT(*) FROM `CRMtemplateaccessrules` WHERE `template`=" . $primary_id . " AND `group` > 0") == 0) {
							$output .= '<span class="fa fa-exclamation-triangle" title="No access configured"></span>';
							if (db_GetValue("SELECT COUNT(*) FROM `CRMtemplateaccessrules` WHERE `template`=" . $primary_id . " AND `group` = 0") > 0) {
								 	$output .= '<span class="fa fa-exclamation-triangle" title="Only administrators can access this template. Can\'t be right!"></span>';
							}
						}
					}
					else if ($custom_extra_column['id'] == "calendars-noaccessrestrictionwarning")
					{
						if (db_GetValue("SELECT COUNT(*) FROM `CRMcalendaraccessrules` WHERE `calendar`=" . $primary_id) == 0)
						{
							$output .= '<span class="fa fa-exclamation-triangle" title="No access configured"></span>';
						}
					}
					else if ($custom_extra_column['id'] == "modules-run")
					{
						$output .= '<a href="modules.php?action=run&amp;mid=' . $primary_id . '"><span class="fa fa-play"></span></a>';
					}
					else if ($custom_extra_column['id'] == "files-icon")
					{
							$output .= GetImageFileTypeIcon($primary_id, $listid);
					}
					else if ($custom_extra_column['id'] == "files-version")
					{
						if (!$files_isfolder)
						{
							$earlier_versions = '';
							foreach (GetFileEarlierVersions($primary_id) as $earlier_version)
							{
								$earlier_versions .= '<tr><td>' . $earlier_version['version_no'] . '</td><td><a href="csv.php?fileid=' . $earlier_version['fileid'] . '">' . htme($earlier_version['filename']) . '</a></td><td>' . from_iso_datetime($earlier_version['timestamp_last_change']) . '</td><td>' . htme(GetUserName($earlier_version['username'])) . '</td></tr>';

							}
							if ($earlier_versions != '')
							{
								//dit zou eigenlijk een recursieve call naar make_list() moeten zijn
								$output .= '<a href="fileversions.php?fileid=' . $primary_id . '&amp;nonavbar=1" onclick="PopFancyBoxLarge(\'File revisions\',this.href); return false;"><span class="fa fa-history" title="Earlier versions of this file exist. Click to view." id="ver' . $listid . 'L' . $primary_id . '"></span></a>';
							}
							else
							{
								$output .= '<span class="fa fa-history"></span>';
							}
						}
					}
					else if ($custom_extra_column['id'] == "files-revert")
					{
						//TODO lang tag
						$output .= '<input type="button" onclick="if (confirm(\'Are you sure?\')) { window.location.href=\'fileversions.php?revert_to=' . $primary_id . '&amp;nonavbar=1\'; }" value="Revert">';
					}
					else if ($custom_extra_column['id'] == "users-bethisuser")
					{
						$output .= '<input type="button" onclick="window.location.href=\'admin.php?BeThisUser='. $primary_id . '\'; return(false);" value="Be">';
					}
					else if ($custom_extra_column['id'] == "locks-delete")
					{
						//het is een compound key; AL deze kolommen MOETEN geselecterd zijn in de listlayout om het deleten van locks te laten werken
						$output .= '<a href="listadmin.php?showlocks=1&amp;sameresultandpage=1&amp;deletelock=1&amp;lockon=' . $record['locks-lockon'] . '&amp;lockby=' . $record['locks-lockby'] . '&amp;locktable=' . $record['locks-locktable'] . '" onclick="return confirm(\'Are you sure?\')"><span class="fa fa-times"></span></a>';;
					}

					if ($outputtype == "table")
					{
						$output .= '</td>';
					}
					else if ($outputtype == "tile")
					{
						$output .= '</dd>';
					}
				}

				if ($outputtype == "table")
				{
					for ($filluptd = $columncounter; $filluptd < $num_cols; $filluptd++)
					{
						//lege cellen worden achteraan gezet om voor valide html te zorgen (iedere <tr> moet evenveel <td>'s hebben)
						$output .= '<td></td>';
					}
				}

				//rules, ruleconditions, ruleactions en ruleactionconditions
				//uitgezet 2015-10-19 te traag
				/*
				if (preg_match("/^rule/", $table))
				{
					//EXPERIMENTAL: beeld recursief alle childs af van dit record
					//de noclick doet het nog niet goed
					$output .= '<td class="noclick listinlist">';
					foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["childs"]) as $child_table)
					{
						//TODO: checken op circulaire joins
						foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["childs"][$child_table] as $child_field)
						{
							if (!$GLOBALS['UC']['PersonalDataStructure'][$child_table]["fields"][$child_field]["multiselect"])
							{
								$listoptions_child = array();
								$listoptions_child["table"] = $child_table;
								$listoptions_child["hiddenfilter"][$child_table . "-" . $child_field][] = $primary_id;
								$listoptions_child["tabletype"] = "plain"; //beeldt ALLE record af; kan veel te veel zijn
								$listoptions_child["type_of_list"] = "sublist";
								$listoptions_child["internallistcall"] = true;

								//todo interactive fields aan

								list($child_number_of_records, $child_output) = make_list($listoptions_child);

								if ($child_number_of_records > 0)
								{
									//kan een plain list niet met een header worden afgebeeld?
									$output .= '<h2>' . htme($GLOBALS['UC']['PersonalDataStructure'][$child_table]["plural"]) . ' (' . FormatNumberNEW($child_number_of_records). ')</h2>';
									$output .= $child_output;
								}
							}
						}
					}
					$output .= '</td>';
				}
				*/

				if ($outputtype == "table")
				{
					$output .= '</tr>';
				}

				if (($outputtype == "table") && ($tabletype == "normal"))
				{
					if (($GLOBALS['UC']['PersonalDataStructure'][$table]["headerrepeat"] > 0) && ($recordcounter % $GLOBALS['UC']['PersonalDataStructure'][$table]["headerrepeat"] == 0) && ($recordcounter < $number_of_records_on_this_page))
					{
						$output .= '<tr class="headerrow headerrepeat">';
						foreach ($listlayout as $key)
						{
							$output .= '<th>' . htme($full_fieldnames_displaylistname[$key]) . '</th>';
						}
						foreach ($custom_extra_columns as $custom_extra_column)
						{
							$output .= '<th';
							if ($custom_extra_column['cssclass'] != "")
							{
								$output .= ' class="' . $custom_extra_column['cssclass'] . '"';
							}
							$output .= '>' . $custom_extra_column['header'] . '</th>';
						}
						$output .= '</tr>';
					}
				}

				if ($outputtype == "tile")
				{
					$output .= '</dl></div>';
				}
				else if ($outputtype == "xml")
				{
					if ($xml_ms_excel)
					{
						$output .= '</Row>';
					}
					else
					{
						$output .= '</record>';
					}
				}
				else if ($outputtype == "tsv")
				{
					$output .= "\n";
				}
				else if ($outputtype == "template")
				{
					//bijna gelijk aan tile
					if ($tabletype != "plain")
					{
						$output .= '<div';
						if ($watch > 0) //2017-02-19 misschien ALTIJD data-id er in zetten als we dit voor meer zaken kunnen gebruiken?
						{
							$output .= ' data-id="' . $primary_id . '"';
						}
						if ($onclick_action != "")
						{
							$output .= ' onclick="' . $onclick_action . '"';
						}
						if (count($classes_record) > 0)
						{
							$output .= ' class="' . implode(" ", $classes_record) . '"';
						}
						$output .= '>';
					}
					//JH 2017-04-22
					$templated_list_html_this = ParseTemplateListTags($templated_list_html_this);
					$output .= unescape_interleave_characters($templated_list_html_this);
					if ($tabletype != "plain")
					{
						$output .= '</div>';
					}
				}

				if (!$buffering)
				{
					print $output;
					$output = '';
					ob_flush();
				}
			}

			if ($html_output)
			{
				//to hide loadmore button; this should also be added to tile and table views. TODO
				if (($continuous_scroll) && ($outputtype == "template") && ($number_of_records_after_this_page == 0))
				{
					$output .= '<div id="' . $listid . 'thisisthelastpage"></div>';
				}

				if (($number_of_records_on_this_page == 0) && ($watch == 0))
				{
					$noresults = $lang['noresults'];
					if (preg_match("/^t(\d+)$/", $table, $match))
					{
						$noresults2 = GetAttribute("flextable", "NoResultsMessage", $match[1]);
						if (($noresults2 != "") && ($noresults2 != "{{language default}}"))
						{
							$noresults = $noresults2;
						}
					}
					if ($outputtype == "table")
					{
						$output .= '<tr><td colspan="' . $num_cols . '" class="noclick noresults">';
					}
					else
					{
						$output .= '<div class="noresults">';
					}
					$output .= htme($noresults);
					if ($outputtype == "table")
					{
						$output .= '</td></tr>';
					}
					else
					{
						$output .= '</div>';
					}
				}
			}
			//rechtstreekse call mysqli
			mysqli_free_result($ResultSet);

			if (!$only_data)
			{
				if ($outputtype == "table")
				{
					$output .= '</tbody>';

					//vergelijkbare code staat 3x in lf; 1x helemaal onderaan in de tfoot, 1x in de add-up-per-group BINNEN DE LUS, 1x in de add-up-per-group BUITEN DE LUS. dit kan beter
					if (!$hide_add_up_total)
					{
						if ($lastgroup != "")
						{
							//vergelijkbare code staat 3x in lf; 1x helemaal onderaan in de tfoot, 1x in de add-up-per-group BINNEN DE LUS, 1x in de add-up-per-group BUITEN DE LUS. dit kan beter
							if (!$hide_add_up_total)
							{
								$display_footer = false;

								$footer = '<tr class="liketfoot noclick">';
								foreach ($listlayout as $key)
								{
									//het veld waarop gegroupt wordt beelden we niet af in de lijst
									if (($key == $group) && ($hide_groupedcolumn))
									{
										continue;
									}

									$table_parts = explode("-", $key);
									$last_field = $table_parts[count($table_parts)-1];
									$last_table = $table_parts[count($table_parts)-2];

									$footer .= '<td';

									if (array_key_exists($key,$totals_per_group))
									{
										$footer .= ' class="num">';
										$display_footer = true;

										$value = FormatNumberNEW($totals_per_group[$key], $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["decimals"], $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["prefix"], $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["postfix"]);

										//de span zit er om alleen een lijn boven het getal te zetten
										$footer .= '<span>' . $value . '</span>';
									}
									else
									{
										$footer .= '>';
									}
									$footer .= '</td>';
								}
								$footer .= str_repeat('<td></td>', count($custom_extra_columns));

								$footer .= '</tr>';

								if ($display_footer)
								{
									$output .= $footer;
								}
							}
						}


						if ($number_of_records_on_this_page > 1)
						{
							$display_footer = false;

							$footer = '<tr>';
							foreach ($listlayout as $key)
							{
								//het veld waarop gegroupt wordt beelden we niet af in de lijst
								if (($key == $group) && ($hide_groupedcolumn))
								{
									continue;
								}

								$table_parts = explode("-", $key);
								$last_field = $table_parts[count($table_parts)-1];
								$last_table = $table_parts[count($table_parts)-2];

								$footer .= '<td';

								if (array_key_exists($key,$totals))
								{
									$footer .= ' class="num">';
									$display_footer = true;

									$value = FormatNumberNEW($totals[$key], $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["decimals"], $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["prefix"], $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["postfix"]);

									//de span zit er om alleen een lijn boven het getal te zetten
									$footer .= '<span>' . $value . '</span>';
								}
								else
								{
									$footer .= '>';
								}
								$footer .= '</td>';
							}
							$footer .= str_repeat('<td></td>', count($custom_extra_columns));

							$footer .= '</tr>';

							if ($display_footer)
							{
								$output .= '<tfoot>';
								$output .= $footer;
								$output .= '</tfoot>';
							}
						}
					}
					if (!$hide_addrow && $GLOBALS['UC']['PersonalDataStructure'][$table]["addrow_on_bottom"])
					{
						$output .= $addrow_html;
					}
					$output .= '</table>';
				}
				else if ($outputtype == "tile")
				{
					if ($lastgroup != "")
					{
						$output .= '</div>'; //listgroupdiv
					}
					$output .= '</div>'; //class=tiles
				}
				else if ($outputtype == "template")
				{
					if ($lastgroup != "")
					{
						$output .= '</div>'; //listgroupdiv
					}
					if ($tabletype != "plain")
					{
						$output .= '</div>'; //container
					}
				}

				if (($outputtype == "table") || ($outputtype == "template"))
				{
					$output .= '</div>'; //horscroll
					$output .= '</div>'; //horscrollouter
				}

				if ($html_output)
				{
					if (($tabletype == "normal") || ($tabletype == "select"))
					{
						$output .= '<div class="listoptions listoptionsfooter">';
						$output .= '<div class="listoptionsleft">';
						//deze staat ook boven de lijst
						//dit is de listfooter; staat alleen de paginator in; misschien in de toekomst nog meer?
						if ((!$hide_bottom_pagination) && ($paginator != ''))
						{
							$output .= $paginator;
						}

						$output .= '</div>';

						$output .= '<div class="listoptionsright">';
						//if (in_array("massupdate", $custom_extra_columns))
						if (count($massupdate_modules) > 0)
						{
							$output .= '<div class="massupdate">';
							foreach ($massupdate_modules as $module_id => $module_name)
							{
								$popupsize = 0;
								$popupsize_txt = GetAttribute("module", "MassUpdatePopup", $module_id);
								if ($popupsize_txt == "Small")
								{
									$popupsize = 1;
								}
								else if ($popupsize_txt == "Medium")
								{
									$popupsize = 2;
								}
								else if ($popupsize_txt == "Large")
								{
									$popupsize = 3;
								}
								else if ($popupsize_txt == "Extra large")
								{
									$popupsize = 4;
								}
								//"Full screen (no popup)"

								$output .= '<input type="button" value="' . htme($module_name) . '" onclick="massupdate(\'' . $module_id . '\',' . $popupsize . ',\'' . $listid . '\',' . $dqs_stash_id . ')"> ';
							}
							$output .= '</div>';
						}
						$output .= '</div>';
						$output .= '</div>';
					}

					if ($continuous_scroll)
					{
						if ($last_record_on_this_page < $number_of_records)
						{
							$output .= '<div id="' . $listid . 'loadmore" class="loadmore"><input type="button" onclick="ListContinuousScroll(\'' . $listid . '\')" value="' . htme($lang["loadmore"]) . '"></div>';
						}
					}
				}

				if ($close_listinner_div)
				{
					$output .= '</div>';
				}
			}
		}

		if (!$only_data)
		{
			if ($html_output)
			{
				//dus niet plain
				if (($tabletype == "normal") || ($tabletype == "select"))
				{
					if ($GLOBALS['SesMem']['debug'])
					{
						//deze roept divsubmit aan met debug (laatste var = 1), die doet dan een alert van alle hidden form elements hieronder
						$output .= '<input type="button" onclick="divsubmit(document.getElementById(\'' . $listid . 'divform\'), \'GET\', \'\', \'\', false, false, true, false, false)" value="Show list options">';
					}

					$output .= '<div id="' . $listid . 'divform">';

		//omdat een sublist al in een form kan staan kunnen we hier geen form meer openen
		//			$output .= '<form method="get" action="list.php" id="form' . $listid . '" onsubmit="$.fancybox.close(); ListReload(\'' . $listid . '\',true,true); return false;">';
					$output .= '<input type="hidden" name="table" id="' . $listid . 'table" value="' . htme($table) . '">';
					$output .= '<input type="hidden" name="search_highlight_only" id="' . $listid . 'search_highlight_only" value="' . htme($search_highlight_only) . '">';
					$output .= '<input type="hidden" name="search" id="' . $listid . 'search" value="' . htme($search) . '">';
					$output .= '<input type="hidden" name="hiddensearch" id="' . $listid . 'hiddensearch" value="' . htme($hiddensearch) . '">';

					//deze moet bestaan zodat het filter ook opgeheven kan worden
					$output .= '<input type="hidden" name="searchperfield[]" id="' . $listid . 'searchperfield[]" value="">';

					$output .= '<input type="hidden" name="hiddensearchperfield[]" id="' . $listid . 'hiddensearchperfield[]" value="">';
					foreach ($hiddensearchperfield as $key => $val)
					{
						$output .= '<input type="hidden" name="hiddensearchperfield[' . htme($key) . ']" value="' . htme($val) . '">';
					}

					$output .= '<input type="hidden" name="start" id="' . $listid . 'start" value="' . htme($start) . '">';
					$output .= '<input type="hidden" name="perpage" id="' . $listid . 'perpage" value="' . htme($orig_perpage) . '">';
					$output .= '<input type="hidden" name="sortby" id="' . $listid . 'sortby" value="' . htme(implode(",",$sortby_arr)) . '">';
					$output .= '<input type="hidden" name="sortdirection" id="' . $listid . 'sortdirection" value="' . htme(implode(",",$sortdirection_arr)) . '">';
					$output .= '<input type="hidden" name="outputtype" id="' . $listid . 'outputtype" value="' . htme($outputtype) . '">';

					$output .= '<input type="hidden" name="type_of_list" id="' . $listid . 'type_of_list" value="' . htme($type_of_list) . '">';

					//deze vars zijn een beetje vreemd; die heeft slechts effect voor 1 pageload in list.php die dan de outputtype var overschrijft. de js functies die de lijst reloaden zetten deze var altijd op ""
					$output .= '<input type="hidden" name="outputtypedownload" id="' . $listid . 'outputtypedownload" value="">';
					$output .= '<input type="hidden" name="outputtypedownload_listname" id="' . $listid . 'outputtypedownload_listname" value="">';

					//deze moet bestaan zodat het filter ook opgeheven kan worden; MAAR NIET MET DUBBELE ARRAY
					//in div_all_filters staan de echte filter checkboxen
					$output .= '<input type="hidden" name="filter[]" id="' . $listid . 'filter[]" value="">';

					//deze moet bestaan zodat het filter ook opgeheven kan worden
					$output .= '<input type="hidden" name="filtertype[]" id="' . $listid . 'filtertype[]" value="">';

					$output .= '<input type="hidden" name="hiddenfilter[]" id="' . $listid . 'hiddenfilter[]" value="">';

					foreach (array_keys($hiddenfilter) as $key)
					{
						foreach ($hiddenfilter[$key] as $key2 => $val)
						{
							$arraykey = "";
							//3x = want key2 kan ook numeriek zijn en dat matcht anders ook. waarom?
							if (($key2 === "from") || ($key2 === "to"))
							{
								$arraykey = $key2;
							}
							$output .= '<input type="hidden" name="hiddenfilter[' . htme($key) . '][' . $arraykey . ']" id="' . $listid . 'hiddenfilter[' . htme($key) . '][' . $arraykey . ']" value="' . htme($val) . '">';
						}
					}

					$output .= '<input type="hidden" name="hiddenfiltertype[]" id="' . $listid . 'hiddenfiltertype[]" value="">';
					foreach ($hiddenfiltertype as $key => $val)
					{
						$output .= '<input type="hidden" name="hiddenfiltertype[' . htme($key) . ']" value="' . htme($val) . '">';
					}

					$output .= '<input type="hidden" name="RemoveExtraSelectCondition" id="' . $listid . 'RemoveExtraSelectCondition" value="' . $RemoveExtraSelectCondition . '">';

					$output .= '<input type="hidden" name="selectionnew" id="' . $listid . 'selectionnew" value="' . htme($selectionnew) . '">';
					$output .= '<input type="hidden" name="saveselectionnew" id="' . $listid . 'saveselectionnew" value="">';
					$output .= '<input type="hidden" name="deleteselectionnew" id="' . $listid . 'deleteselectionnew" value="">';

					$output .= '<input type="hidden" name="disable_click_record_to_open" id="' . $listid . 'disable_click_record_to_open" value="' . htme($disable_click_record_to_open) . '">';

					$output .= '<input type="hidden" name="ignorelastview" id="' . $listid . 'ignorelastview" value="' . htme($ignorelastview) . '">';

					$output .= '<input type="hidden" name="use_adaptive_filters" id="' . $listid . 'use_adaptive_filters" value="' . htme($use_adaptive_filters) . '">';

					//hide_all wordt hier niet ingezet want die is bovenaan al verwerkt naar onderstaande losse vars
					$output .= '<input type="hidden" name="hide_header" id="' . $listid . 'hide_header" value="' . htme($hide_header) . '">';
					$output .= '<input type="hidden" name="customheader" id="' . $listid . 'customheader" value="' . htme($customheader) . '">';

					$output .= '<input type="hidden" name="hide_pagination_if_only_1_page" id="' . $listid . 'hide_pagination_if_only_1_page" value="' . htme($hide_pagination_if_only_1_page) . '">';
					$output .= '<input type="hidden" name="hide_bottom_pagination" id="' . $listid . 'hide_bottom_pagination" value="' . htme($hide_bottom_pagination) . '">';
					$output .= '<input type="hidden" name="hide_perpage" id="' . $listid . 'hide_perpage" value="' . htme($hide_perpage) . '">';
					$output .= '<input type="hidden" name="hide_recordcounter" id="' . $listid . 'hide_recordcounter" value="' . htme($hide_recordcounter) . '">';
					$output .= '<input type="hidden" name="hide_search" id="' . $listid . 'hide_search" value="' . htme($hide_search) . '">';
					$output .= '<input type="hidden" name="hide_selections" id="' . $listid . 'hide_selections" value="' . htme($hide_selections) . '">';
					$output .= '<input type="hidden" name="hide_addlink" id="' . $listid . 'hide_addlink" value="' . htme($hide_addlink) . '">';
					$output .= '<input type="hidden" name="hide_addrow" id="' . $listid . 'hide_addrow" value="' . htme($hide_addrow) . '">';
					$output .= '<input type="hidden" name="hide_id_input" id="' . $listid . 'hide_id_input" value="' . htme($hide_id_input) . '">';
					$output .= '<input type="hidden" name="hide_listlayout" id="' . $listid . 'hide_listlayout" value="' . htme($hide_listlayout) . '">';

					$output .= '<input type="hidden" name="hide_outputtype_table" id="' . $listid . 'hide_outputtype_table" value="' . htme($hide_outputtype['table']) . '">';
					$output .= '<input type="hidden" name="hide_outputtype_tile" id="' . $listid . 'hide_outputtype_tile" value="' . htme($hide_outputtype['tile']) . '">';
					$output .= '<input type="hidden" name="hide_outputtype_tsv" id="' . $listid . 'hide_outputtype_tsv" value="' . htme($hide_outputtype['tsv']) . '">';
					$output .= '<input type="hidden" name="hide_outputtype_xml" id="' . $listid . 'hide_outputtype_xml" value="' . htme($hide_outputtype['xml']) . '">';
					$output .= '<input type="hidden" name="hide_outputtype_excel" id="' . $listid . 'hide_outputtype_excel" value="' . htme($hide_outputtype['excel']) . '">';

					$output .= '<input type="hidden" name="hide_recent" id="' . $listid . 'hide_recent" value="' . htme($hide_recent) . '">';
					$output .= '<input type="hidden" name="hide_modules" id="' . $listid . 'hide_modules" value="' . htme($hide_modules) . '">';
					$output .= '<input type="hidden" name="hide_sortby" id="' . $listid . 'hide_sortby" value="' . htme($hide_sortby) . '">';
					$output .= '<input type="hidden" name="hide_filters" id="' . $listid . 'hide_filters" value="' . htme($hide_filters) . '">';
					$output .= '<input type="hidden" name="hide_add_up_total" id="' . $listid . 'hide_add_up_total" value="' . htme($hide_add_up_total) . '">';

					$output .= '<input type="hidden" name="hide_must_be_must_not_be_filter" id="' . $listid . 'hide_must_be_must_not_be_filter" value="' . htme($hide_must_be_must_not_be_filter) . '">';
					$output .= '<input type="hidden" name="hide_human_readable_filter" id="' . $listid . 'hide_human_readable_filter" value="' . htme($hide_human_readable_filter) . '">';

					$output .= '<input type="hidden" name="hide_listlayoutselector" id="' . $listid . 'hide_listlayoutselector" value="' . htme($hide_listlayoutselector) . '">';

					$output .= '<input type="hidden" name="search_parents" id="' . $listid . 'search_parents" value="' . htme($search_parents) . '">';
					$output .= '<input type="hidden" name="search_childs" id="' . $listid . 'search_childs" value="' . htme($search_childs) . '">';
					$output .= '<input type="hidden" name="search_files" id="' . $listid . 'search_files" value="' . htme($search_files) . '">';
					$output .= '<input type="hidden" name="recent" id="' . $listid . 'recent" value="' . htme($recent) . '">';

					$output .= '<input type="hidden" name="listname" id="' . $listid . 'listname" value="' . htme($listname) . '">';

					//listid bevat alfanumerieke chars; dus hoeft niet door htme()
					$output .= '<input type="hidden" name="listid" id="' . $listid . 'listid" value="' . $listid . '">';
					//alleen op de eerste pageload is deze leeg
					$output .= '<input type="hidden" name="listid_div_exists" id="' . $listid . 'listid_div_exists" value="1">';

					$output .= '<input type="hidden" name="tabletype" id="' . $listid . 'tabletype" value="' . htme($tabletype) . '">';

					$output .= '<input type="hidden" name="listlayoutcolumns" id="' . $listid . 'listlayoutcolumns" value="' . htme($listlayoutcolumns) . '">';

					$output .= '<input type="hidden" name="hidetableheader" id="' . $listid . 'hidetableheader" value="' . htme($hidetableheader) . '">';
					$output .= '<input type="hidden" name="hidetableheaderwhennoresults" id="' . $listid . 'hidetableheaderwhennoresults" value="' . htme($hidetableheaderwhennoresults) . '">';
					$output .= '<input type="hidden" name="hide_tablewhennoresults" id="' . $listid . 'hide_tablewhennoresults" value="' . htme($hide_tablewhennoresults) . '">';

					$output .= '<input type="hidden" name="hide_massupdate" id="' . $listid . 'hide_massupdate" value="' . htme($hide_massupdate) . '">';
					$output .= '<input type="hidden" name="add_class_to_each_cell" id="' . $listid . 'add_class_to_each_cell" value="' . htme($add_class_to_each_cell) . '">';

					//nodig voor de js functie ListSetSearch
					$output .= '<input type="hidden" name="id_input_in_search" id="' . $listid . 'id_input_in_search" value="' . htme($id_input_in_search) . '">';

					$output .= '<input type="hidden" name="reload_list_on_child_update" id="' . $listid . 'reload_list_on_child_update" value="' . htme($reload_list_on_child_update) . '">';

					$output .= '<input type="hidden" name="group" id="' . $listid . 'group" value="' . htme($group) . '">';
					$output .= '<input type="hidden" name="hide_groupedcolumn" id="' . $listid . 'hide_groupedcolumn" value="' . htme($hide_groupedcolumn) . '">';

					$output .= '<input type="hidden" name="hide_empty_cells" id="' . $listid . 'hide_empty_cells" value="' . htme($hide_empty_cells) . '">';
					$output .= '<input type="hidden" name="continuous_scroll" id="' . $listid . 'continuous_scroll" value="' . htme($continuous_scroll) . '">';

					$output .= '<input type="hidden" name="template_id" id="' . $listid . 'template_id" value="' . htme($template_id) . '">';
					$output .= '<input type="hidden" name="template_to_use_open" id="' . $listid . 'template_to_use_open" value="' . htme($template_to_use_open) . '">';
					$output .= '<input type="hidden" name="watch" id="' . $listid . 'watch" value="' . htme($watch) . '">';

					$output .= '<input type="hidden" name="childrecordid" id="' . $listid . 'childrecordid" value="' . htme($childrecordid) . '">';
					$output .= '<input type="hidden" name="selectvia" id="' . $listid . 'selectvia" value="' . htme($selectvia) . '">';
					$output .= '<input type="hidden" name="selectviafilter" id="' . $listid . 'selectviafilter" value="' . htme($selectviafilter) . '">';


					//TODO! filters worden aan en uitgezet in de all filters div en niet boven aan de lijst, hierdoor komt er een hoop html overhead bij. dit moet er uit gehaald worden maar dan moeten ook alle list js functies herschreven worden
					$output .= '<div class="hide_content">';
					$output .= $div_all_filters;

					//2015-06-15 om het niet 2x te doen, anders is het filter nooit meer op te heffen; dankzij deze code kan er een normal (zichtbaar) filter gezet zijn op een kolom die niet wordt afgebeeld in de lijst
					//we zouden dit op een betere manier moeten opmaken door niet meer met de all filter div te werken
					//dit is nodig voor de newselections
						foreach (array_keys($filter) as $key)
						{
							foreach ($filter[$key] as $key2 => $val)
							{
								//3x = want key2 kan ook numeriek zijn en dat matcht anders ook. waarom?
								if (($key2 === "from") || ($key2 === "to"))
								{
									if (!$filter_from_to_printed[$key])
									{
										$id = $listid . 'filter[' . $key . '][' . $key2 . ']';
										$output .= 'filter[' . htme($key) . '][' . $key2 . '] = ' . $val . '<br>';
										$output .= '<input type="text" name="filter[' . htme($key) . '][' . $key2 . ']" id="' . htme($id) . '" value="' . htme($val) . '">';
									}
								}
								else
								{
									if (!$filter_printed[$key])
									{
										$id = $listid . 'filter[' . $key . '][]=' . whitespace_to_underscore($val);
										$output .= '<input type="checkbox" name="filter[' . htme($key) . '][]" id="' . htme($id) . '" value="' . htme($val) . '" checked>';
									}
								}
							}
						}
						foreach ($filtertype as $key => $val)
						{
							if ((!$filtertype_printed[$key]) && ($val != ""))
							{
								$output .= '<input type="checkbox" name="filtertype[' . htme($key) . ']" id="' . $listid . 'filtertype[' . htme($key) . ']=' . $val . '" value="' . htme($val) . '" checked>';
							}
						}
						foreach ($searchperfield as $key => $val)
						{
							if ((!$searchperfield_printed[$key]) && ($val != ""))
							{
								$output .= '<input type="search" name="searchperfield[' . htme($key) . ']" id="' . $listid . 'searchperfield[' . htme($key) . ']" value="' . htme($val) . '">';
							}
						}

					$output .= '</div>';

		//zie comment bij <form hierboven
		//			$output .= '</form>';

					$output .= '</div>';
				}
			}

			if ($close_listid_div)
			{
				if ($watch > 0)
				{
					$output .= '<script>ListWatch(\'' . $listid . '\', true);</script>';
				}
				//einde van de main divLISTID
				$output .= '</div>';
			}
		}

		if ($outputtype == "xml")
		{
			if ($xml_ms_excel)
			{
				$output .= '</Table>';
				$output .= '</Worksheet>';
				$output .= '</Workbook>';
			}
			else
			{
				$output .= '</interleave>';
			}
		}

		if (!$buffering)
		{
			print $output;
			$output = '';
		}

		if ($outputtype == "excel")
		{
			RealExcel($exceldata, "array", GetDownloadFileNameForTable($table));
		}
	}

	$t2_make_list = microtime(true);
	$t_make_list = $t2_make_list - $t1_make_list;
	if ($GLOBALS['SesMem']['debug'])
	{
		$output .= '<div>runtime ' . $listid . ': ' . $t_make_list . '</div>';
		$output .= '<div>size of html: ' . strlen($output) . ' bytes</div>';

		//we printen hem hier ook want als type_of_list == count dan wordt de ouput niet afgedrukt
		print '<h1>' . $listid . ' - ' . htme($GLOBALS['UC']['PersonalDataStructure'][$table]["plural"]) . ' - make_list() - ' . $type_of_list . ' - ' . $t_make_list . ' - countonly ' . $countonly . '</h1>';
	}

	qlog(INFO, "End make_list " . string_r($listoptions));

//	max_one_concurrent_load_finish($unique_command_id);

	if ($outputtype == "array")
	{
		if ($GLOBALS['SesMem']['debug'])
		{
			//output contains debug info
			print $output;
		}
		//1 element: array
		return $arraydata;
	}
	else if ($outputtype == "json")
	{
		if ($GLOBALS['SesMem']['debug'])
		{
			//output contains debug info
			print $output;
		}
		//2 elements: count and json
		return array($number_of_records, json_encode($arraydata));
	}
	else
	{
		//2 elements: count and html
		return array($number_of_records, $output);
	}
}
function from_iso_datetime($datetime, $ShowSeconds = false)
{
	if (($datetime == "") || ($datetime == "0000-00-00") || ($datetime == "0000-00-00 00:00:00"))
	{
		$datetime = "";
	}
	else
	{
		list($date, $time) = explode(" ", $datetime);
		list($year, $month, $day) = explode("-", $date);
		$datetime = $date;
		if ($GLOBALS['UC']['DateFormat'] == "dd-mm-yyyy")
		{
			$datetime = $day . "-" . $month . "-" . $year;
		}
		else if ($GLOBALS['UC']['DateFormat'] == "mm-dd-yyyy")
		{
			$datetime = $month . "-" . $day . "-" . $year;
		}
		if ($time != "")
		{
			list($hour, $minute, $second) = explode(":", $time);
			$meridien = "";
			if ($GLOBALS['UC']['DateFormat'] == "mm-dd-yyyy")
			{
				$meridien = "am";
				if ($hour > 12)
				{
					$hour -= 12;
					if ($hour < 10)
					{
						$hour = "0" . $hour;
					}
					$meridien = "pm";
				}
			}
			$datetime .= " " . $hour . ":" . $minute;
			if ($ShowSeconds)
			{
				$datetime .= ":" . $second;
			}
			if ($meridien != "")
			{
				$datetime .= " " . $meridien;
			}
		}
	}
	return $datetime;
}
function fieldalias_to_fullname($table_alias, $return_displaylistname_if_set = true, $technical = false)
{
	//technical includeert de eigen tabel naam, retourneert html, slaat nooit tags over, en gebruikt - voor de separator ipv spatie
	$full_fieldname = "";

	// HF20150508 Standaard sep ook hier raquo gemaakt, veel vragen over de structuur steeds en dit is veel duidelijker, al is het wat technischer.
	$separator = " » ";
	if ($technical)
	{
		$return_displaylistname_if_set = false;
	}

	$table_parts = explode("-", $table_alias);

	$last_field = $table_parts[count($table_parts)-1];
	$last_table = $table_parts[count($table_parts)-2];
	//deze overrulet de dynamische opbouw van field names. Naar mijn idee is displaylistname beter niet te gebruiken ivm consistentie van naamgeving
	if (($return_displaylistname_if_set) && ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["displaylistname"] != ""))
	{
		$full_fieldname = $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["displaylistname"];
	}
	else
	{
		if ($technical)
		{
			$first_table = $table_parts[0];
			$full_fieldname = $GLOBALS['UC']['PersonalDataStructure'][$first_table]["singular"];
		}

		$table_name = "";
		$field_name = "";
		$this_name_part = "";

		//TODO count -1 of -2?
		for ($x = 0; $x < count($table_parts); $x += 2)
		{
			$table_name = $table_parts[$x];
			$field_name = $table_parts[$x+1];

			$this_name_part = $GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["name"];
			if (($return_displaylistname_if_set) && ($GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["displaylistname"] != ""))
			{
				$this_name_part = $GLOBALS['UC']['PersonalDataStructure'][$table_name]["fields"][$field_name]["displaylistname"];
			}

			//komt alleen voor als [[id]] er af wordt geknipt bij de get_recursive_access_querypart (?) functie in html output
			if ($this_name_part != "")
			{
				if ($full_fieldname != "")
				{
					$full_fieldname .= $separator;
				}
				if ($technical)
				{
					$full_fieldname .= htme($this_name_part);
				}
				else
				{
					$full_fieldname .= $this_name_part;
				}
			}
		}
	}

	//2014-01-17 catch
	if ($full_fieldname == "")
	{
		$full_fieldname = $table_alias;
	}

	return $full_fieldname;
}
function CheckTableRecordAccess($table, $id)
{
	$type_of_access = "nok";

	if ($GLOBALS['SesMem']['ShowReasons'])
	{
		$GLOBALS['DECISIONTRACE'][] = "CheckTableRecordAccess for table " . $table . " record " . $id;
	}

	if (is_array($GLOBALS['UC']['PersonalDataStructure'][$table]))
	{
		if ($GLOBALS['SesMem']['ShowReasons'])
		{
			$GLOBALS['DECISIONTRACE'][] = "Table " . $table . " exists and user has access";
		}

		//id == _new_ moet eruit; daarom vast voorbereiding voor id == 0
		if (!is_positive_integer($id))
		{
			if ($GLOBALS['UC']['PersonalDataStructure'][$table]["allowed_to_add"])
			{
				$type_of_access = "ok";
			}
			if ($GLOBALS['SesMem']['ShowReasons'])
			{
				$GLOBALS['DECISIONTRACE'][] = "allowed_to_add: " . $type_of_access;
			}
		}
		else
		{
			$selectallfields = false;
			$bypass_record_access_by_sql = false;
			//geen select fields array en selectallfields=false; dan wordt alleen het primary id geselecteerd
			$record = getrecord($table, $id, $selectallfields, $bypass_record_access_by_sql);

			if ($record['type_of_access'] == "ok")
			{
				if ($record['record_locked_by'] == 0)
				{
					$type_of_access = "ok";
				}
				else
				{
					$type_of_access = "readonly";
				}
			}
			else if ($record['type_of_access'] == "readonly")
			{
				$type_of_access = "readonly";
			}

			if ($GLOBALS['SesMem']['ShowReasons'])
			{
				$GLOBALS['DECISIONTRACE'][] = "Access result from security: " . $type_of_access;
				if ($type_of_access == "nok")
				{
					$GLOBALS['DECISIONTRACE'][] = "No access which could mean the record does not exists, or record access prevents this user from seeing this record";
				}

				$selectallfields = true;
				$bypass_record_access_by_sql = true;
				$GLOBALS['DECISIONTRACE'][] = "bypass_record_access_by_sql is set to true";
				$GLOBALS['DECISIONTRACE'][] = "selectallfields is set to true";
				$GLOBALS['DECISIONTRACE'][] = "we will now run getrecord() again without sql security and pass through each record access rule in php (instead of sql)";
				//geen select fields array en selectallfields=false; dan wordt alleen het primary id geselecteerd

				$record = getrecord($table, $id, $selectallfields, $bypass_record_access_by_sql);
				$primary_key = $table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"];

				//we defaulteren naar ok; record_access_restrictions neemt namelijk rechten weg
				$type_of_access_php = "ok";

				if ($record[$primary_key] == 0)
				{
					$type_of_access_php = "nok";
					$GLOBALS['DECISIONTRACE'][] = "Record NOT found access " . $type_of_access_php;
				}
				else
				{
					$GLOBALS['DECISIONTRACE'][] = "Record found";

					foreach (array("read", "write") as $type_of_access_check)
					{
						//wanneer een record al nok is dan checken we geen write access meer
						if ($type_of_access_php == "ok")
						{
							$GLOBALS['DECISIONTRACE'][] = "We will now check " . $type_of_access_check . " access";
							foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["allfields"] as $allfield)
							{
								$table_parts = explode("-", $allfield);
								$last_field = $table_parts[count($table_parts)-1];
								$last_table = $table_parts[count($table_parts)-2];

								$addfield = false;
								if (is_array($GLOBALS['UC']['PersonalDataStructure'][$last_table]["record_access_restrictions"][$type_of_access_check][$last_field]))
								{
									//er ligt security op dit veld
									$addfield = true;

									if ($type_of_access_check == "write")
									{
										//het is een veld in een hogergelegen tabel; write access zou wel eens niet van toepassing kunnen zijn
										if (count($table_parts) > 2)
										{
											for ($x = 0; $x < count($table_parts) - 2; $x += 2)
											{
												$last_table2 = $table_parts[$x];
												$last_field2 = $table_parts[$x+1];

												//write access overerft tenzij InheritWriteAccessFromParent == false voor 1 van de tussengelegen relaties
												if (!$GLOBALS['UC']['PersonalDataStructure'][$last_table2]["fields"][$last_field2]["InheritWriteAccessFromParent"])
												{
													$addfield = false;
													$GLOBALS['DECISIONTRACE'][] = "write access is defined for " . $allfield . ", but InheritWriteAccessFromParent == false for " . $last_table2 . "-" . $last_field2 . ", so THIS table (" . $table . ") does not inherit its write access";
													break;
												}
											}
										}
									}

									if ($addfield)
									{
										for ($x = 0; $x < count($table_parts); $x += 2)
										{
											$last_table2 = $table_parts[$x];
											$last_field2 = $table_parts[$x+1];

											if ($GLOBALS['UC']['PersonalDataStructure'][$last_table2]["fields"][$last_field2]["relation_to"] != "")
											{
												$relation_field = "";
												for ($y = 0; $y <= $x + 1; $y++)
												{
													if ($relation_field != "")
													{
														$relation_field .= "-";
													}
													$relation_field .= $table_parts[$y];
												}
												if (($record[$relation_field] == 0) && (!$GLOBALS['UC']['PersonalDataStructure'][$last_table2]["fields"][$last_field2]["required"]))
												{
													$addfield = false;
													$GLOBALS['DECISIONTRACE'][] = $type_of_access_check . " security is defined for " . $allfield . ", but NON required reference " . $relation_field . "==0, therefore security is NOT applied because the parent table is NULL";
													break;
												}
											}
										}
									}
								}

								if ($addfield)
								{
									foreach (array("IS", "NOT") as $operand)
									{
										$conditions_checked = 0;
										$conditions_matched = 0;
										foreach ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["record_access_restrictions"][$type_of_access_check][$last_field][$operand] as $value)
										{
											$conditions_checked++;
											$match = false;
											if ($record[$allfield] == $value)
											{
												$match = true;
											}
											if ($operand == "NOT")
											{
												$match = !$match;
											}

											if ($match)
											{
												$GLOBALS['DECISIONTRACE'][] = $allfield . " " . $type_of_access_check . " check, record value \"" . $record[$allfield] . "\" " . $operand . " compare value \"" . $value . "\" condition PASSED";
												$conditions_matched++;
												//er hoeft er maar 1 te matchen om de access toe te staan
												if ($operand == "IS")
												{
													break;
												}
											}
											else
											{
												$GLOBALS['DECISIONTRACE'][] = $allfield . " " . $type_of_access_check . " check, record value \"" . $record[$allfield] . "\" " . $operand . " compare value \"" . $value . "\" condition NOT passed";
												//er hoeft er maar 1 NIET te matchen om de access NIET toe te staan
												if ($operand == "NOT")
												{
													break;
												}
											}
										}

										if (($operand == "IS") && ($conditions_matched == 0))
										{
											$subquery = unescape_interleave_characters(ParseTemplateUserAndGroupFields($GLOBALS['UC']['PersonalDataStructure'][$last_table]["record_access_restrictions"][$type_of_access_check][$last_field]["query"]));
											if ($subquery != "")
											{
												$GLOBALS['DECISIONTRACE'][] = "running custom subquery sql for record access: " . $subquery;

												//ook als een subquery geen results heeft dan telt het runnen van de query nog steeds als een gecheckte conditie. als de query geen results heeft moet access ontnomen worden
												$conditions_checked++;

												foreach (db_GetArray($subquery) as $allowed_records)
												{
													$conditions_checked++;
													if ($allowed_records[0] == $record[$allfield])
													{
														$GLOBALS['DECISIONTRACE'][] = $allfield . " " . $type_of_access_check . " check, query result \"" . $allowed_records[0] . "\" compare against \"" . $record[$allfield] . "\" condition PASSED";

														$conditions_matched++;
														break;
													}
													else
													{
														$GLOBALS['DECISIONTRACE'][] = $allfield . " " . $type_of_access_check . " check, query result \"" . $allowed_records[0] . "\" compare against \"" . $record[$allfield] . "\" condition NOT passed";
													}
												}
											}
										}

										$access = true;
										if ($conditions_checked > 0)
										{
											//slechts 1 conditie hoeft te machten
											if (($operand == "IS") && ($conditions_matched == 0))
											{
												$access = false;
											}
											//alle condities moeten machten
											else if (($operand == "NOT") && ($conditions_matched < $conditions_checked))
											{
												$access = false;
											}
										}

										if (!$access)
										{
											if ($type_of_access_check == "read")
											{
												$type_of_access_php = "nok";
											}
											else
											{
												$type_of_access_php = "readonly";
											}
											$GLOBALS['DECISIONTRACE'][] = $type_of_access_check . " ACCESS RULE IN EFFECT, RECORD IS " . $type_of_access_php;
											break 2;
										}
									}
								}
							}
						}
					}
					if (($type_of_access_php == "ok") && (!$GLOBALS['UC']['PersonalDataStructure'][$table]["allowed_to_write"]))
					{
						//door overerving van readonlyness van een hele tabel kan het zijn dat security via een multi select field loopt, waarop anders geen security wordt gelegd. deze conditie ondervangt dat.
						$GLOBALS['DECISIONTRACE'][] = "This whole table is readonly, regardless of record access rules. Maybe because InheritWriteAccessFromParent is true for a multiselect reference to table which is readonly itself. Search Data Structure Access Reasons for occurences of InheritWriteAccessFromParent.";
						$type_of_access_php = "readonly";
					}

					$GLOBALS['DECISIONTRACE'][] = "FINAL ACCESS DECISION IS " . $type_of_access_php;
				}

				if ($type_of_access != $type_of_access_php)
				{
					$errortext = "ERROR: CheckTableRecordAccess something is wrong type_of_access (sql) is " . $type_of_access . " and type_of_access_php is " . $type_of_access_php . " BOTH SHOULD BE THE SAME (check for an error of 'Too many tables in join' which could be the cause)";
					$GLOBALS['DECISIONTRACE'][] = $errortext;
					log_msg($errortext);
				}
			}
		}
	}
	else
	{
		if ($GLOBALS['SesMem']['ShowReasons'])
		{
			$GLOBALS['DECISIONTRACE'][] = "Table " . $table . " doet NOT exists OR user has access, see Personal Data Structure Reasons for more information";
		}
	}

	return $type_of_access;
}
function getrecord($table, $id, $selectallfields = true, $bypass_record_access_by_sql = false, $selectfields = array(), $do_not_check_write_access = false)
{
	$record = array();

	$t1 = microtime(true);
	$query = "";

	if ((is_array($GLOBALS['UC']['PersonalDataStructure'][$table])) && (is_positive_integer($id)))
	{
		$search = "";
		$searchperfield = array();
		$start = 0;
		$perpage = 0; //alles; het is er toch maar 1
		$sortby = "";
		$sortdirection = "asc";
		$filtercolumn = "";
		$filter = array();
		$filtertype = array();
		$listlayout = array();
		//TODO als de text om een extra select conditie niet is gezet, dan zou die niet weggehaald moeten mogen worden en zou de persoon dus geen access hebben. DAT MOET omgeschreven worden, ExtraSelectCondition mag NIET een security setting zijn; daarvoor moeten Record access rules gebruikt worden
		$RemoveExtraSelectCondition = true; //belangrijk!
		$connecting_field = "";
		$use_pre_select_records = false;

		$search_parents = true;
		$search_childs = true;
		$search_files = true;

		$pre_select_records = array();

		$filter[$table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"]][] = $id;

		//niet belangrijk
		$type_of_list = "mainlist";

		$tabletype = "normal";
		//allen in combinatie met tabletype = select
		$childtable = "";
		$childfield = "";
		//dit is het callende child record, moet 0 zijn
		$childrecordid = 0;

		if (count($selectfields) > 0)
		{
			$listlayout = $selectfields;
		}
		else if ($selectallfields)
		{
			$listlayout = $GLOBALS['UC']['PersonalDataStructure'][$table]["allfields"];
		}
		else
		{
			//we selecteren alleen het primary field; da's snel
			$listlayout[] = $table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"];
		}

		$return_countquery = false;

		$already_joined = array();

		$hiddenfilter = array();
		$hiddenfiltertype = array();
		$listid = "";
		$order_by_pre_select = false;

		$query = make_query($table, $search, $searchperfield, $start, $perpage, $sortby, $sortdirection, $filtercolumn, $filter, $filtertype, $listlayout, $RemoveExtraSelectCondition, $connecting_field, $use_pre_select_records, $pre_select_records, $order_by_pre_select, $search_parents, $search_childs, $search_files, $return_countquery, $type_of_list, $bypass_record_access_by_sql, $tabletype, $childtable, $childfield, $childrecordid, $already_joined, $hiddenfilter, $hiddenfiltertype, $listid, $do_not_check_write_access);

		$record = db_GetRow($query);
	}

	$t2 = microtime(true);
	$t = $t2 - $t1;

	if ($GLOBALS['SesMem']['debug'])
	{
		print '<h1>getrecord query</h1>';
		print '<div>' . htme($query) . '</div>';
		print '<div>runtime: ' . $t . '</div>';
/*		print '<pre>';
		debug_print_backtrace();
		print '</pre>';*/
	}

	//als je wilt weten of er toegang is, dan moet er gecheckt wordt of $record[primary-key] > 0 is
	return $record;
}
function get_tags($template)
{
	//retourneeert de *UNIEKE* tags; als een tag 2x in een template staat wordt hij dus maar 1x geretourneerd
	//deze functie moet gerunt worden NADAT een template is omgezet met parse_old_template_to_new_template()

	$tags = array();

	$count = preg_match_all("/(\@|\#|\%|\^)([a-zA-Z0-9\-\_]+)\\1/", $template, $matches);
	for ($x = 0; $x < $count; $x++)
	{
		$tag = $matches[2][$x];
		//om te voorkomen dat DATE uit @DATE@ wordt geretourneerd
//		if ((in_array($tag, $GLOBALS['UC']['PersonalDataStructure'][$table]["allfields"])) && (!in_array($tag, $tags)))
		if (!in_array($tag, $tags))
		{
			$tags[] = $tag;
		}
	}

	return $tags;
}
//ZERO not included
function is_positive_integer($num)
{
	$result = false;
	if (preg_match("/^[1-9][0-9]*$/", $num) === 1)
	{
		$result = true;
	}
	return $result;
}
//ZERO included
function is_positive_integer_or_zero($num)
{
	$result = false;
	if ((is_positive_integer($num)) || ($num == "0"))
	{
		$result = true;
	}
	return $result;
}
function CheckIfRecordHasChilds($table, $recordid)
{
	//this function is somewhat similar to descendingdelete_recursive()

	//we moeten hier werken met de SystemDataStructure omdat de huidige user een tabel niet zou kunnen zien die wel moet verhinderen dat dit record gedelete kan worden
	$SystemDataStructure = GetSystemDataStructure();

	$RecordHasChilds = false;

	//deleted users remain visible in Interleave, they are logically-logically deleted. For this reason we do not descend from the users table. This also means we don't have to exclude uselog, journal, datajournal and mailqueue from the child tables
	if ($table != "users")
	{
		foreach (array_keys($SystemDataStructure[$table]["childs"]) as $child_table)
		{
			foreach ($SystemDataStructure[$table]["childs"][$child_table] as $child_field)
			{
				$table_in_database = $SystemDataStructure[$child_table]["name_in_database"];
				//geen support voor "calculated on display"; moet nog omzetten naar return_database_fieldname()
				$field_in_database = $SystemDataStructure[$child_table]["fields"][$child_field]["name_in_database"];

				$query = "SELECT `" . $SystemDataStructure[$child_table]["primary"] . "` FROM `CRM" . $table_in_database . "` WHERE ";
				if ((preg_match("/^t\d+$/", $child_table)) && ($GLOBALS['UC']['PersonalDataStructure'][$child_table]["logical_delete"]))
				{
					$query .= "`deleted`='n' AND ";
				}
				$query .= "`" . $field_in_database . "`='" . mres($recordid) . "'";

				$childrecord = db_GetValue($query);
				if ($childrecord > 0)
				{
					$RecordHasChilds = true;
					//we breaken uit de lus
					break 2;
				}
			}
		}
	}

	return $RecordHasChilds;
}
function select_refer_field_options($childtable, $childfield, $childrecordid = 0, $check_if_only_one_record = false, $check_if_exists_recordids = array())
{
	//selecteert de mogelijke parent records waaraan gerefereert wordt vanuit de child; de door de child gewenste refer field layout wordt gehonoreerd
	$refer_field_options = array();

	$table = $GLOBALS['UC']['PersonalDataStructure'][$childtable]["fields"][$childfield]["relation_to"];
	if (($table != "") && (is_array($GLOBALS['UC']['PersonalDataStructure'][$table])))
	{
		$search = "";
		$searchperfield = array();
		$start = 0;
		$perpage = 0;

		$sortby = $GLOBALS['UC']['PersonalDataStructure'][$table]["default_sortby"];
		$sortdirection = $GLOBALS['UC']['PersonalDataStructure'][$table]["default_sortdirection"];

		if ($check_if_only_one_record)
		{
			//op 0 krijg je alle records terug; in form-functions wordt dit gecallt met een limit van 2 om te bepalen of er slechts *1* mogelijke keus is, die dan gepreselect wordt
			$perpage = 2;
			//sort interesseert ons dan niet; dit versneld de query op grote tabellen
			$sortby = $table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"];
			$sortdirection = "asc";
		}

		$filter = array();
		$filtertype = array();

		//we willen weten OF dit record voorkomt in de lijst van mogelijk records; voor de validatie functie; daarom filteren we er op voor snelheid
		foreach ($check_if_exists_recordids as $check_if_exists_recordid)
		{
			$filter[$table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"]][] = $check_if_exists_recordid;
		}

		$filtercolumn = "";
		$listlayout = array();
		//TODO als de text om een extra select conditie niet is gezet, dan zou die niet weggehaald moeten mogen worden en zou de persoon dus geen access hebben. DAT MOET omgeschreven worden, ExtraSelectCondition mag NIET een security setting zijn; daarvoor moeten Record access rules gebruikt worden
		$RemoveExtraSelectCondition = true; //belangrijk!
		$connecting_field = "";
		$use_pre_select_records = false;

		$search_parents = true;
		$search_childs = true;
		$search_files = true;

		$use_pre_select_records = false;
		$pre_select_records = array();

		$type_of_list = "mainlist";

		$tabletype = "select";

		//we selecteren het primary field
		$listlayout[] = $table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"];

		foreach ($GLOBALS['UC']['PersonalDataStructure'][$childtable]["fields"][$childfield]["layout_tags"] as $tag)
		{
			//omdat we 1 tabel hoger selecteren moeten we de huidige tabel en field name van de tags afhalen want die gaat er vanuit dat er vanaf 1 niveau LAGER wordt geparsed. DEZE CODE STAAT MEERDERE KEREN IN DEZE FILE
			$listlayout[] = preg_replace("/^" . $childtable . "\-" . $childfield . "\-/", "", $tag);
		}

		$return_countquery = false;

		$already_joined = array();

		//childrecordid is het callende child record

		$order_by_pre_select = false;
		$bypass_record_access_by_sql = false;

		$hiddenfilter = array();
		$hiddenfiltertype = array();
		$listid = "";
		$do_not_check_write_access = false;

		$query = make_query($table, $search, $searchperfield, $start, $perpage, $sortby, $sortdirection, $filtercolumn, $filter, $filtertype, $listlayout, $RemoveExtraSelectCondition, $connecting_field, $use_pre_select_records, $pre_select_records, $order_by_pre_select, $search_parents, $search_childs, $search_files, $return_countquery, $type_of_list, $bypass_record_access_by_sql, $tabletype, $childtable, $childfield, $childrecordid, $already_joined, $hiddenfilter, $hiddenfiltertype, $listid, $do_not_check_write_access);

		if ($GLOBALS['SesMem']['debug'])
		{
			print '<h1>select_refer_field_options</h1>';
			print '<div>childtable: ' . htme($childtable) . '</div>';
			print '<div>childfield: ' . htme($childfield) . '</div>';
			print '<div>childrecordid: ' . htme($childrecordid) . '</div>';
			print '<div>query: ' . htme($query) . '</div>';
		}

		$primary_key = $table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"];

		foreach (db_GetArray($query) as $record)
		{
			$refer_field_options[$record[$primary_key]] = parse_refer_layout_for_field($childtable, $childfield, $record, true);
		}

		if (is_positive_integer($childrecordid))
		{
			$currently_selected_recordids = array();
			//geen support voor "calculated on display"; moet nog omzetten naar return_database_fieldname()
			$currently_selected = db_GetValue("SELECT `" . $GLOBALS['UC']['PersonalDataStructure'][$childtable]["fields"][$childfield]["name_in_database"] . "` FROM `CRM" . $GLOBALS['UC']['PersonalDataStructure'][$childtable]["name_in_database"] . "` WHERE `" . $GLOBALS['UC']['PersonalDataStructure'][$childtable]["primary"] . "`=" . $childrecordid);
			if ($GLOBALS['UC']['PersonalDataStructure'][$childtable]["fields"][$childfield]["multiselect"])
			{
				$currently_selected_recordids[] = unserialize($currently_selected);
			}
			else
			{
				$currently_selected_recordids[] = $currently_selected;
			}

			foreach ($currently_selected_recordids as $currently_selected_recordid)
			{
				if (is_positive_integer($currently_selected_recordid))
				{
					//om een keuze toe te voegen aan de dropdown die op dit moment niet langer gekozen mag worden zodat oude records hun referentie houden
					if (!array_key_exists($currently_selected_recordid, $refer_field_options))
					{
						$refer_field_options[$currently_selected_recordid] = get_refer_layout_for_record_field($childtable, $childfield, $currently_selected_recordid);
					}
				}
			}
		}
	}

	return $refer_field_options;
}
function getfieldcolors($table, $field, $value)
{
	$bgcolor = "";
	$fgcolor = "";

	if (is_array($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["colors"]))
	{
		$numeric_value = $value;
		if ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["fieldtype"] == "date")
		{
			$numeric_value = NLDate2Epoch($numeric_value);
		}

		foreach (array("IS", "LT", "GT", "HAS") as $operand)
		{
			if (is_array($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["colors"][$operand]))
			{
				if ($operand == "IS")
				{
					$bgcolor = $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["colors"][$operand][$value]['background'];
					$fgcolor = $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["colors"][$operand][$value]['foreground'];
				}
				else if ($operand == "LT")
				{
					foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["colors"][$operand]) as $colorvalue)
					{
						if ($numeric_value < $colorvalue)
						{
							$bgcolor = $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["colors"][$operand][$colorvalue]['background'];
							$fgcolor = $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["colors"][$operand][$colorvalue]['foreground'];
						}
					}
				}
				else if ($operand == "GT")
				{
					foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["colors"][$operand]) as $colorvalue)
					{
						if ($numeric_value > $colorvalue)
						{
							$bgcolor = $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["colors"][$operand][$colorvalue]['background'];
							$fgcolor = $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["colors"][$operand][$colorvalue]['foreground'];
						}
					}
				}
				else if ($operand == "HAS")
				{
					foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["colors"][$operand]) as $colorvalue)
					{
						//kan met strpos
						if (preg_match("/" . preg_quote($colorvalue) . "/", $value))
						{
							$bgcolor = $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["colors"][$operand][$colorvalue]['background'];
							$fgcolor = $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["colors"][$operand][$colorvalue]['foreground'];
						}
					}
				}
			}
		}
	}

	return array($bgcolor, $fgcolor);
}
function inline_jsencode($str)
{
//gebruik deze functie om inline javascript te coderen, zoals hier: print '<input type="button" onclick="alert(\'' . inline_jsencode($_GET['bericht']) . '\');" value="klik mij">';

	// \ -> \\
	$str = str_replace('\\', '\\\\', $str);

	// ' -> \'
	$str = str_replace('\'', '\\\'', $str);

	// " -> \"
//	$str = str_replace('"', '\\"', $str);

	//we gebruiken geen htme() want die codeert ook single quotes naar &#039; en dan klapt javascript er weer uit

	$str = htmlspecialchars($str);

	return $str;
}
function get_recursive_access_querypart($table, $table_alias = "", $type_of_access = "read", $already_joined = array(), $human_readable = false)
{
	if ($table_alias == "")
	{
		$table_alias = $table;
	}

	$securityquery = "";
	//we hoeven verder niks te checken want we weten al dat de hele tabel readonly is
	if (($type_of_access == "write") && (!$GLOBALS['UC']['PersonalDataStructure'][$table]["allowed_to_write"]))
	{
		$securityquery = "FALSE";
	}
	else
	{
		foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]) as $field)
		{

//2017-08-21; fields in the user table can no longer be displayed in other tables; they ARE still displayed when viewing the user table directly. this code is duplicated in make_query_parts() and get_recursive_access_querypart()
//TODO: instead of looping through the fields manually in get_recursive_access_querypart() we should use the allfields array in the DS.
if ((!$GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["hide_field_when_in_child_table"]) || ($table_alias == $table))
{

			$select_field = return_database_fieldname($table, $field, $table_alias);

			//er ligt security op dit veld
			if (is_array($GLOBALS['UC']['PersonalDataStructure'][$table]["record_access_restrictions"][$type_of_access][$field]))
			{
				$full_alias = $table_alias;
				//als ie op primary field ligt dan hoeven we [[id]] niet weer te geven want we parsen de layout toch nog
				if ($GLOBALS['UC']['PersonalDataStructure'][$table]["primary"] != $field)
				{
					$full_alias .= "-" . $field;
				}

				foreach (array("IS", "NOT") as $operand)
				{
					$restrictions_this_field = array();
					if (count($GLOBALS['UC']['PersonalDataStructure'][$table]["record_access_restrictions"][$type_of_access][$field][$operand]) > 0)
					{
						$querypart = "";

						if ($human_readable)
						{
							$langtag = '=';
							if ($operand == "NOT")
							{
								$langtag = '!=';
							}
							$querypart = fieldalias_to_fullname($full_alias, false, true) . " " . $langtag . " ";
						}
						else
						{
							$select_field = convert_database_date_and_number_column($table, $field, $select_field);

							$querypart = $select_field . " ";
							if ($operand == "NOT")
							{
								$querypart .= "NOT ";
							}
							$querypart .= "IN (";
						}

						$count = 0;
						foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["record_access_restrictions"][$type_of_access][$field][$operand] as $value)
						{
							if ($count > 0)
							{
								$querypart .= ",";
								if ($human_readable)
								{
									$querypart .= " ";
								}
							}

							if ($human_readable)
							{
								$targettable = "";
								if ($GLOBALS['UC']['PersonalDataStructure'][$table]["primary"] == $field)
								{
//									$querypart .= ' PRIMARY';
									$targettable = $table;
								}
								else if ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["relation_to"] != "")
								{
//									$querypart .= ' RELATION TO ';
									$targettable = $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["relation_to"];
								}

								if ($targettable != "")
								{
									//uitzondering omdat de group comparison inlogt als een user, dat leidt dan tot "verschillen" in de groups omdat de fullname niet overeen komt, daarom USERID
									if ($targettable == "users")
									{
										$value = "USERID";
									}
									else
									{
										//deze wordt altijd geparsed op basis van de layout van de bovenliggende tabel en dus NIET op basis van de gewenste layout van het refer field
										$value = get_refer_layout_for_record($targettable, $value);
									}
								}

								if ($value == "")
								{
									//IS NULL?
									$querypart .= '<em>empty</em>';
								}
								else
								{
									$querypart .= htme($value);
								}
							}
							else
							{
								if ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["fieldtype"] == "id")
								{
									//om dan maar geen results te hebben
									if (!is_positive_integer_or_zero($value))
									{
										//IS NULL?
										$value = -1;
									}
									//hoeft niet door mres() en hoeven ook geen quotes om heen
									$querypart .= $value;
								}
								else
								{
									$querypart .= "'" . mres($value) . "'";
								}
							}

							$count++;
						}

						if (!$human_readable)
						{
							$querypart .= ")";
						}
						$restrictions_this_field[] = $querypart;
					}

					if ($operand == "IS")
					{
						//er is maximaal 1 sql_subselect_security en deze retourneert altijd de primary id's van de tabel waar we in zitten waar je WEL in mag; als er vaste nummer worden opgegeven dan moeten die voorafgaan door een select, dus: "SELECT 712"
						if ($GLOBALS['UC']['PersonalDataStructure'][$table]["record_access_restrictions"][$type_of_access][$field]["query"] != "")
						{
							//let op in reasons staat dezelfde code

							$querypart = unescape_interleave_characters(ParseTemplateUserAndGroupFields($GLOBALS['UC']['PersonalDataStructure'][$table]["record_access_restrictions"][$type_of_access][$field]["query"]));

							//hij zou nu leeg kunnen zijn door de php parse
							if ($querypart != "")
							{
								if ($human_readable)
								{
									$langtag = '=';
									$restrictions_this_field[] = fieldalias_to_fullname($full_alias, false, true) . " " . $langtag . " " . $querypart;
								}
								else
								{
									//2016-11-09 we materializen alleen voor flextables; zonder deze uitzondering zou ook journal, datajournal en files worden gematerialized omdat er op die 3 tabellen voor iedere niet admin user ook sql security ligt; omdat die queries heel veel records kunnen retourneren lopen we daardoor out of memory
									//...........als we in de toekomst een keer a) een flextable hebben met heel veel records b) met daarop sql security die c) heel veel records retourneert dan zouden we een attribute moeten maken op flextable niveau DoNotMaterializeSecuritySubQuery
									if ((preg_match("/^t\d+$/", $table)) && (!$GLOBALS["DO_NOT_USE_php_subquery_materialization"]))
									{
										if ($GLOBALS["PageCache"]["php_subquery_materialization"][$table]["record_access_restrictions"][$type_of_access][$field] == "")
										{
											//deze access queries worden slechts 1x per pageload uitgevoerd; dit is heel veel sneller dan door mysql want
											//a) die materialized ze niet altijd waardoor je suboptimale join volgorde kan krijgen
											//b) de materialization werkt nu gedurende een hele page load; zeker als er filters op een lijst aan staan kan hiermee voorkomen worden dat dezelfde subquery tig keer wordt uitgevoerd
											//naast snelheid is een ander voordeel dat dit garandeert dat er met dezelfde result set wordt gewerkt gedurenden een pageload; als er halverwege de pageload een record wordt aangepast dan leidt dit niet tot inconsistente resulaten (bv filters op waarden die niet meer bestaan)
											$GLOBALS["PageCache"]["php_subquery_materialization"][$table]["record_access_restrictions"][$type_of_access][$field] = implode(",",db_GetFlatArray($querypart));
											if ($GLOBALS["PageCache"]["php_subquery_materialization"][$table]["record_access_restrictions"][$type_of_access][$field] == "")
											{
												$GLOBALS["PageCache"]["php_subquery_materialization"][$table]["record_access_restrictions"][$type_of_access][$field] = "0";
											}
											if ($GLOBALS['SesMem']['debug'])
											{
												print "php_subquery_materialization " . $table . " " . $type_of_access . " " . $field . " = " . $GLOBALS["PageCache"]["php_subquery_materialization"][$table]["record_access_restrictions"][$type_of_access][$field] . "<br>";
											}
										}
										$restrictions_this_field[] =  $select_field . " IN (" . $GLOBALS["PageCache"]["php_subquery_materialization"][$table]["record_access_restrictions"][$type_of_access][$field] . ")";
									}
									else
									{
										$restrictions_this_field[] = $select_field . " IN (" . $querypart . ")";
									}
								}
							}
						}
					}

					if (count($restrictions_this_field) > 0)
					{
						if ($securityquery != "")
						{
							if ($human_readable)
							{
								$securityquery .= "<br>";
							}
							else
							{
								$securityquery .= " AND ";
							}
						}
						//dit is alleen zo wanneer er standaard IS security op dit veld ligt EN custom sql security
						if (count($restrictions_this_field) > 1)
						{
							$securityquery .= "(";
						}
						$securityquery .= implode(" OR ", $restrictions_this_field);
						if (count($restrictions_this_field) > 1)
						{
							$securityquery .= ")";
						}
					}
				}
			}


//fields in the user table are not joined upwards from child tables; to prevent bloated (and potentially circular) datastructures. this code is duplicated in make_query_parts() and get_recursive_access_querypart()
if (($table == "users") && ($table_alias != "users") && (preg_match("/^f\d+$/", $field)))
{
}
else
{

			$relation_to = $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["relation_to"];
			//geen security op multi select relation velden

			if (($relation_to != "") && (!$GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["multiselect"]) && (($type_of_access == "read") || ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["InheritWriteAccessFromParent"])))
			{
				//nieuwe manier om circulair joins te voorkomen. ik *denk* dat dit betrouwbaar is
	//			if (!preg_match("/" . $field . "-" . $relation_to . "/", $table_alias))

				if ($already_joined[$table][$field] == 0)
				{
					$table_alias_parent = $table_alias . "-" . $field . "-" . $relation_to;
					$already_joined_parent = $already_joined;
					$already_joined_parent[$table][$field]++;

					$securityquery_on_parent = get_recursive_access_querypart($relation_to, $table_alias_parent, $type_of_access, $already_joined_parent, $human_readable);
					if ($securityquery_on_parent != "")
					{
						if ($securityquery != "")
						{
							if ($human_readable)
							{
								$securityquery .= "<br>";
							}
							else
							{
								$securityquery .= " AND ";
							}
						}
						//eigenlijk willen we dit alleen op required velden doen, maar als er dan meerdere referenties naar (bijvoorbeeld) CMRflextable3.recordid liggen dan zit de =0 op het ene veld wel en op het andere veld niet. hierdoor wordt de query VEEL trager vandaar dat we de =0 altijd toevoegen
						//update 2014-02-27: we moeten niet ALTIJD =0 toevoegen, want dan betekent dat als er een keer 0 staat in een required veld (bijvoorbeeld een rfc die om een of andere reden niet aan een klant is toegewezen), dan is die rfc voor alle klanten inzichtelijk. NIET GOED.
						//dit hadden we eerder opgelost met ALSOALLOWNULL

						if (!$GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["required"])
						{
							//hij is niet required dus mag hij ook 0 zijn, in dat geval moet de parent security er niet opgelegd worden want al die velden zijn dan toch NULL
							if ($human_readable)
							{
								$langtag = '=';
								$securityquery .= fieldalias_to_fullname($table_alias . "-" . $field, false, true) . " " . $langtag . " 0 OR " . $securityquery_on_parent;
							}
							else
							{
								$securityquery .= "((" . $select_field . "=0) OR (" . $securityquery_on_parent . "))";
							}
						}
						else
						{
							$securityquery .= $securityquery_on_parent;
						}
					}
				}
			}
}
}
		}
	}

	return $securityquery;
}
function GetCalendarEvents($calObjId, $caller, $calender_startdate = "1800-01-01 00:00:00", $calender_enddate = "2200-01-01 00:00:00")
{
	//$caller = html | messages | ical -> this is added to the listid; to allow multiple different calendar runs at the same time; interleave_exectute_command_start()

	$calender_items = array();

	if ($calObjId != "")
	{
		$calendars = GetAccessibleCalendars();
		$calObj = $calendars[$calObjId];

		if (!is_array($calObj))
		{
			PrintAD("Unknown calendar object " . $calObjId);
		}
		list($table, $field_ignore) = explode("-", $calObj["ref_end_date_time"]);

		//dit zijn html velden; TODO; dit zouden GEEN html velden moeten zijn
		$to_parse = array();
		$to_parse['summary'] = $calObj['summary'];
		$to_parse['resource'] = $calObj['resource'];
		$to_parse['comment'] = $calObj['comment'];
		$to_parse['location'] = $calObj['location'];

		$filter = array();

		//bezoeken die geeindigd zijn VOOR het begin van de huidige kalender hoeven we niet te zien
		//events zonder startdate in db (0000-00-00), worden hiermee ook weergegeven
		$filter[$calObj["ref_end_date_time"]]["from"] = $calender_startdate;
		//bezoeken die beginnen NA het eind van de huidige kalender hoeven we niet te zien
		$filter[$calObj["ref_start_date_time"]]["to"] = $calender_enddate;

		$RemoveExtraSelectCondition = true; //belangrijk!

		$type_of_list = "mainlist";

		$primary_key = $table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"];

		//we selecteren het primary field
		$listlayout = array();
		$listlayout[] = $primary_key;
		$listlayout[] = $calObj["ref_start_date_time"];
		$listlayout[] = $calObj["ref_end_date_time"];

		if ($calObj["ref_organiser"] != "")
		{
			$listlayout[] = $calObj["ref_organiser"];
		}

		$to_parse_tags = array();
		foreach (array_keys($to_parse) as $field)
		{
			$to_parse[$field] = ParseTemplateGeneric($to_parse[$field]);
			$to_parse[$field] = parse_old_template_to_new_template($to_parse[$field], $table);
			$to_parse_tags[$field] = get_tags($to_parse[$field]);
			foreach ($to_parse_tags[$field] as $field_to_select)
			{
				if (!in_array($field_to_select, $listlayout))
				{
					$listlayout[] = $field_to_select;
				}
			}
		}

		$calender_startdate_numeric = str_replace("-", "", $calender_startdate);
		$calender_enddate_numeric = str_replace("-", "", $calender_enddate);

		$listoptions = array();
		$listoptions["table"] = $table;
		$listoptions["listid"] = $table . "cal" . $caller;
		$listoptions["start"] = 0;
		$listoptions["perpage"] = 0;
		$listoptions["internallistcall"] = true; //om te voorkomen dat perpage = 0 wordt overrulet in make_list()
		$listoptions["filter"] = $filter;
		$listoptions["filtertype"] = array(); //default IS
		$listoptions["listlayoutcolumns"] = implode(",", $listlayout);
		$listoptions["RemoveExtraSelectCondition"] = $RemoveExtraSelectCondition;
		$listoptions["type_of_list"] = $type_of_list;
		$listoptions["outputtype"] = "array";
		//dit moet nog beter
		$listoptions["ignorelastview"] = true;

		$listoptions["do_not_check_write_access"] = true;

		$ExtraListOptions_Packed = $calObj['listoptions'];
		if (trim($ExtraListOptions_Packed) != "")
		{
			$ExtraListOptions_Packed = unescape_interleave_characters(ParseTemplateUserAndGroupFields($ExtraListOptions_Packed));
			if (ipi($_GET["NUMBERFROMGET"]))
			{
				$ExtraListOptions_Packed = str_replace("@NUMBERFROMGET@", $_GET["NUMBERFROMGET"], $ExtraListOptions_Packed);
			}

			// Converteer deze naar array
			$ExtraListOptions = ConvertTagOptionsToListOptions($ExtraListOptions_Packed);

			// Merge deze met de huidige listoptions, maar de huidige listoptions prevaleren!
			$listoptions = array_merge($ExtraListOptions, $listoptions);
		}
		if ($GLOBALS['SesMem']['debug'])
		{
			print '<div>Calendar list options:</div>';
			DA($listoptions);
		}

		//make_list retourneert een array
		foreach (make_list($listoptions) as $record)
		{
			$start = $record[$calObj["ref_start_date_time"]];
			//het is een date field en niet een date/time
/*			if (strlen($start) == 10)
			{
				$start .= " 00:00:00";
			}*/

			$end = $record[$calObj["ref_end_date_time"]];
			//het is een date field en niet een date/time
/*			if (strlen($end) == 10)
			{
				$end .= " 23:59:59";
			}*/

			//de startdatum kan ook leeg zijn, in dat geval is de einddatum wel gevuld; we moeten dan nog wel kijken of die binnen de huidige kalender valt
			$add_this_calender_item = false;

			//het is een "allday" event
			if ($start == "0000-00-00 00:00:00")
			{
				list($enddate, $endtime) = explode(" ", $end);
				$enddate_numeric = str_replace("-", "", $enddate);
				if (($enddate_numeric >= $calender_startdate_numeric) && ($enddate_numeric <= $calender_enddate_numeric))
				{
					$add_this_calender_item = true;
	//				$start = $enddate . " 00:00:00";
					$start = $end;
				}
			}
			else
			{
				$add_this_calender_item = true;
			}

//2015-03-31 dit moet in de renderer niet in de get events!
/*
			if ((strlen($start) == 19) && (strlen($end) == 19))
			{
				list($start_date, $start_time) = explode(" ", $start);
				list($end_date, $end_time) = explode(" ", $end);
				if (($start_date != $end_date) && ($end_time == "00:00:00"))
				{
					//fix 2015-03-030 om gelijk te lopen met Thunderbird en Apple kalender
					//een kalender item wat begint op 29-03-2015 00:00 en duurt tot 30-03-2015 00:00 zie ik op 29-03 maar NIET op 30-03.
					//echter, een kalender item wat begint op 30-03-2015 00:00 en duurt tot 30-03-2015 00:00 (een enkel tijdstip dus) zie ik op WEL op 30-03.
					$end = db_GetValue("SELECT DATE_ADD('" . $end . "', INTERVAL -1 SECOND)");
				}
			}
*/

			if ($add_this_calender_item)
			{
				$calender_item = array();

				//niet in gebruik maar misschien wel handig
				$calender_item['recordid'] = $record[$primary_key];
				$calender_item['openlink'] = $GLOBALS['UC']['PersonalDataStructure'][$table]["openlink"] . $record[$primary_key];
				$calender_item['table'] = $table;

				$calender_item['iso_start'] = $start;
				$calender_item['iso_end'] = $end;

				foreach (array_keys($to_parse) as $field)
				{
					$calender_item[$field] = $to_parse[$field];
					if ($calender_item[$field] != "")
					{
						//TODO PARSING CENTRALISEREN?
						foreach ($to_parse_tags[$field] as $tag)
						{
							$calender_item[$field] = str_replace("@" . $tag . "@", escape_interleave_characters(formatvalue($record[$tag], $tag)), $calender_item[$field]);
						}
						$calender_item[$field] = unescape_interleave_characters($calender_item[$field]);
					}
				}

				$calender_item['user'] = "";
				if ($record[$calObj["ref_organiser"]] > 0)
				{
					//2017-08-25: isn't the user's fullname already selected in the query?
					//dit is de Organiser
					$calender_item['user'] = GetUserEmail($record[$calObj["ref_organiser"]]);
				}
				$calender_item['calendarname'] = $calObj['description'];
				$calender_item['reminder'] = $calObj['reminder_in_minutes'];
				$calender_item['id'] = $calObj['id'];

				$calender_items[] = $calender_item;
			}
		}

		//niet zo'n mooie code maar het werkt
		//http://php.net/manual/en/function.array-multisort.php
		// Obtain a list of columns
		//sorteren niet nodig
		/*
		$start = array();
		$id = array();
		foreach ($calender_items as $key => $row)
		{
			$start[$key] = $row['start'];
			$id[$key] = $row['id'];
		}

		// Sort the data on startdate and id
		// Add $data as the last parameter, to sort by the common key
		array_multisort($start, SORT_ASC, $id, SORT_ASC, $calender_items);
		*/
	}

	return $calender_items;
}
function ParseTemplateRecordData($template, $table, $recordid, $recordid_to_preload = 0)
{
	$selectallfields = false;
	$bypass_record_access_by_sql = false;

	$template = parse_old_template_to_new_template($template, $table);

	$template_tags = array();
	foreach (get_tags($template) as $tag)
	{
		if (in_array($tag, $GLOBALS['UC']['PersonalDataStructure'][$table]["allfields"]))
		{
			$template_tags[] = $tag;
		}
	}

	if (count($template_tags) > 0)
	{
		$template_tags_to_select = $template_tags;

		foreach ($template_tags_to_select as $tag)
		{
			//we willen altijd ook het primary id van iedere tabel selecteren omdat we die nodig hebben in return_ifield() slordige code dit
			//we hebben deze ook nodig voor het colorizen van de ref_value -> dit is lomp
			//ONDERSTAANDE CODE STAAT 2X IN PTRD()
			$table_parts = explode("-", $tag);
			//$last_field = $table_parts[count($table_parts)-1];
			$last_table = $table_parts[count($table_parts)-2];
			$table_parts[count($table_parts)-1] = $GLOBALS['UC']['PersonalDataStructure'][$last_table]["primary"];
			$newtag = implode("-", $table_parts);
			if (!in_array($newtag, $template_tags_to_select))
			{
				$template_tags_to_select[] = $newtag;
			}
		}

		$record = array();
		if (is_positive_integer($recordid))
		{
			$record = getrecord($table, $recordid, $selectallfields, $bypass_record_access_by_sql, $template_tags_to_select);
		}
		//_new_
		else
		{
			//we dupliceren een record

			//LET OP: het parsen van ListOptionsForParentTable is heel exotisch, we callen ParseTemplateRecordData() zonder recordid maar MET recordid_to_preload. Het gevolg is dat de huidige database waarde worden ingeladen van de recordid_to_preload (== het record) en dat die worden aangevuld/overschreven met waarden uit de GET. Hierdoor kan een ListOptionsForParentTable zowel data uit de db gebruiken als ook een nieuw gekozen waarde op het formulier wat je aan het invullen bent (bijvoorbeeld afhankelijk van de keuze van contactpersonen die werken bij het bedrijf wat tevens geselecteerd staat) N.B.: dit zou beter omgezet kunnen worden tot een selectvia constructie

			if (is_positive_integer($recordid_to_preload))
			{
				$record = getrecord($table, $recordid_to_preload, $selectallfields, $bypass_record_access_by_sql, $template_tags_to_select, true); //do_not_check_write_access=true
				//het gedupliceerde record moet writable zijn; ook wanneer het te dupliceren record readonly is
				$record['type_of_access'] = "ok";
				$record['record_locked_by'] = 0;
			}

			//prefill values vanuit de url
			foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]) as $field)
			{
				//wordt uit de request gehaald; er worden alleen vars uit de request gehaald voor de EIGEN tabel waar we nu inzitten; het heeft geen functie om voor bovenliggende tabellen waardes mee te geven
				//LET OP: we doen dit alleen op NEW forms

				$val = "";
				//record.php uses FQN names, check these first
				if (isset($_REQUEST[$table . '-' . $field]))
				{
					$val = $_REQUEST[$table . '-' . $field];
				}
				//FQN tag + recordid (wat in dit geval _new_ is)
				else if (isset($_REQUEST[$table . '-' . $field . '-_new_']))
				{
					$val = $_REQUEST[$table . '-' . $field . '-_new_'];
				}
				//dit is niet mooi; wordt gebruikt voor addrow om waardes door te geven; we zouden parsetemplaterecorddata uit moeten bereiden met een extra argument voor prefills; dan ook bovenstaande REQUEST checks eruit
				//VERY DIRTY: we save this prefill in a global. ParseTemplateRecordData needs this in order to pre-parse ListOptionsForParentTable. We should pass this as a function argument.
				else if (isset($GLOBALS['prefill_fields'][$table . '-' . $field]))
				{
					$val = $GLOBALS['prefill_fields'][$table . '-' . $field];
				}

				if ($val != "")
				{
					$fullpath = $table . "-" . $field;
					$relation_to = $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["relation_to"];

					$ok = true;

					if ($relation_to != "")
					{
						//dit is input validatie... onlogisch om op deze plek te doen, toch?
						if (is_positive_integer($val))
						{
							//dit moeten we doen voor de return_ifield functie
							$record[$fullpath . "-" . $relation_to . "-" . $GLOBALS['UC']['PersonalDataStructure'][$relation_to]["primary"]] = $val;

							//dit moeten we doen voor full qualified tags op dit form, als je bv het mail adres van de parent wilt afbeelden
							$template_tags_parent = array();
							foreach ($template_tags as $tag)
							{
								if (preg_match("/^" . preg_quote($fullpath) . "\-(.*)$/", $tag, $match))
								{
									//we hebben de prefix eraf geknipt
									$tag_in_parent = $match[1];
									$template_tags_parent[] = $tag_in_parent;

									//we willen altijd ook het primary id van iedere tabel selecteren omdat we die nodig hebben in return_ifield() slordige code dit
									//we hebben deze ook nodig voor het colorizen van de ref_value -> dit is lomp
									//ONDERSTAANDE CODE STAAT 2X IN PTRD()
									$table_parts_in_parent = explode("-", $tag_in_parent);
									//$last_field_in_parent = $table_parts_in_parent[count($table_parts_in_parent)-1];
									$last_table_in_parent = $table_parts_in_parent[count($table_parts_in_parent)-2];
									$table_parts_in_parent[count($table_parts_in_parent)-1] = $GLOBALS['UC']['PersonalDataStructure'][$last_table_in_parent]["primary"];
									$newtag_in_parent = implode("-", $table_parts_in_parent);
									if (!in_array($newtag_in_parent, $template_tags_parent))
									{
										$template_tags_parent[] = $newtag_in_parent;
									}
								}
							}
							if (count($template_tags_parent) > 0)
							{
								$parent = getrecord($relation_to, $val, $selectallfields, $bypass_record_access_by_sql, $template_tags_parent, true); //do_not_check_write_access=true
								foreach ($template_tags_parent as $parent_tag)
								{
									//we plakken de prefix er weer voor
									$record[$fullpath . "-" . $parent_tag] = $parent[$parent_tag];
								}
							}
						}
						else
						{
							$ok = false;
						}
					}

					if ($ok)
					{
						$record[$fullpath] = $val;
					}
				}
			}

			//backwards compat; of in het geval dat het primary id is overschreven uit de request, dan wel uit de recordid_to_preload
			$recordid = "_new_";
			$record[$table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"]] = "_new_";
		}

		foreach ($template_tags as $tag)
		{
			$table_parts = explode("-", $tag);
			$last_field = $table_parts[count($table_parts)-1];
			$last_table = $table_parts[count($table_parts)-2];

			if (strpos($template, "@" . $tag . "@") !== false)
			{
				$value = formatvalue($record[$tag], $tag);
				$template = str_replace("@" . $tag . "@", escape_interleave_characters($value), $template);
			}

			if (strpos($template, "%" . $tag . "%") !== false)
			{
				$template = str_replace("%" . $tag . "%", escape_interleave_characters(htme($record[$tag])), $template);
			}

			if (strpos($template, "^" . $tag . "^") !== false)
			{
				//deze ^ tags zijn vooral in gebruik voor mailto links, rawurlencode maakt geen + van spatie, maar %20 (Thunderbird accepteert geen +, en andere mail clients misschien ook niet)
				$template = str_replace("^" . $tag . "^", escape_interleave_characters(rawurlencode($record[$tag])), $template);
			}

			if (strpos($template, "#" . $tag . "#") !== false)
			{
				$value = "";
				//het is een IField; een veld wat aanpasbaar is in een hoger gelegen tabel; we weten niet of dit veld writable is want make_query zegt alleen iets over de writability van het EIGEN record; return_ifield mag dit dus zelf uitzoeken
				if (count($table_parts) > 2)
				{
					//omdat we de ref niet hebben selecteren we de primary id uit de eigen tabel; DIE hebben we wel vanwege de ifield add code hierboven. dit is lomp
					$ref_value = $record[implode("-", array_slice($table_parts,0,count($table_parts)-2)) . "-" . $last_table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$last_table]["primary"]];
					$value = return_ifield($tag, $ref_value, $record[$tag]);
				}
				//het is een veld uit de eigen tabel
				else
				{
					//als je een veld helemaal niet mag zien vanwege een conditie, dan mag je hem dus ook niet write-n, en krijg je dus een lege box terug (readonly)
					$readonly = true;
					//2015-09-18 conditionele write access tot velden komt nu ook uit de database
					if (is_positive_integer($recordid))
					{
						if (($record['type_of_access'] == "ok") && ($record['record_locked_by'] == 0) && ($record['*' . $tag] == 1))
						{
							$readonly = false;
						}
					}
					//op een nieuw record zijn alle velden writable, behalve diegene die dat nooit zijn. er wordt niet gekeken naar conditionele write access op een nieuw record
					else
					{
						if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["allowed_to_write"])
						{
							$readonly = false;
						}
					}

					$recordid = $record[$table . "-" . $GLOBALS['UC']['PersonalDataStructure'][$last_table]["primary"]];

					//als dit veld readonly wordt afgebeeld dan printen we hem zelf
					if (($readonly) && ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype_in_extrafields_table"] != "comment"))
					{
						//als het een button is dan beelden we niks af
						if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype_in_extrafields_table"] != "Button")
						{
							//data-unparsedvalue is added when a editable form element is requested (#t1-f2#), but we decide this element must be readonly (for this user), so we display undeditable text instead of the form element. data-unparsedvalue is used in the jquery Visual only record access rules (extendedprivsgroup) => update: data-unparsedvalue gaat niet werken.

							$value = '<div id="ffc-' . $last_table . '-' . $last_field . '" class="formfieldcontainer';
							if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["multiline"])
							{
								$value .= ' formfieldcontainermultiline';
							}
							else if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype_in_extrafields_table"] == "diary")
							{
								$value .= ' formfieldcontainerdiary';
							}
							$value .= '"><div class="formfieldlabel">' . htme($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["name"]) . '</div><div class="formfield">';

							/* HF20150729 als het een calc is, zet er een span omheen zodat we 'm kunnen reloaden. */
							if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["calculation"] != "")
							{
								$value .= '<div id="--' . $last_table . '-' . $last_field . '--' . $recordid . '-UPV" style="display: none;">' . htme($record[$tag]) . '</div>';
								$value .= '<span class="calculation" id="--' . $last_table . '-' . $last_field . '--' . $recordid . '">';
							}

							//multiselect relations are formatted with formatvalue(), single select relations we format here. This is illogical.
							if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["relation_to"] != "") && (!$GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["multiselect"]))
							{
								if ($record[$tag] > 0)
								{
									//deze zelfde code staat ook in record-functions
									//we gebruiken get_refer_layout_for_record() dus dat is een extra call naar getrecord(), het zou eigenlijk mooier zijn om in 1x alle benodigde velden met getrecord hierboven te selecteren. punt is dat we hierboven nog niet weten of het record readonly is, en dat parse_old_template_to_new_template() dus niet de component parts van een relation veld retourneert. dat zouden we aan kunnen passen, maar voor nu werkt dit prima

									$close_link = false;
									//we maken alleen links naar de bovenliggende tabel wanneer "display as text" aan staat
									if ((preg_match("/^f(\d+)$/", $last_field, $match)) && ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["formelementtype"] == "list"))
									{
										$close_link = true;

										//geen this.href gebruiken want die maakt er een URL van en dat gaat mis bij beautification (https://admin.interleave.nl/admin/https://admin.interleave.nl/admin/Rules/0?&duplicate=2)
										$value .= '<a href="' . $GLOBALS['UC']['PersonalDataStructure'][$GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["relation_to"]]["openlink"] . $record[$tag];

										//TODO: door ilor() laten doen?
										if (!strstr($GLOBALS['UC']['PersonalDataStructure'][$GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["relation_to"]]["openlink"], "?"))
										{
											$value .= '?';
										}
										else
										{
											$value .= '&amp;';
										}
										$starttable = $_GET['starttable'];
										$startrecordid = $_GET['startrecordid'];
										if (($starttable == "") || (!is_positive_integer($startrecordid)))
										{
											$starttable = $table;
											$startrecordid = $recordid;
										}
										$value .= 'starttable=' . $starttable . '&amp;startrecordid=' . $startrecordid;
										$value .= '" onclick="ilor($(this).attr(\'href\'),' . $GLOBALS['UC']['PersonalDataStructure'][$table]["open_records_in_popup_mainlist"] . ',\'\',\'\',\'\'); return false;" class="toparent' . $last_field . '">';
									}
									$value .= get_colorspan_and_htme($last_table, $last_field, get_refer_layout_for_record_field($last_table, $last_field, $record[$tag]), $record[$tag]);
									if ($close_link)
									{
										$value .= '</a>';
									}
								}
							}
							else
							{
								/* HF20150729 als het een calc is, zet er een span omheen zodat we 'm kunnen reloaden. */
/*								if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["calculation"] != "")
								{
									$value .= '<div id="--' . $last_table . '-' . $last_field . '--' . $recordid . '-UPV" style="display: none;">' . htme($record[$tag]) . '</div>';
									$value .= '<span class="calculation" id="--' . $last_table . '-' . $last_field . '--' . $recordid . '">';
								}*/
								//formatvalue maakt geen referenties op, alleen enkele velden, vandaar de uitzondering hierboven
								$value .= formatvalue($record[$tag], $tag);

								if ($record[$tag] != "")
								{
									//code gejat uit form-functions; de iconen moeten er namelijk ook staan op een readonly #mail/link/tel# veld
									if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]['fieldsubtype'] == "mail")
									{
										$value .= '&nbsp;<a href="mailto:' . htme($record[$tag]) . '"><span class="fa fa-envelope"></span></a>';
									}
									else if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]['fieldsubtype'] == "link")
									{
										$value .= '&nbsp;<a href="' . htme($record[$tag]) . '" onclick="window.open(this.href); return false;"><span class="fa fa-external-link"></span></a>';
									}
									else if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]['fieldsubtype'] == "tel")
									{
										$value .= '&nbsp;<a href="tel:' . htme($record[$tag]) . '"><span class="fa fa-phone"></span></a>';
									}
								}
							}

							if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["calculation"] != "")
							{
								$value .= '</span>';
							}

							$value .= '</div></div>';
						}
					}
					else
					{
						//2016-02-16 het is een comment; als je read access hebt dan wordt de letterlijke waarde uit de db geselecteerd "CommentReadAccess", anders is die leeg. als hij leeg is dan heb je geen read access door conditionele access rules en dan beelden we dus niks af
						//2016-11-02 we doen dit alleen op een bestaand record; want alleen DAN kan er CommentReadAccess uit de db komne. op een nieuw record wordt dit form comment dus altijd weergegeven
						if (($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype_in_extrafields_table"] == "comment") && ($record[$tag] != "CommentReadAccess") && (is_positive_integer($recordid)))
						{
						}
						else
						{
							$classes = array();
							$name_as_placeholder = false;
							$interactive_field_list_element = false;
							$parentlistid = false;
							//dit is de nieuwe functie; die met fully qualified names werkt
							$value = '<div id="ffc-' . $tag . '"';
							if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype_in_extrafields_table"] != "comment")
							{
								$value .= ' class="formfieldcontainer';
								if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype_in_extrafields_table"] == "Button")
								{
									$value .= ' formfieldcontainerbutton';
								}
								else if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype_in_extrafields_table"] == "diary")
								{
									$value .= ' formfieldcontainerdiary';
								}
								else if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["multiline"])
								{
									$value .= ' formfieldcontainermultiline';
								}
								if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["required"])
								{
									$value .= ' requiredfield';
								}
								$value .= '"';
							}
							$value .= '>' . get_form_element($tag, $record[$tag], $recordid, $classes, $name_as_placeholder, $interactive_field_list_element, $parentlistid) . '</div>';
						}
					}
				}
				$template = str_replace("#" . $tag . "#", escape_interleave_characters($value), $template);
			}
		}

		//slordige uitzondering; er moet minimaal 1 template_tag in staan om hier te komen
		foreach ($record as $tag => $value)
		{
			if (preg_match("/^ShowThisField\-/", $tag))
			{
				$template = str_replace("@" . $tag . "@", escape_interleave_characters($value), $template);
			}
		}
	}

	return $template;
}
function return_diary($diary_array, $outputtype = "table")
{
	$val = '';
	$row = 1;
	foreach (array_reverse($diary_array) as $part)
	{
		$datetime = date("Y-m-d H:i:s", $part[0]);
		$username = GetUserName($part[1]);
		$message = $part[2];

		$datetime = from_iso_datetime($datetime);

		if (($outputtype == "table") || ($outputtype == "tile") || ($outputtype == "template"))
		{
			$val .= '<tr class="diaryrowcolor' . $row . '"><td>' . $datetime . '</td><td>' . htme($username) . '</td><td>' . nl2br(htme($message)) . '</td></tr>';
		}
		else
		{
			//tsv, excel en xml
			$val .= $datetime . "," . $username . "," . $message . "\n";
		}
		if ($row == 1)
		{
			$row = 2;
		}
		else
		{
			$row = 1;
		}
	}
	if (($val != '') && (($outputtype == "table") || ($outputtype == "tile") || ($outputtype == "template")))
	{
		$val = '<div class="horscrollouter"><div class="horscroll"><table class="diarytable">' . $val . '</table></div></div>';
	}
	return $val;
}
function formatvalue($value, $tag)
{
	$table_parts = explode("-", $tag);
	$last_field = $table_parts[count($table_parts)-1];
	$last_table = $table_parts[count($table_parts)-2];

	//deze functie is alleen voor @ tags wordt gebruikt op de template parser
	//wordt NIET gebruikt om waardes in de lijst te formatteren; daar is deze logica herhaald. we moeten dit consistent maken. code duplicatie in make_list() op dit moment dus! we zouden alle te printen waardes in list ook hierdoor heen kunnen halen, maar dat is wel trager
	//TODO: wordt ook niet gebruik voor refer fields; dat zou wel moeten

	if ($value != "")
	{
		if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "date")
		{
			$value = from_iso_datetime($value, $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["ShowSeconds"]);
		}
		else if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype"] == "number")
		{
			$value = FormatNumberNEW($value, $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["decimals"], $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["prefix"], $GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["postfix"]);
		}
		else if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["fieldtype_in_extrafields_table"] == "diary")
		{
			$outputtype = "table";
//jh 2016-05-27 om de diary kommagescheiden af te beelden in exports. was dit uberhaupt in gebruik?
//			$outputtype = "csv";
			$value = return_diary(unserialize($value), $outputtype);
		}
		else if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["multiselect"])
		{
			$newvalue = '';
			foreach (unserialize($value) as $val)
			{
				//multi dropdown en sql nog niet ondersteund
				if ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["relation_to"] != "")
				{
					//TODO DIT IS NIET EFFICIENT: getrecord wordt voor ieder multiselect waarde gerunt. dat kan dus enkele 100-en keren zijn wanneer pagination hoog is ingesteld. Als je al records opvraagt uit een grote lijst dan duurt het extreem lang. een alternatief is eerst een array opbouwen met alle mogelijke refers. dat deed ik eerst met multiselect_data_array, nadeel daarvan is dat DAT heel erg duur is op korte lijsten
					//FIX 2014-01-06 we cachen eenmaal geparsde records; dit levert vooral veel op, bij weinig keuzes die vaak zijn geselecteerd

/*
					if ($multiselect_data_array[$key][$val] != "")
					{
						$val = $multiselect_data_array[$key][$val];
					}
					else
					{*/
						$selectallfields = false;
						$bypass_record_access_by_sql = false;
						$selectfields = array();

						foreach ($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["layout_tags"] as $tag)
						{
							$selectfields[] = $tag;
						}

						$one_record = getrecord($GLOBALS['UC']['PersonalDataStructure'][$last_table]["fields"][$last_field]["relation_to"], $val, $selectallfields, $bypass_record_access_by_sql, $selectfields, true); //do_not_check_write_access=true

						$val = parse_refer_layout_for_field($last_table, $last_field, $one_record, false);
//					}
				}

				if ($newvalue != '')
				{
					$newvalue .= ', ';
				}
				$newvalue .= $val;
			}
			$value = $newvalue;
		}
	}

	//JH 2017-12-14 we zouden datums, nummers en diarys ook ongeformateerd deze functie in moeten doen
	$value = get_colorspan_and_htme($last_table, $last_field, $value);

	return $value;
}
function get_colorspan_and_htme($table, $field, $textvalue, $rawvalue = "")
{
	//merge this function with getfieldcolors

	//rawvalue will be set when it is a refer field
	if ($rawvalue == "")
	{
		$rawvalue = $textvalue;
	}

	//input value is plain text; output is htme encoded
	$output = '';

	$close_span = false;

	$optionclass = $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["optionclass"][$rawvalue];
	if ($optionclass != "")
	{
		$output .= '<span class="colorspan ' . $optionclass . '">';
		$close_span = true;
	}
	else
	{
		list($bgcolor, $fgcolor) = getfieldcolors($table, $field, $rawvalue);
		if ($bgcolor != "")
		{
			$output .= '<span style="background-color:' . $bgcolor . ';';
			if ($fgcolor != "")
			{
				$output .= 'color:' . $fgcolor . ';';
			}
			$output .= '" class="colorspan">';
			$close_span = true;
		}
	}

	if ((!$GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["ishtml"]) && ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["fieldtype_in_extrafields_table"] != "diary"))
	{
		$textvalue = htme($textvalue);
		if ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["multiline"])
		{
			$textvalue = nl2br($textvalue);
		}
	}
	$output .= $textvalue;

	if ($close_span)
	{
		$output .= '</span>';
	}

	return $output;
}
function getpopupsize($table)
{
	//default Large
	$popupsize_num = 3;
	if (preg_match("/^t(\d+)$/", $table, $match))
	{
		$popupsize_txt = GetAttribute("flextable", "PopUpSize", $match[1]);
		if ($popupsize_txt == "Small")
		{
			$popupsize_num = 1;
		}
		else if ($popupsize_txt == "Medium")
		{
			$popupsize_num = 2;
		}
/*		else if ($popupsize_txt == "Large")
		{
			$popupsize_num = 3;
		}*/
		else if ($popupsize_txt == "Extra large")
		{
			$popupsize_num = 4;
		}
	}
	return $popupsize_num;
}
function get_ds_tablename_for_field($efid)
{
	if ($GLOBALS["ds_tablename_for_field"][$efid] == "")
	{
		foreach (db_GetArray("SELECT `id`, `tabletype` FROM `CRMextrafields` WHERE `deleted`='n'") as $row)
		{
			$table = "";
			if (is_positive_integer($row['tabletype']))
			{
				$table = "t" . $row['tabletype'];
			}
			else if ($row['tabletype'] == "user")
			{
				$table = "users";
			}
			else if ($row['tabletype'] == "group")
			{
				$table = "groups";
			}
			else if ($row['tabletype'] == "files")
			{
				$table = "files";
			}
			$GLOBALS["ds_tablename_for_field"][$row['id']] = $table;
		}
	}

	return $GLOBALS["ds_tablename_for_field"][$efid];
}
function whitespace_to_underscore($str)
{
	return preg_replace("/\s/", "_", $str);
}
function FormatNumberNEW($number, $decimals = 0, $prefix = "", $postfix = "")
{
	$number_str = "";
	//!== is nodig omdat we "0" wel willen formatteren en "" niet
	if ($number !== "")
	{
		$decimal_separator = ",";
		$thousands_separator = ".";

		if ($GLOBALS['UC']['DateFormat'] != "dd-mm-yyyy")
		{
			$decimal_separator = ".";
			$thousands_separator = ",";
		}

		$number_str = $prefix;
		if ($number_str != "")
		{
	//FormatNumberNEW wordt alleen gerunt voor html pagina's
			$number_str .= " ";
	//		$number_str .= "&nbsp;";
		}

		$number_str .= number_format($number, $decimals, $decimal_separator, $thousands_separator);

		if ($postfix != "")
		{
	//FormatNumberNEW wordt alleen gerunt voor html pagina's
			$number_str .= " " . $postfix;
	//		$number_str .= "&nbsp;";
		}
	}

	return $number_str;
}
//deze functie kan mogelijk gaan dienen ipv de caching van de allfields wat veel data vraagt
/*
function get_allfields($table, $table_alias = "", $already_joined = array(), $get_parents = true, $get_childs = true)
{
	$allfields = array();
	$table_alias .= $table;
	foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]) as $field)
	{
		$allfields[] = $table_alias . "-" . $field;
	}
	if ($get_parents)
	{
		foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["parents"]) as $parent_table)
		{
			foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["parents"][$parent_table] as $parent_field)
			{
				if ($already_joined[$table][$parent_field] == 0)
				{
					if (preg_match("/^t\d+$/", $parent_table))
					{
						$already_joined2 = $already_joined;
						$already_joined2[$table][$parent_table]++;
						$allfields = array_merge($allfields, get_allfields($parent_table, $table_alias . "-" . $parent_field . "-", $already_joined2, true, false));
					}
				}
			}
		}
	}
	if ($get_childs)
	{
		foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["childs"]) as $child_table)
		{
			foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["childs"][$child_table] as $child_field)
			{
				if ($already_joined[$child_table][$child_field] == 0)
				{
					if (preg_match("/^t\d+$/", $child_table))
					{
						$already_joined2 = $already_joined;
						$already_joined2[$child_table][$child_field]++;
						$allfields = array_merge($allfields, get_allfields($child_table, $table_alias . "|CHILD:" . $child_table . "-" . $child_field . "|", $already_joined2, true, true));
					}
				}
			}
		}
	}
	return $allfields;
}
*/
function ParseAutoForms($template, $table)
{
	/*
		Types of tags

		@AUTO@                      Displays all fields and the file upload box, and the header
		@AUTOHEADER@                Displays the autoheader
		@AUTOFIELDS@                Displays all fields
		@AUTO_WITH_CHILD_TABLES@    Header, fields, filebox and all child tables
		@AUTOCHILDTABLESINNAV@      Header, fields, child tables and files in nav



	*/

	if ($GLOBALS['UC']['PersonalDataStructure'][$table]["HasFileAttachments"]) {
		$template = str_replace("@AUTO@", "@AUTOHEADER@@AUTOFIELDS@", $template);
	} else {
		$template = str_replace("@AUTO@", "@AUTOHEADER@@AUTOFIELDS@", $template);
	}
	$template = str_replace("@AUTO_WITH_CHILD_TABLES@", "@AUTOHEADER@@AUTOFIELDS@@AUTOCHILDTABLES@", $template);

	$show_childtablenav = false;

	if (strstr($template, "@AUTOCHILDTABLESINNAV@")) {
		if (strstr($template, '@AUTOHEADER@')) {
			$template = str_replace("@AUTOHEADER@", "@AUTOHEADER@@AUTOCHILDTABLENAV@", $template);
			$show_childtablenav = true;
		} else {
			log_msg('ERROR: Autoform for table ' . $table . ' contains AUTOCHILDTABLESINNAV tag but no AUTOHEADER which is mandatory. Form will be incomplete.');
		}
	}
	//print $template;
	$template = str_replace("@AUTOHEADER@", ReturnDefaultFormForTable($table, false, true, false, false), $template);

	$template = str_replace("@AUTOCHILDTABLES@", ReturnDefaultFormForTable($table, true, false, false, false), $template);

	if (strstr($template, "@AUTOCHILDTABLESINNAV@")) {

		$template = str_replace("@AUTOCHILDTABLESINNAV@", ReturnDefaultFormForTable($table, true, false, true, true, false, true), $template);
	} else {
		$template = str_replace("@AUTOFIELDS@", ReturnDefaultFormForTable($table, false, false, true, false), $template);
	}

	$template = str_replace("@AUTOCHILDTABLENAV@", ReturnDefaultFormForTable($table, false, false, false, false, true), $template);
	return $template;
}
function ReturnDefaultFormForTable($table, $show_childtables = false, $show_header = true, $show_fields = true, $childtablesindiv = false, $show_childtablenav = false, $skip_files = false)
{
	global $lang;

	//@AUTO@ = @AUTOHEADER@ @AUTOFIELDS@
	//@AUTO_WITH_CHILD_TABLES@ = @AUTOHEADER@ @AUTOFIELDS@ @AUTOCHILDTABLES@

	//@AUTOHEADER@
	//@AUTOFIELDS@
	//@AUTOCHILDTABLES@
	$html = '';

	if (is_array($GLOBALS['UC']['PersonalDataStructure'][$table]))
	{
		if ($show_header)
		{
			//admin table
			if (!$GLOBALS['UC']['PersonalDataStructure'][$table]["datatable"])
			{
				$html .= '@FAMILYTREE@';
			}

			$html .= '<div class="autoformheader autoformheader-' . $table . '">';
			$html .= '<div>';
			$html .= '@XONLY@#LOCKICON# #JOURNALICON# ';
			if ($GLOBALS['UC']['PersonalDataStructure'][$table]["datatable"])
			{
				$html .= ' #ARROWS#';
			}
			$html .= '@ENDXONLY@';
			$html .= '<h1 class="showinlineblock formtitle">';
			$html .= '@XONLY@' . htme($GLOBALS['UC']['PersonalDataStructure'][$table]["singular"]) . ' ' . $GLOBALS['UC']['PersonalDataStructure'][$table]["layout"] . '@ENDXONLY@';
			$html .= '@NEWONLY@' . htme($GLOBALS['UC']['PersonalDataStructure'][$table]["addlinktext"]) . '@ENDNEWONLY@';
			$html .= '</h1>';
			$html .= '</div>';
			$html .= '@LOCKEDONLY@<div class="warnmsg">[[recordislocked1]] @LOCKBY@[[recordislocked2]]</div>@ENDLOCKEDONLY@';
			$html .= '</div>';
		}

		if ($show_fields)
		{
			if ($childtablesindiv) {
				$html .= '<div class="iltab" id="subnav-main_content">';
			}

			$html .= '<div class="autoformfields autoformfields-' . $table . '">';
			$html .= '@XONLY@<div><div id="ffc-' . $table . '-' . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"] . '" class="formfieldcontainer"><div class="formfieldlabel">' . htme($lang['id']) . '</div><div class="formfield">@' . $table . '-' . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"] . '@</div></div></div>@ENDXONLY@';

			foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]) as $field)
			{
				//timestamp_last_change heeft geen hide_from_form
				if (($field != $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"]) && ($field != "timestamp_last_change"))
				{
					if (!$GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["hide_from_form"])
					{
						$xonly = false;

						if ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["calculation"] != "")
						{
							//the primary id (recordid) is used in a calculation; this id is not yet available so we hide this calculation on a new form
							if ((preg_match("/@table_alias@." . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"] . "\b/", $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["calculation"])) || (preg_match("/@table_alias@.`" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"] . "`/", $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["calculation"])))
							{
								$xonly = true;
							}
						}
						//on a new form we will not display fields that can't be written to
						else if (!$GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["allowed_to_write"])
						{
							$xonly = true;
						}

						if ($xonly)
						{
							$html .= '@XONLY@';
						}
						if ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["fieldtype_in_extrafields_table"] != "Button")
						{
							$html .= '<div class="container-' . $field . '">#' . $table . '-' . $field . '#</div>';
						}
						if ($xonly)
						{
							$html .= '@ENDXONLY@';
						}
					}
				}
			}
			$html .= '</div>';

			$html .= '<div class="autoformbuttons autoformbuttons-' . $table . '">#SAVEBUTTON#';
			foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]) as $field)
			{
				//timestamp_last_change heeft geen hide_from_form
				if (($field != $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"]) && ($field != "timestamp_last_change"))
				{
					if (!$GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["hide_from_form"])
					{
						if ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["fieldtype_in_extrafields_table"] == "Button")
						{
							$html .= ' &nbsp; #' . $table . '-' . $field . '#';
						}
					}
				}
			}
			$html .= ' &nbsp; #DELETEBUTTON#</div>';

			if ($GLOBALS['UC']['PersonalDataStructure'][$table]["HasFileAttachments"])
			{
				/* the file list is shown as a sublist if $show_childtables is enabled, but not on new forms */
				$html .= '@NEWONLY@';
				$html .= '<div class="autoformfiles autoformfiles-' . $table . '">#FILEBOX#</div>';
				$html .= '@ENDNEWONLY@';
			} elseif ($GLOBALS['UC']['PersonalDataStructure'][$table]["HasFileAttachments"] && !$skip_files && $show_childtablenav) {
				$html .= '<div class="autoformfiles autoformfiles-' . $table . '">#FILEBOX#</div>';
			}

			if ($childtablesindiv) {
				$html .= '</div>'; /* end subnav-main_content div */
			}


		}

		//dit wordt snel te zwaar en te langzaam voor hooggelegen tabellen
		if ($show_childtables)
		{
			$html .= '@XONLY@';

			$html .= '<div class="autoformchildtables autoformchildtables-' . $table . '">';
			foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["childs"]) as $child_table)
			{
				foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["childs"][$child_table] as $child_field)
				{
					//er wordt nu een autoheader opgemaakt voor tabellen zonder headerhtml en die zet hem er al in een h2 tag boven inclusief aantal records tussen haken
//LET OP LISTID moet terug want nu kan je dubbele id's op 1 page hebben als een tabel 2 dezelfde childs heeft; dit hefet alleen tot gevolg dat sortering op lists verloren gaat. dat is vervelend. we zouden pas een id aan de TWEEDE en volgende list moeten geven
//					$html .= '<div>#LIST|table=' . $child_table . '|listid=' . $child_table . $child_field . 'sublist|type_of_list=sublist|hiddenfilter_' . $child_table . '-' . $child_field . '=%' . $table . '-'. $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"] . '%';
					if ($childtablesindiv) {
						$html .= '<div class="hide_content iltab" id="subnav-' . $child_table . '-' . $child_field . '_content">';
					}
					$html .= '<div>#LIST|table=' . $child_table . '|type_of_list=sublist|hiddenfilter_' . $child_table . '-' . $child_field . '=%' . $table . '-'. $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"] . '%';
					if ($childtablesindiv) {
						$html .= '|hide_header=1';
					}
					if ($GLOBALS['UC']['PersonalDataStructure'][$child_table]["AddRowEnabledOnSublistsOnAutoForms"]) {
						$html .= '|hide_addrow=0|hide_addlink=1';
					} else {

						$show_addrow_on_tables = array("calendaraccessrules");
						if (in_array($child_table, $show_addrow_on_tables))
						{
							$html .= '|hide_addrow=0';
						}
					}
					$html .= '#</div>';
					if ($childtablesindiv) {
						$html .= '</div>';
					}

				}


			}
			if ($GLOBALS['UC']['PersonalDataStructure'][$table]["HasFileAttachments"] && !$childtablesinnav)  {
				if ($childtablesindiv) {
						$html .= '<div class="hide_content iltab" id="subnav-files_content">';
				}
				$html .= '<div>#FILEBOX#</div>';
				if ($childtablesindiv) {
					$html .= '</div>';
				}

			}
			$html .= '</div>';

		}
		$html .= '@ENDXONLY@';



		if ($show_childtablenav && (count($GLOBALS['UC']['PersonalDataStructure'][$table]["childs"]) > 0 || $GLOBALS['UC']['PersonalDataStructure'][$table]["HasFileAttachments"])) {


			/* create a nav element and javascript to show / hide these elements. */
			$html = '@XONLY@<nav class="submenu iltabs"><ul>';

			/* the "main" div holding the autoformfields hide/show */
			$html .= '<li><a id="subnav-main">' . htme($GLOBALS['UC']['PersonalDataStructure'][$table]["singular"]) . ' ' . $GLOBALS['UC']['PersonalDataStructure'][$table]["layout"] . '</a></li>';

			/* for each of the child tables */
			// dat kan waarschijnlijk netter met een href, '#' etc. de highlight in het menu doet het ook nog niet
			foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["childs"]) as $child_table) {
				foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["childs"][$child_table] as $child_field) {

					/* name, link */
					$html .= '<li><a id="subnav-' . $child_table . '-' . $child_field . '">';
					$html .= htme($GLOBALS['UC']['PersonalDataStructure'][$child_table]["plural"]);
					if (count($GLOBALS['UC']['PersonalDataStructure'][$table]["childs"][$child_table]) > 1)
					{
						$html .= ' - ' . htme($GLOBALS['UC']['PersonalDataStructure'][$child_table]["fields"][$child_field]["name"]);
					}

					/* the count */
					$html .= ' (#LIST|table=' . $child_table . '|type_of_list=sublist|hiddenfilter_' . $child_table . '-' . $child_field . '=%' . $table . '-'. $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"] . '%|countonly=1#)</a></li>';
				}
			}
			if ($GLOBALS['UC']['PersonalDataStructure'][$table]["HasFileAttachments"]) {
				$html .= '<li><a id="subnav-files">';
				$html .= htme($GLOBALS['UC']['PersonalDataStructure']['files']["plural"]);


				/* the count */
				$html .= ' (#LIST|table=files|type_of_list=sublist|hiddenfilter_files-koppelid=%' . $table . '-'. $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"] . '%|hiddenfilter_files-type=' . ds_tablename_to_binfiles_table($table) . '|hiddenfilter_files-version_belonging_to=0|countonly=1#)</a></li>';
			}
			$html .= '</ul></nav>@ENDXONLY@';


		}
	}
	else
	{
		$html .= "Error: table \"" . $table . "\" does not exist or no access";
		if ($GLOBALS['SesMem']['ShowReasons']) {
			$html .= GetBackTrace();
		}
	}

	return $html;
}
function ReturnDefaultPDFTemplateForTable($table, $record, $include_child_tables = false) {
	$pdf = '<h1>' . get_refer_layout_for_record($table, $record) . '</h1>';
	$pdf .= '<table>';
	foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"] AS $field => $props) {
			$pdf .= '<tr><th>';
			if ($field == 'recordid') {
				$pdf .= '[[id]]';
			} elseif ($field == 'timestamp_last_change') {
				$pdf .= '[[lastupdate]]';
			} else {
				$pdf .= '[[' . $field . ']]';
			}
			$pdf .= '</th><td>@' . $table . '-' . $field . '@</td></tr>';
	}
	$pdf .= '</table>';
	foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["childs"]) as $child_table) {
		foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["childs"][$child_table] as $child_field) {

			$pdf .= '<p>#LIST|table=' . $child_table . '|type_of_list=sublist|hiddenfilter_' . $child_table . '-' . $child_field . '=%' . $table . '-'. $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"] . '%|tabletype=plain#</p>';
		}
	}
	return ($pdf);
}

function highlight($str_html_encoded, $highlight)
{
	//hotfix voor slash; een slash in de string gaat mis met deze regexp. moet nog gefixt worden
	foreach (explode("/", $highlight) as $highlight_zonder_slash)
	{
		//we willen niet teksten in html tags vervangen, daarom:
		//de searchpart kan ook html bevatten < > " &, daarom moeten we die door htme() halen want str_html_encoded is dat ook al
		if ($highlight_zonder_slash != "")
		{
			$str_html_encoded = preg_replace("/(" . preg_quote(htme($highlight_zonder_slash)) . ")(?!([^<]+)?>)/i", '<span class="SearchHighlight">\\1</span>', $str_html_encoded);
		}
	}
	return $str_html_encoded;
}
function list_filebox($parent_table, $parent_recordid, $folder = 0, $version_belonging_to = 0)
{
//#LIST|table=files|type_of_list=sublist|hiddenfilter_files-type=flextable16|hiddenfilter_files-koppelid=@t2-recordid@|listlayoutcolumns=files-fileid,files-filename,files-timestamp_last_change,files-username,files-filetype,files-filesize#

	$output = '';
	if ($GLOBALS['UC']['PersonalDataStructure'][$parent_table]["HasFileAttachments"])
	{
		$search = "";
		//dit is smerig; staat ook in initiate.php
		if (GetSetting("HighlightSearchesInForms") == "Yes")
		{
			if ((is_positive_integer($_REQUEST['record'])) && ($_REQUEST['listid'] != ""))
			{
				$attribsql = "SELECT value FROM CRMattributes WHERE entity='" . mres($GLOBALS['USERID']) . "' AND identifier='user' AND attribute LIKE 'List-lastview-%-" . mres($_REQUEST['listid']) . "'";

				//onhandig in de lastview attrib zit ook de tabel naam opgeslagen. dat is niet nodig want listid moet al uniek zijn. om die reden deze lompe query
				//$lastview_attribute = "List-lastview-" . $table . "-" . $listid;
				//$lastview = GetAttribute("user", $lastview_attribute, $GLOBALS['USERID']);
				$lastview = unserialize(db_GetValue($attribsql));

				$search = $lastview['search'];
			}
		}

		$table = "files";

		$listid = $parent_table . $table;
		if (is_positive_integer($version_belonging_to))
		{
			$listid .= "revisions";
		}

		$listoptions = array();
		$listoptions["listid"] = $listid;
		$listoptions["table"] = $table;
		$listoptions["type_of_list"] = "sublist";
		$listoptions["hiddenfilter"] = array();
		$listoptions["hiddenfilter"]["files-type"][] = $GLOBALS['UC']['PersonalDataStructure'][$parent_table]["name_in_database"];
		$listoptions["hiddenfilter"]["files-koppelid"][] = $parent_recordid;

		if (is_array($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]["version_no"]))
		{
			//toegevoegd voor revert list 2014-11-28
			$listoptions["hiddenfilter"]["files-version_belonging_to"][] = $version_belonging_to;
		}
		if (is_array($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]["folder"]))
		{
			$listoptions["hiddenfilter"]["files-folder"][] = $folder;
		}

		//(extra) fields op files niveau altijd toevoegen
		$listlayoutcolumns = array();

		if (is_positive_integer($version_belonging_to))
		{
			//laatste boveaan van de Revert list
			$listoptions["sortby"] = "files-fileid";
			$listoptions["sortdirection"] = "desc";
		}

		$listlayoutcolumns[] = "files-fileid";
		//meestal gelijk voor revert list
		$listlayoutcolumns[] = "files-filename";
		$listlayoutcolumns[] = "files-filesize";
		$listlayoutcolumns[] = "files-timestamp_last_change";
		$listlayoutcolumns[] = "files-username";
		foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]) as $field)
		{
			//ALLE fields worden weergegeven. we zouden hier in de toekomst nog onderscheid per tabel in willen kunnen maken
			if (preg_match("/^f\d+$/", $field))
			{
				$listlayoutcolumns[] = $table . "-" . $field;
			}
		}

		$listoptions["listlayoutcolumns"] = implode(",", $listlayoutcolumns);

		$listoptions["search"] = $search;
		$listoptions["search_highlight_only"] = true;
		$listoptions["start"] = 0;

		$last_element = count($GLOBALS['UC']['PersonalDataStructure'][$table]["perpage_options"]) - 1;
		$listoptions["perpage"] = $GLOBALS['UC']['PersonalDataStructure'][$table]["perpage_options"][$last_element];

		$listoptions["hide_recordcounter"] = true;
		//2015-05-20 zou misschien op false kunnen voor autoheader (zodat aantal files ook worden weergegeven), maar dit heeft impact op alle installaties dus niet doen
		$listoptions["hide_header"] = true;
		$listoptions["hide_pagination_if_only_1_page"] = true;
		$listoptions["hide_perpage"] = true;
		$listoptions["hide_selections"] = true;
		$listoptions["hide_id_input"] = true;
		$listoptions["hide_listlayout"] = true;
		$listoptions["hide_outputtype_excel"] = true;
		$listoptions["hide_add_up_total"] = true;
		$listoptions["hide_filters"] = true;
		//hier staat toch alleen maar de search in
		$listoptions["hide_human_readable_filter"] = true;

		/* HF20171030 uitgezet; anders kan geen folders en documenten aanmaken in een lege lijst */
//		$listoptions["hide_tablewhennoresults"] = true;

		list($number_of_records, $output) = make_list($listoptions);
	}
	else
	{
		log_msg("ERROR: file list tag found on template but HasFileAttachements is No");
	}

	return $output;
}
function disable_mysql_triggers()
{
	//geldt alleen voor de huidige sessie
	//triggers hoeven niet weer aangezet te worden; de volgende connect zijn ze dat automatisch
	//we doen dit af en toe in modules dus deze code moet aan de "voorkant" zitten
	mcq("SET @interleave_disable_trigger = TRUE");
}
function enable_mysql_triggers()
{
	mcq("SET @interleave_disable_trigger = NULL");
}
/*
//kan gebruikt worden in listlayout.php maar staat gehasht
function ascend_to_parents($table, $prefix = "")
{
	$str = '';
	if ($prefix != "")
	{
		$prefix .= "-";
	}
	$prefix .= $table . "-";
	foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]) as $field)
	{
		if (!$GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["hide_from_list"])
		{
			$myself = $prefix . $field;
			$parent = '';
			if ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["relation_to"] != "")
			{
				$parent .= ascend_to_parents($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["relation_to"], $myself);
			}
			$str .= '<div>';
			if ($parent != "")
			{
				$str .= '<a onclick="if (document.getElementById(\'parent-' . $myself . '\').style.display == \'none\') { document.getElementById(\'parent-' . $myself . '\').style.display = \'block\'; this.innerHTML = \'&boxminus;\'; } else { document.getElementById(\'parent-' . $myself . '\').style.display = \'none\'; this.innerHTML = \'&boxplus;\';}">&boxplus;</a> ';
			}
			else
			{
				$str .= '&boxtimes;';
			}
			$str .= ' <a onclick="ListLayoutAddColumn(\'' . $myself . '\',this.innerHTML);">' . htme($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["name"]);
			if (($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["displaylistname"] != "") && ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["displaylistname"] != $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["name"]))
			{
				$str .= ' (' . htme($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["displaylistname"]) . ')';
			}
			$str .= '</a>';
			if ($parent != "")
			{
				$str .= '<div id="parent-' . $myself . '" class="hide_content">' . $parent . '</div>';
			}
			$str .= '</div>';
		}
	}

	return $str;
}
*/
function escape_interleave_characters($string)
{
	//escapet alle Interleave control karakters met een tilde; @ -> ~2, om dubbele parsing te voorkomen, let op volgorde is belangrijk de tilde zelf moet als EERSTE worden geescapet. str_replace() is 7x zo snel als preg_replace()
	//vraagtekens worden gecodeerd om < ? php open tags in html velden (die niet door htme() worden gehaald) te voorkomen dat ze worden uitgevoerd, we zouden ook < en > kunnen coderen maar dat zijn er veel meer (meer cpu tijd dus)
	return str_replace(array("~", "@", "#", "%", "[", "]", "^", "?"), array("~1", "~2", "~3", "~4", "~5", "~6", "~7", "~8"), $string);
}
function unescape_interleave_characters($string)
{
	//UNescapet alle Interleave control karakters ~2 -> @, om dubbele parsing te voorkomen, let op volgorde is belangrijk de tilde zelf moet als LAATSTE worden geUNescapet. str_replace() is 7x zo snel als preg_replace()
	//vraagtekens worden gecodeerd om < ? php open tags in html velden (die niet door htme() worden gehaald) te voorkomen dat ze worden uitgevoerd, we zouden ook < en > kunnen coderen maar dat zijn er veel meer (meer cpu tijd dus)
	return str_replace(array("~2", "~3", "~4", "~5", "~6", "~7", "~8", "~1"), array("@", "#", "%", "[", "]", "^", "?", "~"), $string);
}
function return_valid_utf8($str, $input_charset)
{
//iconv should do this as well but doesn't remove invalid chars; see comments "The "//ignore" option doesn't work with recent versions of the iconv library. So if you're having trouble with that option, you aren't alone." http://php.net/manual/en/function.iconv.php
//	return iconv($input_charset, "UTF-8//IGNORE", $str);
	return mb_convert_encoding($str, "UTF-8", $input_charset);
}
//deze functie wordt gebruikt om de refer layout te parsen zoals een referentie naar die tabel de tabel gepresenteerd wil zien
//remove_first_part_from_tag == false voor multiselect velden wat een rare uitzondering is, zo niet dan is die true
function parse_refer_layout_for_field($table, $field, $data, $remove_first_part_from_tag)
{
	$layout = $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["layout"];
	foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["layout_tags"] as $tag)
	{
		$replacetag = $tag;
		if ($remove_first_part_from_tag)
		{
			//omdat we 1 tabel hoger selecteren moeten we de huidige tabel en field name van de tags afhalen want die gaat er vanuit dat er vanaf 1 niveau LAGER wordt geparsed. DEZE CODE STAAT MEERDERE KEREN IN DEZE FILE
			$replacetag = preg_replace("/^" . $table . "\-" . $field . "\-/", "", $replacetag);
		}
		//in een refer layout kunnen alleen @ tags zitten
		$layout = str_replace("@" . $tag . "@", $data[$replacetag], $layout);
	}
	return $layout;
}
//deze functie wordt gebruikt om de refer layout te parsen zoals de tabel zichzelf wil presenteren
function parse_refer_layout_for_table($table, $data)
{
	$layout = $GLOBALS['UC']['PersonalDataStructure'][$table]["layout"];
	foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["layout_tags"] as $tag)
	{
		//in een refer layout kunnen alleen @ tags zitten
		$layout = str_replace("@" . $tag . "@", $data[$tag], $layout);
	}
	return $layout;
}
//heeft een table en een recordid als input, retourneert de geparsde refer layout; gebruikt parse_refer_layout_for_table() voor de daadwerkelijke parsing die hij voedt met een getrecord()
function get_refer_layout_for_record($table, $id)
{
	$layout = "";
	//als de ref niet gevuld is (id == 0) dan is de layout ""
	if (is_positive_integer($id))
	{
		$layout = parse_refer_layout_for_table($table, getrecord($table, $id, false, false, $GLOBALS['UC']['PersonalDataStructure'][$table]["layout_tags"], true)); //do_not_check_write_access=true
	}
	return $layout;
}
//wordt aangeroepen met de table en field in de child table en het id van het record in de parent waarna hij verwijst
function get_refer_layout_for_record_field($table, $field, $id_in_parent_table)
{
	$layout = "";
	//als de ref naar de parent niet gevuld is (id_in_parent_table == 0) dan is de layout ""
	if (is_positive_integer($id_in_parent_table))
	{
		$selectfields = array();
		foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["layout_tags"] as $tag)
		{
			//omdat we 1 tabel hoger selecteren moeten we de huidige tabel en field name van de tags afhalen want die gaat er vanuit dat er vanaf 1 niveau LAGER wordt geparsed. DEZE CODE STAAT MEERDERE KEREN IN DEZE FILE
			$selectfields[] = preg_replace("/^" . $table . "\-" . $field . "\-/", "", $tag);
		}

		$layout = parse_refer_layout_for_field($table, $field, getrecord($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["relation_to"], $id_in_parent_table, false, false, $selectfields, true), true); //do_not_check_write_access=true
	}
	return $layout;
}
function backtolist($fromlist)
{
	$str = '';
	if ($fromlist != "")
	{
		$fromlist = base64_decode($fromlist);
		while (preg_match("/fromlist=([a-zA-Z0-9\+\/\=]+)/", $fromlist, $match))
		{
			$fromlist = $match[1];
			$fromlist = base64_decode($fromlist);
		}
		//<span class="fa fa-list-alt"></span>
		$str = '<div class="backtolist"><a href="' . htme($fromlist) . '"><span class="fa fa-chevron-left"></span></a></div>';
	}
	return $str;
}

function familytree($table, $recordid, $starttable = "", $startrecordid = 0, $fromlist = "", $custom_current_url = "", $custom_current_text = "")
{
	$str = '';

	$arr = array();
	$x = 0;
	$y = 0;

	//in geval van _new_
	if (!is_positive_integer($recordid))
	{
		$recordid = 0;
		//call vanuit modules kan zonder table en recordid
		if ($table != "")
		{
			//geen href zodat er nooit weggenavigeerd wordt
			$arr[$x][$y] = '<a class="active">' . htme($GLOBALS['UC']['PersonalDataStructure'][$table]["addlinktext"]) . '</a>';
			$x++;
		}
	}

	if (($starttable == "") || (!is_positive_integer($startrecordid)))
	{
		$starttable = $table;
		$startrecordid = $recordid;
	}

	if ($fromlist != "")
	{
		$str .= backtolist($fromlist);
	}

	if (($custom_current_url != "") && ($custom_current_text != ""))
	{
		//jh 2017-05-02 this is not a link anymore
		//$link = '<a href="' . htme($custom_current_url) . '" class="active">' . htme($custom_current_text) . '</a>';
		$link = '<a class="active">' . htme($custom_current_text) . '</a>';
		$arr[$x][$y] = $link;
		$x++;
	}

	//to prevent Customer / Add customer, but still allow Customer / Invoices / Add invoice
	if ((is_positive_integer($recordid)) || ($starttable != $table))
	{
		list($maxx, $maxy, $arr) = familytree_recursive($starttable, $startrecordid, $starttable, $startrecordid, $table, $recordid, $fromlist, "", $x, $y, $arr);
	}

	$maxx = count($arr);
	$maxy = 1;
	for ($x = 0; $x < $maxx; $x++)
	{
		$thisy = max(array_keys($arr[$x])) + 1;
		if ($thisy > $maxy)
		{
			$maxy = $thisy;
		}
	}

//	$str .= '<div class="horscrollouter clearboth">';
//	$str .= '<div class="horscroll">';

	if ($maxy > 1)
	{
		$str .= '<table class="familytree">';
	}
	else
	{
		$str .= '<div class="familytree">';
	}
	for ($y = $maxy - 1; $y >= 0; $y--)
	{
		//er vallen soms gaten in de $arr die helemaal leeg zijn; de oorzaak ligt in het recursieve gedeelte van deze functie $y onnodig ophoogt; als dat verbeterd is kan deze code weg
		$filled = 0;
		for ($x = $maxx - 1; $x >= 0; $x--)
		{
			if ($arr[$x][$y] != "")
			{
				$filled++;
			}
		}
		if ($filled == 0)
		{
			continue;
		}

		if ($maxy > 1)
		{
			$str .= '<tr>';
		}
		for ($x = $maxx - 1; $x >= 0; $x--)
		{
			if ($maxy > 1)
			{
				$str .= '<td>';
			}
			if ($arr[$x][$y] != "")
			{
//				$str .= '<div>' . $x . ',' . $y . ' ' . htme($arr[$x][$y]) . '</div>';
				$str .= $arr[$x][$y];
			}
			if ($maxy > 1)
			{
				$str .= '</td>';
			}
			if ($x > 0)
			{
				if ($maxy > 1)
				{
					$connectornum = 0;

					//streep van midden naar links
					if ($arr[$x][$y] != "")
					{
						$connectornum += 1;
					}

					//streep van midden naar rechts
					if ($arr[$x - 1][$y] != "")
					{
						$connectornum += 2;
					}
					else
					{
						//streep van midden naar onder
						for ($z = $y - 1; $z >= 0; $z--)
						{
							//de cel hieronder is gevuld
							if ($arr[$x][$z] != "")
							{
								$connectornum += 8;
								break;
							}
						}
					}

					//streep van midden naar boven
					for ($z = $y + 1; $z < $maxy; $z++)
					{
						//de cel hierboven is gevuld
						if ($arr[$x][$z] != "")
						{
							//de cel rechts hierboven is NIET gevuld
							if ($arr[$x - 1][$z] == "")
							{
								$connectornum += 4;
							}
							break;
						}
					}

					//http://www.theasciicode.com.ar/extended-ascii-code/box-drawing-character-single-line-lower-left-corner-ascii-code-192.html
					$connector = "";
					if ($connectornum == 3)
					{
						$connector = "─";
					}
					else if ($connectornum == 7)
					{
						$connector = "┴";
					}
					else if ($connectornum == 12)
					{
						$connector = "│";
					}
					else if ($connectornum == 9)
					{
						$connector = "┐";
					}
					else if ($connectornum == 13)
					{
						$connector = "┤";
					}

					if ($connector != "")
					{
						$str .= '<td class="connector">' . $connector . '</td>';
					}
					else
					{
						$str .= '<td></td>';
					}
				}
				else
				{
					$str .= ' / ';
				}
			}
		}
		if ($maxy > 1)
		{
			$str .= '</tr>';
		}
	}
	if ($maxy > 1)
	{
		$str .= '</table>';
	}
	else
	{
		$str .= '</div>';
	}

//	$str .= '</div></div>';

	return $str;
}
function familytree_recursive($table, $recordid, $starttable = "", $startrecordid = 0, $currentable = "", $currentrecord = 0, $fromlist, $prefix = "", $x = 0, $y = 0, $arr = array(), $already_joined = array())
{
//fromlist not in use anymore; because ilor() is used

	if ($prefix == "")
	{
		$prefix = $GLOBALS['UC']['PersonalDataStructure'][$table]["singular"];
	}

	$url = $GLOBALS['UC']['PersonalDataStructure'][$table]["openlink"] . $recordid;
	if (!strstr($url, "?"))
	{
		$url .= '?';
	}
	else
	{
		$url .= '&amp;';
	}
	//if we're at the starttable/startrecord at this moment; we do not need to add this info to the url
	if (($table != $starttable) || ($recordid != $startrecordid))
	{
		$url .= 'starttable=' . $starttable . '&amp;startrecordid=' . $startrecordid;
	}

	$link = '<a href="' . $url . '" onclick="ilor(this.href); return false;" class="famtree-' . $table;
	if (($table == $currentable) && ($recordid == $currentrecord))
	{
		$link .= ' active';
	}
	$link .= '"><span class="familytree_singularname">' . htme($prefix) . '</span> <span class="familytree_referlayout">' . htme(get_refer_layout_for_record($table, $recordid)) . '</span></a>';

	$arr[$x][$y] = $link;

	if (!$already_joined[$table][$recordid])
	{
		$already_joined[$table][$recordid] = true;
		$x++;
		foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["parents"]) as $parent_table)
		{
			if ($GLOBALS['UC']['PersonalDataStructure'][$parent_table]["MainTable"])
			{
				foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["parents"][$parent_table] as $field_in_current_table_referring_to_parent)
				{
					//TODO: getrecord gebruiken want deze persoon heeft misschien conditioneel geen toegang tot dit veld
					//geen support voor "calculated on display"; moet nog omzetten naar return_database_fieldname()
					$sql = "SELECT `" . $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field_in_current_table_referring_to_parent]["name_in_database"] . "` FROM `CRM" . $GLOBALS['UC']['PersonalDataStructure'][$table]["name_in_database"] . "` WHERE `" . $GLOBALS['UC']['PersonalDataStructure'][$table]["primary"] . "`=" . $recordid;
					$parent_id = db_GetValue($sql);
					if (is_positive_integer($parent_id))
					{
						list($x2, $y, $arr) = familytree_recursive($parent_table, $parent_id, $starttable, $startrecordid, $currentable, $currentrecord, $fromlist, $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field_in_current_table_referring_to_parent]["name"], $x, $y, $arr, $already_joined);
						$y++;
					}
				}
			}
		}
	}
	return array($x, $y, $arr);
}
function human_readable_tablename_to_ds_tablename($human_readable_table_name)
{
	//voor UseURLBeautification

	foreach (array_keys($GLOBALS['UC']['PersonalDataStructure']) as $table)
	{
		//case SENSITIVE
		if ($GLOBALS['UC']['PersonalDataStructure'][$table]["plural"] == $human_readable_table_name)
		{
			return $table;
		}
	}

	$num = db_GetValue("SELECT `recordid` FROM `CRMflextabledefs` WHERE `tablename`='" . mres($human_readable_table_name) . "'");
	if ($num > 0)
	{
		$table = "t" . $num;
		if (is_array($GLOBALS['UC']['PersonalDataStructure'][$table]))
		{
			return $table;
		}
	}

	//dit is traag want er kunnen geen indexen gebruikt worden
	foreach (db_GetArray("SELECT `TEXTID`, `TEXT` FROM `CRMlanguages` WHERE `TEXTID` REGEXP('flextable[0-9]+plural')") as $row)
	{
		//case SENSITIVE, we doen dit buiten mysql want anders matcht mysql case insensitive en Clientele matcht ook op Clientèle
		if (($row['TEXT'] == $human_readable_table_name) && (preg_match("/^flextable(\d+)plural$/", $row['TEXTID'], $match)))
		{
			$table = "t" . $match[1];
			//security; geen tables retourneren waar de gebruiker geen toegang tot heeft
			if (is_array($GLOBALS['UC']['PersonalDataStructure'][$table]))
			{
				return $table;
			}
		}
	}
}
function max_one_concurrent_load_start($unique_command_id)
{
	//NOG NIET IN GEBRUIK
	return true;

	//uit config.inc.php
	global $database;

	$filename = GetSetting("TMP_FILE_PATH") . "/INTLV_CMD_" . $database[$GLOBALS['repository']] . "_" . $unique_command_id;
	$success = false;
	if (!file_exists($filename))
	{
		touch($filename);
		$success = true;
	}
	else
	{
		$errortext = "unique_command_id: " . $unique_command_id . " is already running";
		//PrintAD($errortext);
		log_msg("ERROR: " . $errortext);
	}

	return $success;
}
function max_one_concurrent_load_finish($unique_command_id)
{
	//NOG NIET IN GEBRUIK
	return true;

	//uit config.inc.php
	global $database;

	$filename = GetSetting("TMP_FILE_PATH") . "/INTLV_CMD_" . $database[$GLOBALS['repository']] . "_" . $unique_command_id;
	$success = unlink($filename);
	return $success;
}
function max_one_concurrent_load_cleanup()
{
	//NOG NIET IN GEBRUIK
	return true;
}

//TODO the functions return_database_fieldname() and convert_database_date_and_number_column() and apply_conditional_read_access_to_field() can probably be merged!
function return_database_fieldname($table, $field, $table_alias)
{
	$database_fieldname = "";
	//just in time calculation
	if ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["calculation_type"] == "calculated on display")
	{
		$database_fieldname = "(" . $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["calculation"] . ")";
		$database_fieldname = str_replace("@table_alias@", "`" . $table_alias . "`", $database_fieldname);
		$database_fieldname = EvaluateTemplatePHP($database_fieldname);
	}
	else
	{
		$database_fieldname = "`" . $table_alias . "`.`" . $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["name_in_database"] . "`";
	}
	return $database_fieldname;
}
//TODO the functions return_database_fieldname() and convert_database_date_and_number_column() and apply_conditional_read_access_to_field() can probably be merged!
function convert_database_date_and_number_column($table, $field, $select_field, $format_empty_numbers = true, $always_cast_number_to_decimal = false)
{
	if ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["fieldtype"] == "date")
	{
		if ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["dateformat_in_database"] == "dd-mm-yyyy")
		{
			$select_field = "STR_TO_DATE(" . $select_field . ",'%d-%m-%Y')";
		}
		else if ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["dateformat_in_database"] == "epoch")
		{
			$select_field = "FROM_UNIXTIME(" . $select_field . ")";
		}
	}
	else if ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["fieldtype"] == "number")
	{
		//always_cast_number_to_decimal is only true for reload reload_calculations_with_request_data()
		if (($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["cast_number_to_decimal"]) || ($always_cast_number_to_decimal))
		{
			$tmp = "CAST(" . $select_field . " AS DECIMAL(12," . $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["decimals"] . "))";
			//to not format empty (varchar column) numbers; only in the display of list is this variable FALSE; for the sort column it must be TRUE; otherwise the sorting will not be numeric
			if (!$format_empty_numbers)
			{
				$tmp = "IF(" . $select_field . "='',''," . $tmp . ")";
			}
			$select_field = $tmp;
		}
	}
	return $select_field;
}
//TODO the functions return_database_fieldname() and convert_database_date_and_number_column() and apply_conditional_read_access_to_field() can probably be merged!
function apply_conditional_read_access_to_field($table, $field, $table_alias, $select_field)
{
	//je mag de inhoud van dit veld in principe zien, behalve wanneer een ander veld een bepaalde waarde heeft
	if (count($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["field_access_restrictions"]["read"]) > 0)
	{
		$field_access_restrictions = array();
		foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["field_access_restrictions"]["read"]) as $restricting_field)
		{
			foreach (array("IS", "NOT") as $operand)
			{
				$field_access_restriction = return_database_fieldname($table, $restricting_field, $table_alias);
				$field_access_restriction = convert_database_date_and_number_column($table, $restricting_field, $field_access_restriction);

				if ($operand == "NOT")
				{
					$field_access_restriction .= " NOT";
				}
				$field_access_restriction .= " IN (";
				$tel = 0;
				foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["field_access_restrictions"]["read"][$restricting_field][$operand] as $value)
				{
					if ($tel > 0)
					{
						$field_access_restriction .= ",";
					}
					if ($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$restricting_field]["fieldtype"] == "id")
					{
						//om dan maar geen results te hebben
						if (!is_positive_integer_or_zero($value))
						{
							//IS NULL?
							$value = -1;
						}
						//hoeft niet door mres() en hoeven ook geen quotes om heen
						$field_access_restriction .= $value;
					}
					else
					{
						$field_access_restriction .= "'" . mres($value) . "'";
					}
					$tel++;
				}
				$field_access_restriction .= ")";
				if ($tel > 0)
				{
					$field_access_restrictions[] = $field_access_restriction;
				}
			}
		}
		$select_field = "IF(" . implode(" AND ", $field_access_restrictions) . "," . $select_field . ",'')";
	}
	return $select_field;
}

function add_sessionid_to_url($url)
{
	if ($_GET['sessionid'] != "")
	{
		if (strpos($url, "?") === false)
		{
			$url .= "?";
		}
		else
		{
			$url .= "&amp;";
		}
		$url .= "sessionid=" . urlencode($_GET['sessionid']);
	}
	return $url;
}
function datastructuregraphic() {
	$dsg = '';
	$title = "Graphic data structure";
	$dsg .= '<h1>' . htme($title) . '</h1>';

	// HF20150507 Support voor show_admintables=1 ingebouwd
	$show_datatables = true;
	$show_usertable = true;
	$show_grouptable = true;
	$show_admintables = false;
	if ($_GET['showadmintables'] == 1) {
	    $show_admintables = true;
	}
	$dsg .= directed_graph_data_schema($show_datatables, $show_usertable, $show_grouptable, $show_admintables);

	$dsg .= '<h2>Graphml download</h2>';
	$dsg .= '<form method="get" action="listadmin.php">';
	$dsg .= '<input type="hidden" name="graphml" value="1">';
	$dsg .= '<div><input type="checkbox" name="graphmldatatable" value="yes" id="graphmldatatable" checked> <label for="graphmldatatable">show data tables</label></div>';
	$dsg .= '<div><input type="checkbox" name="graphmlusertable" value="yes" id="graphmlusertable"> <label for="graphmlusertable">show user and group tables</label></div>';
	$dsg .= '<div><input type="checkbox" name="graphmladmintable" value="yes" id="graphmladmintable"> <label for="graphmladmintable">show admin tables</label></div>';
	$dsg .= '<div><input type="checkbox" name="graphmltechnicalname" value="yes" id="graphmltechnicalname" checked> <label for="graphmltechnicalname">show technical names (e.g. t15 or f23)</label></div>';
	$dsg .= '<div><input type="checkbox" name="graphmlhumanname" value="yes" id="graphmlhumanname" checked> <label for="graphmlhumanname">show human names (e.g. Orders or Customer)</label></div>';
	$dsg .= '<div><input type="submit" value="Download xml"></div>';
	$dsg .= '</form>';


	$dsg .= '<div><a href="listadmin.php?datastructuregraphic=1&amp;datman=1&amp;showadmintables=1">Show admin tables as well</a></div>';

	$maxdepth = 0;
	$tabledepth = array();
	foreach (array_keys($GLOBALS['UC']['PersonalDataStructure']) as $table)
	{
	    if (($GLOBALS['UC']['PersonalDataStructure'][$table]["datatable"]) || ($_GET['showadmintables'] == 1))
	    {
	        $depth = get_table_depth($table);
	        $tabledepth[$table] = $depth;
	        if ($depth > $maxdepth)
	        {
	            $maxdepth = $depth;
	        }
	    }
	}

	$number_of_child_tables = array();
	$number_of_parent_tables = array();
	foreach (array_keys($GLOBALS['UC']['PersonalDataStructure']) as $table)
	{
	    if (($GLOBALS['UC']['PersonalDataStructure'][$table]["datatable"]) || ($_GET['showadmintables'] == 1))
	    {
	        $number_of_parent_tables[$table] = 0;
	        $number_of_child_tables[$table] = 0;
	    }
	}

	foreach (array_keys($GLOBALS['UC']['PersonalDataStructure']) as $table)
	{
	    if (($GLOBALS['UC']['PersonalDataStructure'][$table]["datatable"]) || ($_GET['showadmintables'] == 1))
	    {
	        foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]) as $field)
	        {
	            $relation_to = $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["relation_to"];
	            if ($relation_to != "")
	            {
	                $number_of_parent_tables[$table]++;
	                $number_of_child_tables[$relation_to]++;
	            }
	        }
	    }
	}

	$tablecolors = array();
	foreach (array_keys($GLOBALS['UC']['PersonalDataStructure']) as $table)
	{
	    if (($GLOBALS['UC']['PersonalDataStructure'][$table]["datatable"]) || ($_GET['showadmintables'] == 1))
	    {
	        $tablecolors[$table] = get_pastel_color_by_hash($table);
	    }
	}

	/*
	$js = "";
	$link = 0;
	for ($x = 0; $x <= $maxdepth; $x++)
	{
	    foreach (array_keys($GLOBALS['UC']['PersonalDataStructure']) as $table)
	    {
	        if ($tabledepth[$table] == $x)
	        {
	            //if (($GLOBALS['UC']['PersonalDataStructure'][$table]["datatable"]) || ($_GET['showadmintables'] == 1))
	            if (($GLOBALS['UC']['PersonalDataStructure'][$table]["datatable"]) || ($table == "users") || ($table == "groups"))
	            {
	                $dsg .= '<div style="margin: 10px; border: 1px solid #000; padding: 5px; display: inline-block" id="container' . $table . '">' . htme($GLOBALS['UC']['PersonalDataStructure'][$table]["plural"]) . '</div>';

	                foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]) as $field)
	                {
	                    $relation_to = $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["relation_to"];
	                    if ($relation_to != "")
	                    {
	                        $link++;
	                        $source = "e" . $table . $link;
	                        $js .= "var " . $source . " = jsPlumb.addEndpoint('container" . $table . "');\n";

	                        $link++;
	                        $target = "e" . $relation_to . $link;
	                        $js .= "var " . $target . " = jsPlumb.addEndpoint('container" . $relation_to . "');\n";

	                        $js .= "jsPlumb.connect({ source:" . $source . ", target:" . $target . " });\n";
	                    }
	                }
	            }
	        }
	    }
	    $dsg .= '<div class="clear: both;">&nbsp;</div>';
	}

	$dsg .= '<script src="https://jsplumbtoolkit.com/js/dom.jsPlumb-1.7.0-min.js"></script>';
	$dsg .= '<script type="text/javascript">';
	$dsg .= 'jsPlumb.bind("ready", function() {';
	$dsg .= $js;
	$dsg .= '});';
	$dsg .= '</script>';
	*/


	for ($x = 0; $x <= $maxdepth; $x++)
	{
	    $dsg .= '<div>';
	    $dsg .= '<h2>Level ' . ($x + 1) . '</h2>';
	    foreach (array_keys($GLOBALS['UC']['PersonalDataStructure']) as $table)
	    {
	        if ($tabledepth[$table] == $x)
	        {
	            if (($GLOBALS['UC']['PersonalDataStructure'][$table]["datatable"]) || ($_GET['showadmintables'] == 1))
	            {
	                $dsg .= '<div class="okmsg split" style="background-color: #' . $tablecolors[$table] . ';">';
	                $dsg .= '<h3>';
	                $closetag = false;
	                if (preg_match("/^t(\d+)$/", $table, $match))
	                {
	                    $dsg .= '<a href="flextable.php?EditFlexTable=' . $match[1] . '">';
	                    $closetag = true;
	                }
	                $dsg .= htme($table . ' ' . $GLOBALS['UC']['PersonalDataStructure'][$table]["plural"]);
	                if ($closetag)
	                {
	                    $dsg .= '</a>';
	                }
	                $dsg .= '</h3>';
	                $dsg .= '<ul>';
	                $dsg .= '<li>Number of parent tables: ' . $number_of_parent_tables[$table] . '</li>';
	                $dsg .= '<li>Number of links from child tables: ' . $number_of_child_tables[$table] . '</li>';

	                if (table_is_part_of_circular_join($table))
	                {
	                    $dsg .= '<li>this table is part of a <strong>circular join</strong></li>';
	                }
	                else if (circular_join_in_parents($table))
	                {
	                    $dsg .= '<li>this table has a parent that is part of a <strong>circular join</strong></li>';
	                }
	                $dsg .= '</ul>';
	                $dsg .= '<ol>';
	                foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]["fields"]) as $field)
	                {
	                    $dsg .= '<li>';
	                    $closetag = false;

	                    if (preg_match("/^f(\d+)$/", $field, $match))
	                    {
	                        $dsg .= '<a href="extrafields.php?editextrafield=' . $match[1] . '">';
	                        $closetag = true;
	                    }
	                    $dsg .= htme($field . ' ' . $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["name"]);
	                    if ($closetag)
	                    {
	                        $dsg .= '</a>';
	                    }
	                    $relation_to = $GLOBALS['UC']['PersonalDataStructure'][$table]["fields"][$field]["relation_to"];
	                    if ($relation_to != "")
	                    {
	                        $dsg .= ' <span class="colorspan" style="background-color: #' . $tablecolors[$relation_to] . ';">';
	                        if (preg_match("/^t(\d+)$/", $relation_to, $match))
	                        {
	                            $dsg .= '<a href="flextable.php?EditFlexTable=' . $match[1] . '">';
	                            $closetag = true;
	                        }
	                        $dsg .= htme($relation_to . ' ' . $GLOBALS['UC']['PersonalDataStructure'][$relation_to]["plural"]);
	                        if ($closetag)
	                        {
	                            $dsg .= '</a>';
	                        }
	                        $dsg .= '</span>';
	                    }
	                    $dsg .= '</li>';
	                }
	                $dsg .= '</ol>';
	                $dsg .= '</div>';
	            }
	        }
	    }
	}
	return $dsg;
}
function directed_graph_data_schema($show_datatables = true, $show_usertable = true, $show_grouptable = false, $show_admintables = false)
{
    //deze code is een zooitje, dit kan veel netter
        $str = '
<script src="lib/graphic_DS_js/d3.v3.js"></script>
<script src="lib/graphic_DS_js/graphlib-dot.js"></script>
<script src="lib/graphic_DS_js/dagre-d3.js"></script>

<style>
svg {
border: 1px solid #999;
overflow: hidden;
}

.node {
white-space: nowrap;
}

.node rect,
.node circle,
.node ellipse {
stroke: #333;
fill: #fff;
stroke-width: 1.5px;
}

.cluster rect {
stroke: #333;
fill: #000;
fill-opacity: 0.1;
stroke-width: 1.5px;
}

.edgePath path.path {
stroke: #333;
stroke-width: 1.5px;
fill: none;
}
</style>

<textarea id="inputGraph" class="hide_content">
/* Example */
digraph {
';

    foreach (array_keys($GLOBALS['UC']['PersonalDataStructure']) as $table) {
        $show_this_table = false;
        if (($GLOBALS['UC']['PersonalDataStructure'][$table]['datatable']) && ($show_datatables)) {
            $show_this_table = true;
        }
        if ((!$GLOBALS['UC']['PersonalDataStructure'][$table]['datatable']) && ($show_admintables)) {
            $show_this_table = true;
        }
        if (($table == 'users') && ($show_usertable)) {
            $show_this_table = true;
        }
        if (($table == 'groups') && ($show_grouptable)) {
            $show_this_table = true;
        }

        if ($show_this_table) {
            $str .= $table.' [labelType="html" label="<a href=list.php?table='.$table.'>'.$GLOBALS['UC']['PersonalDataStructure'][$table]['plural']."</a>\"];\n";
            foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]['childs']) as $child_table) {
                $show_this_table = false;
                if (($GLOBALS['UC']['PersonalDataStructure'][$child_table]['datatable']) && ($show_datatables)) {
                    $show_this_table = true;
                }
                if ((!$GLOBALS['UC']['PersonalDataStructure'][$child_table]['datatable']) && ($show_admintables)) {
                    $show_this_table = true;
                }
                if (($child_table == 'users') && ($show_usertable)) {
                    $show_this_table = true;
                }
                if (($child_table == 'groups') && ($show_grouptable)) {
                    $show_this_table = true;
                }

                if ($show_this_table) {
                    foreach ($GLOBALS['UC']['PersonalDataStructure'][$table]['childs'][$child_table] as $child_field) {
                        $str .= $table.'->'.$child_table;
                        if (count($GLOBALS['UC']['PersonalDataStructure'][$table]['childs'][$child_table]) > 1) {
                            $str .= ' [label="'.$GLOBALS['UC']['PersonalDataStructure'][$child_table]['fields'][$child_field]['name'].'"]';
                        }
                        $str .= "\n";
                    }
                }
            }
        }
    }

    $str .= '
}
</textarea>

<svg width="100%" height=800>
<g/>
</svg>

<script>
// Input related code goes here

function graphToURL() {
var elems = [window.location.protocol, \'//\',
window.location.host,
window.location.pathname,
\'?\'];

var queryParams = [];
if (debugAlignment) {
queryParams.push(\'alignment=\' + debugAlignment);
}
queryParams.push(\'graph=\' + encodeURIComponent(inputGraph.value));
elems.push(queryParams.join(\'&\'));

return elems.join(\'\');
}

var inputGraph = document.querySelector("#inputGraph");

var graphLink = d3.select("#graphLink");

var oldInputGraphValue;

var graphRE = /[?&]graph=([^&]+)/;
var graphMatch = window.location.search.match(graphRE);
if (graphMatch) {
inputGraph.value = decodeURIComponent(graphMatch[1]);
}
var debugAlignmentRE = /[?&]alignment=([^&]+)/;
var debugAlignmentMatch = window.location.search.match(debugAlignmentRE);
var debugAlignment;
if (debugAlignmentMatch) debugAlignment = debugAlignmentMatch[1];

// Set up zoom support
var svg = d3.select("svg"),
inner = d3.select("svg g"),
zoom = d3.behavior.zoom().on("zoom", function() {
inner.attr("transform", "translate(" + d3.event.translate + ")" +
"scale(" + d3.event.scale + ")");
});
svg.call(zoom);

// Create and configure the renderer
var render = dagreD3.render();

function tryDraw() {
var g;
if (oldInputGraphValue !== inputGraph.value) {
inputGraph.setAttribute("class", "");
oldInputGraphValue = inputGraph.value;
try {
g = graphlibDot.read(inputGraph.value);
} catch (e) {
inputGraph.setAttribute("class", "error");
throw e;
}

// Save link to new graph
graphLink.attr("href", graphToURL());

// Set margins, if not present
if (!g.graph().hasOwnProperty("marginx") &&
!g.graph().hasOwnProperty("marginy")) {
g.graph().marginx = 20;
g.graph().marginy = 20;
}

g.graph().transition = function(selection) {
return selection.transition().duration(500);
};

// Render the graph into svg g
d3.select("svg g").call(render, g);
}
}

tryDraw();
</script>';

    return $str;
}
function get_table_depth($table, $already_joined = array())
{
    //een tabel die direct naar zichtzelf joint (t1 -> t1) telt als depth = 0
        //een circular die t1 -> t2 -> t1 doet telt voor zowel t1 als t2 als depth = 1
        $already_joined[$table] = true;
    $depth = 0;
    foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]['fields']) as $field) {
        $relation_to = $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'];
        if (($relation_to != '') && (!$already_joined[$relation_to])) {
            $thisdepth = get_table_depth($relation_to, $already_joined) + 1;
            if ($thisdepth > $depth) {
                $depth = $thisdepth;
            }
        }
    }

    return $depth;
}

function table_is_part_of_circular_join($table, $firsttable = '', $already_joined = array())
{
    if ($firsttable == '') {
        $firsttable = $table;
    }
    $already_joined[$table] = true;
    $table_is_part_of_circular_join = false;
    foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]['fields']) as $field) {
        $relation_to = $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'];
        if ($relation_to != '') {
            if ($already_joined[$relation_to]) {
                if ($relation_to == $firsttable) {
                    $table_is_part_of_circular_join = true;
                }
            } elseif (table_is_part_of_circular_join($relation_to, $firsttable, $already_joined)) {
                $table_is_part_of_circular_join = true;
            }
        }
    }

    return $table_is_part_of_circular_join;
}

function circular_join_in_parents($table, $already_joined = array())
{
    $already_joined[$table] = true;
    $circular_join_in_parents = false;
    foreach (array_keys($GLOBALS['UC']['PersonalDataStructure'][$table]['fields']) as $field) {
        $relation_to = $GLOBALS['UC']['PersonalDataStructure'][$table]['fields'][$field]['relation_to'];
        if ($relation_to != '') {
            if ($already_joined[$relation_to]) {
                $circular_join_in_parents = true;
            } elseif (circular_join_in_parents($relation_to, $already_joined)) {
                $circular_join_in_parents = true;
            }
        }
    }

    return $circular_join_in_parents;
}

//we cannot use make_query because we need to rerun ALL calculations with user input data; also the db trigger queries. there is not (easy) support for this in make_query()
//access to this record must have been checked BEFORE calling this function
//this function recalculates the calculations on a form after a field has changed. it ONLY recalculated the calculations that might have change because of $alteredfield
function reload_calculations_after_field_change_with_form_data($table, $recordid, $changedfield, $formvalues)
{
    $result = array();

	$replaces = array();
	$existingrecord = false;
	$getfieldsfromdb = false;
	if (is_positive_integer($recordid))
	{
	    $replaces[$GLOBALS["UC"]["PersonalDataStructure"][$table]["primary"]] = $recordid;
	    $existingrecord = true;
	}
	//not needed; calculations that refer to the primary id field on a new record will remain empty
/*	else
	{
	    //this will never have any results
	    $replaces[$GLOBALS["UC"]["PersonalDataStructure"][$table]["primary"]] = -1;
	}*/

	$replaces["timestamp_last_change"] = "NOW()";

	foreach ($formvalues as $field => $val)
	{
		//only replace this field if this user is allowed to write this field (TODO: check conditional write access?)
		if ($GLOBALS["UC"]["PersonalDataStructure"][$table]["fields"][$field]["allowed_to_write"])
		{
		    //matches "123", "-12", "12.3"
		    if (($GLOBALS["UC"]["PersonalDataStructure"][$table]["fields"][$field]["fieldtype"] == "number") && (preg_match("/^\-?\d+(\.\d+)?$/", $val)))
		    {
			//number
		    }
		    else
		    {
			$val = "'" . mres($val) . "'";
		    }
		    $replaces[$GLOBALS["UC"]["PersonalDataStructure"][$table]["fields"][$field]["name_in_database"]] = $val;
		}
	}

	$changeddbfields = array();
	$changeddbfields[] = $GLOBALS["UC"]["PersonalDataStructure"][$table]["fields"][$changedfield]["name_in_database"];

	$queryparts = array();
	//this code currently reloads ALL the calculations on this record this user has access to; we could also reload only the ones that are on the form; but it takes time for the client to get these from the DOM
	//it only reloads calculations that might have changed (because of $changeddbfields)
	foreach (array_keys($GLOBALS["UC"]["PersonalDataStructure"][$table]["fields"]) as $field)
	{
	    if ($GLOBALS["UC"]["PersonalDataStructure"][$table]["fields"][$field]["calculation"] != "")
	    {
		$calculation = "(" . $GLOBALS["UC"]["PersonalDataStructure"][$table]["fields"][$field]["calculation"] . ")";
		$recalculate_this_calculation = false;
		foreach ($changeddbfields as $changeddbfield)
		{
			if ((preg_match("/@table_alias@." . $changeddbfield . "\b/", $calculation)) || (preg_match("/@table_alias@.`" . $changeddbfield . "`/", $calculation)))
			{
				$recalculate_this_calculation = true;
				break;
			}
		}

		if ($recalculate_this_calculation)
		{
			foreach ($replaces as $dbfield => $val)
			{
			    //converts @table_alias@.EFID12 to @table_alias@.`EFID12` \b for a word boundary so EFID1 does not match EFID12
			    $calculation = preg_replace("/@table_alias@." . $dbfield . "\b/", "@table_alias@.`" . $dbfield . "`", $calculation);
			    //we can use a simple str_replace now because this search pattern CANNOT mismatch
			    $calculation = str_replace("@table_alias@.`" . $dbfield . "`", $val, $calculation);
			}
			if (($existingrecord) && (strstr($calculation, "@table_alias@")))
			{
			    $getfieldsfromdb = true;
			    //any @table_alias@ that are not in $formvalues we get from the db, but ONLY if this is an existing record - otherwise there is no db record
			    $calculation = str_replace("@table_alias@", "`" . $table . "`", $calculation);
			}

			if ($GLOBALS["UC"]["PersonalDataStructure"][$table]["fields"][$field]["calculation_type"] == "calculated on display")
			{
			    $calculation = EvaluateTemplatePHP($calculation);
			}

			//this might happen if the primary auto increment id (recordid) is referred to in a calculation and this is a NEW record (where this id is not yet available) or another field is used in this calculation that is not $formvalues; we simply ignore this calculation
			if (!strstr($calculation, "@table_alias@"))
			{
			    //always_cast_number_to_decimal = true to format "100" as "100.00"
	                    $calculation = convert_database_date_and_number_column($table, $field, $calculation, true, true);
	                    //there is NO conditional field access on new records (TODO: do we need to support this?)
	                    if ($existingrecord)
	                    {
	                	$calculation2 = apply_conditional_read_access_to_field($table, $field, $table, $calculation);
	                        if ($calculation != $calculation2)
	                        {
	                            //we need to query the database because other fields are used - OR DO WE NEED TO VALIDATE CONDITIONAL FIELD ACCESS AGAINST THE POST VARS AS WELL?
	                            $getfieldsfromdb = true;
	                            $calculation = $calculation2;
	                        }
	                    }
			    $queryparts[] = $calculation . " AS `" . $table . "-" . $field . "`";

			    //this calculation might be used in calculations yet to come (with a higher order)
			    //it is not possible to refer to the value of "calculated on display" within other calculations; TODO; we could change this by recursively including the sql statements from other "calculated on display" inside the calculation that uses it; but we need to change this in make_query() as well. for that reason we don't support it here either, to have consistent behaviour
			    if ($GLOBALS["UC"]["PersonalDataStructure"][$table]["fields"][$field]["calculation_type"] == "database trigger")
			    {
				    $replaces[$GLOBALS["UC"]["PersonalDataStructure"][$table]["fields"][$field]["name_in_database"]] = $calculation;
				    $changeddbfields[] = $GLOBALS["UC"]["PersonalDataStructure"][$table]["fields"][$field]["name_in_database"];
			    }
			}
		}
	    }
	}

	if (count($queryparts) > 0)
	{
	    $query = "SELECT " . implode(", ", $queryparts);
	    if ($getfieldsfromdb)
	    {
		$query .= " FROM `CRM" . $GLOBALS["UC"]["PersonalDataStructure"][$table]["name_in_database"] . "` `" . $table . "` WHERE ";
		if ((preg_match("/^t\d+$/", $table)) && ($GLOBALS['UC']['PersonalDataStructure'][$table]["logical_delete"]))
		{
		    $query .= "`" . $table . "`.`deleted`='n' AND ";
		}
		$query .= "`" . $table . "`.`" . $GLOBALS["UC"]["PersonalDataStructure"][$table]["primary"] . "`=" . $recordid;
	    }

//mail("jeroen@atomos.nl", "qry", $query);

	    //stupid code to prevent numeric db indexes; [0] => "abc" AND ["t1-f3"] =]> "abc"
	    foreach (db_GetRow($query) as $key => $val)
	    {
		if (!is_numeric($key))
		{
			$result[$key] = $val;
		}
	    }
//	    $result = db_GetRow($query);
	}

    return $result;
}

?>
